// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCouplPhon.h>
#include <SxConstants.h>
#include <iostream> 
#include <fstream>
#include <SxCLI.h>

#ifndef SX_STANDALONE

double SxCouplPhon::sqOmega (const int iQ, 
                             const double &T)  
{
   SX_CHECK (iQ >= 0);
   SX_CHECK (! (T < 0.));//internal function anyhow
   int X2 = int(qMesh.getSize ()) - 1;
   double qPar = (qMesh(iQ) ^ qMesh(X2)) 
               / qMesh(X2).normSqr (), 
          qOrth = (qMesh(iQ) - qPar * qMesh(X2)).norm () 
                / qMesh(X2).norm (),
          res = sqr (12.3 * qPar / ASQUMEV) 
              + sqr (12.3 * qOrth / ASQUMEV) * 3.//2-4 times larger!
              - 2. * (fq (qPar, T) * (1. - kuz (T)) + fmu (qPar, T) * kuz (T))
              * exp (- sqr (qOrth * 6.) / 2.);
   return res; //\omega^2
}

SxVector<Double> SxCouplPhon::symH (const int iQ, 
                                    const int jQ, 
                                    const double &T)  
{
   SX_CHECK (T > -1.e-4);
   int X2 = int(qMesh.getSize ()) - 1;
   double iQPar = ( qMesh((iQ >= 0 ? iQ : softQ)) ^ qMesh(X2) ) 
                / qMesh(X2).normSqr (), 
          iQOrth = ( qMesh(iQ >= 0 ? iQ : softQ) - iQPar * qMesh(X2) ).norm () 
                 / qMesh(X2).norm (),
          jQPar = ( qMesh((jQ >= 0 ? jQ : softQ)) ^ qMesh(X2) ) 
                / qMesh(X2).normSqr (), 
          jQOrth = ( qMesh(jQ >= 0 ? jQ : softQ) - jQPar * qMesh(X2) ).norm () 
                 / qMesh(X2).norm (),
          softQPar = (qMesh(softQ) ^ qMesh(X2)) / qMesh(X2).normSqr (),
          softRes;
   SxVector<Double> res (2);
 
   if (potType == "real" || potType == "dubReal")  {
      
      // --real space potential, should give nice results?!
      res = 24. * real (iQPar, jQPar, T);//(qq',q'q)/2. correct factor?
      softRes = 24. * real (softQPar, softQPar, T)(0);//g_sh_s/2
   }  else if (potType == "broad")  {

      // ---explicit broadening: symmetric
      res(0) = fqBroad (iQPar, T) * fqBroad (jQPar, 0.) / sqr (scalQ);//(qq',q'q)/2.
      res = SxVector<Double> (SxList<double> () << res(0) << res(0))
          * exp (- sqr (iQOrth / 2.) / 2.) * exp (- sqr (jQOrth / 2.) / 2.);
      softRes = sqr (fqBroad (softQPar, T / sqrt (2.)) / scalQ);
   }  else if (potType == "dubBroad")  {

      // ---explicit broadening: asymmetric
      res = fqfqBroad (iQPar, jQPar, T) / sqr (scalQ)
          * exp (- sqr (iQOrth / 2.) / 2.) * exp (- sqr (jQOrth / 2.) / 2.);
      softRes = fqfqBroad (softQPar, softQPar, T)(0) / sqr (scalQ);//g_sh_s/2
   }  else if (potType == "default")   {
     
      // ---default potential: symmetric; including magnet!
      res(0) = fq (iQPar, T) * fq (jQPar, 0.) / sqr (scalQ) * (1. - kuz (T))
             + fmu (iQPar, T) * fmu (jQPar, 0.) / sqr (scalM) * kuz (T);
      res = SxVector<Double> (SxList<double> () << res(0) << res(0))
          * exp (- sqr (iQOrth * 6.) / 2.) * exp (- sqr (jQOrth * 6.) / 2.);
      softRes = sqr (fq (softQPar, T / sqrt (2.)) / scalQ) * (1. - kuz (T))
              + sqr (fmu (softQPar, T / sqrt (2.)) / scalM) * kuz (T);
   }  else  {
     
      // ---potType = "double":asymmetric, including magnet!
      res = (fqfq (iQPar, jQPar, T) / sqr (scalQ) * (1. - kuz (T))
            + fmufmu (iQPar, jQPar, T) / sqr (scalM) * kuz (T))
          * exp (- sqr (iQOrth * 6.) / 2.) * exp (- sqr (jQOrth * 6.) / 2.);
      softRes = fqfq (softQPar, softQPar, T)(0) / sqr (scalQ) * (1. - kuz (T))
              + fmufmu (softQPar, softQPar, T)(0) / sqr (scalM) * kuz (T);
   }
   if (iQ == -1 && jQ == -1)  {//both are softQ
      return SxList<double> () << 3. * softRes << 3. * softRes;
   }
   return res / softRes;//at least one not softQ
}

SxCouplPhon::SxCouplPhon (const SxArray<Coord> &qMesh_, 
                          const SxString &potType_,
                          const bool debug_)  
{
   qMesh = qMesh_;
   int nQ = int(qMesh.getSize ());
   if (nQ == 0)  {
      cout << "Error initializing SxCouplPhon, no qMesh provided!" << endl;
      SX_EXIT;
   }
   
   potType = potType_;
   if ( potType != "default" && potType != "double" && potType != "broad"
     && potType != "dubBroad" && potType != "real" && potType != "dubReal")
   {
      cout << "Error initializing SxCouplPhon! wrong potType provided:" 
           << potType << endl;
      SX_EXIT;
   }
   
   debug = debug_;
   if (debug)  {
      cout << "debug mode set" << endl;
   }
   
   //set scaling units once (2.044, .0862)
   ASQUMEV = 1.e+11 * HPLANCK * sqrt (HA2EV * JL2HA * 10. / U2KG) / TWO_PI;
   K2MEV = KB * JL2HA * HA2EV * 1000.;
   scalQ = 4.1;
   scalM = 3.29;
   Tc = 376.;
   cout << "units set " << ASQUMEV << K2MEV << endl;

   /* determine (prelim.) softQ, very simple! take ~.37 X_2
   softQ = 1;
   for (iQ = 2; iQ < nQ; iQ++)  {//for all pots!
      softQ = ( (qMesh(iQ) - qMesh(nQ - 1) * .37).norm () 
            < (qMesh(softQ) - qMesh(nQ - 1) * .37).norm () 
            ? iQ : softQ );
   }*/
   setSoftQ (0.);
   if (debug)  {
      cout << "maximal coupling @qPoint= " << qMesh(softQ) << endl;
   }
}

void SxCouplPhon::setSoftQ (const double &T)  {
   bool done = false;
   int add = 1;
   double oldRes = sqOmega (softQ, T) / symH (softQ, -1, T).sum () / symH (-1, -1, T).sum (),
          res = sqOmega (softQ+1, T) / symH (softQ+1, -1, T).sum () / symH (-1, -1, T).sum ();
   if (res > oldRes)  { add = -1; }
   while (! done)  {
     softQ += add;
     res = sqOmega (softQ, T) / symH (softQ, -1, T).sum () / symH (-1, -1, T).sum ();
     if (res > oldRes)  { done = true; }
     oldRes = res;
   }
   softQ -= add;
}

SxVector<Double> SxCouplPhon::getFreq (const double &softAmp, 
                                       const SxVector<Double> &avCouplAmp,
                                       const double &T)  
{
   if (softAmp < 0.)  {
      cout << "Error in SxCouplPhon::getFreq! negative softAmp: " 
           << softAmp << endl;
      SX_EXIT;
   }
   SX_CHECK (T > 0.);
/* if (! (avCouplAmp(0) > 0. && avCouplAmp(1) > 0.))  {
      cout << "Error in SxCouplPhon::getFreq! negative avCouplAmp: " 
           << avCouplAmp << endl;
      SX_EXIT;
   }*/

   int iQ, 
       X2 = int(qMesh.getSize ()) - 1;
   double softCouplAmp = symH (-1, -1, T).sum () * sqr (softAmp);
   SxVector<Double> res (X2 + 1); 
   for (iQ = 0; iQ <= X2; iQ++)  {
      if ((qMesh(iQ) - qMesh(softQ)).norm () < .75 * qMesh(X2).norm ())  {
         //only close to softQ!  
          res(iQ) = sqOmega (iQ, T) + symH (iQ, -1, T).sum () * softCouplAmp 
                  + dot (symH (iQ, -1, T), avCouplAmp) / 2.; 
      }  else  {
         res(iQ) = sqOmega (iQ, T);
      }
      if (res(iQ) < 1.e-6)  {
         cout << "Negative frequency " << res(iQ) 
              << "@softCouplAmp: " << softCouplAmp 
              << ", T: " << T << ", qPoint: " << qMesh(iQ) 
              << " put to small (pos) value!" << endl;
         res(iQ) = 1.e-12;
      }
      res(iQ) = sqrt (res(iQ));
   }

   return res * ASQUMEV; //= 2.04454;//in meV! 
} 

SxVector<Double> SxCouplPhon::getAmp (const SxVector<Double> &inAmp, 
                                       const double &softAmp,
                                       const double &T)  {
   SX_CHECK (T > 0.);
   SX_CHECK (softAmp > 0.);
   int iQ, 
       nQ = int(qMesh.getSize ());
   double kT = K2MEV * T,           
          qWeight = 1. / nQ; //BZ/SS scaling?
   
   //get the frequencies; they're small when inAmp too small!
   SxVector<Double> freq = getFreq (softAmp, inAmp, T) / ASQUMEV,
                    res (2);
   res.set (0.);
   for (iQ = 0; iQ < nQ; iQ++)  {
     // sum classical amplitudes; quadratic weights?
         res += symH (-1, iQ, T) / sqr (freq(iQ));
   }
        res *= kT * symH (-1, -1, T).sum () * qWeight;
   return res;
}

SxVector<Double> SxCouplPhon::getAvCouplAmp (const double &softAmp, 
                                            const double &T)  { 
   if (softAmp < 0.)  {
      cout << "Error in SxCouplPhon::getAvCouplAmp! negative softAmp: " 
           << softAmp << endl;
      SX_EXIT;
   }
   if (T < 0)  {
      cout << "Error in SxCouplPhon::getAvCouplAmp! negative temperature: " 
           << T << endl;
      SX_EXIT;
   }
   if (T == 0)  {
      return SxList<double> () << 0. 
                               << 0.;
   }
   SxVector<Double> outAmp (2), 
                    inAmp (2);
   
   // --initialize inAmp to reasonably large value & outAmp smaller 
   inAmp(0) = 2.5;
   inAmp(1) = inAmp(0);
   outAmp = .5 * inAmp;

   // ---scale amp's until they're equal
   while ( inAmp(0) > outAmp(0) || 
           inAmp(1) > outAmp(1) )  
   {
      // get new amplitudes
      outAmp = getAmp (inAmp, softAmp, T);
     
      //scale inAmp: lowering one increases the (output of the) other as well!
      // they nicely balance?!
      inAmp *= exp (1.e-2 * logVec (.98 * outAmp / inAmp));
   }
   inAmp *= exp (1.e-2 * logVec (inAmp / outAmp / .98));//one step back
   outAmp = getAmp (inAmp, softAmp, T);
   if (debug)  {
      cout << inAmp << " inAmp; ratio " << inAmp / outAmp << endl;
   }
   return inAmp;
}

double SxCouplPhon::getAvCouplAmp1 (const double &softAmp, 
                                    const double &T)  
{ 
   SX_CHECK (softAmp >= 0.);
   
   double kT = K2MEV * T, 
          softCouplAmp = sqr (hq (softQ, T) * softAmp),
          inAmp = sqOmega (softQ, T) + 2. * softCouplAmp;
              //min 0! max max fq, must not be too small!!
   if (T < 10.)  {
      return sqrt (kT);//calculation too involved
   }
 
   // initialize large enough inAmp: max ampl + sqrt (kT)
   inAmp = (inAmp < 0. ? -2. * inAmp + sqrt (kT) : sqrt (kT)); 
   if (debug)  {
      cout << "initial coupling amplitude " << inAmp;
   }
   
   int iQ, nQ = int(qMesh.getSize ());
   double minFreq = 0., 
          sqFreq, 
          res = 0., //init smaller
          resO = .1; //if scaled to much
   while (res < inAmp)  {
      res = 0.;
      minFreq = 10.;
      for (iQ = 0; iQ < nQ; iQ++)  {
         sqFreq = sqOmega (iQ, T) + hq (iQ, T) / hq (softQ, T) 
                * (2. * softCouplAmp + inAmp); 
         if (sqFreq < minFreq)  {
            minFreq = sqFreq;
         }
         if (sqFreq < 1.e-12)  {//should not happen anymore?!?
            inAmp *= exp (1.01e-0 * log (inAmp / resO / .98));
            res = 0.;//loop shouldn't end now
            cout << iQ;
            break;
         }
         res += hq (iQ, T) * hq (softQ, T) / sqFreq;
      }
      if (iQ == nQ)  {//no break
         res *= kT / nQ;
         inAmp *= exp (1.e-2 * log (.98 * res / inAmp));//improved scaling
         resO = res;
      }
   }
   if (debug)  {
      cout << "final amp" << inAmp << "ratio" << inAmp/res << "minFreq" 
           << minFreq << endl;
   } 
   //return scaled inAmp, factor should be as good as 1!
   return inAmp * exp (1.e-2 * log (inAmp / res / .98));
}

SxVector<Double> SxCouplPhon::getFreq1 (const double &softAmp, 
                                        const double &avCouplAmp, 
                                        const double &T)  
{
   if (softAmp < 0.)  {
      cout << "Error in SxCouplPhon::getFreq! negative X: " 
           << softAmp << endl;
      SX_EXIT;
   }
   if (avCouplAmp < 0.)  {
      cout << "Error in SxCouplPhon::getFreq! negative avCouplAmp: " 
           << avCouplAmp << endl;
      SX_EXIT;
   }

   int iQ, 
       X2 = int(qMesh.getSize ()) - 1;
   SxVector<Double> res (X2 + 1);
   for (iQ = 0; iQ <= X2; iQ++)  {
      if ((qMesh(iQ) - qMesh(softQ)).norm () < .75 * qMesh(X2).norm ())  {
         //only qPoints close to softQ 
         res(iQ) = sgnSqrt ( sqOmega(iQ, T) + hq (iQ, T) / hq (softQ, T)
                 * (2. * sqr (hq (softQ, T) * softAmp) + avCouplAmp));
      }  else  {//TODO weighing with distance (to get a smooth dependence)!
         res(iQ) = sgnSqrt (sqOmega (iQ, T));
      } 
      if (res(iQ) < 1.e-12)  {
         cout << "smallFreq!" << res(iQ) << endl;
      }
   }
   return res * ASQUMEV; // meV!
}

double SxCouplPhon::getFreeEn (const SxVector<Double> &freq, 
                               const double &T)  
{ 
   int nFreq = int(freq.getSize ());
   double qWeight = 1. / nFreq,//scaling
          kT = K2MEV * T;//K->meV 

   if (nFreq == 0)  {
      cout << "Error in SxCouplPhon::getFreeEn: no freq provided!" << endl;
      SX_EXIT;
   }
   if (T < 1.)  {
      cout << "Error in SxCouplPhon::getFreeEn: no freq provided!" << endl;
      SX_EXIT;
   }
   return kT * logVec (freq / sqrt (PI * kT)).sum () * qWeight;
}

void SxCouplPhon::writeToFile (const SxArray<SxVector<Double> > &freq, 
                               const SxArray<SxVector<Double> > &freeEn, 
                               const double &dT, 
                               const double &dX,
                               const SxArray<double> &minX)  
{
    // --total energy output to file
   ofstream totEout;
   //SxString file ("couplPhonE"+potType);//does not work???
   totEout.open ("couplPhonE");
   cout << "couplPhonE opened" << endl;
   totEout << "#q-point amplitude energy (meV/atom)" << endl;
   int iQ, iX, iT;
   for (iQ = 0; iQ < qMesh.getSize (); iQ++)  {
      for (iX = 0; iX < 2. * minX(0) / dX; iX++)  {
         totEout << iQ << " " << iX * dX << " " 
                 << (sqOmega (iQ, 0.) * sqr (iX * dX) / 2. 
                  + sqr ( fq (iQ, 0.) / scalQ) * quart (iX * dX) ) << endl;
      }// getTotE no q-dep!
   } 
   totEout.close ();
   cout << "couplPhonE closed" << endl;

   // --magnetic E output to file
   ofstream magnEnOut;
   magnEnOut.open ("couplPhonM");
   cout << "couplPhonM opened" << endl;
   magnEnOut << "#softAmp (scaled) magnEn (meV/atom) vs. T in steps of"
             << dT << "K";
   for (iX = 0; iX < freeEn(0).getSize (); iX++)  {
      magnEnOut << endl << iX * dX;
      for (iT = 0; iT < freeEn.getSize (); iT++)  {
         magnEnOut << " " << getTotE (iX * dX, iT * dT);
      }
   }
   magnEnOut.close ();
   cout << "couplPhonM closed" << endl;

   // --free energy output to file
   ofstream freeEnOut;
   freeEnOut.open ("couplPhonF");
   cout << "couplPhonF opened" << endl;
   freeEnOut << "#minX(T) in steps of " << dT << " K, for potType:" 
             << potType << endl << "#";
   for (iT = 0; iT < minX.getSize (); iT++)  {
      freeEnOut << minX(iT) << " ";
   }
   freeEnOut << endl << "#softAmp (scaled) freeEn (meV) vs. T in steps of"
             << dT << "K";
   for (iX = 0; iX < freeEn(0).getSize (); iX++)  {
      freeEnOut << endl << iX * dX;
      for (iT = 0; iT < freeEn.getSize (); iT++)  {
         freeEnOut << " " << (freeEn(iT)(iX) - freeEn(iT)(0));
      }
   }
   freeEnOut.close ();
   cout << "couplPhonF closed" << endl;

   // --frequency output to file
   ofstream freqOut;
   freqOut.open ("couplPhonFreq");
   cout << "couplPhonFreq opened" << endl;
   freqOut << "#q-point freq(T) up to " << double(freq.getSize ()) * dT << endl;
   for (iQ = 0; iQ < freq(0).getSize (); iQ++)  {
      freqOut << iQ;
      for (iT = 0; iT < freq.getSize (); iT++)  {
         freqOut << " " << freq(iT)(iQ);
      }
      freqOut << endl;
   }
   freqOut.close ();
   cout << "couplPhonFreq closed" << endl;
}

#else /* SX_STANDALONE */

int main (int argc,
          char **argv)  
{
   // --command line interface
   SxCLI cli (argc, argv);
   bool oneD = cli.option ("-D|--oneD", "bool", 
                           "toggle 1D grid").toBool (),
        debug = cli.option ("-db|--debugg", "bool", 
                            "toggle debug mode").toBool ();
   if (oneD)  {
      cout << "1D qGrid" << endl;
   }
   if (debug)  {
      cout << "debug mode" << endl;
   }
   int iX, iT, iQ, 
       nQ = cli.option ("-nQ", "number", "#q-points").toInt (24);
   double dX = cli.option ("-dX", "decimal", 
                           "soft amplitude step").toDouble (.3),
          maxT = cli.option("-maxT","decimal", 
                            "max. temperature (K)").toDouble (375.),
          dT = cli.option ("-dT", "decimal", 
                           "temperature step (K)").toDouble (75.),
          maxX = cli.option ("-maxX", "decimal", 
                             "max. soft amplitude").toDouble (1.5);
   SxString potType = cli.option ("-pot", "string", 
                                 "type of potential used").toString ("default");
   cli.finalize (); //defaults are reasonably conservative
   cout << "magnetism included" << endl;

   //--- initialize 1D/3D mesh
   int nQpar = (oneD ? nQ : int (cbrt (8. * nQ))), 
       nQorth = int (sqrt (nQ / nQpar));//per dir!
   // nQ reset to nice product
   nQ = (nQ / (nQpar * sqr (nQorth))) * nQpar * sqr (nQorth);
   double dQ = 1. / nQ, 
          dQpar = 1. / nQpar, 
          dQorth = 1. / nQorth;
   SxArray<Coord> qMesh (nQ);
   for (iQ = 1; iQ < nQ; iQ++)  {
      qMesh (iQ - 1) = Coord ((iQ % nQpar) * dQpar, 
                             ((iQ / nQpar) % nQorth) * dQorth,
                             (iQ / (nQpar * nQorth)) * dQ * nQpar * nQorth);
   }//first nQpar are [110]!
   qMesh(nQ - 1) = Coord (1., 0., 0.); //last==X2
   /*TODO import data C00 & 3M (+red!), calc. phon,F for non-soft modes, reorder soft
    * phon 3M (!well defined grid!); add interpol F to Etot param(+e(T),mu(T)):
    * calc softAmp, avCoupAmp (F for this V!), print/ save for phonSpectrum!
    for phonSpectr. import data again, new q-points -> interpolated spectrum +
    correct (low) soft freq
    Finally, combine F(V) with murn curve for QH approx... (-xV??)
    grid e.g. 36x18x12->12x9x12
    However: decreasing mu!? just like V dependence? no!
    conclusion: in principle possible, but too involved and harmonic/volume dependence is small anyhow!
    What about increasing softQ??!
    */
  
   //-- initialize class and some parameters
   int nX = int(maxX / dX), 
       nT = int(maxT / dT);
   cout << "nX" << nX << "nT" << nT << endl;
   double temp;
   SxArray<double> minX (nT);
   SxVector<Double> freqTX, 
                    avCouplAmp;
   SxArray<SxVector<Double> > freq (nT), 
                              freeEn (nT);
   SxCouplPhon couplPhon (qMesh, potType, debug); 
   
   // ---loop over temperatures
   for (iT = 0; iT < nT; iT++)  {
      cout << "iT" << iT << endl;
      temp = (iT + .5) * dT;
      freeEn(iT).resize (nX);
      
      // ---loop over amplitudes
      for (iX = nX - 1; iX >= 0; iX--)  {//TODO vary softQ+/-1!!?
         cout << "iX" << iX << endl;

         // get average coupling amplitude of all phonons 
         avCouplAmp = ( 
              potType == "default" || potType == "real" || potType == "broad" ?
             SxList<double> () << couplPhon.getAvCouplAmp1 (iX * dX, temp) << 0.
                                : couplPhon.getAvCouplAmp (iX * dX, temp) );
         
         // get the frequencies (meV)
         freqTX = ( 
               potType == "default" || potType == "real" || potType == "broad" ?
               couplPhon.getFreq1 (iX * dX, avCouplAmp(0), temp) :
               couplPhon.getFreq (iX * dX, avCouplAmp, temp) ); 
         
         // get the total energy (meV/atom, all potTypes) 
         freeEn(iT)(iX) = couplPhon.getTotE (iX * dX, temp);
         cout << "totEnX" << freeEn(iT)(iX) << endl;

         // add the vibr. freeEn (meV/mode) only 1 out of 3*4 modes!
         freeEn(iT)(iX) += (oneD? .1 : .25) * couplPhon.getFreeEn (freqTX, temp);
         cout << "freeEnX" << freeEn(iT)(iX) << endl;

         // --set freq and amplitude of minimum freeEn
         if (iX == nX - 1 || freeEn(iT)(iX) < freeEn(iT)(iX + 1))  {
            // now avCouplAmp counts double?? no!!!
       /*   freqTX = ( 
               potType == "default" || potType == "real" || potType == "broad" ?
               couplPhon.getFreq1 (iX * dX, 2. * avCouplAmp(0), temp) : 
               couplPhon.getFreq (iX * dX, 2. * avCouplAmp, temp) );
*/
            freq(iT) = freqTX; //only 1st nQPar!!
            minX(iT) = iX * dX; 
         }
      }
   }
   couplPhon.writeToFile (freq, freeEn, dT, dX, minX);
   return 0;
}  
#endif /* SX_STANDALONE */
