// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_COUPLPHON_H_
#define _SX_COUPLPHON_H_

#include <SxAtomicStructure.h>
#include <SxMatrix.h>
#include <SxVector.h>
#include <SxArray.h>
#include <SxMath.h> 
#include <SxExt.h>

/** \brief calculate the free energy and frequencies of a coupled phonon system.
 * The self-couplings are taken from the anharmonic single mode potentials. The
 * couplings with the soft mode come from the frequency shift. Those couplings
 * are combined to ones of type fq*gq'. The couplings together with the macroscopic
 * displacement of the 'soft mode' stabelizes all frequencies and so the partition sum
 * can be evaluated.  :useless description, it doesn't explain anything!
 1D version of a coupled phonon system
 needed:freq(func), coupl(multiFunc), nQ, dT, maxT, dX, maxX
 output freq(T,q), freeEn(T,X), minX(T)
 done:getAvCoupAmpl (softAmpl, iT); getXmin(iT); getFreq(X, iT);
 getFreeEn(freq, iT); writeToFile(freq, freeEn, minX).
UNITS: meV/atom!, \AA, u/atom!
NO QUANTUM SUPPORT!
TODO: use functions of SxCouplPhon1

   \b SxCouplPhon = S/PHI/nX compute the thermodynamics of a coupled phonon system.

   \author Matth\'e Uijttewaal, uijttewaal@mpie.de 
 */
class SX_EXPORT_EXT SxCouplPhon
{
   public:
      /* x to power a
         */
      inline double power (const double &x, 
                           const double &a)  
      {
         if (fabs (x) > 1.e-6)  {
            return exp (a * log (fabs (x)));
         }
         return 0.;
      }

      inline double sqr (const double &x)  
      {//SxMath
         return x * x;
      }
      
      inline double cube (const double &x)  
      {//SxMath
         return x * x * x;
      }

      inline double quart (const double &x)  
      {//SxMath
         return sqr (sqr (x)); //or faster: x * x * x * x?
      }
      
      inline double sgnSqrt (const double &x)  {//SxMath
         return sqrt (fabs (x)) * (x < 0. ? -1. : 1.);
      }

      inline SxVector<Double> logVec (const SxVector<Double> &vec) {//SxMath
         ssize_t nElt = vec.getSize ();
         SxVector<Double> res (nElt);
         for (ssize_t iElt = 0; iElt < nElt; iElt++)  {
            if (vec(iElt) < 1.e-16)  {
               cout << "Error in logVec, too small value: " << vec(iElt) 
                    << " for iElt: " << iElt << endl;
               SX_EXIT;
            }
            res(iElt) = log (vec(iElt));
         }
         return res; 
      }
      
      inline double dot (const SxVector<Double> &vec, 
                         const SxVector<Double> &vec1)  
      {//>SxVec.h
         SxVector<Double> res (vec);
         res *= vec1;
         return res.sum ();
      }
   protected:
      bool debug; //toggle debug mode
      int softQ; //index soft q-point
      SxString potType; //selects type of potential:
      //"default", "broad", "real", "double"
      double Tc,    //Curie temperature
             scalQ, //scaling factor quartic coupling
             scalM, //scaling factor reduced magnetization
             ASQUMEV,//angstroem sqrt (u meV) 
             K2MEV; //kelvin->meV
      SxArray<Coord> qMesh; //mesh with 1-points; last == X2
  
      //---------various INTERNAL potential functions------------
      /*kuz'min function for T dependent mu, achieves maximum of 1 @Tc
        */
      inline double kuz (const double &T)
      {
         SX_CHECK (T > -1.e-6);
         if (T > Tc) { return 1.; }//faster
         double T1 = power (T / Tc, 1.5),
                T2 = power (T / Tc, 2.8),
                res = (1. - 1. * power ((1. - .1 * T1 - .9 * T2), .333));
         return res;
      }

      /* fit to non-acoustic harmonic / anharmonic: 
         T_c=\infty, f(q)=-651x^2q^4.6/(1+exp(24*(q-.41))), 
         T^c=1776, a=20.1 => softQ = .371 X_2; sqr(fq/a) is 4th order term 
         @param Q q-point in qMesh
         @param T electronic temperature, quartic term: T/sqrt(2.)!
         @return (f_qf'_q',f_q' f'_q)/2.;
       */
      inline double fq (const double &Q,
                        const double &T = 0.)
      {
         return sqrt (fqfq (Q, Q, sqrt (2.) * T)(0));
      }
      /*asymmetric!*/
      inline SxVector<Double> fqfq (const double &iQPar,
                                    const double &jQPar, 
                                    const double &T = 0.)  
      {
         return sqr (651.) * SxVector<Double> (SxList<double> () 
                           << sqr (cube (iQPar)) * cube (jQPar) //~^4.6*2
                            / sqr (1. + exp (24. * (jQPar - .41)))
                           << sqr (cube (jQPar)) * cube (iQPar) 
                            / sqr (1. + exp (24. * (iQPar - .41)))
                                             ) * exp (- sqr (T / 1776.));
      }
      inline double fmu (const double &Q,
                         const double &T = 0.)
      {
         return sqrt (fmufmu (Q, Q, sqrt (2.) * T)(0));
      }
      inline SxVector<Double> fmufmu (const double &iQPar,
                                      const double &jQPar,
                                      const double &/*T = 0.*/)
      { 
         return sqr (61.57) * SxVector<Double> (SxList<double> () 
                           << quart (iQPar) * cube (jQPar) //~^3.4*2
                            / sqr (1. + exp (11.36 * (jQPar - .55)))
                           << quart (jQPar) * cube (iQPar) 
                            / sqr (1. + exp (11.36 * (iQPar - .55))));
                             //NO T dep!       ) * exp (- sqr (T / 1776.));
      }
      
      // Broader coupling function
      inline double fqBroad (const double &Q, 
                             const double &T = 0.)  
      { 
         return sqrt (fqfqBroad (Q, Q, sqrt (2.) * T).sum ());
      }
      inline SxVector<Double> fqfqBroad (const double &iQPar, 
                                         const double &jQPar,
                                         const double &T = 0.)  
      {
         return sqr (10.77) / 2. * SxVector<Double> (SxList<double> ()
             << quart (iQPar / .41) * sqr (jQPar / .41) 
                                    / sqr (1. + exp (12. * (jQPar - .41)))
             << quart (jQPar / .41) * sqr (iQPar / .41) 
                                    / sqr (1. + exp (12. * (iQPar - .41)))
                                                    ) * exp (- sqr (T / 1776.));
      }
      
      /**real space couplings 
         @param iQPar 1st q-point (X2) 
         @param jQPar 2nd q-point (X2) 
         @param T temperature (K)
         @return (g_q h_q', h_q g_q')/2., only parallel parts!
       */
      inline SxVector<Double> real (const double &iQPar, 
                                    const double &jQPar,
                                    const double &T = 0.)  
      { 
         return SxVector<Double> (SxList<double> ()
               << .08 * sqr (sin (PI * iQPar) * sin (1.5 * PI * jQPar))
               << .08 * sqr (sin (PI * jQPar) * sin (1.5 * PI * iQPar))
                                                    ) * exp (- sqr (T / 1776.));
      }

      /* harmonic potential (meV) vs. X(\AA sqrt(u)) 
         master formula: sqr(12.3qx/2.0445)*exp(-sqr(T/T_c))/2
                +(sqr( sqr(qx)f(q)-1 )-1)/sqr(a)*exp(-sqr(T/T^c)); (q~[110]) 
         @param iQ index q-point in mesh
         @param T (electronic) temperature
         @return omega_q^2*/
      double sqOmega (const int iQ, 
                      const double &T = 0.); 

      inline double hq (const int iQ, 
                        const double &T)
      {
         return sqrt (symH (iQ, iQ, T)(0) * symH (-1, -1, T).sum ());
      }


      /* \brief 2 mode quartic potential from self- and softQ interaction
         @param iQ index of 1st q-point in mesh
         @param jQ 2nd q-point, negative indicates softQ
         @return (g_q*h_q', g_q'*h_q)/g_sh_s 
              || 3*(g_s*h_s, g_s*h_s)/2 if q=q'=s, s=1/3?! */
      SxVector<Double> symH (const int iQ, 
                             const int jQ = -1, 
                             const double &T = 0.);
      //--------------END-----------------

   public:
      /* \brief constructor: set debug & softQ 
          softQ is min of \omega_q^2 g_sh_s/(g_qh_q)/\omega_s^2
       */
      SxCouplPhon (const SxArray<Coord> &qMesh_, 
                   const SxString &potType_ = "default", 
                   const bool debug_ = false); 
       
      // empty destructor
     ~SxCouplPhon () { /* empty */ }
   
      /* \brief set softQ as minimum of sqrOmega/coupling (varies with T)
         \param T temperature (K) 
       */ 
      void setSoftQ (const double &T);

      // get the freq vs. q-point @X softAmpl and T~K using statPhys.
     // for double potentials
      SxVector<Double> getFreq (const double &softAmp, 
                                const SxVector<Double> &avCouplAmp, 
                                const double &T);

   protected:
     /* get amplitude from input amplitude, uses getFreq
        INTERNAL only for minimizing of getAvCoupAmp 
      */
      SxVector<Double> getAmp (const SxVector<Double> &inAmp, 
                               const double &softAmp, 
                               const double &T);
   public:   
      /* \brief get average coupling amplitude of all phonons (sqrt(meV)) 
         
         formula: CLASSICAL, idem for g<->h: <g_s h\nu^2>_{\nu_s}=k_BT\int dq g_sh_q /freq^2
         @param X soft mode amplitude (times couplings, in sqrt(meV)?) 
         @param T temperature in K
         @return (g_s<h nu^2>,h_s<g nu^2>) or (g_s<h n/omega>,h_s<g n/omega>)*/
      SxVector<Double> getAvCouplAmp (const double &softAmp, 
                                      const double &T);
      double getAvCouplAmp1 (const double &softAmp, 
                             const double &T);


      // calculate correct frequencies vs. q-point @min softAmpl
      //for single potentials 
      SxVector<Double> getFreq1 (const double &softAmp, 
                                 const double &avCouplAmp, 
                                 const double &T);

      /* get the total energy (meV/3M) for all types of potentials
       @param softAmp amplitude of the soft mode
       @param T temperature
       @return total energy 
       */
      inline double getTotE (const double &softAmp, 
                             const double &T)  {
         if (softAmp < 0)  {
            cout << "Error in SxCouplPhon::getTotE! " << endl 
                 << "soft mode amplitude is negative: " << softAmp << endl;
            SX_EXIT;
         }  
         return sqOmega (softQ, T) * sqr (softAmp) / 2. 
              + symH (-1, -1, T).sum () * quart (softAmp) / 6.;
      }

      /* get the vibr. freeEn (meV/mode)
       @param freq provided frequencies
       @param T temperature
       @return the free energy (meV/mode)
      */
      double getFreeEn (const SxVector<Double> &freq, 
                        const double &T);
      
      // write the output data to files
      void writeToFile (const SxArray<SxVector<Double> > &freq, 
                        const SxArray<SxVector<Double> > &freeEn,
                        const double &dT, 
                        const double &dX, 
                        const SxArray<double> &minX);
};

#endif /* _SX_COUPLPHON_H_ */
