// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCouplPhon1.h>
#include <iostream> 
#include <fstream>
#include <SxCLI.h>

#ifndef SX_STANDALONE

inline double SxCouplPhon1::sqOmega (const int iQ, 
                                     const double &T)  
{
   SX_CHECK (iQ >= 0);
   SX_CHECK (! (T < 0.));

   int X2 = int(qMesh.getSize ()) - 1;
   double qPar = (qMesh(iQ) ^ qMesh(X2)) 
                / qMesh(X2).normSqr (), 
          qOrth = (qMesh(iQ) - qPar * qMesh(X2)).norm () 
                 / qMesh(X2).norm ();
   double res = sqr (12.3 * qPar / 2.04454) 
              + sqr (12.3 * qOrth / 2.04454) * 3.//2-4 times larger!?
              - 2. * fq (qPar, T) * exp (- sqr (qOrth * 6.) / 2.);
   return res; //\omega^2
}//CHECK

inline double SxCouplPhon1::hq (const int iQ, 
                                const double &T)  
{
   int X2 = int(qMesh.getSize ()) - 1;
   double qPar = (qMesh(iQ) ^ qMesh(X2)) 
               / qMesh(X2).normSqr (), 
          qOrth = ( qMesh(iQ) - qPar * qMesh(X2) ).norm () 
                / qMesh(X2).norm ();
   double res = fq (qPar, T / sqrt(2.)) / 20.1 
              * exp (- sqr (qOrth / 2.) / 2.);
   //self-coupling
 //double res = coup (qPar, T/sqrt(2.))/20.1 * exp (-sqr(qOrth/2.)/2.);//explicit broadening
                //sqrt(.08)* sin (PI*qPar) * sin (3*PI*qPar/2.);//real-space broadening
   return res * sqrt (6.); //! h_q with combinatorics
}//CHECK

SxCouplPhon1::SxCouplPhon1 (const SxArray<Coord> &qMesh_, //last is X_2!
                            const bool debug_)  
{
   qMesh = qMesh_;
   debug = debug_;
   int X2 = int(qMesh.getSize ()) - 1;
   SX_CHECK (X2 >= 0);
   softQ = 0;
   for (int iQ = 1; iQ <= X2; iQ++)  {
      softQ = (hq (iQ, 0) / qMesh(iQ).normSqr () 
            > hq (softQ, 0) / qMesh(softQ).normSqr ()
            ? iQ : softQ); 
   }
   cout << "max coupling @q= " << qMesh(softQ) << " X2 " << qMesh(X2) << endl; 
}//CHECK

SxCouplPhon1::~SxCouplPhon1 ()  { 
   //empty
}//CHECK

double SxCouplPhon1::getAvCoupAmpl (const double &minX, 
                                    const double &T, 
                                    const bool quant)  
{ 
   SX_CHECK (minX >= 0.);//double of softAmp!
   
   double kT = .08617343 * T, 
          softAmpl = hq (softQ, T) * sqr (minX),
          ampl = /*30;*/(quant ? .67 : 1.) * sqOmega (softQ, T) 
               / hq (softQ, T) + 2. * softAmpl;//min 0! max max fq 
   if (T < 10.)  {
      return sqrt (kT);//calculation too involved
   }
   if (debug)  {
      cout << ampl << "amp/softAmp" << softAmpl << endl
           << "initAmpl" << - sqOmega (softQ, T) 
           << -2. * hq (softQ, T) * softAmpl 
           << "sqrt(kT)" << sqrt (kT) << endl;
   }
   // initialize large enough ampl: max ampl + sqrt (kT)
   ampl =(ampl < 0 ? -2 * ampl + sqrt (kT) : sqrt (kT)); 
   
   int iQ = 0, nQ = int(qMesh.getSize ());
   double //delta = ampl/10., 
          minFreq = 0., 
          freq, 
          res = 0., 
          resO = 0.;//init smaller
   /*while (delta > 5.e-4)  {//alternative minimizer: kink in F?!*/
   while (res < ampl)  {
      res = 0.;
      minFreq = 10.;
      for (iQ = 0; iQ < nQ; iQ++)  {
         if (quant)  { 
            // sum amplitudes for all q-points, quantum case; quadratic weights?
            freq = sqOmega (iQ, T) 
                 + 2. * hq (iQ,T) / 3. * (2. * softAmpl + ampl);
            freq /= sqrt (abs (sqOmega (iQ,T)));
         }  else  { 
            freq = sqOmega (iQ, T) + hq (iQ, T) * (2. * softAmpl + ampl); 
         }//classical!
         if (freq < minFreq)  {
            minFreq = freq;
         }
         if (freq < 1.e-12)  {//Does not happen anymore?!?
         /* ampl += 1.1*delta;
            delta /= 2.;*/
            ampl *= exp (1.01e-2 * log (ampl / resO / .98));
            res = 0.;//loop shouldn't end now
            cout << iQ << endl;
            break;
         }
         if (quant)  {
            res += 2. * hq (iQ, T) / sqrt (abs (2. * sqOmega (iQ, T))) 
                 / (1. - exp (- freq / kT));
         }  else  {
            res += hq (iQ, T) / freq;
         }//freq^2! classical!
      }
      if (iQ == nQ)  {//no break
         res *= kT / nQ;
         ampl *= exp (1.e-2 * log (.98 * res / ampl));//improved scaling
         resO = res;
      }
    /*if (1.2*res < ampl)  {ampl -= delta;}//too large still
      else  {ampl += 1.1*delta; delta /= 2.;}//too small*/
   }
   if (debug)  { 
      cout << ampl << "ratio" << res/ampl << "minFreq" 
           << minFreq << "Q" << iQ << endl; 
   }  
   //return scaled ampl, factor should be as good as 1!
   return ampl * exp (1.e-2 * log (ampl / res / .98));
}//COPY

SxVector<Double> SxCouplPhon1::getFreq (const double &softAmp, 
                                        const double &avCouplAmp, 
                                        const double &T, 
                                        const bool quant)  
{
   if (softAmp < 0.)  {
      cout << "Error in SxCouplPhon::getFreq! negative softAmp: " 
           << softAmp << endl;
      SX_EXIT;
   }
   if (avCouplAmp < 0.)  {
      cout << "Error in SxCouplPhon::getFreq! negative avCoupAmp: " 
           << avCouplAmp << endl;
      SX_EXIT;
   }

   int iQ, 
       X2 = int(qMesh.getSize ()) - 1;
   SxVector<Double> res (X2 + 1);
   for (iQ = 0; iQ <= X2; iQ++)  {
      if (quant)  {
         res(iQ) = sqOmega (iQ, T) + 2. * hq (iQ, T) / 3. 
                      * ( 2. * hq (softQ, T) * sqr (softAmp) + avCouplAmp );
       /*+hq(iQ,T)*hq(softQ,T)*(<(hq/hs-2.)/sqrt(sqOmega)>;//small correction*/
         res(iQ) /= sqrt (abs (sqOmega (iQ, T)));//correct scaling
      }  else  {
         if ((qMesh(iQ) - qMesh(softQ)).norm () < .75 * qMesh(X2).norm ())  {
            //only correct freq for qPoints close to softQ 
            res(iQ) = sgnSqrt ( sqOmega(iQ, T) + hq (iQ, T) 
                    * (2. * hq (softQ, T) * sqr (softAmp) + avCouplAmp) );
         }  else  {
            res(iQ) = sgnSqrt (sqOmega (iQ, T));
         } 
      }
      if (res(iQ) < 1.e-12)  {
         cout << "smallFreq!" << res(iQ) << endl;
      }
   }
   return res * 2.04454; // meV!
}//COPY

double SxCouplPhon1::getTotE (const double &softAmp, 
                              const double &T)  
{
   if (softAmp < 0)  {
      cout << "Error in SxCouplPhon::getTotE! " << endl 
           << "soft mode amplitude is negative: " << softAmp << endl;
      SX_EXIT;
   }  
   return sqOmega (softQ, T) * sqr (softAmp) / 2. 
        + sqr (hq (softQ, T) * sqr (softAmp)) / 6.; 
}//CHECK, dependent of coupling parametrisation

//for 1 T from given freq 
double SxCouplPhon1::getFreeEn (const SxVector<Double> &freq, 
                                const double &T, 
                                const bool quant)  
{   
   int nFreq = int(freq.getSize ());
   double kT = .08617343 * T, 
          qWeight = 1. / nFreq, 
          res = -kT * log (PI * kT);//shift zero
   if (quant)  {//quantum free energy
      // --zero point vibration; quadratic weights?
      res = freq.sum () * qWeight / 2.;
      cout << "0pt" << res;
      
      // --temperature dependent term; quadratic weights?
      if (T > 0)  {
         SxVector<Double> boltzmannT = exp (- freq / kT); 
         res += kT * logVec (1. - boltzmannT).sum () * qWeight;//logVec checks on negative freq.
         cout << "T dep.F " <<  kT * logVec (1. - boltzmannT).sum () * qWeight;
      }
   }  else  {//classical; 
      for (int iQ = 0; iQ < nFreq; iQ++)  {
         if (freq(iQ) < 1.e-8)  {
               cout << "Error in SxCouplPhon1::getFreeEn, too small freq: " << freq(iQ) 
                    << " for iQ: " << iQ << endl;
               SX_EXIT;
         }
         res +=  kT * log (freq(iQ)) * qWeight; 
      }
   }
   return res; 
}//CHECK
   

void SxCouplPhon1::writeToFile (const SxArray<SxVector<Double> > &freq, 
                                const SxArray<SxVector<Double> > &freeEn, 
                                const double &dT, 
                                const double &dX,
                                const SxArray<double> &minX, 
                                const bool quant)  
{   
   // --total energy output to file
   ofstream totEout;
   totEout.open ("couplPhon1E");
   cout << "couplPhon1E opened" << endl;
   totEout << "#q-point amplitude energy" << endl;
   int iQ, iX;
   for (iQ = 0; iQ < qMesh.getSize (); iQ++)  {
      for (iX = 0; iX < minX(0) / dX + 2; iX++)  {
         totEout << iQ << " " << iX * dX << " " 
                 << sqOmega (iQ, 0.) * sqr (iX * dX) / 2. 
                  + sqr ( fq (iQ, 0.) / 20.1 * sqr (iX * dX) ) << endl;
      }
   } 
   totEout.close ();
   cout << "couplPhon1E closed" << endl;

   // --free energy output to file 
   ofstream freeEnOut;
   freeEnOut.open ("couplPhon1F");
   cout << "couplPhon1F opened" << endl;
   freeEnOut << "#minX(T) in steps of " << dT << " K" << endl << "#"; 
   int iT;
   for (iT = 0; iT < minX.getSize (); iT++)  {
      freeEnOut << minX(iT) << " ";
   }
   freeEnOut << "#softAmp (scaled) freeEn (meV) vs. T in steps of " << dT 
             << "K. quant?:" << quant << endl;
   for (iX = 0; iX < freeEn(0).getSize (); iX++)  {
      freeEnOut << endl << iX * dX;
      for (iT = 0; iT < freeEn.getSize (); iT++)  {
         freeEnOut << " " << freeEn(iT)(iX) - freeEn(iT)(0); 
      }
   }
   freeEnOut.close ();
   cout << "couplPhon1F closed" << endl;

   // --frequency output to file
   ofstream freqOut;
   freqOut.open ("couplPhon1Freq");
   cout << "couplPhon1Freq opened" << endl;
   freqOut << "#q-point freq(T) up to " << double(freq.getSize ()) * dT << endl;
   for (iQ = 0; iQ < freq(0).getSize (); iQ++)  {
      freqOut << iQ;
      for (iT = 0; iT < freq.getSize (); iT++)  {
         freqOut << " " << freq(iT)(iQ); 
      }
      freqOut << endl; 
   }
   freqOut.close ();
   cout << "couplPhon1Freq closed" << endl;
}//CHECK

#else /* SX_STANDALONE */
  
int main (int argc, 
          char **argv)  
{
   // --command line interface
   SxCLI cli (argc, argv);
   bool /*quant = cli.option ("-q|--quantum", "bool", 
                            "toggle quantum treatment").toBool (),*/
        debug = cli.option ("-db|--debugg", "bool", 
                            "toggle debug mode").toBool ();
   if (debug)  {
      cout << "debug mode" << endl;
   }
   int iX, iT, iQ, 
       nQ = cli.option ("-nQ", "number", "#q-points").toInt (16); 
   double dX = cli.option ("-dX", "decimal", "amplitude step").toDouble (1.),
          maxT = cli.option ("-maxT", "decimal", 
                             "max. temperature (K)") .toDouble (500.),
          dT = cli.option ("-dT", "decimal", 
                           "temperature step (K)").toDouble (50.),
          maxX = cli.option ("-maxX", "decimal", 
                             "max. soft amplitude").toDouble (10.);
   cli.finalize (); //defaults are reasonably conservative

   //--- initialize 3D mesh
   int nQpar = nQ/*int (cbrt (3. * nQ))*/, 
       nQorth = int (sqrt (nQ / nQpar));
   // nQ reset to nice product
   nQ = (nQ / (nQpar * nQorth)) * nQpar * nQorth;
   SxArray<Coord> qMesh (nQ);
   double dQ = 1. / nQ, 
          dQpar = 1. / nQpar, 
          dQorth = 1. / nQorth;
   for (iQ = 1; iQ < nQ; iQ++)  {
      qMesh (iQ - 1) = Coord ((iQ % nQpar) * dQpar, 
                             ((iQ / nQpar) % nQorth) * dQorth,
                             (iQ / (nQpar * nQorth)) * dQ * nQpar * nQorth);
   }//first nQpar are [110]!
   qMesh(nQ - 1) = Coord (1., 0., 0.); //last==X2

   int nX = int(maxX / dX), 
       nT = int(maxT / dT);
   cout << "nX" << nX << "nT" << nT << "nQ" << nQ << "nQpar" << nQpar << endl;
   double avCAmpl,
          temp;
   SxArray<double> minX (nT);
   SxVector<Double> freqTX;
   SxArray<SxVector<Double> > freq (nT), 
                              freeEn (nT);
   SxCouplPhon1 couPhon (qMesh, debug);
   for (iT = 0; iT < nT; iT++)  { 
      cout << "iT" << iT << endl;
      temp = (iT + .5) * dT;
      freeEn(iT).resize (nX);
      for (iX = 0; iX < nX; iX++)  {
         cout << "iX" << iX << endl;
         avCAmpl = couPhon.getAvCoupAmpl (iX * dX, temp);
         
         // get the (corrected) frequencies (meV) 
         freqTX = couPhon.getFreq (iX * dX, avCAmpl, temp);

         // get the total energy (meV/3M cell)
         freeEn(iT)(iX) = couPhon.getTotE (iX * dX, temp); 
         cout << "totEnX" << freeEn(iT)(iX) << endl;
         
         // add the vibr. free energy (meV/mode)
         freeEn(iT)(iX) += 3 * couPhon.getFreeEn (freqTX, temp);
         cout << "freeEnX" << freeEn(iT)(iX) << endl;

         // --set freq and amplitude of minimum freeEn
         if (iX == 0 || freeEn(iT)(iX) < freeEn(iT)(iX - 1))  {
            // now avCAmpl counts double??
       //   freqTX = couPhon.getFreq (iX * dX, 2. * avCAmpl, temp);
            
            // only select freq in [110], =first nQpar
            freqTX.resize (nQpar, true);
            freq(iT).copy (freqTX);
            minX(iT) = iX * dX; 
         } 
      }
   }
   couPhon.writeToFile (freq, freeEn, dT, dX, minX);
   return 0; 
}  
#endif /* SX_STANDALONE */
