// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_COUPLPHON1_H_
#define _SX_COUPLPHON1_H_

#include <SxAtomicStructure.h>
#include <SxMatrix.h>
#include <SxVector.h>
#include <SxArray.h>
#include <SxMath.h> 
#include <SxExt.h>

/** \brief calculate the free energy and frequencies of a coupled phonon system.
 * The couplings are taken from the anharmonic single mode potentials. These
 * couplings together with the macroscopic displacement of the 'soft mode'
 * stabelises all frequencies and so the partition sum can be evaluated.
 * units: meV, \AA, u.

   \b SxCouplPhon1 = S/PHI/nX compute the thermodynamics of a coupled phonon system.

   \author Matth\'e Uijttewaal, uijttewaal@mpie.de 
 */
class SX_EXPORT_EXT SxCouplPhon1
{
   public:
      inline double sqr (const double &x)  {
         return x * x;
      }
      
      inline double cube (const double &x)  {
         return sqr (x) * x;
      }
      
      inline double sgnSqrt (const double &x)  {
         return sqrt (fabs (x)) * (x < 0. ? -1. : 1.);
      }

      inline SxVector<Double> logVec (const SxVector<Double> &vec)  //>SxMath
      {
         ssize_t nElt = vec.getSize ();
         SxVector<Double> res (nElt);
         for (ssize_t iElt = 0; iElt < nElt; iElt++)  {
            if (vec(iElt) < 1.e-8)  {
               cout << "Error in logVec, too small value: " << vec(iElt) 
                    << " for iElt: " << iElt << endl;
               SX_EXIT;
            }
            res(iElt) = log (vec(iElt));//does log also check for negative values?
         }
         return res; 
      }

      // empty constructor
      SxCouplPhon1 (const SxArray<Coord> &qMesh_, //last is X_2!
                    const bool debug_ = false); 
      
      // destructor
     ~SxCouplPhon1 ();

     protected:
      bool debug; //toggle debug mode
      int softQ; //index softQ in qMesh
      SxArray<Coord> qMesh;
  
     /*master formula: fit non-acoustic harmonic
       T_c=\infty, f(q)=-651*x**2q**2.6/(1+exp(24*(q-.41))), T^c=1776, a=20.1 
        => softQ = .371, relative to X_2 */
      inline double fq (const double &Q, 
                        const double &T = 0.)  
      {
         return 651. * sqr (sqr (Q)) * sqrt (Q) / (1. + exp (24. * (Q - .41)))
            //;
                    * exp (- sqr (T / 1776.)); 
      }

      //broader coupling function
      inline double coup (const double &Q, 
                          const double &T = 0.)  
      {
         return 10.77 * cube (Q / .41) / (1. + exp (12 * (Q - .41)))//;
                      * exp (- sqr (T / 1776.)); //T dep. not sqr for hq!
      }

/**\brief sqr(omega)&sqrt(6pot4): parameters of the potential (meV) vs. x in \AA sqrt(u),
  including electron excitations @ 1q
  master formula: sqr(12.3qx/2.0445)*exp(-sqr(T/T_c))/2
                +(sqr( sqr(qx)f(q)-1 )-1)/sqr(a)*exp(-sqr(T/T^c)); (q~[110])
*/
      double sqOmega (const int iQ, const double &T); 

      double hq (const int iQ, const double &T); 

   public:
      /**\brief average coupling amplitude of all phonons (sqrt(meV)) for a
                given soft coupling amplitude (sqrt(meV))
       */
      double getAvCoupAmpl (const double &minX, 
                            const double &T, 
                            const bool quant = false);

      // calculate correct frequencies vs. q-point @min softAmpl
      SxVector<Double> getFreq (const double &minX, 
                                const double &avCAmpl, 
                                const double &T, 
                                const bool quant = false);
      
      // calculate classical free energy from provided frequencies
      double getFreeEn (const SxVector<Double> &freq, 
                        const double &T, 
                        const bool quant = false);
      
      // get the total energy @minX
      double getTotE (const double &minX, 
                      const double &T);


      // write the output data to files
      void writeToFile (const SxArray<SxVector<Double> > &freq, 
                        const SxArray<SxVector<Double> > &freeEn, 
                        const double &dT, 
                        const double &dX, 
                        const SxArray<double> &minX, 
                        const bool quant = false);
};
#endif /* _SX_COUPLPHON1_H_ */
