// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCouplPhonSimp.h>
#include <SxConstants.h>
#include <iostream> 
#include <fstream>
#include <SxCLI.h>

#ifndef SX_STANDALONE

SxCouplPhonSimp::SxCouplPhonSimp (const double &bz,
                                  const double &corr_,
                                  const int &nQ_, 
                                  const bool debug_, 
                                  const bool printMu)  
{ 
   if (nQ_ == 0)  {
      cout << "Error initializing SxCouplPhonSimp, no qPoints provided!" << endl;
      SX_EXIT;
   }
   nQ = nQ_;
   debug = debug_;
   
   //set scaling units once (2.044, .0862)
   ASQUMEV = 1.e+11 * HPLANCK * sqrt (HA2EV * JL2HA * 10. / U2KG) / TWO_PI;
   K2MEV = KB * JL2HA * HA2EV * 1000.;
   scalQ = 4.08;//4.10 old
   scalM = 3.8;
   scalBZ = bz;
   corr = corr_;
   cout << "units set " << ASQUMEV << K2MEV << endl;

   //set preliminary softQ (decreasing T)
   softQ = nQ / 2;

   if (printMu)  {
      cout << "print kuzmin curve" << endl;
      for (int iT = 1; iT < 500; iT++)  { 
         cout << iT << " " << kuzmin (iT) << endl;
      }
   }
}

double SxCouplPhonSimp::setSoftQ (const double &T)  {

   bool done = false;
   int add = 1;//take sqr (sqOmega) for deepest well!?
   //max of f / q^2!
   double oldRes = f (softQ, T) / sqr (softQ),
          res = f (softQ + 1, T) / sqr (softQ + 1.);
   cout << oldRes << "oldCouplRat, CouplRatio" << res << endl;
   if (res < oldRes)  { add = -1; }
   while ((! done) && (softQ > 0) && (softQ < nQ))  {
     softQ += add;
     res = f (softQ, T) / sqr (softQ);
     cout << "couplRat " << res << endl;
     if (res < oldRes)  { done = true; }
     oldRes = res;
   }
   softQ -= add;
   cout << "softQ " << (softQ + 1.) / nQ << endl;
   return (softQ + 1.) / nQ;
}

SxVector<Double> SxCouplPhonSimp::getStartAmp (const double &T)  {

   double kuz = kuzmin (T),
          scal = scalQ * (1. - kuz) + scalM * kuz,
          fac = K2MEV * T * scalBZ / 4. / nQ / sqr (scal) / 6.,//BZ, SS scaling
          term = sqr (sqOmega (softQ, T)) +  nQ * fac * sqr (f (softQ, T));//was nQ*24??
   SxVector<Double> res = (sqrt (term) - sqOmega (softQ, T)) / fac / f2 (softQ, T);
   cout << "startAmp " << res << endl;
   return res;
}

SxVector<Double> SxCouplPhonSimp::getFreq (const SxVector<Double> &avCouplAmp,
                                           const double &softAmp, 
                                           const double &T)  
{
   if (softAmp < 0.)  {
      cout << "Error in SxCouplPhonSimp::getFreq! negative softAmp: " 
           << softAmp << endl;
      SX_EXIT;
   }
   SX_CHECK (T > 0.);
   if (avCouplAmp(0) < 0. || avCouplAmp(1) < 0.)  {
      cout << "Error in SxCouplPhonSimp::getFreq! negative avCouplAmp: " 
           << avCouplAmp << endl;
      SX_EXIT;
   }  

   int iQ;
   double kuz = kuzmin (T),
          scal = scalQ * (1. - kuz) + scalM * kuz,       
          fact =  K2MEV * T * scalBZ / 4. / nQ / sqr (scal);
          //scaling with temp, BZ size and coupling scaling

   SxVector<Double> res (nQ),
                    softCouplAmp = f1 (softQ, T) * sqr (softAmp / scal); 
   for (iQ = 0; iQ < nQ; iQ++)  {
      if (abs (iQ - softQ) < .37 * nQ)  { //only close to softQ!  
          res(iQ) = sqOmega (iQ, T) + dot (f2 (iQ, T), softCouplAmp) *  6. //=12.  
                                    + dot (f2 (iQ, T), avCouplAmp) * 3. * fact;//=6. 
      }  else  { res(iQ) = sqOmega (iQ, T); }

      if (res(iQ) < 1.e-6)  {
         cout << "Negative frequency " << res(iQ) 
              << "@softAmp: " <<  softAmp 
              << ", T: " << T << ", qPoint: " << (iQ / (nQ - 1.)) 
              << " put to small (pos) value!" << endl;
         res(iQ) = 1.e-12;
      }
      res(iQ) = sqrt (res(iQ));
   }

   return res * ASQUMEV; //= 2.04454;//in meV! 
} 

SxVector<Double> SxCouplPhonSimp::getAmp (const SxVector<Double> &inAmp, 
                                          const double &softAmp,
                                          const double &T)  {
   SX_CHECK (T > 0.);
   SX_CHECK (softAmp > -1.e-4);
   
   //get the frequencies; they're small when inAmp too small!
   SxVector<Double> freq = getFreq (inAmp, softAmp, T) / ASQUMEV, //in potential units again
                    res (2);
   res.set (0.);
   for (int iQ = 0; iQ < nQ; iQ++)  {
     // sum classical amplitudes; no quadratic weights for increase BZ radius!
         res += f1 (iQ, T) / sqr (freq(iQ));
   }
   return res;
}

SxVector<Double> SxCouplPhonSimp::getAvCouplAmp (
      const SxVector<Double> &startAmp,//decreasing
      const double &softAmp, 
      const double &T)  { 
   if (softAmp < 0.)  {
      cout << "Error in SxCouplPhonSimp::getAvCouplAmp! negative softAmp: " 
           << softAmp << endl;
      SX_EXIT;
   }
   if (T < 0)  {
      cout << "Error in SxCouplPhonSimp::getAvCouplAmp! negative temperature: " 
           << T << endl;
      SX_EXIT;
   }
   if (T == 0)  {
      return SxList<double> () << 0. 
                               << 0.;
   }
   SxVector<Double> outAmp (2), inAmp (2);
   
   // --initialize inAmp to reasonably large value & outAmp smaller 
   inAmp <<= startAmp; //no ref!
   outAmp = .5 * inAmp;

   // ---scale amp's until they're equal
   while ( inAmp(0) > outAmp(0) || inAmp(1) > outAmp(1) )  {

      // get new amplitudes
      outAmp = getAmp (inAmp, softAmp, T);
     
      //scale inAmp: lowering one increases the (output of the) other as well!
      // they nicely balance?!
      inAmp *= exp (1.e-2 * logVec (.98 * outAmp / inAmp));
   }
   inAmp *= exp (1.e-2 * logVec (inAmp / outAmp / .98));//one step back
   outAmp = getAmp (inAmp, softAmp, T);
   if (debug)  
   { cout << inAmp << " avCouplAmp; ratio(~1?) " << (inAmp / outAmp) << endl; }

   return inAmp;
}

double SxCouplPhonSimp::getFreeEn (const SxVector<Double> &freq, 
                               const double &T)  
{ 
   int nFreq = int(freq.getSize ());
   double qWeight = scalBZ / 4. / nFreq; //scaling to proper weight (/mode)

   if (nFreq == 0)  {
      cout << "Error in SxCouplPhonSimp::getFreeEn: no freq provided!" << endl;
      SX_EXIT;
   }
   if (T < 1.)  {
      cout << "Error in SxCouplPhonSimp::getFreeEn: no freq provided!" << endl;
      SX_EXIT;
   }
   return K2MEV * T * logVec (freq / sqrt (PI * K2MEV * T)).sum () * qWeight;
}

void SxCouplPhonSimp::writeToFile (const SxArray<SxVector<Double> > &freq, 
                               const SxArray<SxVector<Double> > &freeEn, 
                               const double &dT, 
                               const double &dX,
                               const SxArray<double> &minX,
                               const SxArray<double> &sQ)  
{
    // --total energy output to file
   ofstream totEout;
   totEout.open ("couplPhonE");
   cout << "couplPhonE opened" << endl;
   totEout << "#q-point amplitude energy (meV/atom)" << endl;
   int iQ, iX, iT;
   for (iQ = 0; iQ < nQ; iQ++)  {
      for (iX = 0; iX < 2. * minX(0) / dX; iX++)  {
         totEout << (iQ+1.)/nQ << " " << iX * dX << " " 
                 << (sqOmega (iQ, 0.) * sqr (iX * dX) / 2. 
                  + sqr ( f (iQ, 0.) / scalQ) * quart (iX * dX) ) << endl;
      }// getTotE no q-dep!
   } 
   totEout.close ();
   cout << "couplPhonE closed" << endl;

   // --magnetic E output to file
   ofstream magnEnOut;
   magnEnOut.open ("couplPhonM");
   cout << "couplPhonM opened" << endl;
   magnEnOut << "#softAmp (scaled) magnEn (meV/atom) vs. T in steps of"
             << dT << "K";
   for (iX = 0; iX < freeEn(0).getSize (); iX++)  {
      magnEnOut << endl << iX * dX;
      for (iT = 0; iT < freeEn.getSize (); iT++)  {
         magnEnOut << " " << getTotE (iX * dX, iT * dT);//correct?
      }
   }
   magnEnOut.close ();
   cout << "couplPhonM closed" << endl;

   // --free energy output to file
   ofstream freeEnOut;
   freeEnOut.open ("couplPhonF");
   cout << "couplPhonF opened" << endl;
   freeEnOut << "#softQ(T) in steps of " << dT << " K " << endl << "#";
   for (iT = 0; iT < sQ.getSize (); iT++)  {
      freeEnOut << sQ(iT) << " ";
   }
   freeEnOut << endl << "#minX(T) in steps of " << dT << " K" << endl << "#";
   for (iT = 0; iT < minX.getSize (); iT++)  {
      freeEnOut << minX(iT) << " ";
   }
   freeEnOut << endl << "#softAmp (scaled) freeEn (meV) vs. T in steps of"
             << dT << "K";
   for (iX = 0; iX < freeEn(0).getSize (); iX++)  {
      freeEnOut << endl << iX * dX;
      for (iT = 0; iT < freeEn.getSize (); iT++)  {
         freeEnOut << " " << (freeEn(iT)(iX) - freeEn(iT)(0));
      }
   }
   freeEnOut.close ();
   cout << "couplPhonF closed" << endl;

   // --frequency output to file
   ofstream freqOut;
   freqOut.open ("couplPhonFreq");
   cout << "couplPhonFreq opened" << endl;
   freqOut << "#q-point freq(T) up to " << double(freq.getSize ()) * dT << endl;
   for (iQ = 0; iQ < freq(0).getSize (); iQ++)  {
      freqOut << (iQ+1.)/nQ;
      for (iT = 0; iT < freq.getSize (); iT++)  {
         freqOut << " " << freq(iT)(iQ);
      }
      freqOut << endl;
   }
   freqOut.close ();
   cout << "couplPhonFreq closed" << endl;
}

#else /* SX_STANDALONE */

int main (int argc,
          char **argv)  
{
   // --command line interface
   SxCLI cli (argc, argv);
   bool debug = cli.option ("-db|--debugg", "bool", 
                            "toggle debug mode").toBool (),
        printMu = cli.option ("-mu|--printMu", "bool",
                              "print magnetization curve").toBool (),
        uncoupl = cli.option ("-uncoupl", "bool",
                              "untoggle couplings").toBool () ;
   cout << "1D qGrid" << endl;
   if (debug)  { cout << "debug mode" << endl; }
   if (uncoupl)  { cout << "no couplings" << endl; }
   int iX, nQ = cli.option ("-nQ", "number", "#q-points").toInt (24);
   double bz = cli.option ("-bz", "decimal", 
                           "BZ volume").toDouble (.1),
          corr = cli.option ("-corr", "decimal", 
                           "correlation strength").toDouble (.4),
          dX = cli.option ("-dX", "decimal", 
                           "soft amplitude step").toDouble (.3),
          maxT = cli.option("-maxT","decimal", 
                            "max. temperature (K)").toDouble (375.),
          dT = cli.option ("-dT", "decimal", 
                           "temperature step (K)").toDouble (75.),
          maxX = cli.option ("-maxX", "decimal", 
                             "max. soft amplitude").toDouble (1.5);
   cli.finalize (); //defaults are reasonably conservative
   cout << "magnetism included" << endl;

   //-- initialize class and some parameters
   int nX = int(maxX / dX) + 1, //include both 0 and maxX
       nT = int(maxT / dT);
   cout << "nX" << nX << "nT" << nT << "bz" << bz << "corr" << corr << endl;
   double temp;
   SxArray<double> minX (nT),
                   sQ (nT);
   SxVector<Double> oldFreq,
                    freqTX, 
                    startAmp, //to increase speed
                    avCouplAmp;
   SxArray<SxVector<Double> > freq (nT), 
                              freeEn (nT);
   SxCouplPhonSimp couplPhon (bz, corr, nQ, debug, printMu); 
   
   // ---loop over temperatures
   for (int iT = nT - 1; iT >= 0; iT--)  {//decreasing avCouplAmp
      cout << "iT" << iT << endl;
      temp = (iT + .5) * dT;
      freeEn(iT).resize (nX);
      sQ(iT) = couplPhon.setSoftQ (temp);//reset
      startAmp = couplPhon.getStartAmp (temp);//redundant if uncoupl

      // ---loop over amplitudes
      for (iX = 0; iX < nX; iX++)  {//decreasing avCouplAmp
         cout << "iX" << iX << endl;

         // get average coupling amplitude of all phonons 
         avCouplAmp = (uncoupl ? SxVector<Double> (SxList<double> () << 0. << 0.) 
                               : couplPhon.getAvCouplAmp (startAmp, iX * dX, temp));
         startAmp = avCouplAmp; //previous amplitude
         
         // get the frequencies (meV)
         if (iX > 0)  { oldFreq = freqTX; }
         freqTX = couplPhon.getFreq (avCouplAmp, iX * dX, temp); 
         
         // get the total energy (meV/atom) 
         freeEn(iT)(iX) = couplPhon.getTotE (iX * dX, temp);
         cout << "totEn" << freeEn(iT)(iX) << endl;

         // add the vibr. freeEn (meV/mode) scaled
         freeEn(iT)(iX) += couplPhon.getFreeEn (freqTX, temp);
         cout << "freeEn" << freeEn(iT)(iX) << endl;

         // --set freq and amplitude of minimum freeEn
         if (iX == 0 || freeEn(iT)(iX) < freeEn(iT)(iX - 1))  {
            freq(iT) = freqTX; 
            minX(iT) = iX * dX; 
         }
      }
   }
   couplPhon.writeToFile (freq, freeEn, dT, dX, minX, sQ);
   return 0;
}  
#endif /* SX_STANDALONE */
