// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_COUPLPHONSIMP_H_
#define _SX_COUPLPHONSIMP_H_

#include <SxAtomicStructure.h>
#include <SxMatrix.h>
#include <SxVector.h>
#include <SxArray.h>
#include <SxMath.h> 
#include <SxExt.h>

/** \brief calculate the free energy and frequencies of a coupled phonon system.
 * The self-couplings are taken from the anharmonic single mode potentials. The
 * couplings with the soft mode come from the frequency shift. Those couplings
 * are combined to ones of type fq*gq'. The couplings together with the macroscopic
 * displacement of the 'soft mode' stabelizes all frequencies and so the partition sum
 * can be evaluated.  :useless description, it doesn't explain anything!
 1D version of a coupled phonon system with one correlated potential (including mu)
 needed:freq(func), coupl(multiFunc), nQ, dT, maxT, dX, maxX
 output freq(T,q), freeEn(T,X), minX(T)
 done:getAvCoupAmpl (softAmpl, iT); getXmin(iT); getFreq(X, iT);
 getFreeEn(freq, iT); writeToFile(freq, freeEn, minX).
UNITS: meV/atom!, \AA, u/atom!
NO QUANTUM SUPPORT!

   \b SxCouplPhonSimp = S/PHI/nX compute the thermodynamics of a simple coupled phonon system.

   \author Matth\'e Uijttewaal, uijttewaal@mpie.de 
 */
class SX_EXPORT_EXT SxCouplPhonSimp
{
   public:
      /* x to power a
         */
      inline double power (const double &x, const double &a)  {
         if (fabs (x) > 1.e-6)  {
            return exp (a * log (fabs (x)));
         }
         return 0.;
      }

      inline double sqr (const double &x)  {//SxMath
         return x * x;
      }
      
      inline double cube (const double &x)  {//SxMath
         return x * x * x;
      }

      inline double quart (const double &x)  {//SxMath
         return sqr (sqr (x)); //or faster: x * x * x * x?
      }
      
      inline double sgnSqrt (const double &x)  {//SxMath
         return sqrt (fabs (x)) * (x < 0. ? -1. : 1.);
      }

      inline SxVector<Double> logVec (const SxVector<Double> &vec) {//SxMath
         int iElt, nElt = int(vec.getSize ());
         double min = vec.minval (&iElt);
         if (min < 1.e-16)  {
            cout << "Error in logVec, too small value: " << min 
                 << " for iElt " << iElt << endl;
            SX_EXIT;
         }
         SxVector<Double> res (nElt);
         for (iElt = 0; iElt < nElt; iElt++)  {
                      res(iElt) = log (vec(iElt));
         }
         return res; 
      }
      
      inline double dot (const SxVector<Double> &vec, 
                         const SxVector<Double> &vec1)  {//>SxVec.h
         SxVector<Double> res (vec);
         res *= vec1;
         return res.sum ();
      }
   protected:
      bool debug; //toggle debug mode
      int softQ; //index soft q-point
      double scalQ, //scaling factor quartic coupling
             scalM, //scaling factor reduced magnetization
             scalBZ, //scaling factor for BZ volume
             corr, //correlation strength
             ASQUMEV,//angstroem sqrt (u meV) 
             K2MEV; //kelvin->meV
      int nQ; //1D mesh with q-points; last == X2
  
      /*kuz'min function for T dependent mu, achieves maximum of 1 @Tc !
        */
      inline double kuzmin (const double &T)
      {
         SX_CHECK (T > -1.e-6);
         double Tc = 376.,
                max = 2.;
         if (T > Tc) { return max; }//faster
         double T1 = power (T / Tc, 1.5),
                T2 = power (T / Tc, 2.8),
                res = max * (1. - power ((1. - .1 * T1 - .9 * T2), .333));
         return res;
      }

      /* fit to non-acoustic harmonic / anharmonic: FIXME description
         T_c=\infty, f(q)=-651x^2q^4.6/(1+exp(24*(q-.41))), 
         T^c=1300, a=4.08 => softQ = .376 X_2; sqr(fq/a) is 4th order term 
         @param Q q-point in qMesh
         @param T electronic temperature, quartic term: T/sqrt(2.)!
         @return (f_qf'_q',f_q' f'_q)/2.;
         */
      inline double f (const int &iQ, const double &T = 0.) 
      {  return sqrt (f2 (iQ, T).product ()); }

      /*asymmetric!*/
      inline SxVector<Double> f1 (const int &iQ, const double &T)  
      {  return f2 (iQ, T).product () / f2 (iQ, T); }

      SxVector<Double> f2 (const int &iQ, const double &T = 0.)  {

         SX_CHECK (iQ >= 0);
         SX_CHECK (T > -1.e-6);

         double kuz = kuzmin (T), //store for speed, interp. param, not function!
                a = 1. / 280. * (1. - kuz) + 1. / 62. * kuz, //651 old
                b = 3.9 * (1. - kuz) + 3.3 * kuz, //4.6 old
                b1 = (1. - corr) * b, //asymmetry from correlation, sum 2!
                b2 = (1. + corr) * b,
                c = 1. / 27. * (1. - kuz) + 1. / 11. * kuz,  //24 old
                d = .43 * (1. - kuz) + .55 * kuz,  //.41 old
                e =  T / 1300. * (1. - kuz),
                q1 = (iQ + 1.) / nQ;
         return  SxVector<Double> (SxList<double> () 
                 << fit (a, q1, b1, c, d, e) << fit (a, q1, b2, c, d, e));
      }

      inline double fit (const double &a, const double &q, const double &b,
                         const double &c, const double &d, const double &e)  
      { return power (q, b) / a / (1. + exp ((q - d) / c)) * exp (- sqr (e) / 2.); }
      
      /* harmonic potential (meV) vs. X(\AA sqrt(u)) 
         master formula: sqr(12.3qx/2.0445)*exp(-sqr(T/T_c))/2
                +(sqr( sqr(qx)f(q)-1 )-1)/sqr(a)*exp(-sqr(T/T^c)); (q~[110]) 
         @param iQ index q-point in mesh
         @param T (electronic) temperature
         @return omega_q^2*/
      double sqOmega (const int iQ, const double &T = 0.)  { 
         SX_CHECK (iQ >= 0);
         SX_CHECK (! (T < 0.));//internal function anyhow

         return sqr (12.3 * (iQ + 1.) / nQ / ASQUMEV) - 2. * f (iQ, T);
      }  //exclude Gamma!      should be sqrt2T for el. part, but not magn!

   public:
      /* \brief constructor: set debug & softQ 
          softQ is min of \omega_q^2 g_sh_s/(g_qh_q)/\omega_s^2
       */
      SxCouplPhonSimp (const double &bz,
                       const double &corr_,
                       const int &nQ_, 
                       const bool debug_ = false,
                       const bool printMu = false); 
       
      // empty destructor
     ~SxCouplPhonSimp () { /* empty */ }
   
      /* \brief set softQ as minimum of sqrOmega/coupling (varies with T)
         @param T temperature (K) 
         @return softQ (X_2)
       */ 
      double setSoftQ (const double &T);

      /* \brief set (too large) starting amplitude at zero softAmp for faster 
         convergence of getAvCouplAmp: take only the contribution from softQ 
         with some scaling factor (solve quadratic equation)
       @param T temperature
       @return startAmp
       */
      SxVector<Double> getStartAmp (const double &T);

      // get the freq vs. q-point @X softAmpl and T~K using statPhys.
     // for double potentials
      SxVector<Double> getFreq (const SxVector<Double> &avCouplAmp, 
                                const double &softAmp, 
                                const double &T);

   protected:
     /* get amplitude from input amplitude, uses getFreq
        INTERNAL only for minimizing of getAvCoupAmp 
      */
      SxVector<Double> getAmp (const SxVector<Double> &inAmp, 
                               const double &softAmp, 
                               const double &T);
   public:   
      /* \brief get average coupling amplitude of all phonons (sqrt(meV)) 
         
         formula: CLASSICAL, idem for g<->h: <g_s h\nu^2>_{\nu_s}=k_BT\int dq g_sh_q /freq^2
         @param X soft mode amplitude (times couplings, in sqrt(meV)?) 
         @param startAmp initial amplitude (decreasing)
         @param T temperature in K
         @return (g_s<h nu^2>,h_s<g nu^2>) or (g_s<h n/omega>,h_s<g n/omega>)*/
      SxVector<Double> getAvCouplAmp (const SxVector<Double> &startAmp,
                                      const double &softAmp, 
                                      const double &T);

      /* get the total energy (meV/3M) for all types of potentials
       @param softAmp amplitude of the soft mode
       @param T temperature
       @return total energy 
       */
      inline double getTotE (const double &softAmp, 
                             const double &T)  {
         if (softAmp < 0)  {
            cout << "Error in SxCouplPhonSimp::getTotE! " << endl 
                 << "soft mode amplitude is negative: " << softAmp << endl;
            SX_EXIT;
         }  
         double kuz = kuzmin (T),
                scal = scalQ * (1. - kuz) + scalM * kuz;
         return sqOmega (softQ, T) * sqr (softAmp) / 2. 
               + sqr (f (softQ, T) * sqr (softAmp) / scal);
      }

      /* get the vibr. freeEn (meV/mode) scaled to BZ volume!
       @param freq provided frequencies
       @param T temperature
       @return the free energy (meV/mode)
      */
      double getFreeEn (const SxVector<Double> &freq, 
                        const double &T);
      
      // write the output data to files
      void writeToFile (const SxArray<SxVector<Double> > &freq, 
                        const SxArray<SxVector<Double> > &freeEn,
                        const double &dT, 
                        const double &dX, 
                        const SxArray<double> &minX,
                        const SxArray<double> &sQ);
};

#endif /* _SX_COUPLPHONSIMP_H_ */
