// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

// 
//  SxHist2Corr by Lars Ismer, ismer@fhi-berlin.mpg.de
//

#include <SxTimeCorr.h>
#include <SxUtil.h>
#include <SxError.h>
#include <stdio.h>
#include <iostream>
#include <SxError.h>
#include <SxHist2Corr.h>

#ifndef SX_STANDALONE

// empty


SxHist2Corr::SxHist2Corr () 
{
   //empty
}

SxHist2Corr::~SxHist2Corr () 
{
   //empty
}

void SxHist2Corr::printHelp() {
   cout 
      << "This tools loads a moldynHist.dat file to analyse "
      "the trajectory and derive mascroscopic observables" << endl;
   
   cout << "Example: \n\n"
      << "sxhist2corr -i moldynHist.dat -o corr.dat -r 1 1000 -m 1\n";
   exit(0);
}

void SxHist2Corr::printUsage () 
{
   cout << "usage: sxhist2corr.x -i [hist file] -o [output file]"
         <<" -m [modus]"<< endl;
   cout << "optional: -r [start] [end]" << endl;
   cout << endl << "-m 1     ->     "
      << "velocity-velocity autocorrelation function "
      << endl << "-m 2     ->     "
      << "generalized frequency spectrum             " 
      << endl << "-m 3     ->     "
      << "generalized hessian and frequencies        " 
      << endl;
      
   cout     << " r                 "
      << "->     observation range on time axis" << endl; 
   cout << "help: sxhist2corr.x -h" <<
      endl;	
}

int SxHist2Corr::saveMatrixSx (const SxMatrix<Double> &matrix, 
      const SxString &fileName)
{
   FILE *fp = NULL;
   ssize_t i, j;
   if ( !(fp = fopen (fileName.ascii (), "w")) ) {
      sxprintf ("Can't open file %s", fileName.ascii ()); 
      SX_QUIT; 
   }
   ssize_t nDoF = matrix.nRows ();
   fprintf (fp, "format matrix;\nmatrix = [\n");
   SxString line;
   for (i = 0; i < nDoF; i++) {
      line = SxString("[");
      for (j = 0; j < nDoF; j++) {
         line += SxString (matrix(i, j), SxString("%13.12e"));
         if (j < (nDoF - 1))
            line += SxString (", ");
      }
      line += SxString("]");
      if (i < (nDoF - 1))
         line += SxString(",");
      fprintf (fp, "%s\n", line.ascii ());
   }   
   fprintf (fp, "];\n");
   fclose(fp);
   return 1;
}   
   
#else /* SX_STANDALONE */


int main (int argc, char **argv)
{
   int i, outputMode = 0, j;
   SxBinIO io;
   SxString          commandLine;
   SxList<SxString>  commandList;
   SxList<SxString>  subCommandList;
   SxString          outputFile, inputFile, inModus, outModus;
   //bool hiddenCommand = false;
   int startObsIndex = 0, endObsIndex = 0;

   
   initSPHInXMath ();
   SxHist2Corr h2c;

   cout << SEPARATOR;
   cout << "| SxHist2Corr \n";
   cout << SEPARATOR;
   // --- Reading in command-Line 
   for (i = 1; i < argc; i++) {
      commandLine += SxString(argv[i]);
      commandLine += SxString(" ");
   }
 
   commandList = commandLine.tokenize ('-');

   if ((commandList.getSize () < 3) ){
      h2c.printUsage ();
      SX_QUIT;
   } 
   // --- Parsing command-Line	
   bool correctCommand;	
   subCommandList = commandList(0).tokenize (' ');
   if (commandList.getSize () == 1) {
      if (subCommandList(0) == SxString ("h")) {
         h2c.printHelp ();
      }
   }
   if (commandList.getSize () >= 3) {
      for (i = 0; i < commandList.getSize (); i++) {
         correctCommand = false;
         subCommandList = commandList(i).tokenize (' ');
         
         if (subCommandList(0) == SxString("i")) {
            correctCommand = true;
            inputFile = subCommandList(1);
         }
         if (subCommandList(0) == SxString("o")) {
            correctCommand = true;
            outputFile = subCommandList(1);
         }
         if (subCommandList(0) == SxString("m")) {
            correctCommand = true;
            outputMode = subCommandList(1).toInt ();
         }
         if (subCommandList(0) == SxString("r")) {
            correctCommand = true;
            startObsIndex = subCommandList(1).toInt ();
            endObsIndex   = subCommandList(2).toInt ();
         }
         if (correctCommand == false) {
            cout << "Unknown command: " << subCommandList(0) << endl;
            exit (0);
         }
      }
   } 	

   // --- loading in velocity-trajectory
   SxTimeCorr tcVel;
   cout << "Loading trajectory ...";
   tcVel.loadTrajectory (inputFile, SxString("velocity"));
   cout << "ready !" << endl;
   cout << "Start Time: " << tcVel.startTime << endl;
   cout << "End Time: " << tcVel.endTime << endl;
   cout << "dt " << tcVel.dt << endl;

   double startObs = (double)startObsIndex*tcVel.dt + tcVel.startTime;
   double endObs = (double)endObsIndex*tcVel.dt + tcVel.startTime;
   SxArray<double> G;
 
   cout << "Start Obs: " << startObs << endl;
   cout << "End Obs: " << endObs << endl;
  
   //--- Calculation of auto-correlation function
   if (outputMode == 1) {
      double time = 0;
      FILE *fp = fopen (outputFile.ascii (), "w");
      tcVel.setObservationRange (startObsIndex, endObsIndex);
      G = tcVel.getAveragedAutoCorrelationFunction ();
      
      for (i = 0; i < G.getSize (); i++) {
         fprintf (fp, "%10.5f %e\n", time, G(i));
         time += tcVel.dt;
      }
      fclose (fp);
   }
   
   //--- calculation of generalized frequency spectrum
   if (outputMode == 2) {
      double dfreq;
      double freq = 0.;
      FILE *fp = fopen (outputFile.ascii (), "w");
      
      tcVel.setObservationRange (startObsIndex, endObsIndex);
      G = tcVel.getGeneralizedFrequencySpectrum ();
      dfreq = tcVel. getDFreq  ();
      
      for (i = 0; i < G.getSize (); i++) {
         if (freq < 4000.) {
            fprintf (fp, "%10.5f %e\n", freq, G(i));
            freq += dfreq;
         }
      }
      fclose (fp);
   }

   //--- Calculation of effective dynamical matrix
   if (outputMode == 3) {
      tcVel.setObservationRange (startObsIndex, endObsIndex);
      SxTimeCorr tcForce;
      tcForce.loadTrajectory (inputFile, SxString("force"));
      tcForce.setObservationRange (startObsIndex, endObsIndex);
      SxMatrix<Double> hessian = tcVel.getHessian (tcVel, tcForce);
      int nDoF = int(hessian.nRows ());
     
      io.open (outputFile, SxBinIO::SX_WRITE_ONLY);
      io.write ("matrix", hessian, "dummy", "dummy");
      io.close();

      SxMatrix<Double>::Eigensystem eig;
      eig = hessian.eigensystem ();
      io.open (outputFile, SxBinIO::SX_APPEND_ONLY);
      io.write ("eigenvalues", eig.vals, "dummy", 0);
      io.close();

      for (i = 0; i < nDoF; i++) {
         for (j = 0; j < nDoF; j++) 
            hessian(i, j) = hessian(i, j)
                               /sqrt(tcVel.masses(i/3)*tcVel.masses(j/3));
      }
      eig = hessian.eigensystem ();
       
      SxVector<Complex16> freqs (nDoF);
      for (i = 0; i < nDoF; i++) 
         freqs(i) = sqrt(eig.vals(i).re) * 5123.75;

      io.open (outputFile, SxBinIO::SX_APPEND_ONLY);
      io.write ("frequencies", freqs, "dummy", 0);
      io.close();
       
      eig = (tcVel.getCovarianceMatrix ()).eigensystem ();
      cout << eig.vals << endl;
   }
   return 0;
}

  
#endif /* SX_STANDALONE */
   
