// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxMadelung.h>
#include <SxCLI.h>

#ifndef SX_STANDALONE

void SxMadelung::autoGauss (const SxCell &cellIn)
{
   SX_CHECK_VAR (cellIn.volume > 0., cellIn.volume);
   rGauss = pow(cellIn.volume, 1./3.) * sqrt(0.5);
}

double SxMadelung::screenedReal (double dist) const
{
   SX_CHECK_DIV (rGauss);
   return (dist > 1e-3)
          ? ( derfc(dist/rGauss) / dist )
          : (-2./sqrt(PI) / rGauss);            // dist == 0.
}

double SxMadelung::screenedReciprocal (const Coord &latVec,
                                       const Coord &delta) const
{
   double nrm2 = latVec.normSqr ();
   if (nrm2 > 1e-10)
      return 4. * PI / nrm2
             * exp(-0.25 * nrm2 * rGauss * rGauss)
             * cos(latVec ^ delta);
   else
      return -PI * rGauss * rGauss; // nrm2 == 0.
}

SxArray<SxVector3<Int> > SxMadelung::getShellVecs (int shell)
{
   SX_CHECK_VAR (shell > 0, shell);
   int d1, d2, d3;
   SxVector3<Int> i;

   int nVec = 8
            + 12 * (2*shell - 1)
            + 6 * (2 * shell - 1) * (2 * shell - 1);
   int iVec = 0;
   SxArray<SxVector3<Int> > latVecs(nVec);
   // corners
   for (i(0) = -1; i(0) <= 1; i(0)+=2)
      for (i(1) = -1; i(1) <=1; i(1)+=2)
         for (i(2) = -1; i(2) <=1; i(2)+=2)
            latVecs(iVec++) = shell * i;
   // --- permute dimensions
   for (d1 = 0; d1 < 3; d1++)  {
      d2 = (d1 + 1) % 3;
      d3 = (d1 + 2) % 3;
      // edges (along d1)
      for (i(d1) = 1-shell; i(d1) < shell; i(d1)++)
         for (i(d2) = -shell; i(d2) <= shell; i(d2)+=2*shell)
            for (i(d3) = -shell; i(d3) <=shell; i(d3)+=2*shell)
               latVecs(iVec++) = i;
      // sides (d1 x d2)
      for (i(d1) = 1-shell; i(d1) < shell; i(d1)++)
         for (i(d2) = 1-shell; i(d2) < shell; i(d2)++)
            for (i(d3) = -shell; i(d3) <=shell; i(d3)+=2*shell)
               latVecs(iVec++) = i;
   }
   SX_CHECK_VARS (iVec == nVec, iVec, nVec);

   return latVecs;
}


double SxMadelung::recSpaceContrib(const Coord &delta, int nShells) const
{
   SxVector3<Int> i;
   double res = 0.;
   Coord latVec;
   SxCell recCell = cell.getReciprocalCell ();
   for (i(0) = -nShells; i(0) <= nShells; i(0)++)  {
      for (i(1) = -nShells; i(1) <= nShells; i(1)++)  {
         for (i(2) = -nShells; i(2) <= nShells; i(2)++)  {
            latVec = recCell.relToCar(i);
            res += screenedReciprocal(latVec, delta);
         }
      }
   }
   return res / cell.volume;
}

double SxMadelung::dRecSpaceContrib(const Coord &delta, int shell) const
{

   SxArray<SxVector3<Int> > latVecs = getShellVecs (shell);
   int iVec, nVec = (int)latVecs.getSize ();
   SxCell recCell = cell.getReciprocalCell ();
   double res = 0.;
   for (iVec = 0; iVec < nVec; ++iVec)
      res += screenedReciprocal ( recCell.relToCar(latVecs(iVec)), delta );

   return res / cell.volume;
}

double SxMadelung::realSpaceContrib(const Coord &delta, int nShells) const
{
   SxVector3<Int> i;
   double res = 0.;
   Coord latVec;
   for (i(0) = -nShells; i(0) <= nShells; i(0)++)  {
      for (i(1) = -nShells; i(1) <= nShells; i(1)++)  {
         for (i(2) = -nShells; i(2) <= nShells; i(2)++)  {
            latVec = cell.relToCar(i);
            res += screenedReal ((delta + latVec).norm ());
         }
      }
   }
   return res;
}

double SxMadelung::dRealSpaceContrib(const Coord &delta, int shell) const
{

   SxArray<SxVector3<Int> > latVecs = getShellVecs (shell);
   int iVec, nVec = (int)latVecs.getSize ();
   double res = 0.;
   for (iVec = 0; iVec < nVec; ++iVec)
      res += screenedReal( ( delta + cell.relToCar(latVecs(iVec)) ).norm () );
   return res;
}

double SxMadelung::compute(const Coord &delta) const
{
   SX_CHECK_VAR(rGauss > 0., rGauss);
   int minShell = 3;
   double E1, E2, dE;
   // cout << "delta = " << delta << endl;

   // cout << "real space" << endl;
   E1 = realSpaceContrib(delta,minShell-1);
   if (fabs(E1) > 1e-14)  {
      for (int shell = minShell; shell < 100; ++shell) {
         E1 += dE = dRealSpaceContrib(delta,shell);
         // cout << "shell " << shell << ": " << dE << endl;
         if (fabs(dE) < 1e-8 * fabs(E1)) break;
      }
   }
   // cout << "dE = " << dE1 << endl;

   // cout << "reciprocal space" << endl;
   E2 = recSpaceContrib(delta,minShell-1);
   if (fabs(E2) > 1e-14)  {
      for (int shell = minShell; shell < 100; ++shell) {
         E2 += dE = dRecSpaceContrib(delta,shell);
         // cout << "shell " << shell << ": " << dE << endl;
         if (fabs(dE) < 1e-8 * fabs(E2)) break;
      }
   }
   // cout << "dE = " << dE2 << endl;

   return E1 + E2;

}

double SxMadelung::compute(const SxAtomicStructure &structure,
                           const SxVector<Double> &charges)
{
   int nSpecies = structure.getNSpecies ();
   SX_CHECK_VARS(nSpecies == charges.getSize (), nSpecies, charges.getSize ());

   // --- set up internal data
   setCell(structure.cell);
   if (rGauss < 0.) autoGauss ();

   // --- sum Madelung energy over atom pairs
   int is, js, ia, ja;
   double E = 0., fac;
   Coord delta;
   for (is = 0; is < nSpecies; is++)  {
      for (ia = 0; ia < structure.getNAtoms(is); ia++)  {
         for (js = 0; js < nSpecies; js++)  {
            fac = 0.5 * charges(is) * charges(js);
            for (ja = 0; ja < structure.getNAtoms(js); ja++)  {
               delta = structure.getAtom(is,ia) - structure.getAtom(js,ja);
               E += fac * compute(delta);
            }
         }
      }
   }
   return E;
}

double SxMadelung::computeQPol ()  const
{
   double E, dE;

   SxArray<SxVector3<Int> > latVecs;
   int iVec, nVec;

   E = 0.;
   for (int shell = 1; shell < 500; ++shell) {
      latVecs = getShellVecs (shell);
      nVec = (int)latVecs.getSize ();
      dE = 0.;
      for (iVec = 0; iVec < nVec; ++iVec)
         dE += pow(cell.relToCar(latVecs(iVec)).normSqr (), -2.5); // r^-5
      E += dE;
      // cout << "shell " << shell << ": " << dE << endl;
      if (fabs(dE) < 1e-6 * fabs(E)) break;
   }

   return E;

}

#else

int main (int argc, char **argv)
{
   // --- init S/PHI/nX Utilities
   initSPHInXMath ();

   // --- parse command line
   SxCLI cli (argc, argv);
   cli.authors = "C. Freysoldt";

   SxString inFile
      = cli.option ("-i|--input", "input file", "structure input file")
        .toString ("input.sx");

   double rGauss = cli.option ("-a|--alpha","number","Gauss radius in bohr")
                   .toDouble (-1.);

   SxList<double> chargeList
      = cli.option ("--charges", "q1:q2:q3...",
                    "set nominal charges for each species")
        .toDoubleList ();
   cli.last ().defaultValue = "default: automatic";

   cli.newGroup ("single charge");
   bool single = cli.option ("--single","single charge").toBool ();
   cli.last ().required ();

   bool borderPot;
   Coord borderPoint;
   cli.option ("--point", "vector", "compute potential at specified point");
   if ((borderPot = cli.last ().exists ()))  {
      borderPoint = Coord(cli.last ().toList3 ());
   }

   cli.finalize ();

   // --- read input
   SxAtomicStructure structure;
   SxParser parser;
   SxConstPtr<SxSymbolTable> tree;
   tree = parser.read (inFile, "std/structure.std");
   structure = SxAtomicStructure (&*tree);
   SxSpeciesData specData;
   if (!single)
      specData = SxSpeciesData(&*tree);

   structure.fprint (stdout);

   SxVector<Double> charges;

   if (!single)  {
      int nSpecies = specData.getNSpecies ();
      if (chargeList.getSize () == nSpecies)
         charges = SxVector<Double> (chargeList);
      else
         charges = specData.valenceCharge;
   }


   // --- compute
   SxMadelung madelung;
   if (rGauss > 0.)  {
      madelung.rGauss = rGauss;
   } else {
      madelung.setCell (structure.cell);
      madelung.autoGauss ();
      cout << "rGauss = " << madelung.rGauss << endl;
   }

   cout.precision (10);
   if (single)  {
      madelung.setCell (structure.cell);
      cout << "E = " << 0.5 * madelung.compute (Coord(0.,0.,0.))
           << "* q^2" << endl;
      //cout << "E(QPol) = " << 0.5 * madelung.computeQPol () << "* Q^2" << endl;
      if (borderPot)  {
         cout << "potential at " << borderPoint << " : "
              << madelung.compute(structure.cell.relToCar(borderPoint))
              << endl;
      }
   } else {
      double Eprim = 0.;
      for (int ia = 0; ia < structure.getNAtoms (); ++ia)  {
         for (int ja = ia+1; ja < structure.getNAtoms (); ++ja)  {
            Eprim +=  charges(structure.getISpecies(ia))
                    * charges(structure.getISpecies(ja))
                    / structure.cell.getMapped(structure(ia) - structure(ja),
                                               SxCell::WignerSeitz)
                      .norm ();
         }
      }
      cout << "Intracell energy (Wigner-Seitz):" << Eprim << endl;
      double Emad = madelung.compute (structure, charges);
      cout << "Madelung energy = " << Emad << endl;
      cout << "Intercell Madelung energy = " << (Emad-Eprim) << endl;
   }

}
#endif
