// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_MADELUNG_H_
#define _SX_MADELUNG_H_

#include <SxConfig.h>
#include <SxAtomicStructure.h>
#include <SxSpeciesData.h>
#include <SxExt.h>

/** This class computes the Madelung energy of a point charge array
    by an Ewald summation.
  */
class SX_EXPORT_EXT SxMadelung  {
   public:
      /** \brief Gauss radius for Ewald summation */
      double rGauss;

   protected:
      /** \brief Unit cell (set automatically by compute) */
      SxCell cell;

   public:
      /** Constructor */
      SxMadelung() : rGauss(-1.) { /* empty */ }

      /** \brief Guess good Gauss radius */
      void autoGauss(const SxCell &cellIn);

      /** \brief Guess good Gauss radius */
      void autoGauss() { autoGauss(cell); }

      /** \brief Compute Madelung potential contribution */
      double compute(const Coord &delta) const;

      /** \brief Compute Madelung energy */
      double compute(const SxAtomicStructure &structure,
                     const SxVector<Double> &charges);

      /** \brief Compute quadrupole-quadrupole interaction */
      double computeQPol () const;

      /** \{ \name Internal functions */

      /** \brief Screened interaction (real space) */
      inline double screenedReal (double dist) const;

      /** \brief Gaussian potential (reciprocal space) */
      double screenedReciprocal (const Coord &latVec, const Coord &delta) const;

      /** Set cell */
      void setCell (const SxCell &cellIn)
      {
         SX_CHECK_VAR(cellIn.volume > 0., cellIn.volume);
         cell = cellIn;
      }

   protected:

      /** \brief Compute realSpaceContribution */
      double realSpaceContrib(const Coord &delta, int nShells) const;
      /** \brief Compute reciprocal space contribution */
      double recSpaceContrib(const Coord &delta, int nShells) const;

      /** \brief Compute reciprocal space contribution for a shell */
      double dRecSpaceContrib(const Coord &delta, int shell) const;
      /** \brief Compute reciprocal space contribution for a shell */
      double dRealSpaceContrib(const Coord &delta, int shell) const;

      /** \brief Get vectors of belonging to a certain shell */
      static SxArray<SxVector3<Int> > getShellVecs (int shell);

      /// \}
};

#endif
