// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------
#include <SxUtil.h>
#include <SxError.h>
#include <stdio.h>
#include <iostream>
#include <SxParser.h>
#include <SxVector.h>
#include <SxMatrix.h>
#include <SxBinIO.h>
#include <SxCLI.h>
#include <SxMatrixConv.h>

#ifndef SX_STANDALONE
SxMatrixConv::SxMatrixConv () 
{
   //empty
}

SxMatrixConv::~SxMatrixConv () 
{
   //empty
}

int SxMatrixConv::loadMatrixSx  (const SxString &fileName) 
{
   SxParser parser;
   SxParser::Table table = parser.read (fileName);
   matrix = SxMatrix<Double> (table->get("matrix")->toList ());
   // --- TODO: the following columns will be replaced by getDims ()
   int size = (int)(sqrt) ((double)matrix.getSize ());
   matrix.reshape (size, size);
   matrix = matrix.transpose ();
   return 1;
}

int SxMatrixConv::loadMatrixOut (const SxString &fileName)
{
   SxBinIO io;
   io.open (fileName, SxBinIO::ASCII_READ_ONLY);
   matrix = SxMatrix<Double>(nDoF, nDoF);
   io.read ("dummy", &matrix, nDoF, nDoF, 0, 0);
   io.close();
   return 1;
}  

int SxMatrixConv::saveMatrixOut (const SxString &fileName)
{
   SxBinIO io;
   io.open (fileName, SxBinIO::ASCII_WRITE_ONLY);
   io.write ("dummy", matrix, "dummy", "dummy");
   io.close();
   return 1;
}

int SxMatrixConv::saveMatrixSx (const SxString &fileName)
{
   SxBinIO io;
   io.open (fileName, SxBinIO::SX_WRITE_ONLY);
   io.write ("matrix", matrix, "dummy", "dummy");
   io.close();
   return 1;
}   

  
#else /* SX_STANDALONE */


int main (int argc, char **argv)
{
   initSPHInXMath ();
   
   //--- command line parsing
   SxCLI cli (argc, argv);
   SxString inputFile  = cli.option ("-i|--input", "file",
                                     "input file (extensions: .sx, .out)"
                                    ).required().toString();
   SxString outputFile = cli.option ("-o|--output", "file",
                                     "output file (extensions: .sx, .out)"
                                    ).required().toString();
   int nDoF            = cli.option ("-n|-d", "int",
                                     "matrix size / degrees of freedom"
                                    ).required().toInt ();
   cli.finalize ();

   
   SxString inFormat, outFormat;
   SxMatrixConv mc;
   mc.nDoF = nDoF;
   
   SxList<SxString> sides = inputFile.tokenize ('.');
   SxString extension;
   
   //--- validity of input-file format is checked
   if (sides.getSize() == 1)  {
      cout << "Please provide a consumerable file extension for ";
      cout << "input file: .sx or .out" << endl;
      SX_QUIT;
   }	
   if (sides.getSize() >= 2) {
      extension = sides(sides.getSize() - 1);
      if ((extension == "sx") || (extension == "out")) {
         inFormat = extension;
      } else {
         cout << "Format " << extension << " not supported !";
         SX_QUIT;
      }
   }

   //--- validity of output-file format is checked
   sides = outputFile.tokenize ('.');
   if (sides.getSize() == 1)  {
      cout << "Please provide a consumerable file extension for ";
      cout << "input file: .sx or .out" << endl;
      SX_QUIT;
   }
   if (sides.getSize() >= 2) {
      extension = sides(sides.getSize() - 1);
      if ((extension == "sx") || (extension == "out") ) {
         outFormat = extension;
      } else {
         cout << "Format " << extension << " not supported !";
         SX_QUIT;
      }
      
   }	
   
   int implemented = 0;
   SxString response;
   
   response = "Converting " + inFormat;
   
   //--- matrix is parsed in
   if (inFormat == "sx")  implemented = mc.loadMatrixSx(inputFile);
   if (inFormat == "out") implemented = mc.loadMatrixOut(inputFile);
   
   response += " to ";
   response += outFormat;
   
   //--- matrix is written with given output format
   if (outFormat == "sx")  implemented *= mc.saveMatrixSx(outputFile);
   if (outFormat == "out") implemented *= mc.saveMatrixOut(outputFile);
   
   if (implemented == 0)   response += " is not implemented yet";
   
   cout << response << endl;
   return 0;
}
#endif /* SX_STANDALONE */
   
