// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------


#include <SxPAWHamiltonian.h>
#include <SxRadBasis.h>
#include <SxProjector.h>
#include <SxNeighbors.h>
#include <SxYlm.h>
#include <SxSphereGrid.h>
#include <SxGaussIJ.h>
#include <SxPAWOverlap.h>
#include <SxForceSym.h>
#include <SxRadialAtom.h>
#include <SxPartialWaveBasis.h>
#include <SxPAWBasis.h>
#include <SxLoopMPI.h>
#include <SxTaskGroup.h>
#include <SxParallelHierarchy.h>
#include <SxPAWSet.h>


double SxPAWHamiltonian::rSoft = 1.0;

/*
References:
 1. P. E. Bloechl, Phys. Rev. B 50, 17953 (1994).
 2. C. Freysoldt, PAW implementation notes
*/

SxPAWHamiltonian::SxPAWHamiltonian ()
   : SxHamiltonian (),
     hContrib(CalcAll),
     eTotal (0.),
     rPtr (NULL)
{
   // empty
}


SxPAWHamiltonian::SxPAWHamiltonian (const SxGBasis  &G,
                      const SxPtr<SxPAWPot> &potPtrIn,
                      const SxGkBasis &Gk,
                      const SxString  &rhoFile,
                      const SxSymbolTable *table)
/*
                      //const SxFermi &fermi,
                      //const SxPtr<SxPWSet> &wavesPtr)
*/
   : hContrib(CalcAll),
     eTotal (0.),
     potPtr(potPtrIn),
     rPtr (NULL),
     pawRho (potPtrIn)
{
   SX_CHECK (G.structPtr);
   structure = *G.structPtr;
   const SxRBasis &R = G.getRBasis ();
   int nSpin = SxHamiltonian::getNSpin (table);
   pBasis    = SxPtr<SxPartialWaveBasis>::create (potPtr, structure);
   rPtr      = const_cast<SxRBasis*>(&R);
   setupProj (Gk);
   setupPSRefG (G);
   xcPtr = SxPtr<SxXC>::create (nSpin);
   try  {
      xcPtr->read (table->getGroup ("PAWHamiltonian"), structure.cell);
   } catch (SxException e)  {
      e.print ();
      SX_EXIT;
   }
   if (!xcPtr->xcBasisPtr)  {
      unsigned oldMode = 0;
      SxFFT::quickFFTPlanner (SxFFT::Estimate, &oldMode);
      SxMesh3D mesh2 = 2 * R.fft3d.mesh;
      xcPtr->xcBasisPtr = SxPtr<SxRBasis>::create (mesh2, structure.cell);
      xcPtr->xcBasisPtr->registerBasis (G);
      SxFFT::restorePlannerMode (oldMode);
      cout << "xc will be computed on " << mesh2 << " mesh.\n";
   }
   xcPtr->origXcBasis = xcPtr->xcBasisPtr;

   if (SxXC::isHybrid (xcPtr->xcFunctional))  {
      // Look at initialGuess and think about how it is done correctly
      SX_EXIT;
   }
   pawRho.pwRho = SxRho(R, nSpin, -1.);
   pawRho.readRho (rhoFile);
   /*
   if (pawHamPtr->pawRho.Dij(0).getNSpecies () != structure.getNSpecies ())  {
      cout << SEPARATOR;
      cout << "| WARNING " << endl;
      cout << "| Species in rho and in structure differ, rho is now calculated from waves!" << endl;
      cout << SEPARATOR;
      pawHamPtr->computeRho (fermi.focc, *wavesPtr);
   }
   */

   computeRhoTerms (pawRho.Dij, pawRho.pwRho.rhoR);

}

SxPAWHamiltonian::~SxPAWHamiltonian ()
{
   // empty
}

void SxPAWHamiltonian::compute (bool, bool)
{
   SX_EXIT; // not yet implemented
}

void SxPAWHamiltonian::printEnergies () const
{
   //SX_EXIT;
}
void SxPAWHamiltonian::read (const SxSymbolTable *table)
{
   const SxSymbolTable *hamGroup = getHamiltonianGroup (table);
   if (hamGroup->contains ("rSoft"))  {
      rSoft = hamGroup->get ("rSoft")->toReal ();
   }
   // --- dipole correction
   if (withDipoleCorrection (hamGroup))  {
      dipoleCorr = SxPtr<SxDipoleCorrZ>::create ();
      try {
         if (hamGroup->contains("fixedDipoleZ"))  {
            double dipolePos = hamGroup->get("fixedDipoleZ")->toReal();
            dipoleCorr->fixedDipoleZ = true;
            dipoleCorr->dipolePos = dipolePos;
         }
         if (hamGroup->contains("zField"))
            dipoleCorr->extraField = hamGroup->get("zField")->toReal() / HA2EV;
      } catch (SxException e) {
         e.print ();
         SX_EXIT;
      }
   }

   // --- external potential
   if (hamGroup->containsGroup("vExt"))  {
      SxString vExtFile = hamGroup->getGroup("vExt")->get("file")->toString();

      try  {
         SxBinIO io (vExtFile, SxBinIO::BINARY_READ_ONLY);
         SX_CHECK (rPtr);
         const SxRBasis &R = *rPtr;
         int nSpinFile = io.getDimension ("nMeshes");
         if (nSpinFile != 1)  {
            cout << "Inconsistency when reading external potential '"
                 << vExtFile << "'.\n";
            cout << "There should be one channel, but found " 
                 << nSpinFile << " in file." << endl;
            SX_QUIT;
         }
         SxMatrix3<Double> cellFile;
         SxVector3<Int>    dim;
         vExt = io.readMesh (&cellFile, &dim)(0);
         if ((cellFile - R.cell).absSqr ().sum () > 1e-7)  {
            cout << "cell inconsistency in external potential." << endl;
            cout << "Expected " << R.cell << endl;
            cout << "Found " << cellFile << endl;
            SX_QUIT;
         }
         if (!(dim == R.fft3d.mesh))  {
            cout << "Mesh inconsistency in external potential." << endl;
            cout << "Expected " << R.fft3d.mesh << ", found " << dim << endl;
            SX_QUIT;
         }
         vExt.setBasis (&R);
      } catch (SxException e) {
         e.print ();
         SX_EXIT;
      }

      //Check if vExt is symmetry consistent
      cout << SEPARATOR;
      cout << "Check for symmetry consistent vExt ... ";
      SxDiracVec<Double> vExtSym = rPtr->symmetrize(vExt);
      if((vExtSym - vExt).normSqr() > 1e-6)   {
         cout << endl;
         cout << SEPARATOR;
         cout << "vExt seems to break symmetry!\nS/PHI/nX quits here!" << endl;
         cout << SEPARATOR;
         SX_QUIT;
      }
      cout << "ok" << endl;
      cout << SEPARATOR;
   }
   if (hamGroup->contains ("hContrib"))  {
      hContrib = lround(hamGroup->get ("hContrib")->toReal ());
      cout << SEPARATOR;
      cout << "| hContrib:" << endl;
      if (hContrib & CalcNuc)
         cout << "|    nuclear potential" << endl;
      if (hContrib & CalcBar)
         cout << "|    local pseudopotential vBar" << endl;
      if (hContrib & CalcXcRad)
         cout << "|    radial xc" << endl;
      if (hContrib & CalcXcPS)
         cout << "|    pseudo xc" << endl;
      if (hContrib & CalcKinRad)
         cout << "|    kinetic energy on-site" << endl;
      if (hContrib & CalcKinPS)
         cout << "|    kinetic energy" << endl;
      if (hContrib & CalcNHatRad)
         cout << "|    compensation charge (radial)" << endl;
      if (hContrib & CalcNHatGHard)
         cout << "|    compensation charge hard" << endl;
      if (hContrib & CalcNHatGSoft)
         cout << "|    compensation charge soft" << endl;
      if (hContrib & CalcCore)
         cout << "|    core energy" << endl;
      if (hContrib & CalcHartreeRad)
         cout << "|    Hartree (radial)" << endl;
      if (hContrib & CalcHartreePS)
         cout << "|    Hartree (pseudo)" << endl;
      if (hContrib & CalcU)
         cout << "|    \"U\" (compensation hard/soft)" << endl;
      if (hContrib & CalcCorValExch)
         cout << "|    core-valence exchange (if any)" << endl;
      if (hContrib & CalcValExch)
         cout << "|    valence exchange (if any)" << endl;
      cout << SEPARATOR;
   }
}

bool SxPAWHamiltonian::rereadTable (const SxSymbolTable *table)
{
   SX_CHECK (table);
   SX_CHECK (xcPtr);
   bool changed;
   // xc mesh
   changed = xcPtr->rereadTable (table);

   // --- dipole correction
   if (table->contains ("dipoleCorrection"))  {
      if (!dipoleCorr)  {
         cout << "WARNING: no dipole correction in PAWHamiltonian"
              << endl;
         cout << "WARNING: From here on, there will be a dipole correction."
              << endl;
         dipoleCorr = SxPtr<SxDipoleCorrZ>::create ();
         hContrib &= ~CalcDipoleC; // until now, it was suppressed
      }
      changed = true;
      bool dipole = table->get ("dipoleCorrection")->toAttribute ();
      cout << "| Dipole correction is now switched "
           << (dipole ? "on" : "off") << "." << endl;

      bool recompute = (dipole ? 0 : 1) != ((hContrib & CalcDipoleC) ? 0 : 1);
      if (dipole)
         hContrib |= CalcDipoleC; // switch on
      else
         hContrib &= ~CalcDipoleC; // switch off

      // is dipole correction being switched?
      if (recompute) {
         cout << "| Recomputing potentials ..." << endl;
         computeRhoTerms (pawRho.Dij, pawRho.pwRho.rhoR);
      }
   }
   return changed;
}

void SxPAWHamiltonian::backToDefault (const SxSymbolTable *)
{
   SX_CHECK (xcPtr);
   // xc basis
   xcPtr->rereadTable (NULL);

   // --- dipole correction
   if (dipoleCorr && !(hContrib & CalcDipoleC))  {
      cout << "| Dipole correction is now switched on." << endl;
      cout << "| Recomputing potentials ..." << endl;
      hContrib |= CalcDipoleC;
      computeRhoTerms (pawRho.Dij, pawRho.pwRho.rhoR);
   }
   hContrib |= CalcDipoleC;

}

const SxSymbolTable *
SxPAWHamiltonian::getHamiltonianGroup (const SxSymbolTable *table)
{
   SxSymbolTable *hGroup = NULL;
   if (table->name == "PAWHamiltonian") return table;
   try  {
      hGroup = table->getGroup("PAWHamiltonian");
   }  catch (SxException) {  }
   return hGroup;
}

void SxPAWHamiltonian::setupProj (const SxGkBasis &gk)
{
   SX_CHECK (gk.getNk() > 0);
   SX_CHECK (potPtr);

   SxArray<SxDiracMat<Double> > projRad(potPtr->getNSpecies ());
   for (int is = 0; is < potPtr->getNSpecies (); ++is)  {
      if (    potPtr->pPsFine.getSize () > 0
           && potPtr->pPsFine(is).getSize () > 0)
      {
         projRad(is) = potPtr->pPsFine(is);
      } else {
         projRad(is) = potPtr->pPS(is);
      }
   }
   projBasis = SxPtr<SxAOBasis>::create (gk, projRad, potPtr->lPhi);

   // --- setup partial wave basis
   pBasis = SxPtr<SxPartialWaveBasis>::create (potPtr, structure);
   pBasis->projectors = projBasis;

   // --- Setup symmetry group
   int lmax = potPtr->getLMax ();
   ylmRot = SxYlm::computeYlmRotMatrices 
      (structure.cell.symGroupPtr->getSymmorphic (), lmax);
}

SxArray<SxRadMat> SxPAWHamiltonian::computeDij (const SxPWSet &waves,
                                                const Focc &focc) const
{
   return computeDij (waves, focc, potPtr, *pBasis, ylmRot);
}

SxArray<SxRadMat>
SxPAWHamiltonian::computeDij (const SxPWSet &waves,
                              const Focc &focc,
                              const SxConstPtr<SxPAWPot> &potPtr,
                              const SxPartialWaveBasis &pBasis,
                              const SxYlmRotGroup &ylmRot)
{
   CLOCK (Timer::Dij);

   const SxVector<Double> &weights = waves.getGkBasis ().weights;
   SX_CHECK (waves.getGkBasis ()(0).structPtr);
   const SxAtomicStructure &structure = *waves.getGkBasis ()(0).structPtr;
   SX_CHECK (potPtr);
   const SxPAWPot &pot = *potPtr;
   int nSpin = waves.getNSpin ();
   SxArray<SxRadMat> res(nSpin);
   for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
      res(iSpin).resize (structure.atomInfo, pot);
      res(iSpin).set (0.);
   }

   SX_NEW_LOOP (waves);
   for (int ik = 0; ik < waves.getNk (); ++ik) {
      for (int iSpin = 0; iSpin < nSpin; ++iSpin) {
         SX_MPI_LEVEL("waves-k");
         if (SxLoopMPI::myWork(ik))
         {  // LoopMPI
            double weight = weights(ik);
            SxDiracVec<Complex16> p = pBasis | waves(iSpin,ik);
            int offset = 0;
            for (int is = 0; is < structure.getNSpecies (); ++is)  {
               int nProjLocal = pot.getNProj (is);

               // --- set up i^l
               SxVector<Complex16> il(nProjLocal); // i^(l)  (:ip)
               for (int ipt = 0, ipl = 0; ipt < pot.lPhi(is).getSize (); ++ipt)  {
                  int l = pot.lPhi(is)(ipt);
                  SxComplex16 c = (l & 1) ? I : SxComplex16(1.);
                  if (l & 2) c=-c;
                  for (int m = -l; m <=l; ++m, ++ipl)  {
                     il(ipl) = c;
                  }
               }

               for (int ia = 0; ia < structure.getNAtoms(is); ++ia, offset+=nProjLocal)  {
                  SxMatrix<Double> &D = res(iSpin)(is,ia);
                  for (int ipl = 0; ipl < nProjLocal; ++ipl)  {
                     for (int jpl = 0; jpl < nProjLocal; ++jpl)  {
                        SxComplex16 phase = il(ipl) * il(jpl).conj ();
                        for (int iState = 0; iState < p.nCols (); ++iState)  {
                           // ref. 1, Eq. (52)
                           // + projector i^l phase-correction (Ref. 2)
                           // + use k-dependent weights
                           // + exploit k/-k symmetry (=> .re)
                           D(ipl, jpl) += weight * focc(iState,iSpin,ik)
                                           * (  p(offset+ipl,iState).conj ()
                                                 * p(offset+jpl,iState) * phase).re;
                        }
                     }
                  }
               }
            }
         }   // LoopMPI
      }
   }

   // --- symmetrize
   SX_MPI_SOURCE ("waves-k", TaskGroupMaster);
   SX_MPI_TARGET (TopLevel, TaskGroupAll);
   for (int iSpin = 0; iSpin < nSpin; ++iSpin)
   {
      res(iSpin).sumMPI (); // LoopMPI
      res(iSpin).symmetrize (structure, pot, ylmRot);
   }

   return res;
}

enum LocTimer { Setup, TimeYlmGl, Loop1, Loop2, Loop3, PsStuff, NHatG,
                PsStuff1, PsStuff2, PsStuff3, PsStuff4, PsStuff5, PsStuff6 };
REGISTER_TIMERS (LocTimer)
{
   regTimer (Setup, "setup");
   regTimer (TimeYlmGl, "YlmGl");
   regTimer (Loop1, "loop 1");
   regTimer (NHatG, "nHatG");
   regTimer (Loop2, "loop 2");
   regTimer (Loop3, "loop 3");
   regTimer (PsStuff, "ps stuff");
   regTimer (PsStuff1, "ps stuff 1");
   regTimer (PsStuff2, "ps stuff 2");
   regTimer (PsStuff3, "ps stuff 3");
   regTimer (PsStuff4, "ps stuff 4");
   regTimer (PsStuff5, "ps stuff 5");
   regTimer (PsStuff6, "ps stuff 6");
}


void SxPAWHamiltonian::computeRhoTerms (const SxArray<SxRadMat> &Dij,
                                        const RhoR &rhoPSR)
{
   CLOCK (Timer::ComputeHamRho);
   START_TIMER (Setup);
   const double SQRT_4PI = sqrt(FOUR_PI);
   double rNorm = FOUR_PI / sqrt(structure.cell.volume);
   SX_CHECK_NUM (rNorm);

   SX_CHECK (potPtr);
   SX_CHECK (xcPtr);
   SX_CHECK_VAR (rhoPSR.getSize () > 0, rhoPSR.getSize ());
   const SxPAWPot &pawpot = *potPtr;

   // --- get Dirac bases
   const SxRadBasis &rad  = pawpot.getRadBasis ();
   const SxRBasis &rBasis = rhoPSR(0).getBasis<SxRBasis> ();
   const SxGBasis &gBasis = rBasis.getGBasis ();
   // indices without G=0 component
   SxIdx noZero(1,gBasis.ng - 1);

   int nSpin = int(Dij.getSize ());
   SX_CHECK_VAR  (nSpin > 0 && nSpin <= 2, nSpin);
   SX_CHECK_VARS (rhoPSR.getSize () == nSpin, rhoPSR.getSize (), nSpin);

   // compute total rhoPS in G-space
   PsiG rhoPSG = gBasis | ((nSpin == 1) ? rhoPSR(0)
                                        : rhoPSR(0) + rhoPSR(1));

   double eXcRadial      = 0.;
   double eHartreeRadial = 0.;
   double eBarRadial     = 0.;
   double eR0            = 0.;
   double vHatGZero = 0.;
   eDoubleCounting = 0.;

   //double eVnc = 0.;
   Vij.resize (nSpin);
   for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
      Vij(iSpin).resize (Dij(iSpin));
      Vij(iSpin).set (0.);
   }

   // Ref 1, Eq. 35
   SxArray<SxArray<SxVector<Double> > > dEdQrl(structure.getNSpecies ());
   Qrl.resize (structure.getNSpecies ());

   // \hat n, \hat n' in G-space
   PsiG nHatG(gBasis), nHatpG(gBasis);
   nHatG.set (0.);     nHatpG.set (0.);

   // |G|^l * Ylm(G)
   START_TIMER (TimeYlmGl);
   int lMaxRho = pawpot.lMaxRho.maxval ();
   SxDiracMat<Double> YlmGl(gBasis.ng, sqr(lMaxRho + 1));
   YlmGl.setBasis (gBasis);
   for (int lm = 0, l = 0; l <= lMaxRho; ++l)  {
      SxDiracVec<Double> gl = pow (gBasis.g2, 0.5 * l);
      for (int m = -l; m <= l; ++m,++lm)
         YlmGl.colRef(lm) <<= gBasis.getYlm (l,m) * gl;
   }
   STOP_TIMER (TimeYlmGl);

   // Hartree potential from rhoPSG (G != 0)
   PsiG vRhoPSG = FOUR_PI / gBasis.g2(noZero) * rhoPSG(noZero);
   if (!(hContrib & CalcHartreePS)) vRhoPSG.set (0.);

   STOP_TIMER (Setup);

   SxVector<Double> MSpin(structure.getNAtoms()) ;
   MSpin.set(0.);
   SX_CHECK(potPtr);
   const SxPAWPot &pawPot = *potPtr;

   // --- loop over atoms
   START_TIMER (Loop1);
   for (int is = 0; is < structure.getNSpecies (); ++is)
   {

      SxRadialMesh v1AE,v1PS;
      SxArray<SxRadialMesh> vXcAE,vXcPS;
      SxArray<SxDiracVec<Double> > gRl;

      double logDr = rad.logDr(is);
      int lMax = pawpot.lMaxRho (is);
      double r0cub = pow(rad.radFunc(is)(0), 3.);

      // --- resize objects
      Qrl(is).resize (structure.getNAtoms (is));
      dEdQrl(is).resize (structure.getNAtoms (is));
      v1AE.resize (rad, is, lMax);
      v1PS.resize (rad, is, lMax);

      // --- set up generalized Gaussians
      gRl.resize (lMax+1);
      for (int l = 0; l <= lMax; ++l)  {
         gRl(l) = pawpot.getGrl (is, l);
      }
      // --- Gaussians in G space
      double rc2 = sqr(pawpot.rc(is)), rcp2 = sqr(rSoft);
      SxDiracVec<Double> gaussHard = exp ( -(0.25 * rc2 ) * gBasis.g2),
            gaussSoft = exp ( -(0.25 * rcp2) * gBasis.g2);

      SxRadialAtom radial(xcPtr->xcFunctional, pawpot.aGridType(is));

		// --- temporary variables for parallelization (SxParallelHierarchy)
      double eXcRadialPrime       = 0.;
      double eHartreeRadialPrime  = 0.;
      double eBarRadialPrime      = 0.;
      double eR0Prime             = 0.;
      double vHatGZeroPrime       = 0.;
      double eDoubleCountingPrime = 0.;
      PsiG nHatGPrime(gBasis);
      PsiG nHatpGPrime(gBasis);
      nHatGPrime.set(0.);
      nHatpGPrime.set(0.);
      // ---
      /*
      SxDiracVec<Double> VncAE
         = -(SQRT_4PI * pawpot.nuclearCharge(is))/rad.radFunc(is)
         + radial.getHartreePotential (pawpot.rhoCoreAE(is));
      SxDiracVec<Double> VncPS;
      {
         SxDiracVec<Double> dRhoCore = pawpot.rhoCorePS(is).getCopy ();
         dRhoCore -= gRl(0) * (tr (pawpot.rhoCorePS(is)) + SQRT_1_4PI * pawpot.valenceCharge(is));
         VncPS = radial.getHartreePotential (dRhoCore);
      }
      */
      
      // --- setup omegaij(ipl,jpl) for spin constraints
      SxMatrix<Double> omegaij(pawPot.getNProj(is),pawPot.getNProj(is));
      omegaij.set (0.);
      SX_LOOP2(ipt, jpt)  {
         int li = pawPot.lPhi(is)(ipt);
         int lj = pawPot.lPhi(is)(jpt);
         if (li == lj)  {
            int iOff = pawPot.offset(is)(ipt),
                jOff = pawPot.offset(is)(jpt);
            for (int im = 0; im < 2 * li + 1; ++im)
               omegaij(im+iOff,im+jOff) = pawPot.omegaPAW(is)(ipt,jpt);
         }
      }

      for (int ia = 0; ia < structure.getNAtoms (is); ++ia)
      {
         int iTlAtom = structure.getIAtom (is, ia);
         if (SxLoopMPI::myWork(iTlAtom)) // SxParallelHierarchy
         {

            // Compute 1-center density
            SxArray<SxRadialMesh> rhoRadPS(nSpin), rhoRadAE(nSpin);
            for (int iSpin = 0; iSpin <  nSpin; ++iSpin)  {
               rhoRadPS(iSpin) = pawpot.computeRhoPS (Dij(iSpin)(is,ia),is,nSpin);
               rhoRadAE(iSpin) = pawpot.computeRhoAE (Dij(iSpin)(is,ia),is,nSpin);
            }

            // --- xc part
            if (hContrib & CalcXcRad)  {
               if (   pawpot.kjXcShape.getSize () > 0
                     && pawpot.kjXcShape(is).getSize () > 0)  {
                  // --- Kresse-Joubert variant: include compensation charge
                  // in pseudo-xc
                  SxArray<SxRadialMesh> rhoXcKj(nSpin);
                  double Q = 0.;
                  for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
                     Q += ((rhoRadAE(iSpin)(0,0) - rhoRadPS(iSpin)(0,0))
                           * rad.radFunc(is).cub ()).integrate (logDr);
                  }
                  // remove core charges (not included in Kresse-Joubert xc)
                  Q -= ((pawpot.rhoCoreAE(is) - pawpot.rhoCorePS(is))
                        * rad.radFunc(is).cub ()).integrate (logDr);
                  // cout << "Q=" << Q << endl;
                  for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
                     rhoXcKj(iSpin).resize (rhoRadPS(iSpin));
                     rhoXcKj(iSpin).meshData.copy (rhoRadPS(iSpin).meshData);
                     rhoXcKj(iSpin)(0,0) += Q/nSpin * pawpot.kjXcShape(is);
                  }

                  vXcPS = radial.computeXC (rhoXcKj);
                  eXcRadialPrime -= radial.eXc;
               } else {
                  vXcPS = radial.computeXC (rhoRadPS);
                  eXcRadialPrime -= radial.eXc;
               }
               // --- double counting correction PS xc
               eDoubleCountingPrime -= radial.eXc;
               for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
                  for (int l = 0; l <= lMax; ++l)  {
                     for (int m = -l; m <= l; ++m)  {
                        eDoubleCountingPrime +=  tr(  rhoRadPS(iSpin)(l,m)
                                                    * vXcPS(iSpin)(l,m));
                     }
                  }
                  // rhoRadPS contains pseudo-core electrons, which are not 
                  // double-counted
                  if (pawpot.rhoCorePS(is).getSize () > 0)
                     eDoubleCountingPrime -=  tr(pawpot.rhoCorePS(is)
                                                 * vXcPS(iSpin)(0,0)) / nSpin;
               }
               // --- all electron xc
               vXcAE = radial.computeXC (rhoRadAE);
               eXcRadialPrime += radial.eXc;

               // --- double counting correction AE xc
               eDoubleCountingPrime += radial.eXc;
               for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
                  for (int l = 0; l <= lMax; ++l)  {
                    for (int m = -l; m <= l; ++m)  {
                      eDoubleCountingPrime -=  tr(rhoRadAE(iSpin)(l,m)
                                             * vXcAE(iSpin)(l,m));
                    }
                  }
                  // rhoRadAE contains core electrons, which are not 
                  // double-counted
                  if (pawpot.rhoCoreAE(is).getSize () > 0)
                     eDoubleCountingPrime +=  tr(pawpot.rhoCoreAE(is)
                                            * vXcAE(iSpin)(0,0)) / nSpin;
               }
            }

            if (nSpin == 2)  {
               // switch to total densities, save in iSpin=0
               rhoRadPS(0) += rhoRadPS(1);
               rhoRadPS(1) = SxRadialMesh ();
               rhoRadAE(0) += rhoRadAE(1);
               rhoRadAE(1) = SxRadialMesh ();
            }
            // --- vBar part (Ref. 1, Eq. 21) without ps core
            if (hContrib & CalcBar)  {
               bool hasCore = pawpot.rhoCorePS(is).getSize () > 0;
               SxDiracVec<Double> rhoVal 
                  = hasCore ? rhoRadPS(0)(0,0) - pawpot.rhoCorePS(is)
                            : rhoRadPS(0)(0,0);
               eBarRadialPrime += tr (rhoVal * pawpot.vBar(is));
               eR0Prime += rhoVal(0) * pawpot.vBar(is)(0) * r0cub / 3.;
               //cout << "eBarR0=" << rhoVal(0) * pawpot.vBar(is)(0) * r0cub / 3. << endl;
            }

            // --- compute QRL
            SxRadialMesh rhoPAWcorr = rhoRadAE(0) - rhoRadPS(0);
            SxVector<Double> &Q = Qrl(is)(ia);
            Q.resize(rhoPAWcorr.meshData.nCols ());
            for (int l = 0; l <= rhoPAWcorr.lmax; ++l)  {
               SxDiracVec<Double> rl = pow(rad.radFunc(is),double(l+3));
               for (int m = -l; m <= l; ++m)  {
                  int lm = SxYlm::combineLm(l,m);
                  // Ref. 1, Eq. (24)
                  Q(lm) = (rhoPAWcorr(l,m) * rl).integrate (rad.logDr(is));
                  if (l == 0) Q(lm) -= SQRT_1_4PI * pawpot.nuclearCharge(is);
                  //cout << "Q(l=" << l << "; m=" << m << ")=" << Q(lm) << endl;

                  // add compensation charge to pseudo-density
                  if (hContrib & CalcNHatRad)
                     rhoRadPS(0)(l,m).plus_assign_ax(Q(lm), gRl(l));
               }
            }

            // --- Hartree energy and potentials
            dEdQrl(is)(ia).resize(rhoPAWcorr.meshData.nCols ());
            if (hContrib & CalcHartreeRad)  {
               for (int l = 0; l <= rhoPAWcorr.lmax; ++l)  {
                  for (int m = -l; m <= l; ++m)  {
                     v1AE(l,m) <<= radial.getHartreePotential (rhoRadAE(0)(l,m));
                     v1PS(l,m) <<= radial.getHartreePotential (rhoRadPS(0)(l,m));
                     // Ref. 1, Eq. 20, Hartree term without n^Z
                     double E =  0.5 * tr(rhoRadAE(0)(l,m) * v1AE(l,m));
                     eHartreeRadialPrime += E;
                     eDoubleCountingPrime -= E;
                     // Ref. 1, Eq. 21, Hartree term
                     E = 0.5 * tr(rhoRadPS(0)(l,m) * v1PS(l,m));
                     eHartreeRadialPrime -= E;
                     eDoubleCountingPrime += E;
                     if (hContrib & CalcNHatRad)  {
                        int lm = SxYlm::combineLm(l,m);
                        // Ref. 1, Eq. 36, 3rd term
                        dEdQrl(is)(ia)(lm) = -tr(gRl(l) * v1PS(l,m));
                     }
                  }
               }
               if (pawpot.rhoCoreAE(is).getSize () > 0)
                  eDoubleCountingPrime += tr(pawpot.rhoCoreAE(is) * v1AE(0,0));
               if (pawpot.rhoCorePS(is).getSize () > 0)
                  eDoubleCountingPrime -= tr(pawpot.rhoCorePS(is) * v1PS(0,0));
               eR0Prime += rhoRadAE(0)(0,0,0) * v1AE(0,0,0) * r0cub / 3.;
               //cout << "H r0 AE="
               //     << (rhoRadAE(0)(0,0,0) * v1AE(0,0,0) * r0cub / 3.) << endl;
               eR0Prime -= rhoRadPS(0)(0,0,0) * v1PS(0,0,0) * r0cub / 3.;
               //cout << "H r0 PS="
               //     << (-rhoRadPS(0)(0,0,0) * v1PS(0,0,0) * r0cub / 3.) << endl;
               if (!(hContrib & CalcNHatRad)) dEdQrl(is)(ia).set (0.);
            } else {
               v1AE.set (0.);
               v1PS.set (0.);
               dEdQrl(is)(ia).set (0.);
            }

            if (hContrib & CalcNuc)  {
               // Ref. 1, Eq. 20, Hartree n^Z parts
               // \int r^2 dr rho_AE(r) * (-Z)/r = -Z \int d(log r) r^2 rho_AE(r)
               eHartreeRadialPrime -= SQRT_4PI * pawpot.nuclearCharge(is)
                                 * (rhoRadAE(0)(0,0)  * rad.radFunc(is).sqr ())
                                 .integrate (logDr);
               eDoubleCountingPrime -= SQRT_4PI * pawpot.nuclearCharge(is)
                                 * (pawpot.rhoCoreAE(is) * rad.radFunc(is).sqr ())
                                 .integrate (logDr);
               // Ref. 1, Eq. (38): n^Z part
               v1AE(0,0) -= (SQRT_4PI * pawpot.nuclearCharge(is))
                               / rad.radFunc(is);

               // --- Estimate integral 0..r0 from rho(r0) and rho'(r0)
               double r0  = rad.radFunc(is)(0),
                     r1  = rad.radFunc(is)(1),
                     r01 = r0 - r1,
                     w0  = logDr / 3.,      // includes Simpson weight
                     w1  = 4. * logDr / 3., // includes Simpson weight
                     rho0 = rhoRadAE(0)(0,0,0),
                     rho1 = rhoRadAE(0)(1,0,0),
                     Z    = pawpot.nuclearCharge(is);
               eHartreeRadialPrime -= SQRT_4PI * Z * (rho0 * 0.5 * r0 * r0
                     - (rho0  -rho1 ) / r01 * r0cub/6.);
               eDoubleCountingPrime -= SQRT_4PI * Z *
                                  (pawpot.rhoCoreAE(is)(0) * 0.5 * r0 * r0
                     - (pawpot.rhoCoreAE(is)(0)  -pawpot.rhoCoreAE(is)(1) ) / r01 * r0cub/6.);
               v1AE(0,0,0) -= SQRT_4PI * Z *
                     (0.5 / r0 - 1./(6. * r01)) / w0;
               v1AE(1,0,0) -= SQRT_4PI * Z * 1./(6. * r01)
                                 *r0cub / (r1*r1*r1) / w1;
            }
            //eVnc += tr (VncAE * (rhoRadAE(0)(0,0) - pawpot.rhoCoreAE(is)));
            //eVnc -= tr (VncPS * (rhoRadPS(0)(0,0) - pawpot.rhoCorePS(is)));

            // Missing part vBar in Ref. 1, Eq. (39) (cf. Eq. 21)
            if (hContrib & CalcBar)
               v1PS(0,0) += pawpot.vBar(is);

            if (hContrib & CalcXcRad)  {
               // Ref. 1, Eq. 38/39: Hartree + xc part
               for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
                  // Ref 1, Eq. 40 with Eq. 38/39 without vR0 (see below)
                  Vij(iSpin)(is,ia)
                        = pawpot.getVMatEl (v1AE + vXcAE(iSpin), pawpot.phiAE(is))
                        - pawpot.getVMatEl (v1PS + vXcPS(iSpin), pawpot.phiPS(is));
               }
            } else  {
               // Ref. 1, Eq. 38/39: Hartree, but not xc part
               // Ref 1, Eq. 40 with Eq. 38/39 without vR0 (see below)
               Vij(0)(is,ia) = pawpot.getVMatEl (v1AE, pawpot.phiAE(is))
                                - pawpot.getVMatEl (v1PS, pawpot.phiPS(is));
               if (nSpin == 2)
                  Vij(1)(is,ia) <<= Vij(0)(is,ia);
            }

            bool spinConstraint = nuA.getSize () > 0; // TODO: fix condition
            if (nSpin == 2) {
               SX_LOOP2(i,j)
                  MSpin(iTlAtom) += omegaij(i,j) * (  Dij(0)(is,ia)(i,j) 
                                                    - Dij(1)(is,ia)(i,j));
            }
            if (spinConstraint)  {
               Vij(0)(is,ia) -= nuA(iTlAtom) * omegaij;
               Vij(1)(is,ia) += nuA(iTlAtom) * omegaij;
               eDoubleCountingPrime += nuA(iTlAtom) * MSpin(iTlAtom);
            }
            // --- compute compensation charges in G-space
            if (hContrib & CalcNHatG)  {
               CLOCK (NHatG);
               // see also Ref. 2, Eq. 3
               double prefac = 1.;
               PsiG T = gBasis.getPhaseFactors (is,ia);
               for (int l = 0; l <= lMax; ++l)  {
                  prefac /= 2. * l + 1.;
                  for (int m = -l; m <= l; ++m)  {
                     int lm = SxYlm::combineLm(l,m);
                     double N = rNorm * SxYlm::getYlmNormFactor(l,m) * prefac;
                     double weight = N * Qrl(is)(ia)(lm);
                     // i^l factor
                     SxComplex16 il = 1.;
                     if ((l & 3) == 1) il = I;
                     else if ((l & 3) == 2) il = -1.;
                     else if ((l & 3) == 3) il = -I;
                     ssize_t ng = T.getSize ();
                     double dEdQrlm = 0.;
#ifdef USE_OPENMP
#pragma omp parallel for reduction(+:dEdQrlm) if (ng > sxChunkSize) 
#endif
                     // this ugly single ig loop for nHatG, nHatpG, and dEdQrl
                     // improves memory access pattern and parallelizability 
                     // for openMP => much faster than elegant way
                     for (ssize_t ig = 0; ig < ng; ++ig)  {
                        SxComplex16 c = YlmGl(ig,lm) * il * T(ig);
                        if (hContrib & CalcNHatGHard) {
                           nHatGPrime(ig) += weight * gaussHard(ig) * c;
                           if (ig > 0) {
                               dEdQrlm += N * (vRhoPSG(ig-1).conj () * c).re 
                                        * gaussHard(ig);
                           }
                        }
                        if (hContrib & CalcNHatGSoft)
                           nHatpGPrime(ig) += weight * gaussSoft(ig) * c;
                     }
                     dEdQrl(is)(ia)(lm) += dEdQrlm;
                  }
               }
               prefac = 1.;

               if ((hContrib & CalcNHatGHard) && (hContrib & CalcNHatGSoft))  {
                  // absolute alignment of vHat (depends on rc2/rcp2)
                  // 4 PI/G^2 * Qrl * ( grl - gprl)
                  // G -> 0: 4 PI/G^2 * Qrl * N * G^L (1 - rc2/4 G^2 - 1 + rcp2/4 G^2)
                  //         = PI * N * Qrl * G^l (rcp2 - rc2)
                  // non-zero limit for l==0
                  prefac = PI * rNorm * SxYlm::getYlmNormFactor(0,0) * (rcp2 - rc2);
                  // for (int ia = 0; ia < structure.getNAtoms (is); ++ia)  {
                  vHatGZeroPrime         += prefac * Qrl(is)(ia)(0);
                  if (hContrib & CalcHartreePS)
                     dEdQrl(is)(ia)(0) += prefac * rhoPSG(0).re;
                  // }
               }
            }
         } // LoopMPI
      } // ia loop

      {
         // --- MPI summations
         eXcRadialPrime       = SxLoopMPI::sum (eXcRadialPrime);
         eHartreeRadialPrime  = SxLoopMPI::sum (eHartreeRadialPrime);
         eBarRadialPrime      = SxLoopMPI::sum (eBarRadialPrime);
         eR0Prime             = SxLoopMPI::sum (eR0Prime);
         vHatGZeroPrime       = SxLoopMPI::sum (vHatGZeroPrime);
         eDoubleCountingPrime = SxLoopMPI::sum (eDoubleCountingPrime);
         SxLoopMPI::sum (nHatGPrime);
         SxLoopMPI::sum (nHatpGPrime);
         SxLoopMPI::sum (MSpin);
      }
      vHatGZero       += vHatGZeroPrime;
      eXcRadial       += eXcRadialPrime;
      eHartreeRadial  += eHartreeRadialPrime;
      eBarRadial      += eBarRadialPrime;
      eR0             += eR0Prime;
      eDoubleCounting += eDoubleCountingPrime;
      nHatG           += nHatGPrime;
      nHatpG          += nHatpGPrime;
      // ---
   } // is loop
   STOP_TIMER(Loop1);

   if (nuA.getSize ()) SX_LOOP(ia) {
      cout << "nu(" << ia << ") = " << nuA(ia) << endl;
      cout << "Spin of atom " << ia << " = " << MSpin(ia) << endl;
   }

//   eVnc = SxLoopMPI::sum (eVnc);
//   sxprintf ("eVnc = %.12f\n", eVnc);

#ifdef USE_LOOPMPI
   // sync' Qrl between MPI processes
   if (SxLoopMPI::nr () > 1)
   {
      for (int is = 0; is < structure.getNSpecies (); ++is) {
         for (int ia = 0; ia < structure.getNAtoms (is); ++ia) {
            int size = 0;
            int iatom = structure.getIAtom (is, ia);
            int sender = SxLoopMPI::whoseWork (iatom);
            if (SxLoopMPI::myWork (iatom)) {
               size = (int)Qrl(is)(ia).getSize ();
            }
            size = SxLoopMPI::bcast (size, sender);
            if ((!SxLoopMPI::myWork (iatom)) && (size != Qrl(is)(ia).getSize ())) {
               Qrl(is)(ia).resize (size);
            }
            SxLoopMPI::bcast (Qrl(is)(ia), sender);
         }
      }
   }
#endif

   // START_TIMER (PsStuff1);
  
   START_TIMER (PsStuff);

   // --- the pseudo-density related parts
   PsiG rhoHartreeHard = rhoPSG + nHatG,
        rhoHartreeSoft = rhoPSG + nHatpG;

   double sqrtVol = sqrt(structure.cell.volume);
   // cout << "sqrtVol = " << sqrtVol << endl;

   // STOP_TIMER (PsStuff1);
   // doesn't take much time

   cout << "rhoHartreeHard(G=0) = " << (rhoHartreeHard(0) * sqrtVol) << endl;
   cout << "rhoHartreeSoft(G=0) = " << (rhoHartreeSoft(0) * sqrtVol) << endl;

   // remove G=0 component
   //rhoHartreeHard(0) = 0;
   //rhoHartreeSoft(0) = 0;

   // START_TIMER (PsStuff2);
   PsiG rhoCoreG;
   if (hContrib & (CalcHartreePS | CalcBar | CalcNHatG))
      rhoCoreG = gBasis | computeCorePS ();

   double eHartreePS = 0.;

   PsiG vHartreeHard(gBasis);
   // Ref. 1, Eq. 34, first & second term (cf. Eq. 27)
   if (hContrib & CalcHartreePS)  {
      vHartreeHard(0) = vHatGZero;
      if (hContrib & CalcNHatGHard)
         vHartreeHard(noZero) = FOUR_PI * rhoHartreeHard(noZero)
                              / gBasis.g2(noZero);
      else
         vHartreeHard(noZero) = FOUR_PI * rhoPSG(noZero)
                              / gBasis.g2(noZero);
   } else {
      vHartreeHard.set (0.);
   }

   // --- vBar terms (Ref. 1, Eq. 34, third term)
   PsiG vBarTotal(gBasis);
   vBarTotal.set (0.);
   if (hContrib & CalcBar)  {
      for (int is = 0; is < structure.getNSpecies (); ++is)
         vBarTotal += vBarG(is) * gBasis.structureFactors(is);

      // --- Ref. 1, Eq. 19/21 vBar term (without ps core)
      eBar = dot(rhoPSG - rhoCoreG, vBarTotal) - eBarRadial;
   } else {
      eBar = 0.;
   }

   // Hartree terms in Ref. 1, Eq. 34
   SxMeshR vPSR = (vHartreeHard + vBarTotal).to (rBasis);

   // Initialization of dipole potential   
   double eDipole = 0.;
   SxMeshR vDipole;
   double vAlign = 0.;
   if (dipoleCorr && (hContrib & CalcDipoleC))  {
      vDipole = SxMeshR(rBasis);
      vDipole.set (0.);
      dipoleCorr->charge = rhoHartreeSoft(0) * sqrtVol;
      if (fabs(dipoleCorr->charge) < 1e-6)
         dipoleCorr->charge = 0.;
#ifdef USE_LOOPMPI
      dipoleCorr->charge = SxLoopMPI::bcast (dipoleCorr->charge, 0);
#endif
      SxMeshR rhoHartreeR = rhoHartreeSoft.to (rBasis);
      dipoleCorr->update (rhoPSG.to (rBasis), rhoHartreeR);
      dipoleCorr->correctPotential (&vDipole);
      // correction to potential
      vPSR += vDipole;

      // correction to energy
      eDipole =  0.5 * tr(rhoHartreeSoft.to (rBasis) * vDipole);
      eDoubleCounting -= eDipole;
      eDoubleCounting += tr(rhoCoreG.to (rBasis) * vDipole);

      if (fabs(dipoleCorr->charge) > 1e-6)  {
         // --- align to right virtual electrode at z=0
         vAlign = dipoleCorr->getVRZero (vPSR);
         cout << "vAlign=" << vAlign * HA2EV << " eV" << endl;
         vPSR -= vAlign;
         vDipole -= vAlign;
         eDipole         -= 0.5 * rhoHartreeSoft(0).re * sqrtVol * vAlign;
         eDoubleCounting += 0.5 * rhoHartreeSoft(0).re * sqrtVol * vAlign;
         eDoubleCounting -= rhoCoreG(0).re       * sqrtVol * vAlign;
         double electrodeEnergy = dipoleCorr->getElectrodeEnergy (rBasis.cell);
         eDipole += electrodeEnergy;
         eDoubleCounting += electrodeEnergy;
      } else {
         double eField = 0.5 * dipoleCorr->extraField * dipoleCorr->dipole;
         eDipole += eField;
         eDoubleCounting += eField;

      }
      sxprintf ("eDipole         = %18.12f H\n", eDipole);
   }


   STOP_TIMER(PsStuff);

   if (hContrib & CalcNHatGSoft)
   {
      CLOCK (Loop2);
      // --- compute dE/dQRl terms that depend on nHatp
      PsiG vHatpG(gBasis);
      vHatpG(0) = 0.;
      vHatpG(noZero) = FOUR_PI / gBasis.g2(noZero) * nHatpG(noZero);
      if (vExt.getSize () > 0) {
         vHatpG += gBasis | vExt;
      }
      if (dipoleCorr && (hContrib & CalcDipoleC))  {
         vHatpG += gBasis | vDipole;
      }
      double rcp2 = sqr(rSoft);
      SxDiracVec<Double> gaussSoft = exp ( -(0.25 * rcp2) * gBasis.g2);
      for (int is = 0; is < structure.getNSpecies (); ++is) {
         int lMax = pawpot.lMaxRho (is);
         for (int ia = 0; ia < structure.getNAtoms (is); ++ia) {
            if (SxLoopMPI::myWork(structure.getIAtom (is,ia))) { // SxParallelHierarchy
               double prefac = 1.;
               for (int l = 0; l <= lMax; ++l)  {
                  prefac /= 2 * l + 1;
                  PsiG T = gBasis.getPhaseFactors (is,ia);
                  for (int m = -l; m <= l; ++m)  {
                     double N = rNorm * SxYlm::getYlmNormFactor(l,m) * prefac;
                     int lm = SxYlm::combineLm(l,m);
                     PsiG gprl = (YlmGl.colRef(lm) * gaussSoft) * T;
                     // i^l factor
                     SxComplex16 il = 1.;
                          if ((l & 3) == 1) il = I;
                     else if ((l & 3) == 2) il = -1.;
                     else if ((l & 3) == 3) il = -I;

                     // Ref. 1 Eq. 36, first line, 2nd term
                     dEdQrl(is)(ia)(lm) += N * (il * dot(vHatpG, gprl)).re;
                  }
               }
            } // LoopMPI
         }
      }
   }


   // compute U and U-related part of dE/dQrl
   double eHartreeU = 0.;
   if (hContrib & CalcU)  {
      eHartreeU = computeU (&dEdQrl);
      cout << "U = " << eHartreeU << endl;
   }


   /// --- Compute vR0 contribution
   START_TIMER (Loop3);
   for (int is = 0; is < structure.getNSpecies (); ++is)
   {
      double eDoubleCountingPrime = 0.;

      {
         double Zps = -pawpot.nuclearCharge(is) / sqrt(FOUR_PI);
         if (pawpot.rhoCoreAE(is).getSize () > 0)
            Zps += tr(pawpot.rhoCoreAE(is));
         if (pawpot.rhoCorePS(is).getSize () > 0)
            Zps -= tr(pawpot.rhoCorePS(is));
         
         for (int ia = 0; ia < structure.getNAtoms (is); ++ia) {
            if(SxLoopMPI::myWork(structure.getIAtom (is,ia))) { // SxParallelHierarchy
               eDoubleCountingPrime += dEdQrl(is)(ia)(0) * Zps;
            }
         }
      }

      int lMax = pawpot.lMaxRho (is);
      // int np = pawpot.getNProj (is);
      int nProjType = pawpot.getNProjType (is);

      SxRadialMesh vR0;
      vR0.resize (rad, is, lMax);

      for (int ia = 0; ia < structure.getNAtoms (is); ++ia) {
         if (SxLoopMPI::myWork(structure.getIAtom (is,ia))) { // SxParallelHierarchy

            for (int lm = 0, l = 0; l <= lMax; ++l)  {
               SxDiracVec<Double> rl = pow(rad.radFunc(is), double(l));

               // Ref. 1, Eq. 37
               for (int m = -l; m <= l; ++m, ++lm) {
                  vR0(l,m) <<= dEdQrl(is)(ia)(lm) * rl;
               }
            }


            // Ref 1, Eq. 40 with Eq. 38/39: vR0 term and kinetic energy
            SxMatrix<Double> vR0ij =   pawpot.getVMatEl (vR0, pawpot.phiAE(is))
                                     - pawpot.getVMatEl (vR0, pawpot.phiPS(is));
            for (int iSpin = 0; iSpin < nSpin; ++iSpin)   {
               SxMatrix<Double> &VijR = Vij(iSpin)(is,ia);
               // add VR0 term
               VijR += vR0ij;
               // --- add kinetic Energy
               if (hContrib & CalcKinRad)  {
                  for (int ipt = 0; ipt < nProjType; ++ipt)  {
                     int l = pawpot.lPhi(is)(ipt), nm = 2 * l + 1;
                     int offsetI = pawpot.offset(is)(ipt);
                     for (int jpt = 0; jpt < nProjType; ++jpt)  {
                        if (l == pawpot.lPhi(is)(jpt))  {
                           int offsetJ = pawpot.offset(is)(jpt);
                           double Tij = pawpot.deltaKin(is)(ipt,jpt);
                           for (int m = 0; m < nm; ++m)
                              VijR(offsetI + m, offsetJ + m) += Tij;
                        }
                     }
                  }
               }
               // --- add exact core-valence exchange
               if ((hContrib & CalcCorValExch) && exchangePtr) {
                  for (int ipt = 0; ipt < nProjType; ++ipt)  {
                     int l = pawpot.lPhi(is)(ipt), nm = 2 * l + 1;
                     int offsetI = pawpot.offset(is)(ipt);
                     for (int jpt = 0; jpt < nProjType; ++jpt)  {
                        if (l == pawpot.lPhi(is)(jpt))  {
                           int offsetJ = pawpot.offset(is)(jpt);
                           double Xij = SxXCFunctional::alphaHybrid
                                      * pawpot.coreX(is)(ipt,jpt);
                           for (int m = 0; m < nm; ++m)
                              VijR(offsetI + m, offsetJ + m) += Xij;
                        }
                     }
                  }
               }
            }
         } // LoopMPI
      }

#ifdef USE_LOOPMPI
      if (SxLoopMPI::nr () > 1) {
         for (int ia = 0; ia < structure.getNAtoms (is); ++ia) {
            for (int iSpin = 0; iSpin < nSpin; ++iSpin) {
               SxMatrix<Double> &VijR = Vij(iSpin)(is,ia);
               int sender = SxLoopMPI::whoseWork (structure.getIAtom (is, ia));
               SxLoopMPI::bcast (VijR, sender);
            }
         }
         eDoubleCountingPrime = SxLoopMPI::sum (eDoubleCountingPrime);
      }
#endif

      eDoubleCounting += eDoubleCountingPrime;
   } // end is loop
   STOP_TIMER (Loop3);


   CLOCK (PsStuff);

   if (hContrib & CalcHartreePS)  {
      // Ref. 1, Eq. 25, first term via Ref. 1, Eq. 26
      eHartreePS = TWO_PI * (rhoHartreeSoft.absSqr ()(noZero)
                             / gBasis.g2(noZero)).sum ();
      eDoubleCounting -= eHartreePS;
      eDoubleCounting += FOUR_PI * dot(rhoCoreG(noZero), rhoHartreeSoft(noZero)
                                                       / gBasis.g2(noZero)).re;
   } else if (hContrib & CalcNHatGSoft)  {
      eHartreePS = TWO_PI * (nHatpG.absSqr ()(noZero)
                             / gBasis.g2(noZero)).sum ();
      eDoubleCounting -= eHartreePS;
      eDoubleCounting += FOUR_PI * dot(rhoCoreG(noZero), nHatpG(noZero)
                                                       / gBasis.g2(noZero)).re;
   }
   sxprintf ("eHartreePS      = %18.12f H\n", eHartreePS);
   
   // Ref. 1, Eq. 27 in G-space
   PsiG vHatG(gBasis);
   vHatG(0) = vHatGZero; // see above
   vHatG(noZero) = FOUR_PI/gBasis.g2(noZero) * (nHatG - nHatpG)(noZero);

   if (hContrib & CalcHartreePS)  {
      // Ref. 1, Eq. 25, second term
      double eVHat = dot(rhoPSG, vHatG).re;
      sxprintf ("eVHat           = %18.12f H\n", eVHat);
      eHartreePS += eVHat;
      eDoubleCounting -= eVHat;
      eDoubleCounting += dot(rhoCoreG, vHatG).re;
   }

   if (hContrib & CalcU)  {
      // Ref. 1, Eq. 25, third term
      sxprintf ("eHartreeU       = %18.12f H\n", eHartreeU);
      eHartreePS += eHartreeU;
   }


   // compose total Hartree energy
   sxprintf ("eHartreeRadial  = %18.12f H\n", eHartreeRadial);
   eHartree = eHartreePS + eHartreeRadial + eDipole;

   // START_TIMER (PsStuff4);
   // doesn't take much time

   if (vExt.getSize () > 0)  {
      // --- external potential
      eExt = tr (rhoHartreeSoft.to (rBasis) * vExt);
      eDoubleCounting += tr(rhoCoreG.to (rBasis) * vExt);
      vPSR += vExt;
   } else {
      eExt = 0.;
   }

#ifdef USE_LOOPMPI
   SX_MPI_MASTER_ONLY  {
      vPSR = rBasis.symmetrize (vPSR);
   }
   SxLoopMPI::bcast (vPSR, 0);
#else
   vPSR = rBasis.symmetrize (vPSR);
#endif

   // STOP_TIMER (PsStuff4);
   // doesn't take much time


   // START_TIMER (PsStuff5);

   SX_MPI_MASTER_ONLY
   {
      RhoR vElStat(1);
      vElStat(0) = vPSR * HA2EV;
      SxRho(vElStat).writeRho ("vElStat-eV.sxb");
   }

   // STOP_TIMER (PsStuff5);
   // doesn't take much time



   START_TIMER (PsStuff1);

   vPS.resize (nSpin);
   // --- xc part of pseudo-density
   if (hContrib & CalcXcPS)  {
      if (pawpot.kjXcShape.getSize () > 0)  {

         START_TIMER (PsStuff2);

         PsiG rhoXcG(gBasis);
         rhoXcG.set (0.);

         for (int is = 0; is < structure.getNSpecies (); ++is)  {
            if (pawpot.kjXcShape(is).getSize () > 0)  {
               cout << "Using Kresse-Joubert formulation of xc for "
                    << pawpot.prettyName(is) << endl;
               double Qcore = -pawpot.nuclearCharge(is) / sqrt(FOUR_PI)
                            +  ((pawpot.rhoCoreAE(is) - pawpot.rhoCorePS(is))
                                * rad.radFunc(is).cub ()).integrate (rad.logDr(is));
               PsiG kjxcG = gBasis | pawpot.kjXcShape(is);
               for (int ia = 0; ia < structure.getNAtoms(is); ++ia)  {
                  double Q = Qrl(is)(ia)(0) - Qcore;
                  //cout << "Q=" << Q << endl;
                  PsiG T = gBasis.getPhaseFactors(is,ia);
                  rhoXcG.plus_assign_ax (Q, T * kjxcG);
               }
            }
         }
         rhoXcG /= nSpin;
         RhoR rhoXc(nSpin);
         rhoXc(0) = rBasis | rhoXcG;
         //cout << "nXcKj=" << structure.cell.volume * rhoXc(0).sum ()
         //        / rhoXc(0).getSize () << endl;
         if (nSpin == 2)
            rhoXc(1) = rhoXc(0) + rhoPSR(1);
         rhoXc(0) += rhoPSR(0);
         //cout << "nPs=" << structure.cell.volume * rhoPSR(0).sum ()
         //        / rhoPSR(0).getSize () << endl;
         cout << "nXc=" << structure.cell.volume * rhoXc(0).sum ()
                 / (PrecTauR)rhoXc(0).getSize () << endl;
         xcPtr->updateXC (rhoXc);

         STOP_TIMER (PsStuff2);

      } else {
         START_TIMER (PsStuff3);
         xcPtr->updateXC (rhoPSR);
         STOP_TIMER (PsStuff3);
      }
      {
         SxMeshR rhoCoreR = computeCorePS ();
         if (nSpin == 2) rhoCoreR *= 0.5;
         SX_LOOP(iSpin)
            eDoubleCounting -= dot(rhoPSR(iSpin) - rhoCoreR,
                                   xcPtr->vXc(iSpin)) * rBasis.dOmega;
      }
      eDoubleCounting += xcPtr->eXc;

      START_TIMER (PsStuff4);
      eXc = eXcRadial + xcPtr->eXc;
      STOP_TIMER (PsStuff4);

      START_TIMER (PsStuff5);
      // Ref. 1, Eq. 34, last term
      for (int iSpin = 0; iSpin < nSpin; ++iSpin)
         vPS(iSpin) = vPSR + xcPtr->vXc(iSpin);
      STOP_TIMER (PsStuff5);

   } else {
      START_TIMER (PsStuff6);
      eXc = eXcRadial;
      for (int iSpin = 0; iSpin < nSpin; ++iSpin) vPS(iSpin) = vPSR;
      STOP_TIMER (PsStuff6);
   }

   STOP_TIMER (PsStuff1);

   eCore = 0.;
   if (hContrib & CalcCore)
      eCore = (structure.atomInfo->nAtoms * potPtr->coreEnergy).sum ();
   eDoubleCounting -= eCore;

#ifdef USE_LOOPMPI
   {
      double eDCmaster = SxLoopMPI::bcast (eDoubleCounting, 0);
      if (eDCmaster != eDoubleCounting)  {
         cout << "eDC mismatch: " << (eDoubleCounting - eDCmaster) << endl;
      }
      eDoubleCounting = eDCmaster;
   }
#endif

   //cout << "eXc = " << eXc << endl;
   //cout << "eExt = " << eExt << endl;
   //cout << "eHartree = " << eHartree << endl;
   //cout << "eR0 = " << eR0 << endl;
}


PrecEnergy SxPAWHamiltonian::getEnergy (const SxPsiSet &pw,
                                        const SxFermi &fermi)
{
   // TODO: rPtr => SxPtr
   SX_CHECK (rPtr);
   SX_CHECK(dynamic_cast<const SxPWSet *>(&pw));
   compute (dynamic_cast<const SxPWSet &>(pw), fermi, false);
   return eTotal;
}

double SxPAWHamiltonian::getHarrisEnergy (const SxFermi &fermi)
{
   double eHarrisFoulkes = eDoubleCounting;
   cout << "eDoubleCounting = " << eDoubleCounting << endl;
   SX_LOOP2(iSpin,ik)
      eHarrisFoulkes += dot(fermi.eps(iSpin,ik),
                            fermi.focc(iSpin,ik)) * fermi.kpPtr->weights(ik);
   cout << "eBand = " << (eHarrisFoulkes - eDoubleCounting) << endl;
   return eHarrisFoulkes;
}

void SxPAWHamiltonian::compute (const SxPWSet &waves, const SxFermi &fermi,
                                bool computeNewRho)
{
   CLOCK(Timer::ComputeHam);
   SX_CHECK (potPtr);
   SX_CHECK (xcPtr);

   // compute new density
   if (computeNewRho) computeRho (fermi.focc, waves);

   sxprintf ("eHarrisFoulkes = %20.16f\n", getHarrisEnergy (fermi));

   // --- Hartree & xc parts
   computeRhoTerms (pawRho.Dij, pawRho.pwRho.rhoR);


   if (exchangePtr && (hContrib & CalcValExch))  {
      double eX = exchangePtr->compute (waves.getThis (), fermi.focc,
                                        pawRho.Dij);
      cout << "eX = " << eX << endl;
      eXc += SxXCFunctional::alphaHybrid * eX;
      eDoubleCounting -= SxXCFunctional::alphaHybrid * eX;
   }


   // kinetic energy
   double eKin = 0.;
   // --- pseudo-kinetic enery
   int nSpin = waves.getNSpin ();


   if (hContrib & CalcKinPS)  {
      SX_NEW_LOOP (waves);
      for (int ik = 0; ik < waves.getNk (); ++ik)  {
         for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
            // Ref. 1, Eq. 19, kinetic energy
            SX_MPI_LEVEL("waves-k");
            if (SxLoopMPI::myWork(ik)) {
               double weight = fermi.kpPtr->weights(ik);
               const SxGBasis &gk = waves.getGkBasis ()(ik);
               for (int iState = 0; iState < waves.getNStates(ik); ++iState)  {
                  PsiG psiG = gk | waves(iState,iSpin,ik);
                  eKin += 0.5 * weight * fermi.focc(iState,iSpin,ik)
                                   * (psiG.absSqr () * gk.g2).sum ();
               }
            } // LoopMPI
         }
      }
      SX_MPI_SOURCE ("waves-k", TaskGroupMaster);
      SX_MPI_TARGET (TopLevel, TaskGroupAll);
      eKin = SxLoopMPI::sum(eKin);
   }

   bool corValEx = exchangePtr && (hContrib & CalcCorValExch);
   bool kin = (hContrib & CalcKinRad );

   if (kin || corValEx)  {
      // --- 1-center correction of kinetic energy
      const SxPAWPot &pot = *potPtr;
      for (int is = 0; is < structure.getNSpecies (); ++is)  {
         // Ref. 1, Eq. 20/21 with Eq. 52
         // \sum_ij Dij * ( <phiAE_j| -1/2 nabla^2 | phiAE_i>
         //                -<phiPS_j| -1/2 nabla^2 | phiPS_i>)
         int nProjType = pot.getNProjType (is);
         for (int ipt = 0; ipt < nProjType; ++ipt)  {
            int l = pot.lPhi(is)(ipt);
            int offsetI = pot.offset(is)(ipt) + l;
            for (int jpt = 0; jpt < nProjType; ++jpt)  {
               if (l == pot.lPhi(is)(jpt))  {
                  int offsetJ = pot.offset(is)(jpt) + l;
                  double mSumD = 0.;
                  for (int iSpin = 0; iSpin < nSpin; ++iSpin) {
                     for (int ia = 0; ia < structure.getNAtoms(is); ++ia)  {
                        for (int m = -l; m <= l; ++m)  {
                           mSumD += pawRho.Dij(iSpin)(is,ia)(offsetI + m,
                                                             offsetJ + m);
                        }
                     }
                  }
                  if (kin)
                     eKin += mSumD * pot.deltaKin(is)(jpt, ipt);
                  if (corValEx)
                     eXc += SxXCFunctional::alphaHybrid
                          * mSumD * pot.coreX(is)(jpt, ipt);
               }
            }
         }
      }
   }


   sxprintf ("eKin      = % 19.12f H\n", eKin);
   sxprintf ("eHartree  = % 19.12f H\n", eHartree);
   sxprintf ("eXc       = % 19.12f H\n", eXc);
   sxprintf ("eBar      = % 19.12f H\n", eBar);
   sxprintf ("eCore     = % 19.12f H\n", eCore);
   if (vExt.getSize () > 0)
      sxprintf ("eExt      = %19.12f H\n", eExt);
   eTotal = eKin + eHartree + eXc + eBar - eCore + eExt;
   sxprintf ("eTot(Val) = % 19.12f H\n", eTotal);

}

SxRho SxPAWHamiltonian::computeRhoPSCore (const SxRho &rhoPS, int nSpin)
{
   SX_CHECK(&rhoPS);
   SX_CHECK(rPtr);

   SxRho res;
   res = rhoPS.getCopy();
   // --- add pseudo-core densities
   SxMeshR coresR = computeCorePS ();
   for (int iSpin = 0; iSpin < nSpin; ++iSpin)
      res(iSpin).plus_assign_ax (1. / nSpin, coresR);

   return res;
}


SxAtomicStructure SxPAWHamiltonian::computeNumF1 (const SxPWSet &waves, const SxFermi &fermi)
{
   SX_CHECK(&structure);
   SX_CHECK (potPtr);
   SX_CHECK (xcPtr);
   SX_CHECK(rPtr);

   SX_CHECK(structure.cell.symGroupPtr->getSize() == 1);
   
   // compute 1-center density matrix
   cout << "Compute Dij" << endl;
   SxArray<SxRadMat> Dij = computeDij (waves, fermi.focc);

   // --- compute pseudo-density valence
   cout << "Compute rho_ps" << endl;
   int nSpin = waves.getNSpin ();
   SxRho rhoPSVal(*rPtr, nSpin, 0. ); // 0. = no renormalization
   rhoPSVal.computeRho (fermi.focc, waves);
   const SxRBasis &rBasis = rhoPSVal.rhoR(0).getBasis<SxRBasis> ();
   // debug exlude F13
   // SxRho rhoPS = computeRhoPSCore(rhoPSVal, nSpin); 

   // declare force Vector
   SxAtomicStructure numForces = structure.getNewStr ().set(Coord(0.0,0.0,0.0));
   // Save structure
   SxGBasis &gBasis = const_cast <SxGBasis &> (rBasis.getGBasis());
   SxAtomicStructure orgStructure;
   orgStructure.copy(structure);

   //change structure +h
   int ns = structure.getNSpecies ();
   for(int is = 0; is < ns; is++)   {
      int na = structure.getNAtoms (is);
      for(int ia = 0; ia < na; ia++)   {
         double h = 1e-6;
         double hInv = 1e6;
         for(int iDir = 0; iDir < 3; iDir++)   {
            //change structure +h
            structure.ref(is,ia)(iDir) +=  h;
            // Structure changed !!!! Update Tau
            gBasis.changeTau(structure);
            // debug include f13
            SxRho rhoPS = computeRhoPSCore(rhoPSVal, nSpin);
         
            computeRhoTerms(Dij, rhoPS.rhoR);
            //cout << is << ia << "+h: " << eHartree << " " << eBar << " " << eXc << endl; 
            numForces.ref(is,ia)(iDir) += eHartree + eBar + eXc;
         
            // reset structure
            structure.copy(orgStructure);

            //change structure -h
            structure.ref(is,ia)(iDir) -=  h;
            // Structure changed !!!! Update Tau
            gBasis.changeTau(structure);
            // debug include f13
            rhoPS = computeRhoPSCore(rhoPSVal, nSpin);
         
            computeRhoTerms(Dij, rhoPS.rhoR);
            //cout << is << ia << "-h: " << eHartree << " " << eBar << " " << eXc << endl; 
            numForces.ref(is,ia)(iDir) -= eHartree + eBar + eXc;

            // divide difference
            numForces.ref(is,ia)(2) *= 0.5 * hInv;

            // reset gBasis,structure and dEdQrl
            structure.copy(orgStructure);
            gBasis.changeTau(structure);
         }
      }
   }

   for(int is = 0; is < ns; is++)   {
      int na = structure.getNAtoms (is);
      for(int ia = 0; ia < na; ia++)   {
         cout << "numForce f1 " << ia << ": " << numForces.ref(is,ia) << endl;
      }
   }

   return numForces;
}

void SxPAWHamiltonian::computeNumF1F2 (const SxPW &waves, const SxFermi &fermi)
{
   SX_CHECK(&structure);
   SX_CHECK(structure.cell.symGroupPtr->getSize() == 1);

   SxGBasis &gBasis = const_cast <SxGBasis &> (rPtr->getGBasis());
   SxGkBasis &gkBasis = const_cast <SxGkBasis &> (waves.getGkBasis());

   // declare force Vector
   SxAtomicStructure numForces = structure.getNewStr ().set(Coord(0.0,0.0,0.0));
   // Save structure
   SxAtomicStructure orgStructure;
   orgStructure.copy(structure);

   int ns = structure.getNSpecies ();
   for(int is = 0; is < ns; is++)   {
      int na = structure.getNAtoms (is);
      for(int ia = 0; ia < na; ia++)   {
         double h = 1e-6;
         double hInv = 1e6;
         for(int iDir = 0; iDir < 3; iDir++)   {
            //change structure +h
            structure.ref(is,ia)(iDir) += h;
            // structure changed !!! Update gkBasis
            gkBasis.changeTau(structure);
            gBasis.changeTau(structure);
            compute(waves,fermi);
            numForces.ref(is,ia)(iDir) += eTotal;

            //reset Structure
            structure.copy(orgStructure) ;

            //change structure -h
            structure.ref(is,ia)(iDir) -= h;
            // structure changed !!! Update gkBasis
            gBasis.changeTau(structure);
            gkBasis.changeTau(structure);
            compute(waves,fermi);
            numForces.ref(is,ia)(iDir) -= eTotal;

            // divide difference
            numForces.ref(is,ia)(iDir) *= 0.5 * hInv;

            //reset Structure
            structure.copy(orgStructure);
            gBasis.changeTau(structure);
            gkBasis.changeTau(structure);
         }
      }
   }

   for(int is = 0; is < ns; is++)   {
      int na = structure.getNAtoms (is);
      for(int ia = 0; ia < na; ia++)   {
         cout << "numForce f1+f2 " << ia << ": " << numForces.ref(is,ia) << endl;
      }
   }
}

double SxPAWHamiltonian::computeU (SxArray<SxArray<SxVector<Double> > > *dEdQrl)
{
   CLOCK(Timer::ComputeU);
   SX_CHECK (potPtr);
   const SxPAWPot &pot = *potPtr;
   
   // --- for neighbor search
   SxGrid grid (structure, 10);
   SxNeighbors neighbors;
   int neighborParams = SxNeighbors::StoreRel
                      | SxNeighbors::IncludeZeroDistance;
   // note: for rcut=13, 1 bohr wide Gaussians behave like point multipoles
   // within 1e-16
   // we assume that multipole Gaussians are always below 1 bohr wide
   double rcut = 13. * rSoft;

   double U = 0.;
   const SxYlm::SxClebschTable &cg = pot.clebschGordan;
   SxGaussIJ gaussIJ;

   // --- loop over atoms (index i => R in Bloechl notation)
   for (int is = 0; is < structure.getNSpecies (); ++is)  {
      int lmaxI = pot.lMaxRho(is);
      for (int ia = 0; ia < structure.getNAtoms(is); ++ia)  {
         if(SxLoopMPI::myWork(structure.getIAtom (is, ia)))
         {
            // compute neighbors within rcut
            neighbors.compute (grid, structure, structure.getAtom(is,ia),
                  rcut, neighborParams);

            // --- loop over neighbors j (R' in Bloechl notation)
            for (int js=0; js < structure.getNSpecies (); js++)  { // --- species j
               int lmaxJ = pot.lMaxRho(js);
               double a2Hard = sqr(pot.rc(is)) + sqr(pot.rc(js));
               double a2Soft = 2. * sqr(rSoft); // sqr(1.) + sqr(1.);
               for (int ja = 0; ja < neighbors.relPositions.getNAtoms(js); ++ja)
               {
                  // i => R
                  // j => R'
                  // r = R'-R
                  Coord r = neighbors.relPositions.getAtom(js, ja);
                  // Ref. 1 Eq. (28): kernel
                  gaussIJ.setDelta (r, a2Hard, a2Soft, lmaxI + lmaxJ);
                  SxMatrix<Double> delta = gaussIJ.compute (lmaxI, lmaxJ, cg);
                  // sum over moments of j=R'
                  int jaGlobal = neighbors.relPositions.getIAtom(js,ja);
                  int jRefAtom = neighbors.idx(jaGlobal)
                               - structure.atomInfo->offset(js);
                  SxVector<Double> sumJ = delta ^ Qrl(js)(jRefAtom);

                  // Ref. 36, 2nd term (dU/d Qrl)
                  if (dEdQrl) (*dEdQrl)(is)(ia) += sumJ;
                  // Ref. 1 Eq. (28)
                  U += 0.5 * (Qrl(is)(ia) * sumJ).sum ();
               }
            }
         } // LoopMPI
      }
   }

//   eDoubleCounting -= U;
//   U = SxLoopMPI::sum(U);
   
   U = SxLoopMPI::sum(U);
   eDoubleCounting -= U;
   
   return U;
}

void SxPAWHamiltonian::setupPSRefG(const SxGBasis &gBasis)
{
   SX_CHECK(potPtr);
   // pot must have a radial basis
   SX_CHECK(potPtr->getBasisPtr ());
   int nSpecies = potPtr->getNSpecies ();
   vBarG.resize (nSpecies);
   rhoCorePSG.resize (nSpecies);
   for (int is = 0; is < nSpecies; ++is)  {
      vBarG(is)      = (gBasis | potPtr->vBar(is));
      rhoCorePSG(is) = (gBasis | potPtr->rhoCorePS(is));
   }
}

PsiG SxPAWHamiltonian::apply (const PsiG &psi) const
{
   CLOCK (Timer::PAW_H_psi);
   SX_CHECK (psi.handle);
   int iSpin = psi.handle->auxData.iSpin;
   SX_CHECK_VAR (psi.nCols () > 0, psi.nCols ());
   int nStates = (int)psi.nCols ();
   const SxGBasis *gkPtr = dynamic_cast<const SxGBasis *> (psi.getBasisPtr ());
   if (!gkPtr)  {
      const SxPAWBasis *paw 
         = dynamic_cast<const SxPAWBasis *> (psi.getBasisPtr ());
      SX_CHECK (paw);
      gkPtr = paw->gBasis.getPtr ();
   }
   SX_CHECK (gkPtr);
   const SxGBasis &gk = *gkPtr;
   SX_CHECK (rPtr);
   const SxRBasis &R = *rPtr;
   SX_CHECK (pBasis);
   const SxPartialWaveBasis &p = *pBasis;

   // Ref. 1, Eq. 40, |pi>(...)<pj term
   START_TIMER(Timer::PAW_H_nl);
   PsiG res = (gk | p) * (p | (Vij(iSpin) ^ (p | psi) ));
   STOP_TIMER(Timer::PAW_H_nl);

#ifdef USE_FFT2d1d
   bool direct = true;
#else
   bool direct = false;
#endif
   if (direct && (hContrib & CalcVPS))  {
      CLOCK(Timer::PAW_H_loc);
      res += gk.convolute (psi, vPS(iSpin));
   }

   for (int iState = 0; iState < nStates; ++iState)  {
      PsiG psiG = gk | psi.colRef(iState);
      if (hContrib & CalcKinPS)  {
         // pseudo-kinetic energy (Ref. 1, Eq. 40, first term)
         res.colRef(iState).plus_assign_ax(0.5, gk.g2 * psiG);
      }
      if ((hContrib & CalcVPS) && !direct)  {
         CLOCK(Timer::PAW_H_loc);
         // local potential       (Ref. 1, Eq. 40, second term)
         res.colRef(iState) += gk | (vPS(iSpin) * (R | psiG));
      }

   }
   if (exchangePtr && (hContrib & CalcValExch))  {
      PsiG xPsi = exchangePtr->apply (psi);
      res.plus_assign_ax (SxXCFunctional::alphaHybrid, xPsi);
   }

   res.handle->auxData  = psi.handle->auxData;
   res.setBasis (&gk);
   return res;
}

SxDiracVec<TPrecCoeffG::TReal> SxPAWHamiltonian::preconditioner (const PsiG &dPsi) const
{
   const SxGBasis *Gk = dynamic_cast<const SxGBasis *>(dPsi.getBasisPtr());
   if (!Gk)  {
      const SxPAWBasis* pawBasis
         = dynamic_cast<const SxPAWBasis *>(dPsi.getBasisPtr ());
      SX_CHECK(pawBasis);
      Gk = pawBasis->gBasis.getPtr ();
   }
   SX_CHECK (Gk);

   double kin = (*Gk | dPsi).laplacian();
   /* --- beautiful and slow
   SxDiracVec<TPrecCoeffG::TReal> x, x2, x3, x4, n, K;
   x  = Gk->g2 / kin;
   x2 = x.sqr();
   x3 = x.cub();
   x4 = x2.sqr();
   n  = 27. + 18.*x + 12.*x2 + 8.*x3;
   K  = n / (n + 16.*x4);
   K /= kin; // make K approach 1/(G^2 - kin)
   */
   // --- less beautiful, but much faster
   double kinInv = 1. / kin;
   SxDiracVec<TPrecCoeffG::TReal> K;
   K.resize (Gk->ng);
   K.setBasis (Gk);
#ifdef USE_OPENMP
#pragma omp parallel for
#endif
   for (ssize_t ig = 0; ig < Gk->ng ; ++ig)   {
      double x_ = Gk->g2(ig) * kinInv;
      double x2_ = x_ * x_;
      double x3_ = x_ * x2_;
      double x4_ = x2_ * x2_;
      double n_ = 27. + 18. * x_ + 12. * x2_ + 8. * x3_;
      K(ig) = n_ / (kin * (n_ + 16. * x4_));
   }
   return K;
}

SxDiracMat<Complex16>
SxPAWHamiltonian::getHnm (const PsiGI &waveBlock) const
{
   ssize_t nStates = waveBlock.nCols ();
   int bSize = 64;
   if (waveBlock.getSize () < 1000000)  {
      PsiGI HPsi = apply(waveBlock);
      return waveBlock.overlap (HPsi, HPsi.nRows ());
   }
   SxDiracMat<Complex16> Hnm(nStates, nStates);
   ssize_t ngPAW = waveBlock.nRows ();
   for (ssize_t i = 0; i < nStates; i += bSize)  {
      ssize_t ni = i + bSize < nStates ? bSize : (nStates - i);
      SxDiracMat<Complex16> blockI = waveBlock.getBlock (0, i, ngPAW, ni);
      blockI.setBasis (waveBlock.getBasisPtr ());
      PsiGI HPsi = apply(blockI);
      Hnm.setBlock (waveBlock.overlap (HPsi, HPsi.nRows ()), 0, i);
   }
   return Hnm;
}

void SxPAWHamiltonian::computeForces (const SxPWSet &waves, SxFermi &fermi)
{
   CLOCK(Timer::Force);
   bool printForceParts = false;
   SX_CHECK(rPtr);
   SX_CHECK(potPtr);

   forces = structure.getNewStr ().set(Coord(0.0,0.0,0.0));
   const SxGBasis &gBasis      = rPtr->getGBasis();
   const SxGkBasis &gk         = waves.getGkBasis();

   SxIdx noZero(1,gBasis.ng - 1);
   const SxDiracMat<Double> G  = SxDiracMat<Double>(gBasis.gVec);
   const SxDiracVec<Double> &G2 = gBasis.g2;

   int nSpin = waves.getNSpin();
   const SxRho &rhoPSR = pawRho.pwRho;

   PsiG rhoPSG = gBasis | ((nSpin == 1) ? rhoPSR(0)
                                        : rhoPSR(0) + rhoPSR(1));

   // |G|^l * Ylm(G)
   START_TIMER (TimeYlmGl);
   int lMaxRho = potPtr->lMaxRho.maxval ();
   SxDiracMat<Double> YlmGl(gBasis.ng, sqr(lMaxRho + 1));
   YlmGl.setBasis (gBasis);
   for (int lm = 0, l = 0; l <= lMaxRho; ++l) {
      SxDiracVec<Double> gl = pow (gBasis.g2, 0.5 * l);
      for (int m = -l; m <= l; ++m,++lm)
         YlmGl.colRef(lm) <<= gBasis.getYlm (l,m) * gl;
   }
   STOP_TIMER (TimeYlmGl);

   double rcp2 = sqr(rSoft);

   // khr:  the calculations of f11...f14 do not need to be parallelized

   SxAtomicStructure f11 = structure.getNewStr ().set(Coord(0.0,0.0,0.0));
   SxAtomicStructure f12 = structure.getNewStr ().set(Coord(0.0,0.0,0.0));
   SxAtomicStructure f13 = structure.getNewStr ().set(Coord(0.0,0.0,0.0));

   START_TIMER (Timer::F11);
   // --- calculate vHartreeSoft
   PsiG rhoHatPG (gBasis);
   rhoHatPG.set(0.0);
   if (hContrib & CalcNHatGHard) {
      for(int is = 0; is < structure.getNSpecies (); is++)
         for(int ia = 0; ia < structure.getNAtoms (is); ia++)
            rhoHatPG += computeRhoHatG (is,ia,YlmGl, rcp2);
   }

   PsiG vHartreeSoft(gBasis);
   vHartreeSoft(0)        = 0.;
   vHartreeSoft(noZero) <<= FOUR_PI * rhoHatPG(noZero) / G2(noZero);
   if (hContrib & CalcHartree)
      vHartreeSoft(noZero) += FOUR_PI * rhoPSG(noZero) / G2(noZero);

   if (dipoleCorr && (hContrib & CalcDipoleC))  {
      // --- add dipole field to potential
      SxMeshR vDipole(*rPtr);
      vDipole.set (0.);
      dipoleCorr->correctPotential (&vDipole);
      vHartreeSoft += vDipole.to (gBasis);
   }

   if (vExt.getSize () > 0)  {
      // add external field to potential
      vHartreeSoft += vExt.to (gBasis);
   }
   STOP_TIMER (Timer::F11);

   SxMeshG vPSG = (gBasis | ((nSpin == 1) ? vPS(0)
                                          : vPS(0) + vPS(1)));

   // --- calculate f11 f12 f13 (Ref. 1, Eq. (50))
   SX_LOOP2(is,ia) {
      PsiG rhoHatRPG = computeRhoHatG ((int)is,(int)ia,YlmGl, rcp2);
      PsiG T = gBasis.getPhaseFactors ((int)is,(int)ia);
      // --- Ref.1 Eq (50) 1st term,
      {  CLOCK (Timer::F11);
         f11.ref(is,ia) -= ((I * vHartreeSoft.conj() * rhoHatRPG) ^ G)
                           .real ().toVector3 ();
      }
      // --- Ref.1 Eq (50) 2nd term,
      {  CLOCK (Timer::F12);
         PsiG rhoHatRG(gBasis), v(gBasis);
         double rc2 = sqr(potPtr->rc(is));
         // NHatG part
         if (hContrib & CalcNHatGHard)
            rhoHatRG = computeRhoHatG ((int)is, (int)ia, YlmGl, rc2);
         else
            rhoHatRG.set(0.0);
         if (!(hContrib & CalcNHatGSoft)) rhoHatRPG.set (0.);

         v(0)        = 0.;
         v(noZero) <<= FOUR_PI * (rhoHatRG(noZero)-rhoHatRPG(noZero)) / G2(noZero);
         // vBar part
         if (hContrib & CalcBar) v += vBarG(is) * T;
         f12.ref(is,ia) -= ((-I * rhoPSG * v.conj()) ^ G).real ().toVector3 ();
      }
      // Ref.1 Eq (50) 3rd term, contribution due to pseudopot and core density
      {  CLOCK (Timer::F13);
         f13.ref(is,ia) += 1./nSpin * ((rhoCorePSG(is) * T * vPSG.conj()) ^ G)
                                      .imag ().toVector3 ();
      }
   }
   if (printForceParts) cout << "F11" << endl << f11 << endl;
   if (printForceParts) cout << "F12" << endl << f12 << endl;
   if (printForceParts) cout << "F13" << endl << f13 << endl;

   // Ref.1 Eq (50) 4th term, U-Term
   START_TIMER (Timer::F14);
   SxAtomicStructure f14 = structure.getNewStr ().set(Coord(0.0,0.0,0.0));
   if (hContrib & CalcU)  f14 -= computeGradU ();
   STOP_TIMER (Timer::F14);
   if (printForceParts) cout << "F14" << endl << f14 << endl;

   forces = f11 + f12 + f13 + f14;

   if (printForceParts) {
      SxAtomicStructure numF1 = computeNumF1(waves, fermi);
      cout << "NumDiff F1" << endl << forces - numF1 << endl;
   } 


   // Ref.1 Eq (58)
   // khr:  parallelized over the k points
   START_TIMER (Timer::F3);
   SxAtomicStructure f3 = structure.getNewStr ().set(Coord(0.0,0.0,0.0));
   int nStates = waves.getNStates ();
   SxNArray<SxRadMat,2> gradDij(3,nSpin);
   for (int iDir = 0; iDir < 3; iDir++) {
      for (int iSpin = 0; iSpin < nSpin; iSpin++)  {
         gradDij(iDir,iSpin).resize (structure.atomInfo, *potPtr);
         gradDij(iDir,iSpin).set (0.0);
      }
   }

   // k loop, parallelize
   for (int ik = 0; ik < waves.getNk (); ++ik)  {
      SX_MPI_LEVEL("waves-k");
      if (!SxLoopMPI::myWork(ik)) continue;
      for (int iSpin = 0; iSpin < waves.getNSpin (); ++iSpin) {
         SxDiracVec<Complex16> P = *pBasis | waves(iSpin,ik);
         SxArray<SxDiracMat<Complex16> > gradP
         = projBasis->gradProject (gk(ik) | waves(iSpin,ik));
         SxArray<SxRadMat> gradDijBAK =
               computeUnSymmGradDij (fermi, iSpin, ik, P, gradP);
         SX_LOOP3(iDir,is,ia) {
            gradDij(iDir,iSpin)(is,ia) += gradDijBAK(iDir)(is,ia);
         }

         // --- setup Hmn
         SxDiracMat<Complex16> Hnm;
         if (   dynamic_cast<const SxPW*>(&waves) 
             || dynamic_cast<const SxPAWSet*>(&waves))
         {
            Hnm = getHnm (waves(iSpin, ik));
         } else {
            Hnm.reformat (nStates,nStates);
            // Hmn is a hermitian matrix
            for (int jState = 0; jState < nStates; ++jState) {
               PsiG HPsiJ = apply(waves(jState,iSpin,ik));
               for (int iState = jState; iState < nStates; ++iState) {
                  Hnm(iState,jState) = dot(gk(ik)|waves(iState,iSpin,ik),
                                           HPsiJ);
                  Hnm(jState,iState) = Hnm(iState,jState).conj();
               }
            }
         }

         SX_LOOP3(iDir, is, ia) {
            double sum = 0;
            SX_LOOP2(iState, jState) {
               double focc = 0.5 * (fermi.focc((int)iState,iSpin,ik) 
                                    + fermi.focc((int)jState,iSpin,ik));
               sum += (focc * Hnm(iState, jState)
                      * pawGradDot ((int)is, (int)ia, (int)jState, (int)iState,
                                    P, gradP(iDir))).re;
            }
            f3.ref(is,ia)(iDir) -= gk.weights(ik) * sum;
         }
      }
   }

   if (SxLoopMPI::nr () > 1) {
      SX_MPI_SOURCE("waves-k", TaskGroupMaster);
      SX_MPI_TARGET(TopLevel, TaskGroupAll);
      // --- gradDij
      for(int iSpin = 0; iSpin < nSpin; iSpin++) {
         for (int iDir = 0; iDir < 3; iDir++) {
            SX_LOOP2 (is,ia) {
               // TODO:  reduce the number of MPI calls
               SxLoopMPI::sum (gradDij(iDir,iSpin)(is,ia));
            }
         }
      }
      // f3
      SxLoopMPI::sum (f3);
   }

   STOP_TIMER (Timer::F3);
   if (printForceParts) cout << "F3" << endl << f3 << endl;

   forces += f3;

   // Ref.1 Eq (57)
   // khr:  does not need to be parallelized
   START_TIMER (Timer::F2);
   gradDij = symmetrizeGradDij(gradDij);
   SxAtomicStructure f2 = structure.getNewStr ().set(Coord(0.0,0.0,0.0));
   SX_LOOP4 (iDir, iSpin, is, ia)  {
      double sum = 0.0;
      SX_LOOP2 (jpl,ipl) {
         sum += gradDij(iDir,iSpin)(is,ia)(ipl,jpl)
              * Vij(iSpin)(is,ia)(ipl,jpl);
      }
      f2.ref(is,ia)(iDir) += sum;
   }
   STOP_TIMER (Timer::F2);

   forces += f2;
   if (printForceParts) cout << "F2" << endl << f2 << endl;


   // force is - dE/dR
   forces *= -1.0;

   // --- final symmetrize forces
   SxForceSym forceSymmetrizer;
   forceSymmetrizer.setup(structure);
   forces = forceSymmetrizer * forces;
   if (SxLoopMPI::nr () > 1)  {
      // make sure that forces are consistent
      SxLoopMPI::bcast (forces.coords, 0);
   }

   cout << SEPARATOR;
   cout << "Total Forces:" << endl;
   cout << SEPARATOR;
   SX_LOOP2(is,ia) {
      cout << "Species: " << is << "\t"
           << "Atom: "<< ia << "\t"
           << forces.getAtom(is,ia) << "\t"
           << forces.getAtom(is,ia).norm() << endl;
   }
}


SxAtomicStructure SxPAWHamiltonian::computeNumGradU ()
{

   SxAtomicStructure res = structure.getNewStr ().set(Coord(0.0,0.0,0.0));

   //change structure +h
   int ns = structure.getNSpecies ();
   for(int is = 0; is < ns; is++)   {
      int na = structure.getNAtoms (is);
      for(int ia = 0; ia < na; ia++)   {
         double h = 1e-6;
         double hInv = 1e6;
         //change structure +h
         SxAtomicStructure orgStructure;
         orgStructure.copy(structure);
         structure.ref(is,ia) +=  Coord(0.0,0.0,h);
         res.ref(is,ia)(2) += computeU(NULL);
         structure.copy(orgStructure);

         //change structure -h
         structure.ref(is,ia) -=  Coord(0.0,0.0,h);
         res.ref(is,ia)(2) -= computeU(NULL);
         structure.copy(orgStructure);

         // divide difference
         res.ref(is,ia)(2) *= 0.5 * hInv;

      }
   }

   return res;
}

SxAtomicStructure SxPAWHamiltonian::computeGradU ()   
{
   // --- computes last summand Ref 1. eq (50)
   CLOCK(Timer::ComputeGradU);
   SX_CHECK (potPtr);
   const SxPAWPot &pot = *potPtr;

   // --- for neighbor search
   SxGrid grid (structure, 10);
   SxNeighbors neighbors;
   // note: for rcut=13, 1 bohr wide Gaussians behave like point multipoles
   // within 1e-16
   // we assume that multipole Gaussians are always below 1 bohr wide
   double rcut = 13.0 * rSoft;

   SxAtomicStructure GradU = structure.getNewStr ().set(Coord(0.0,0.0,0.0));

   const SxYlm::SxClebschTable &cg = pot.clebschGordan;
   SxGaussIJ gaussIJ;

   // --- loop over atoms (index i => R in Bloechl notation)
   for (int is = 0; is < structure.getNSpecies (); is++)   {
      int lmaxI = pot.lMaxRho(is);
      for (int ia = 0; ia < structure.getNAtoms(is); ia++)   {
         // compute neighbors within rcut
         neighbors.compute (grid, structure, structure.getAtom(is,ia),
                            rcut, SxNeighbors::StoreRel);
         // --- loop over neighbors j (R' in Bloechl notation)
         for (int js = 0; js < structure.getNSpecies (); js++)   {
            int lmaxJ = pot.lMaxRho(js);
            double a2Hard = sqr(pot.rc(is)) + sqr(pot.rc(js));
            double a2Soft = 2.0 * sqr(rSoft);
            for (int ja = 0; ja < neighbors.relPositions.getNAtoms(js); ja++)
            {
               // i => R
               // j => R'
               // r = R' - R
               Coord r = neighbors.relPositions.getAtom(js, ja);
               int jaGlobal = neighbors.relPositions.getIAtom(js,ja);
               int jRefAtom = neighbors.idx(jaGlobal)
                            - structure.atomInfo->offset(js);
               gaussIJ.setDelta (r, a2Hard, a2Soft, lmaxI + lmaxJ + 1);
               gaussIJ.compute (lmaxI, lmaxJ, cg, 1); // prepare forces

               Coord sum(0., 0., 0.);
               SX_LOOP2 (lm2, lm1) {
                  sum += Qrl(is)(ia)(lm1)
                       * Qrl(js)(jRefAtom)(lm2)
                       * gaussIJ.getForce ((int)lm1, (int)lm2, cg);
               }
               sum *= 0.5;
               GradU.ref(is,ia) += sum;
               GradU.ref(js,jRefAtom) -=sum;
            }
         }
      }
   }
   return GradU;
}

PsiG SxPAWHamiltonian::computeRhoHatG (int is, int ia, SxDiracMat<Double> &YlmGl, double rc2)
{  
   double rNorm = FOUR_PI / sqrt(structure.cell.volume);
   SX_CHECK_NUM (rNorm);

   SX_CHECK(potPtr);
  
   const SxPAWPot &pawpot = *potPtr; 
   const SxGBasis &gBasis = YlmGl.getBasis<SxGBasis> ();

   PsiG nHatG (gBasis);
   nHatG.set(0.0);

   int lMax = pawpot.lMaxRho (is);
   SxDiracVec<Double> gauss = exp ( -(0.25 * rc2) * gBasis.g2);
  
   double dFac = 1.;
   PsiG T = gBasis.getPhaseFactors (is,ia);
   for (int l = 0; l <= lMax; l++, dFac *= 2 * l + 1)   {
      for (int m = -l; m <= l; m++)   {
         int lm = SxYlm::combineLm(l,m);
         double N = rNorm * SxYlm::getYlmNormFactor(l,m) / dFac;
         double weight = N * Qrl(is)(ia)(lm);
         // i^l factor
         SxComplex16 il = 1.;
                       if ((l & 3) == 1) il = I; 
                  else if ((l & 3) == 2) il = -1.;
                  else if ((l & 3) == 3) il = -I;
         PsiG grl = (YlmGl.colRef(lm) * gauss) * T;
         nHatG.plus_assign_ax(il * weight,grl);
      }
   }

   return nHatG;
}

SxArray<SxRadMat> SxPAWHamiltonian::computeUnSymmGradDij (const SxFermi &fermi,
                                                          const int iSpin,
                                                          const int ik,
                                                          SxDiracVec<Complex16> &P,
                                                          SxArray<SxDiracMat<Complex16> > &gradP)
{
   CLOCK (Timer::GradDij);
   SX_CHECK (fermi.kpPtr);
   SX_CHECK (potPtr);
   const SxPAWPot &pot = *potPtr;
   int nStates = fermi.getNStates ();
   SxArray<SxRadMat> res(3);
   for (int iDir = 0; iDir < 3; iDir++)   {
      res(iDir).resize (structure.atomInfo, pot);
      res(iDir).set (0.0);
   }

   double weight = fermi.kpPtr->weights(ik);
   for (int iDir = 0; iDir < 3; iDir++)   {
      int offset = 0;
      for (int is = 0; is < structure.getNSpecies (); ++is)  {
         int nProjLocal = pot.getNProj (is);

         // --- set up i^l 
         SxVector<Complex16> il(nProjLocal); // i^(-l)  (:ipl)
         for (int ipt = 0, ipl = 0; ipt < pot.lPhi(is).getSize (); ++ipt)  {
            int l = pot.lPhi(is)(ipt);
            SxComplex16 c = (l & 1) ? I : SxComplex16(1.);
            if (!(l & 2)) c=-c;
            for (int m = -l; m <=l; ++m, ++ipl)  {
               il(ipl) = c;
            }
         }

         for (int ia = 0; ia < structure.getNAtoms(is); ++ia, offset+=nProjLocal)  {
            SxMatrix<Double> &D = res(iDir)(is,ia);
            for (int ipl = 0; ipl < nProjLocal; ++ipl)  {
               for (int jpl = 0; jpl < nProjLocal; ++jpl)  {
                  SxComplex16 phase = il(ipl) * il(jpl).conj ();
                  for (int iState = 0; iState < nStates; ++iState)  {
                     // ref. 1, Eq. (51)
                     // + projector i^l phase-correction (Ref. 2)
                     // + use k-dependent weights
                     // + exploit k/-k symmetry (=> .re)

                     // Somehow a -1 is in the derivative of Dij  
                     D(ipl, jpl) += -1.0 * weight * fermi.focc(iState,iSpin,ik)
                        * (( gradP(iDir)(offset+ipl,iState).conj() * P(offset+jpl,iState)
                                 +   P(offset+ipl,iState).conj() * gradP(iDir)(offset+jpl,iState)) * phase
                          ).re;
                  }
               }
            }
         }
      }
   }
   return res;
}

SxNArray<SxRadMat,2> SxPAWHamiltonian::computeGradDij (const SxPWSet &waves,
                                                       const SxFermi &fermi)
{
   CLOCK (Timer::GradDij);
   SX_CHECK (fermi.kpPtr);
   SX_CHECK (potPtr);
   const SxPAWPot &pot = *potPtr;
   int nSpin = waves.getNSpin ();
   int nStates = waves.getNStates ();
   SxNArray<SxRadMat,2> res(3,nSpin);
   for (int iDir = 0; iDir < 3; iDir++)   {
      for (int iSpin = 0; iSpin < nSpin; iSpin++)   {
         res(iDir,iSpin).resize (structure.atomInfo, pot);
         res(iDir,iSpin).set (0.0);
      }
   }

   for (int ik = 0; ik < waves.getNk (); ++ik)  {
      const SxGBasis &gk = waves.getGkBasis ()(ik);
      double weight = fermi.kpPtr->weights(ik);
      for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
         SxDiracVec<Complex16> P = *pBasis | waves(iSpin,ik);
         SxArray<SxDiracMat<Complex16> > gradP 
            = projBasis->gradProject (gk | waves(iSpin,ik));
         for (int iDir = 0; iDir < 3; iDir++)   {
            int offset = 0;
            for (int is = 0; is < structure.getNSpecies (); ++is)  {
               int nProjLocal = pot.getNProj (is);

               // --- set up i^l 
               SxVector<Complex16> il(nProjLocal); // i^(-l)  (:ipl)
               for (int ipt = 0, ipl = 0; ipt < pot.lPhi(is).getSize (); ++ipt)  {
                  int l = pot.lPhi(is)(ipt);
                  SxComplex16 c = (l & 1) ? I : SxComplex16(1.);
                  if (!(l & 2)) c=-c;
                  for (int m = -l; m <=l; ++m, ++ipl)  {
                     il(ipl) = c;
                  }
               }

               for (int ia = 0; ia < structure.getNAtoms(is); ++ia, offset+=nProjLocal)  {
                  SxMatrix<Double> &D = res(iDir,iSpin)(is,ia);
                  for (int ipl = 0; ipl < nProjLocal; ++ipl)  {
                     for (int jpl = 0; jpl < nProjLocal; ++jpl)  {
                        SxComplex16 phase = il(ipl) * il(jpl).conj ();
                        for (int iState = 0; iState < nStates; ++iState)  {
                           // ref. 1, Eq. (51)
                           // + projector i^l phase-correction (Ref. 2)
                           // + use k-dependent weights
                           // + exploit k/-k symmetry (=> .re)
                         
                           // Somehow a -1 is in the derivative of Dij  
                           D(ipl, jpl) += -1.0 * weight * fermi.focc(iState,iSpin,ik)
                                 * (( gradP(iDir)(offset+ipl,iState).conj() * P(offset+jpl,iState)
                                 +   P(offset+ipl,iState).conj() * gradP(iDir)(offset+jpl,iState)) * phase
                                   ).re;
                        }
                     }
                  }
               }
            }
         }
      }
   }

   // --- symmetrize
   res = symmetrizeGradDij(res);
   return res;
}


SxComplex16
SxPAWHamiltonian::pawGradDot (int iSpecies, int iAtom, int i, int j,
                              const SxDiracMat<Complex16> &P,
                              const SxDiracMat<Complex16> &gradP) const
{
   // CLOCK (Timer::pawGradDot); // don't use: clock takes 50% of time
   SX_CHECK (potPtr);
   const SxArray<int> &lPhi = potPtr->lPhi(iSpecies);
   const SxArray<int> &offset = potPtr->offset(iSpecies);
   const SxDiracMat<Double> &dS = potPtr->deltaS(iSpecies);
   SxComplex16 res = 0.;
   int offsetA = 0;
   for (int js = 0; js < iSpecies; ++js)  {
      offsetA += int(projBasis->refOrbMap(js).getSize ())
                 * structure.getNAtoms (js);
   }
   offsetA += int(projBasis->refOrbMap(iSpecies).getSize ()) * iAtom;
   for (int ipt = 0; ipt < lPhi.getSize (); ++ipt)   {
      int l = lPhi(ipt), nm = 2 * l + 1;
      for(int jpt = 0; jpt < lPhi.getSize (); jpt++)   {
         if (lPhi(jpt) != l) continue;
         // find position in global projector counting 
         int ipg = offset(ipt) + offsetA;
         int jpg = offset(jpt) + offsetA;
         for (int m = 0; m < nm; ++m, ++ipg, ++jpg)  {
            res -= (  gradP(ipg,i).conj() *     P(jpg,j)
                    +     P(ipg,i).conj() * gradP(jpg,j))     
                 * dS(ipt,jpt);
         }
      }
   }
   return res;
}

SxArray<SxArray<SxArray<SxDiracMat<Complex16> > > >  SxPAWHamiltonian::pawNumGradDot (const PsiG &psi)
{
   CLOCK (Timer::pawGradDot);
   SX_CHECK(psi.getSize () > 0);

   int nPsi = (int)psi.nCols ();
   SxPAWOverlap S(pBasis, potPtr);

   SxArray<SxArray<SxArray<SxDiracMat<Complex16> > > > res(3);
   for (int iDir = 0; iDir < 3; iDir++)   {
      int ns = structure.getNSpecies ();
      res(iDir).resize(ns);
      for (int is = 0; is<ns; is++)   {
         int na = structure.getNAtoms(is);
         res(iDir)(is).resize(na);
         for (int ia = 0; ia < na; ia++)   {
            res(iDir)(is)(ia).reformat(nPsi,nPsi);
            res(iDir)(is)(ia).set(0.0);
            // Shift Structure
            double h = 1e-8;
            double hInv =1e8;
            SxAtomicStructure shift;
            shift.copy(structure);
            shift.ref(is,ia)(iDir) += h;
            // Structure changed !!!! 
            // Therefore GkBasis has to be changed
            SxGBasis &gBasis = const_cast <SxGBasis &> (psi.getBasis<SxGBasis>());
            gBasis.changeTau(shift);
            SxDiracMat<Complex16> Ohpnm (nPsi,nPsi);
            for (int iState = 0; iState < nPsi; iState++)   {
               for (int jState = 0; jState < nPsi; jState++)   {
                  Ohpnm(iState,jState) = S.dot(psi.colRef(iState), psi.colRef(jState));
               }
            }
            shift.ref(is,ia)(iDir) -= 2*h;
            // Structure changed !!!! 
            // Therefore GkBasis has to be changed
            gBasis.changeTau(shift);
            SxDiracMat<Complex16> Ohmnm (nPsi,nPsi);
            for (int iState = 0; iState < nPsi; iState++)   {
               for (int jState = 0; jState < nPsi; jState++)   {
                  Ohmnm(iState,jState) = S.dot(psi.colRef(iState), psi.colRef(jState));
               }
            }

            res(iDir)(is)(ia) = 0.5 * hInv * (Ohpnm - Ohmnm);
            // Update GkBasis again
            gBasis.changeTau(structure);
         }
      }
   }

   return res;

}

SxNArray<SxRadMat,2> SxPAWHamiltonian::computeNumGradDij (const SxPW &waves, 
                                                          const SxFermi &fermi) 
{  
   CLOCK (Timer::GradDij);
   SX_CHECK(potPtr);
   const SxPAWPot &pawPot = *potPtr;
   int nSpin = waves.getNSpin ();

   SxNArray<SxRadMat,2> res (3,nSpin);
   for(int iDir = 0; iDir < 3; iDir++)   {
      for(int iSpin = 0; iSpin < nSpin; iSpin++)   {
         res(iDir,iSpin).resize (structure.atomInfo, pawPot);
         res(iDir,iSpin).set (0.0);
         int ns = structure.getNSpecies ();
         for(int is = 0; is < ns; is++)   {
            int na = structure.getNAtoms (is);
            for(int ia = 0; ia < na; ia++)   {
               double h = 1e-8;
               double hInv = 1e8;
               SxAtomicStructure shift;
               shift.copy(structure);
               for (int ja = 0; ja < na; ja++)   {
                  shift.ref(is,ja)(iDir) += pow(-1.0,ja+ia) * 0.5 * h;
               }
               // Structure changed !!!! Update Tau
               SxGkBasis &gkBasis = const_cast <SxGkBasis &> (waves.getGkBasis());
               gkBasis.changeTau(shift);
               SxArray<SxRadMat> Dhpij = computeDij(waves, fermi.focc);
               shift.copy(structure);
               for (int ja = 0; ja < na; ja++)   {
                  shift.ref(is,ja)(iDir) -= pow(-1.0,ja+ia) * 0.5 * h;
               }
               // Structure changed !!!! Update Tau
               gkBasis.changeTau(shift);
               SxArray<SxRadMat> Dhmij = computeDij(waves, fermi.focc);
               res(iDir,iSpin)(is,ia) = hInv * (Dhpij(iSpin)(is,ia)-Dhmij(iSpin)(is,ia));
               gkBasis.changeTau(structure);
            }
         }
      }
   }

   return res;
            
}

void SxPAWHamiltonian::gradTest(SxPW &waves, SxFermi &fermi)
{
   SxArray<SxRadMat> Dij = computeDij (waves,fermi.focc);
   SxNArray<SxRadMat,2> gradDij = computeGradDij (waves, fermi);
   SxNArray<SxRadMat,2> numGradDij = computeNumGradDij (waves, fermi);
   SxDiracVec<Complex16> P = *pBasis | waves(0,0);
   SxArray<SxDiracMat<Complex16> > gradP 
      = projBasis->gradProject (waves.getGkBasis ()(0) | waves(0,0));
   SxArray<SxArray<SxArray<SxDiracMat<Complex16> > > > numGradOnm = pawNumGradDot(waves(0,0));

   SxString file = "Dij.dat";
   SxBinIO out;
   out.open(file, SxBinIO::ASCII_WRITE_ONLY);
   out.writeXYPlot(Dij(0)(0,0));
   out.close();

   for (int iDir = 0; iDir < 3; iDir++)   {
      for (int is = 0; is < structure.getNSpecies (); ++is)  {
         for (int ia = 0; ia < structure.getNAtoms (is); ++ia)  {

            file = "gradDij.dat";
            file += iDir;
            file += is;
            file += ia;
            out.open(file, SxBinIO::ASCII_WRITE_ONLY);
            out.writeXYPlot(gradDij(iDir,0)(is,ia));
            out.close();

            file = "numGradDij.dat";
            file += iDir;
            file += is;
            file += ia;
            out.open(file, SxBinIO::ASCII_WRITE_ONLY);
            out.writeXYPlot(numGradDij(iDir,0)(is,ia));
            out.close();

            file = "gradOnm.dat";
            file += iDir;
            file += is;
            file += ia;
            out.open(file, SxBinIO::ASCII_WRITE_ONLY);

            int nStates = waves.getNStates ();
            SxMatrix<Complex16> gradOnm(nStates, nStates);
            for (int i = 0; i < nStates; ++i)
               for (int j = 0; j < nStates; ++j)
                  gradOnm(i,j) = pawGradDot (is, ia, i, j, P, gradP(iDir));

            out.writeXYPlot(gradOnm);
            out.close();

            file = "numGradOnm.dat";
            file += iDir;
            file += is;
            file += ia;
            out.open(file, SxBinIO::ASCII_WRITE_ONLY);
            out.writeXYPlot(toVector(numGradOnm(iDir)(is)(ia)));
            out.close();
         }
      }
   }
}

SxNArray<SxRadMat,2> SxPAWHamiltonian::symmetrizeGradDij (SxNArray<SxRadMat,2> &GDij)
{
   CLOCK (Timer::GDijSym);
   SX_CHECK (potPtr);
   const SxPAWPot &pot = *potPtr;

   SX_CHECK (structure.cell.symGroupPtr);
   const SxSymGroup & syms = *structure.cell.symGroupPtr;
   int nSyms = syms.getNSymmorphic ();
   SxVector<Int> symAtomIdx(nSyms);
   SxGrid grid (structure,10);
   int nSpin = int(GDij.getDim(1));

   SxNArray<SxRadMat,2> res (3,nSpin);
   for (int iDir = 0; iDir < 3; iDir++)   {
      for (int iSpin = 0; iSpin < nSpin; iSpin++)   {
         res(iDir,iSpin).resize (structure.atomInfo, pot);
         res(iDir,iSpin).set (0.0);
      }
   }


   cout << "GradDij symmetrization has to be checked" << endl;
   // Concept like SxRadMat symmetrization

   for (int iSpin = 0; iSpin < nSpin; iSpin++)   { 
      for (int is = 0; is < structure.getNSpecies (); ++is)  {
         int nProjTypes = pot.getNProjType (is);
         for (int ia = 0; ia < structure.getNAtoms (is); ++ia)  { 
            // --- find equivalent atoms
            for (int iSym = 0; iSym < nSyms; ++iSym)  {
               symAtomIdx(iSym) = structure.find (syms.getSymmorphic(iSym) 
                     ^ structure.getAtom(is,ia), grid)
                     -structure.atomInfo->offset(is);
            }
            // check if symmetrization has been performed already
            if (symAtomIdx.minval () < ia) continue;

            // ---loop over partial wave types
            for (int ipt = 0; ipt < nProjTypes; ++ipt)  {
               int l1 = pot.lPhi(is)(ipt), nm1 = 2 * l1 + 1;
               int offset1 = pot.offset(is)(ipt);
               for (int ipt2 = 0; ipt2 < nProjTypes; ++ipt2)  {
                  int l2 = pot.lPhi(is)(ipt2), nm2 = 2 * l2 + 1;
                  int offset2 = pot.offset(is)(ipt2);
                  // --- construct symmetrized Dij (type1,type2)
                  SxArray<SxMatrix<Double> > sym1 (3);
                  SxMatrix<Double> rotGDijDir (nm1,nm2); 
                  for(int iDir = 0; iDir < 3; iDir++)   {
                     sym1(iDir).reformat(nm1,nm2);
                     sym1(iDir).set (0.0);
                  }
                  for (int iSym = 0; iSym < nSyms; ++iSym)  {
                     SymMat S = syms.getSymmorphic(iSym);
                     for (int iDir = 0; iDir < 3 ; iDir++)   {
                        // --- collect rotgradDij components
                        SxMatrix<Double> &GDijDir = GDij(iDir,iSpin)(is,symAtomIdx(iSym));
                        const SxMatrix<Double> &Dl1 = ylmRot(iSym)(l1);
                        const SxMatrix<Double> &Dl2 = ylmRot(iSym)(l2);
                        for (int m1 = 0; m1 < nm1; ++m1)   {
                           for (int m2 = 0; m2 < nm2; ++m2)   {
                              rotGDijDir(m1,m2) = GDijDir(offset1 + m1, offset2 + m2);
                           }
                        }
                        // debug
                        /*
                        if (ipt == 0 && ipt2 == 0 && iDir == 2)  {
                           cout << iSym << '\t' << rotGDijDir;
                           cout << '\t' << (S(2,2) * rotGDijDir) << endl;
                        }
                        */
                        rotGDijDir = Dl1.transpose () ^ rotGDijDir ^ Dl2;
                        for(int jDir = 0; jDir < 3; jDir++)   {
                           sym1(jDir) += S(iDir, jDir) * rotGDijDir;
                           //sym1(jDir) += S(jDir, iDir) * rotGDijDir;
                        }
                     }
                  }

                  for (int iDir = 0; iDir < 3 ; iDir++)   {
                     sym1(iDir) /= double(nSyms);
                  }

                  // --- distribute symmetrized gradient
                  for (int iSym = 0; iSym < nSyms; ++iSym)  {
                     SymMat S = syms.getSymmorphic(iSym);
                     for (int iDir = 0; iDir < 3 ; iDir++)   {
                        // --- collect rotgradDij component
                        const SxMatrix<Double> &Dl1 = ylmRot(iSym)(l1);
                        const SxMatrix<Double> &Dl2 = ylmRot(iSym)(l2);
                        rotGDijDir.set(0.0);
                        for(int jDir = 0; jDir < 3; jDir++)   {
                           rotGDijDir += S(iDir,jDir) * sym1(jDir);
                           //rotGDijDir += S(jDir,iDir) * sym1(jDir);

                        }
                        rotGDijDir = Dl1 ^ rotGDijDir ^ Dl2.transpose ();

                        // distribute
                        for (int m1 = 0; m1 < nm1; ++m1)   {
                           for (int m2 = 0; m2 < nm2; ++m2)   {
                              res(iDir,iSpin)(is,symAtomIdx(iSym))(offset1 + m1, offset2 + m2) = rotGDijDir(m1,m2);
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   }         

   return res;

}

SxConstPtr<SxOverlapBase> SxPAWHamiltonian::getS () const
{
   return SxPtr<SxPAWOverlap>::create (pBasis, potPtr);
}

SxMeshR SxPAWHamiltonian::computeCorePS () const
{
   SX_CHECK (rPtr);
   const SxGBasis &gBasis = rPtr->getGBasis ();
   PsiG coresG(gBasis);
   coresG.set (0.);
   for (int is = 0; is < structure.getNSpecies (); ++is)
      coresG += rhoCorePSG(is) * gBasis.structureFactors(is);
   VALIDATE_VECTOR (coresG);
   SxMeshR res = coresG.to (*rPtr);
   res = rPtr->symmetrize (res);
   return res;
}

void SxPAWHamiltonian::computeRho (const Focc &focc, const SxPsiSet &psi)
{
   SX_CHECK (rPtr);
   const SxPWSet &waves = dynamic_cast<const SxPWSet &> (psi);
   // compute 1-center density matrix
   cout << "Compute Dij" << endl;
   pawRho.Dij = computeDij (waves, focc);

   // --- compute pseudo-density
   cout << "Compute rho_ps" << endl;
   int nSpin = waves.getNSpin ();
   if (pawRho.pwRho.rhoR.getSize () == 0)  {
      // initialize pwRho
      pawRho.pwRho = SxRho (*rPtr, nSpin, 0. ); // 0. = no renormalization
      // set paw potential
      pawRho.potPtr = potPtr;
   }
   pawRho.pwRho.computeRho (focc, waves);

   // --- add pseudo-core densities
   SxMeshR coresR = computeCorePS ();
   for (int iSpin = 0; iSpin < nSpin; ++iSpin)
      pawRho.pwRho(iSpin).plus_assign_ax (1. / nSpin, coresR);
   //pawRho.psCore = coresR;

}

