// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxRadMat.h>
#include <SxGrid.h>
#include <SxPAWPot.h>
#include <SxPartialWaveBasis.h>

/*
References:
 1. P. E. Bloechl, Phys. Rev. B 50, 17953 (1994).
 2. C. Freysoldt, PAW implementation notes
*/

void SxRadMat::resize (const SxConstPtr<SxAtomInfo> &info,
                       const SxPAWPot &pawPot)
{
   SX_CHECK (info);
   SX_CHECK_VARS (pawPot.getNSpecies () == info->nSpecies,
                  pawPot.getNSpecies (), info->nSpecies);
   int nSpecies = pawPot.getNSpecies ();
   rho.resize (nSpecies);
   for (int is = 0; is < nSpecies; ++is)  {
      int nAtoms = info->nAtoms(is);
      int nOrb = pawPot.getNProj (is);
      rho(is).resize (nAtoms);
      for (int ia = 0; ia < nAtoms; ++ia)  {
         rho(is)(ia) = SxMatrix<Double> (nOrb, nOrb);
      }
   }
}

void SxRadMat::resize (const SxRadMat &in)
{
   rho.resize (in.rho.getSize ());
   for (int is = 0; is < rho.getSize (); ++is)  {
      rho(is).resize (in.rho(is).getSize ());
      if (rho(is).getSize () == 0) continue;
      int np = (int)in.rho(is)(0).nCols ();
      SX_CHECK_VARS (np == in.rho(is)(0).nRows (), np, in.rho(is)(0).nRows ());
      for (int ia = 0; ia < rho(is).getSize (); ++ia)  {
         rho(is)(ia) = SxMatrix<Double> (np, np);
      }
   }
}


void SxRadMat::set (double x)
{
   for (int is = 0; is < rho.getSize (); ++is)
      for (int ia = 0; ia < rho(is).getSize (); ++ia)
         rho(is)(ia).set (x);
}

void SxRadMat::symmetrize (const SxAtomicStructure &str,
                           const SxPAWPot &pot,
                           const SxYlmRotGroup &ylmRot)
{
   SX_CHECK (str.cell.symGroupPtr);
   const SxSymGroup &syms = *str.cell.symGroupPtr;
   int nSym = syms.getNSymmorphic ();
   SxVector<Int> symAtomIdx(nSym);
   SxGrid grid (str, 10);

   for (int is = 0; is < str.getNSpecies (); ++is)  {
      for (int ia = 0; ia < str.getNAtoms (is); ++ia)  {
         // --- find equivalent atoms
         for (int iSym = 0; iSym < nSym; ++iSym)  {
            symAtomIdx(iSym) = str.find (syms.getSymmorphic(iSym) 
                  ^ str.getAtom(is,ia), grid)
                  -str.atomInfo->offset(is);
         }
         // check if symmetrization has been performed already
         if (symAtomIdx.minval () < ia) continue;

         // --- loop of partial wave types
         for (int ipt = 0; ipt < pot.getNProjType (is); ++ipt)  {
            int l = pot.lPhi(is)(ipt), nm = 2 * l + 1;
            int offset = pot.offset(is)(ipt);
            for (int ipt2 = 0; ipt2 < pot.getNProjType (is); ++ipt2)  {
               int l2 = pot.lPhi(is)(ipt2), nm2 = 2 * l2 + 1;
               int offset2 = pot.offset(is)(ipt2);

               // --- construct symmetrized Dij (type1,type2)
               SxMatrix<Double> sym (nm, nm2), rotRho(nm,nm2);
               sym.set (0.);
               for (int iSym = 0; iSym < nSym; ++iSym)  {
                  // --- collect rotRho
                  SxMatrix<Double> &rotD = rho(is)(symAtomIdx(iSym));
                  const SxMatrix<Double> &Dl  = ylmRot(iSym)(l);
                  const SxMatrix<Double> &Dl2 = ylmRot(iSym)(l2);
                  for (int m = 0; m < nm; ++m)
                     for (int m2 = 0; m2 < nm2; ++m2)
                        rotRho(m,m2) = rotD(offset + m, offset2 + m2);
                  // --- symmetrize 
                  sym += Dl.transpose () ^ rotRho ^ Dl2;
               }
               sym /= double(nSym);
               /*
               if (l == 1 && l2 == 1)  {
                  cout << "is=" << is << "; ia=" << ia << endl;
                  cout << "l=1 matrix" << endl;
                  cout << sym << endl;
                  cout << sym.eigenvalues () << endl;
                  cout << sym.eigenvectors () << endl;
               }
               */

               // --- distribute symmetrized matrix
               for (int iSym = 0; iSym < nSym; ++iSym)  {
                  SxMatrix<Double> &rotD = rho(is)(symAtomIdx(iSym));
                  const SxMatrix<Double> &Dl  = ylmRot(iSym)(l);
                  const SxMatrix<Double> &Dl2 = ylmRot(iSym)(l2);
                  // rotate with symmetry
                  rotRho = Dl ^ sym ^ Dl2.transpose ();
                  // distribute
                  for (int m = 0; m < nm; ++m)
                     for (int m2 = 0; m2 < nm2; ++m2)
                        rotD(offset + m, offset2 + m2) = rotRho(m,m2);
               }
            }
         }
      }
   }
}

void SxRadMat::sumMPI ()  // LoopMPI
{
   for (int is = 0; is < rho.getSize (); ++is) {
      for (int ia = 0; ia < rho(is).getSize (); ++ia) {
         SxLoopMPI::sum (rho(is)(ia));
      }
   }
}

SxDiracVec<Complex16>
SxRadMat::operator^ (const SxDiracVec<Complex16> &p) const
{
   // TODO: check loop order with respect to performance
   //       iState loop should probably outmost
   // or try alternative
   // loop is,ia 
   //   SxIdx range(offset, offset+npl-1)   
   //   setup complex Hij matrix
   //   res.rowRef (range) <<= Hij ^ p.rowRef (range)
   // => need rowRef as submatrix class
   SX_CHECK (p.handle);
   const SxPartialWaveBasis &pBasis = p.getBasis<SxPartialWaveBasis> ();
   SxConstPtr<SxPAWPot> potPtr = pBasis.getPotPtr ();

   SxDiracMat<Complex16> res(p.nRows (), p.nCols ());
   res.set (0.);
   res.setBasis (&pBasis);
   res.handle->auxData = p.handle->auxData;

   int offset = 0;
   for (int is = 0; is < getNSpecies (); ++is)  {
      int nProjLocal = potPtr->getNProj (is);

      // --- set up i^l 
      SxVector<Complex16> il(nProjLocal); // i^(-l)  (:ip)
      for (int ipt = 0, ipl = 0; ipt < potPtr->lPhi(is).getSize (); ++ipt)  {
         int l = potPtr->lPhi(is)(ipt);
         SxComplex16 c = (l & 1) ? I : SxComplex16(1.);
         if (l & 2) c=-c;
         for (int m = -l; m <=l; ++m, ++ipl)  {
            il(ipl) = c;
         }
      }

      // --- loop over atoms
      for (int ia = 0; ia < getNAtoms(is); ++ia, offset+=nProjLocal)  {

         // --- loops over projectors
         for (int ipl = 0; ipl < nProjLocal; ++ipl)  {
            for (int jpl = 0; jpl < nProjLocal; ++jpl)  {

               // potential energy * i^l projector phases (Ref. 1)
               SxComplex16 Hij = (*this)(is,ia)(ipl,jpl) 
                               * (il(ipl) * il(jpl).conj ());
               
               // compute sum_j H(R)ij <pj|psi>
               for (int iState = 0; iState < p.nCols (); ++iState)  {
                  // + projector i^l phase-correction 
                  res(offset+ipl, iState) += Hij * p(offset+jpl,iState);
               }
            }
         }
      }
   }
   return res;
}

double tr (const SxRadMat &D, const SxRadMat &A)
{
   double res = 0.;
   for (int is = 0; is < A.getNSpecies (); ++is)
      for (int ia = 0; ia < A.getNAtoms (is); ++ia)
         res += dot(D(is, ia), A(is,ia));
   SX_CHECK_NUM (res);
   return res;
}
