// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_RAD_MAT_H_
#define _SX_RAD_MAT_H_
#include <SxDFT.h>
#include <SxMatrix.h>
#include <SxAtomicStructure.h>
#include <SxYlm.h>

class SxPAWPot;
//template<class T> class SxDiracVec<T>;

/** \brief PAW 1-center matrices

    \b SxClass = S/PHI/nX PAW 1-center matrices


    \author Christoph Freysoldt, freysoldt@mpie.de */
class SX_EXPORT_DFT SxRadMat
{
   public:
      /// 1-center matrices (:is,:ia, (np x np))
      SxArray<SxArray<SxMatrix<Double> > > rho;

      /// Get a 1-center matrix
      SxMatrix<Double> &operator() (int is, int ia)  {
         return rho(is)(ia);
      }

      /// Get a 1-center matrix
      const SxMatrix<Double> &operator() (int is, int ia) const  {
         return rho(is)(ia);
      }

      /// Get a 1-center matrix
      SxMatrix<Double> &operator() (const SxAutoLoop &is, 
                                    const SxAutoLoop &ia)
      {
         is.setLimit (getNSpecies ());
         ia.setLimit (getNAtoms((int)is));
         return rho(is)(ia);
      }

      /// Get a 1-center matrix
      const SxMatrix<Double> &operator() (const SxAutoLoop &is, 
                                          const SxAutoLoop &ia) const
      {
         is.setLimit (getNSpecies ());
         ia.setLimit (getNAtoms((int)is));
         return rho(is)(ia);
      }

      /// Resize
      void resize (const SxConstPtr<SxAtomInfo> &info,
                   const SxPAWPot &pawPot);
      /// Resize according to example
      void resize (const SxRadMat &in);

      void set (double x);

      /// In-place scaling
      void operator *= (double s)
      {
         for (int is = 0; is < rho.getSize (); ++is)
            for (int ia = 0; ia < rho(is).getSize (); ++ia)
               rho(is)(ia) *= s;
      }

      void symmetrize (const SxAtomicStructure &str,
                       const SxPAWPot &pot,
                       const SxYlmRotGroup &ylmRot);

      inline int getNSpecies () const
      {
         return int(rho.getSize ());
      }

      inline int getNAtoms (int is) const
      {
         SX_CHECK_VARS (is >= 0 && is < getNSpecies (), is, getNSpecies ());
         return int(rho(is).getSize ());
      }

      /** \brief Multiply matrices with projections
        @param p   \f$\langle p_i|\psi\rangle\f$

        \f[
        \sum_j A_{ij} <p_i | \psi>
        \f]

        Aij are the matrices stored in rho.
        */
      SxDiracVec<Complex16> operator^ (const SxDiracVec<Complex16> &p) const;

      void sumMPI ();  // LoopMPI
};

/** \brief Compute trace
  \f[
  \sum_{R} \sum_{i,j\,@R} D_{ij} A_{ij}
  \f]
  */
double tr (const SxRadMat &D, const SxRadMat &A);

#endif /* _SX_RAD_MAT_H_ */
