// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxSymMatrix.h>

SxComplex16 expigr (double gr)
{
   SxComplex16 sc;
   sincos (gr, &sc.im, &sc.re);
   return sc;
}

int main (int argc, char **argv)
{
   cout.precision (10);
   SxCLI cli(argc, argv);
   int N = cli.option ("-N", "number", "supercell size").toInt (5,3);
   int nk = cli.option ("--nk", "number", "number of k-points").toInt (10);;
   double Tvv = 1., epsV = -2. * Tvv;
   double lambda = cli.option ("--epsd", "energy", "defect on-site energy")
                  .toDouble (epsV);
   double Tvd = cli.option ("--Tvd", "coupling", "defect coupling")
                .toDouble (Tvv);
   bool sType = cli.option ("--stype", "type of defect").toBool ();
   int iv = cli.option ("--iv", "band id", "band id for mixing").toInt (1);
   if (iv & 1)
      iv = -(iv+1);

   cli.finalize ();

   double alpha = 0., epsD;
   if (lambda - epsV > 2. * Tvv)  {
      alpha = acosh((lambda-epsV)/(2. * Tvv));
      // choose epsD such that user-input becomes defect state energy
      epsD = lambda -  2. * sqr(Tvd)/(lambda - epsV - Tvv * exp(-alpha));
   } else {
      epsD = lambda;
   }
   cout << "# epsD = " << epsD << endl;
   cout << "# exp(-alpha) = " << exp(-alpha) << endl;
   cout << "# exp(-alpha*N) = " << exp(-alpha * N) << endl;

   SxSymMatrix<Complex16> ham (N);

   ham.set (0.);

   ham(0,0) = epsD;
   // ham (-1, 0) = ...

   ham(1,1) = epsV;
   if (sType)
      ham(0,1) = -Tvd;
   else
      ham(0,1) = Tvd;

   for (int i = 2; i < N; ++i)  {
      ham(i,i) = epsV;
      ham(i-1, i) = Tvv;
   }

   SxMatrix<Double> bs(nk+1, N), weightD(nk+1, N);
   SxVector<Double> hamD(nk+1), hamV(nk + 1), sDV(nk + 1), hamvdDiag(nk + 1);
   for (int ik = 0; ik <= nk; ++ik)  {
      
      // k-phase
      SxComplex16 phase = expigr(ik * PI/ nk);

      ham(0,N-1) = Tvd * phase.conj ();

      SxSymMatrix<Complex16>::Eigensystem eig = ham.eigensystem ();

      // --- set up ideal eigenstate
      SxVector<Complex16> dState(N);
      dState.set (0.);
      if (fabs(Tvd) > 0.)  {
         SxComplex16 phaseSum = 0.;
         
         if (alpha > 0.)  {
            phaseSum = 1. / (1. - exp(-alpha * N) * phase);
         }
         //SxComplex16 phaseN = exp(-alpha) * expigr (ik * PI/(N*nk));
         dState(0) = (lambda - epsV - exp(-alpha) * Tvv) / Tvd;
         if (sType)
            dState(0) += exp(-alpha * (N-1)) * (-phaseSum * phase.conj ()
                                                +phaseSum.conj () * phase);
         else
            dState(0) += 2. * exp(-alpha * (N-1)) * (phaseSum * phase.conj ()).re;
         for (int i = 1; i < N; ++i)  {
            if (sType)
               dState(i)   -= exp(-alpha * (i-1)) * phaseSum;
            else
               dState(i)   += exp(-alpha * (i-1)) * phaseSum;
            dState(N-i) += exp(-alpha * (i-1)) * phaseSum.conj () * phase; 
         }
         cout << "# D: " << dState << endl;
         dState.normalize ();
         cout << "# |D|^2: " << dState.absSqr () << endl;
         cout << "# D-loc: " << dState(0).absSqr () + dState(1).absSqr () + dState(N-1).absSqr () << endl;
      } else {
         dState(0) = 1.;
      }

      for (int i = 0; i < N; ++i)  {
         weightD(ik, i) = dot(eig.vecs.colRef(i), dState).absSqr ();
         bs(ik, i) = eig.vals(i);
      }

      //cout << "#" << eig.vecs.colRef(N-1).absSqr () << endl;
      // --- set up pure valence state
      SxVector<Complex16> vState(N);
      for (int i = 0; i < N; ++i)
         vState(i) = expigr (i * (iv * nk + ik) * PI/(nk * N));
      vState /= sqrt(N);

      hamD(ik) = dot (dState, ham.expand () ^ dState);
      hamV(ik) = dot (vState, ham.expand () ^ vState);
      sDV(ik) = dot (vState, dState).absSqr ();
      
      vState.plus_assign_ax (-dot(dState, vState), dState);
      vState.normalize ();
      SxVector<Complex16> hamVorth = ham.expand () ^ vState;
      double hVV = dot (vState, hamVorth).re,
             hDV2 = dot (dState, hamVorth).absSqr (),
             hDD  = hamD(ik);
      hamvdDiag(ik) = 0.5 * (hDD + hVV) + sqrt(0.25*sqr(hVV-hDD) + hDV2);
   }

   for (int i = 0; i < N; ++i)  {
      double avgEps = 0., avgW = 0.;
      for (int ik = 0; ik <= nk; ++ik)  {
         cout << double(ik)/(nk)
              << '\t' << bs(ik, i)
              << '\t' << weightD(ik, i)
              << endl;
         avgEps += bs(ik, i);
         avgW   += weightD(ik,i);
      }
      cout << "#\t" << avgEps/(nk+1) << '\t' << avgW/(nk+1) << endl;
      cout << "&" << endl;
   }
   cout << "# <D(iso) | H| D(iso)>" << endl;
   for (int ik = 0; ik <= nk; ++ik)  {
      cout << double(ik)/(nk)
           << '\t' << hamD(ik)
           << '\t' << 1
           << endl;
   }
   cout << "&" << endl;

   cout << "# <vb | H| vb>" << endl;
   for (int ik = 0; ik <= nk; ++ik)  {
      cout << double(ik)/(nk)
           << '\t' << hamV(ik)
           << '\t' << sDV(ik)
           << endl;
   }
   cout << "&" << endl;

   cout << "# 2-state diag" << endl;
   for (int ik = 0; ik <= nk; ++ik)  {
      cout << double(ik)/(nk)
           << '\t' << hamvdDiag(ik)
           << '\t' << 0.
           << endl;
   }
   cout << "&" << endl;
}

