// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxUtil.h>
#include <SxXYZ.h>
#include <SxCLI.h>
#include <SxAtomicStructure.h>
#include <SxSpeciesData.h>
#include <SxSimpleParser.h>

int main (int argc, char **argv)
{
   // --- parse command line options
   SxCLI cli (argc, argv);

   // --- input and output files
   bool sxbFile = cli.option ("-b|--sxb", "input file is binary waves file "
                              "rather than S/PHI/nX input file")
                  .toBool ();

   SxString inFile
      = cli.option ("-i|--input", "input file",
                    "take original input file")
        .toString (sxbFile ? "waves.sxb" : "input.sx");
   SxString outFile = cli.option("-o","file","XYZ output file to be written")
                      .toString("input.xyz");
   bool withAMat = cli.option ("-a", "prepend lattice vectors (a1, a2, a3)")
                   .toBool();
   bool trajectory = cli.option ("--trajectory", "file is a trajectory")
                     .toBool ();


   cli.finalize ();

   // --- read from input file
   initSPHInXMath ();

   SxAtomicStructure structure;
   SxArray<SxString> chemName;


   if (trajectory && !sxbFile)  {
      FILE *fp =sxfopen (outFile, "w");
      try {
         SxParser parser;
         SxParser::Table table = parser.read (inFile, "std/forces.std");

         chemName = SxSpeciesData::getElements (&*table);
         SX_LOOP(is) chemName(is) = chemName(is).toUpper ();

         int it = 0;
         SYMBOLPARSE(&*table)  {
            FOREACH_SYMBOLGROUP("structure")  {
               structure = SxAtomicStructure(SYMBOLGROUP_TABLE);
               SX_CHECK_VARS(structure.getNSpecies () == chemName.getSize (),
                             structure.getNSpecies (), chemName.getSize ());
               fprintf (fp, "%d\nStep %d\n", structure.getNAtoms (), ++it);
               SX_LOOP2(is,ia)  {
                  const Coord &pos = structure.getAtom (is,ia);
                  fprintf (fp, "%s %.8f %.8f %.8f\n", chemName(is).ascii (),
                           pos(0) / A2B, pos(1) / A2B, pos(2) / A2B);
               }
            }
         }
      } catch (SxException e) {
         e.print ();
         SX_QUIT;
      }
      fclose (fp);
      return 0;
   }

   if (sxbFile)  {
      try {
         SxBinIO io (inFile, SxBinIO::BINARY_READ_ONLY);
         structure.read (io);
         SxString chemNameList;
         io.read("chemNames", &chemNameList);
         chemName = chemNameList.tokenize (',');
         io.close ();
      } catch (SxException e)  {
         e.print ();
         SX_EXIT;
      }
   } else  {
      try {
         SxParser parser;
         SxParser::Table table = parser.read (inFile, "std/structure.std");
         structure = SxAtomicStructure(&*table);
         chemName = SxSpeciesData::getElements (&*table);
      } catch (SxException e) {
         e.print ();
         SX_QUIT;
      }
   }

   SxXYZ (outFile).write (structure, chemName, withAMat);
   return 0;
}

