// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------
#include <SxAtomDFT.h>
#include <SxXCFunctional.h>
#include <SxConstants.h>
#include <SxUtil.h>
#include <SxBinIO.h>
#include <SxCLI.h>
#include <iostream>
#include <fstream>
#include <iomanip>

using namespace std;

//------------------
// Extern

double getDeriv (SxDiracVec<Double> func, double dx, int gridIdx)
// computes extrapolated derivative of function at point with Index gridIdx 
/* 
	ref: "Numerical methods for engineers and scientists - Hoffman, et al. (2001)"
	notes: x) for 1/x [x,3.1,3.9] extrapolation increases the accuracy only for a 
				 second-order difference formula, but not for a 4-th order 
		    x) by extrapolation, the error should be reduced from 2nd to 2nd+m-th 
				 order (m is the increment in the order of the following error terms), 
				 i.e. O(2+3)=O(5)
*/
{
	// TODO: SXCHECK that func(n) n>=5

	SxDiracVec<Double> derivative(2);

	for (int i = 1; i <= 2; i++) 
		
		derivative(i-1) = 
							 (1/(2.*(i*dx)))*(func(gridIdx+i) - func(gridIdx-i));

	return derivative(0) + (1/(4 - 1.))*(derivative(0) - derivative(1));
}
//------------------


// -------------------
// Numerov integration
// solves y''(x) + k(x) * y(x) = s(x) by 3-point integration on grid: 
//		c(i+1) y(i+1) +  c(i-1) y(i-1) = c(i) y(i) + d(i)
/*
	notes: x) results can be significantly depend on the direction of solving 
				 the grid ('rule-of-thumb': algorithm should be applied in 
				 direction of increasing magnitude of y)
			 x) alternative approach: Störmer's rule (see Numerical recipes)
*/

inline bool CheckNodes(double a, double b)
{
	return (a < 0) == (b < 0);
}


// note: Sx* should be always passed by const ... & to a function
SxDiracVec<Double> solveNumerovFwd (const SxDiracVec<Double> &kx, 
											   const SxDiracVec<Double> &sx,
											   double solR0, double solR0dx, double dx, 
											   int *nNodes=NULL)
{
	SxDiracVec<Double> phi(kx.getSize ());
	phi(0) = solR0;
	phi(1) = solR0dx;

	double dxsq = dx*dx;
	double y0, y1 = phi(0), c0, c1, d; 
	int nodeCounter = 0;

	// iterators are about 7 times slower
	for (int i = 2; i < kx.getSize(); i++) {
		y0 = y1;
		y1 = phi(i-1);
		c0 = (1. + (1/12.)*dxsq*kx(i-2))*y0;
		c1 = (2. - ((5./6.)*dxsq)*kx(i-1))*y1;
		d = (dxsq/12.)*(sx(i-2) + 10.*sx(i-1) + sx(i));
		phi(i) = (c1 - c0 + d) / (1. + (1/12.)*dxsq*kx(i));

		if (CheckNodes(phi(i), phi(i-1)) == 0) nodeCounter++;
	}

   if (nNodes)  *nNodes = nodeCounter;
	
	return phi;
}


SxDiracVec<Double> solveNumerovBwd (const SxDiracVec<Double> &kx,
											   const SxDiracVec<Double> &sx,
											   double solRMax, double solRMaxdx, double dx,
											   int *nNodes=NULL)
{	
	int nUpper = int(kx.getSize ()) - 1;

	SxDiracVec<Double> phi(nUpper+1);
	phi(nUpper) = solRMax;
	phi(nUpper-1) = solRMaxdx;
	
	double dxsq = dx*dx;
	double y0, y1 = phi(nUpper), c0, c1, d;
	int nodeCounter = 0;

	for (int i = nUpper - 2; i>=0; i--) {
		y0 = y1;
		y1 = phi(i+1);
		c0 = (1. + (1/12.)*dxsq*kx(i+2))*y0;
		c1 = (2. - ((5./6.)*dxsq)*kx(i+1))*y1;
		d = (dxsq/12.)*(sx(i+2) + 10.*sx(i+1) + sx(i) );
		phi(i) = (c1 - c0 + d)/(1. + (1/12.)*dxsq*kx(i));

		if (CheckNodes(phi(i), phi(i+1)) == 0) nodeCounter++;
	}
	
	if (nNodes)  *nNodes = nodeCounter;

	return phi;
}


SxDiracVec<Double> solveNumerovMatched (const SxDiracVec<Double> &kx,
												    const SxDiracVec<Double> &sx,
													 double solR0,double solR0dx,
													 double solRMax,double solRMaxdx,
													 double dx)
{
	int nUpper = int(kx.getSize ()) - 1;
	int rMatchIdx = nUpper;
	
	SxDiracVec<Double> numerovFwd, numerovBwd, matchedFwdBwd(nUpper+1);

	numerovFwd = solveNumerovFwd (kx, sx, solR0, solR0dx, dx);
	numerovBwd = solveNumerovBwd (kx, sx, solRMax, solRMaxdx, dx);

	SxDiracVec<Double> diff = (numerovFwd - numerovBwd).abs ();

	double diffMin = diff(nUpper);

	for(int i = nUpper; i > 0; i--)  {
		if(diff(i) < diffMin) {
			rMatchIdx = i;
			diffMin = diff(i);
		}
	}

	for(int i = 0; i < rMatchIdx; i++) matchedFwdBwd(i) = numerovFwd(i);

	matchedFwdBwd(rMatchIdx) = 
		0.5*(numerovFwd(rMatchIdx) + numerovBwd(rMatchIdx));
	
	for(int i = rMatchIdx + 1; i <= nUpper; i++) matchedFwdBwd(i) = numerovBwd(i);

	return matchedFwdBwd;
	
}


double getSecondInitialValue (const SxDiracVec<Double> &kx,
									   const SxDiracVec<Double> &sx, double y0,
									   double Dy0, double dx)
// computes second value y1=(y0+dx) from y0 and its derivative y'(0)=Dy0
// see Comp. in Phys. 11, 514 (1997)
{
	double y1;
	double dxsq = dx*dx;

	y1=((1 + (dxsq/24.)*kx(2) )*y0 + ((dxsq*dxsq)/36.)*kx(2)
		*(sx(0) + 2*sx(1) - kx(0)*y0) + (dxsq/24.)
		*(6*sx(1) - sx(2) +  7*(sx(0) - kx(0)*y0)) 
		+ dx*(1 + (dxsq/12.)*kx(2))*Dy0) 
		/(1 + (dxsq/4.)*kx(1) + (dxsq/18.)*kx(1)*kx(2));

	return y1;
}
// -----------------



// -----------------
// Inline functions

inline SxDiracVec<Double> getkx (const SxDiracVec<Double> &r,
										   const SxDiracVec<Double> &vHartree,
											const SxDiracVec<Double> &vXc, 
								 		   double eIn, int z, int l)
{
	SxDiracVec<Double> kx(r.getSize ());

	for (int i = 0; i < r.getSize (); i++) 
		kx(i) = 2.*r(i)*r(i)*(eIn - vHartree(i) - vXc(i)) + 2.*r(i)*z
				  - (l + 1/2.)*(l + 1/2.);
	
	return kx;
}


inline double getFwdBoundCondition (double rVal, double vHartreeVal, 
											   double vXcVal, double eIn, int z, int l)
{
	double phiVal;	
	double ve = vHartreeVal + vXcVal - eIn;

	int lp = (l + 1); 

	phiVal = (pow (8*ve, (lp/2.))*pow (rVal, l))
				*(sqrt (rVal) - ((z*pow (rVal, 3/2.))/lp)
				+ ((ve*lp + z*z)*pow (rVal, 5/2.))
				/((3 + 2*l)*lp)
				- (z*(ve*(4 + 3*l) + z*z)*pow (rVal, 7/2))
				/(3*(6 + 13*l + 9*l*l + 2*l*l*l)));

	return phiVal;
}


inline double getBwdBoundCondition (double x, double vHartreeVal, double vXcVal,
												double eIn)
{
	double phiVal;
	
	phiVal = exp (-sqrt (2.*(vXcVal + vHartreeVal - eIn))*x)/sqrt (x);

	return phiVal;
}

// QN according to Aufbau principle
inline int getQNums (int *n, int *l) 
{
	
	if (*l == 0) {
		
 		switch (*n) {

			case 1: (*n)++; break;

			case 2: case 3: (*l)++; break;

			case 4: case 5: (*n)--;*l+=2; break;

			case 6: case 7: *n-=2;*l+=3; break;
		}
	}
	else {

		(*l)--;
		(*n)++;

	}

	return 0;
}

inline SxDiracVec<Double> getOcc (SxDiracVec<Double> eigenKS, double ekT, 
										    double chemPot)
{
	SxDiracVec<Double> nOccVal(eigenKS.getSize());
		
	int n = 1, l = 0;

	for (int i = 0; i < eigenKS.getSize (); i++) {
		nOccVal(i) = 2*(2*l + 1)/(exp ((eigenKS(i) - chemPot)*ekT) + 1);
		getQNums (&n, &l);
	}

	return nOccVal;
}
// -----------------


// -----------------
// SxAtom
SxAtomDFT::SxAtomDFT (double r0_, double rMax_, int nIntervals_, int nOrb)
// efficient way to initialize member variables (mind order of parameters!)
   :	eigenKS(nOrb),
		r0 (r0_),
    	rMax (rMax_),
	 	nIntervals (nIntervals_)
{
	r = getGrid ();	
}


SxDiracVec<Double> SxAtomDFT::getGrid ()
{
	SxDiracVec<Double> rVal(nIntervals+1);

	// usage of iterators does not seem to increase performance

	dx = log (rMax/r0)/nIntervals; // i.e. rMax=r0*exp(nIntervals*dx)
		
	for (int i = 0; i <= nIntervals; i++) rVal(i) = r0*exp (i*dx);

	return rVal;
}


int SxAtomDFT::computeFwdNodes (const SxDiracVec<Double> &vHartree,
										  const SxDiracVec<Double> &vXc,									
										  double eIn, int z, int l)
{
	int fwdNodes;

	SxDiracVec<Double> kx = getkx (r, vHartree, vXc, eIn, z, l);
	SxDiracVec<Double> sx(nIntervals+1);
	sx = 0.;

	int rMatchIdx = int(kx.getSize ()) - 3;

	for (int i = int(kx.getSize ()) - 3; i >= 2; i--) 
		if (kx(i) > 0) {
			rMatchIdx = i;
			break;
		}

	double solR0 = getFwdBoundCondition (r(0), vHartree(0), vXc(0), 
												    eIn, z, l);
	double solR0dx = getFwdBoundCondition (r(1), vHartree(1), vXc(1),
												      eIn, z, l);

	solveNumerovFwd (kx(SxIdx (0, rMatchIdx)), sx(SxIdx (0, rMatchIdx)), solR0, 
						  solR0dx, dx, &fwdNodes);

	return fwdNodes;									   	
}


double SxAtomDFT::computeMatchFu (const SxDiracVec<Double> &vHartree,
								   	    const SxDiracVec<Double> &vXc,									
									       double eIn, int z, int l)
{
	SxDiracVec<Double> numerovFwdMatch, numerovBwdMatch;

	double matchFu;

	int nBwd, nFwd;

	// redundant with computeFwdNodes: left this way for readability
	SxDiracVec<Double> kx = getkx (r, vHartree, vXc, eIn, z, l);
	SxDiracVec<Double> sx(nIntervals+1);
	sx = 0.;

	int rMatchIdx = int(kx.getSize ()) - 3;

	for (int i = int(kx.getSize ()) - 3; i >= 2; i--) 
		if (kx(i) > 0) {
			rMatchIdx = i;
			break;
		}

	double solR0 = getFwdBoundCondition (r(0), vHartree(0), vXc(0), 
												    eIn, z, l);
	double solR0dx = getFwdBoundCondition (r(1), vHartree(1), vXc(1),
												      eIn, z, l);
	//
	
	numerovFwdMatch = solveNumerovFwd (kx(SxIdx (0, rMatchIdx+2)), 
												  sx(SxIdx (0, rMatchIdx+2)), solR0, 
												  solR0dx, dx, &nFwd);

	if (numerovFwdMatch(rMatchIdx+1)*numerovFwdMatch(rMatchIdx+2) < 0 ||
		 numerovFwdMatch(rMatchIdx)*numerovFwdMatch(rMatchIdx+1) < 0)
			{ nFwd--; }

	nNodes = nFwd;

	double solRMax = getBwdBoundCondition (r(nIntervals), vHartree(nIntervals), vXc(nIntervals), 
														eIn);
	double solRMaxdx = getBwdBoundCondition (r(nIntervals-1), vHartree(nIntervals-1), vXc(nIntervals-1), 
														  eIn);

	numerovBwdMatch = solveNumerovBwd (kx(SxIdx (rMatchIdx-2, nIntervals)), 
												  sx(SxIdx (rMatchIdx-2, nIntervals)), 
												  solRMax, solRMaxdx, dx, &nBwd);

	if (numerovBwdMatch(0)*numerovBwdMatch(1) < 0 || 
		 numerovBwdMatch(1)*numerovBwdMatch(2) < 0)
			{ nBwd--; }	

	nNodes += nBwd;

	matchFu = (getDeriv (numerovFwdMatch, dx, rMatchIdx)
				 		/numerovFwdMatch(rMatchIdx)) 
				 - (getDeriv (numerovBwdMatch, dx, 2)
						/numerovBwdMatch(2));

	return matchFu;							   	
}


SxDiracVec<Double> SxAtomDFT::getKSOrbital (const SxDiracVec<Double> &vHartree, 
												        const SxDiracVec<Double> &vXc,
													     double eIn, int z, int l)
{
	SxDiracVec<Double> numerovFwdMatch, numerovBwdMatch;

	// redundant with computeFwdNodes: left this way for readability
	SxDiracVec<Double> kx = getkx (r, vHartree, vXc, eIn, z, l);
	SxDiracVec<Double> sx(nIntervals+1);
	sx = 0.;

	int rMatchIdx = int(kx.getSize ()) - 3;

	for (int i = rMatchIdx; i >= 2; i--) 
		if (kx(i) > 0) {
			rMatchIdx = i;
			break;
		}

	double solR0 = getFwdBoundCondition (r(0), vHartree(0), vXc(0), 
												    eIn, z, l);
	double solR0dx = getFwdBoundCondition (r(1), vHartree(1), vXc(1),
												      eIn, z, l);
	//
	
	numerovFwdMatch = solveNumerovFwd (kx(SxIdx (0, rMatchIdx)), 
												  sx(SxIdx (0, rMatchIdx)), solR0, 
												  solR0dx, dx);
	
	double solRMax = getBwdBoundCondition (r(nIntervals), vHartree(nIntervals), vXc(nIntervals), 
														eIn);
	double solRMaxdx = getBwdBoundCondition (r(nIntervals-1), vHartree(nIntervals-1), vXc(nIntervals-1), 
														  eIn);

	numerovBwdMatch = solveNumerovBwd (kx(SxIdx (rMatchIdx, nIntervals)), 
												  sx(SxIdx (rMatchIdx, nIntervals)), 
												  solRMax, solRMaxdx, dx);

	SxDiracVec<Double> u(nIntervals+1);

	for(int i = 0; i <= rMatchIdx; i++) u(i) = numerovFwdMatch(i);

	double rescaleBwd = numerovFwdMatch(rMatchIdx)/numerovBwdMatch(0);
	
	for(int i = rMatchIdx + 1, j = 1; i <= nIntervals; i++, j++) 
		u(i) = rescaleBwd*numerovBwdMatch(j);  
	
	// normalize wavefunction
	SxDiracVec<Double> waveFu = (1./sqrt (r))
										 *(u/sqrt(((r.sqr ()*u.sqr ()).integrate (dx))));
	
	// fix nodal structure by comparison with boundary condition
	if (waveFu(0)*getFwdBoundCondition (r(0), vHartree(0), vXc(0), 
												    eIn, z, l) < 0) waveFu = -waveFu;
	
	return waveFu;							   	
}

SxDiracVec<Double> SxAtomDFT::getHartree (const SxDiracVec<Double> &trialRho,
													   int nElectrons, int PoissInitVal)
// TODO: use analytical formula (Gaunt coefficients -> SxRadBasis)
{
	// definition of Poisson equation
	SxDiracVec<Double> kx(nIntervals+1);
	kx = -0.25;

	SxDiracVec<Double> sx = -4*PI*sqrt (pow (r, 5))*trialRho;

	double solR0 = 0., solR0dx = 0.;
	
	switch (PoissInitVal) {

		// approximate density in Poisson equation by n[r0]
		case 0 : solR0 = (3*nElectrons*r(0) - 2*PI*pow (r(0), 3)*rMax*trialRho(0) 
						+ 2*PI*r(0)*pow(rMax, 3)*trialRho(0))/(3*sqrt (r(0))*rMax);
	
					solR0dx = (3*nElectrons*r(1) - 2*PI*pow (r(1), 3)*rMax*trialRho(1)
						+ 2*PI*r(1)*pow(rMax, 3)*trialRho(1))/(3*sqrt (r(1))*rMax);
		// integrate potential for r > r0
		case 1 : SxDiracVec<Double> vHartInt(trialRho.getSize () - 1);
					for (int i = 1; i < trialRho.getSize (); i++) 
                  
                  vHartInt(i-1) = 4 * PI * r(i) * r(i) * r(i) * trialRho(i) / 
										  (r(i) - r(0));
					
               double vHartIntd = vHartInt.integrate (dx);
					solR0 = sqrt (r(0)) * vHartIntd;
					solR0dx = sqrt (r(1)) * vHartIntd;
               
	}

	// integrate Poisson equation & match with second boundary condition
	SxDiracVec<Double> vHartree = solveNumerovFwd (kx, sx, solR0, solR0dx, dx); 
	
	// note: it might improve accuracy, if vHartree is computed on larger grid
	vHartree = (sqrt(r)*vHartree + 
				  (r/rMax)*(nElectrons - sqrt(rMax)*vHartree(nIntervals))) / r;

	return vHartree;
}

void SxAtomDFT::getXc (const SxDiracVec<Double> &rho)
{

	// definition according to Phys. Rev. B 22, 3812 (1982)
	double cX = - pow (3./(2*PI),(2/3.));
	double a = 0.0621814, b = 3.72744, c = 12.9352;
	double q = sqrt (4*c - b*b);
	double x0 = -0.10498;
	double P0 = x0*x0 + b*x0 + c;

	SxDiracVec<Double> rS = pow ((3./(4.*PI*rho)),(1/3.));
	SxDiracVec<Double> x = sqrt (rS);

	SxDiracVec<Double> P = x*x + b*x + c;
	SxDiracVec<Double> Y = atan (q/(2.*x + b)); 

	eX = (3/4.)*(cX/rS);
	vX = (cX/rS);

	eC = (a/2.)*(log (x*x/P) + (2*b/q)*Y - (b*x0/P0) *
					 (log ((x-x0)*(x-x0)/P) + (2*(b + 2*x0)/q)*Y)
					);

	vC = eC - (a/6.)*(c*(x - x0) - b*x*x0)/((x - x0)*P);

}


SxArray<SxDiracVec<Double> > SxAtomDFT::solveKSeqn (const SxDiracVec<Double> &vHartree,
													  const SxDiracVec<Double> &vXc,
                                         const SxArray<SxDiracVec<Int> > &nlPos,
													  double eLower, double eUpper,
													  int z, int nOrb, int nMax, int /*nEl*/)
{

	SxArray<SxDiracVec<Double> > phiKSVal(nOrb);

	//matchFuELoop is set to 1 so that loop is computed at least once
	double matchFuELoopL = 0., matchFuELoopU = 0., matchFuELoop = 0.,
          matchFuEOnNode = 0.; 
	double eLoopL = 0., eLoopU = 0., eLoop, eOnDiffNode, eOnNode;
	double eLowerVar = eLower, eUpperVar = eUpper;
	double eOldLower = eLower, eOldUpper = eUpper,eUpperBound = eUpper;
	double eOldMin = 0., eOldMax = 0.;

	int n = 1, l = 0;
	int orbNode = 0;
	int eVPos = 0;

   int eValCounter = 0;
   int nStart = 2;

	while ( eValCounter <= (nOrb-1) ) {

		// bisection until a value on the right branch is found
		/*
			note: to find 10 eigenvalues of hydrogen atom between -1,0 : 
					310 integrations are required
		*/

		double eMeanLU = (eLowerVar + eUpperVar)/2.;  
	
		int fwdNodes = computeFwdNodes (vHartree, vXc, eMeanLU, z, l);

		if ( fwdNodes == orbNode ) { 

			double matchFuEmeanLU = computeMatchFu (vHartree, vXc, eMeanLU, z, l);
			
			// define interval for another bisection based on minimal possible bounds
			double eMin = eOldLower, eMax = eOldUpper;	
			int nodesELoopL = -1, nodesELoopU = -1;
			bool loopCtrl = true;
			
			// loop to find values bracketing the root
			do {
	
				// restrict the lower boundary until a value on the current branch is found
				if(nodesELoopL != orbNode) {

					eLoopL = (eMeanLU + eMin)/2.; // eMax is set to eLoopM below
					matchFuELoopL =  computeMatchFu (vHartree, vXc, eLoopL, z, l);
					nodesELoopL = nNodes;
					eOldMin = eMin;

				}
				else {
					// when the criterion is fullfilled, the root has been bracketed
					if (matchFuELoopL*matchFuEmeanLU < 0) {
						eMin = eLoopL;
						eMax = eMeanLU;
						break;
					}
								
				}

				// restrict the upper boundary until a value on the current branch is found
				if(nodesELoopU != orbNode) {	

					eLoopU = (eMax + eMeanLU)/2.;
					matchFuELoopU =  computeMatchFu (vHartree, vXc, eLoopU, z, l);
					nodesELoopU = nNodes;
					eOldMax = eMax;

				}
				else {
					// when the criterion is fullfilled, the root has been bracketed
					if (matchFuELoopU*matchFuEmeanLU < 0) {
						eMin = eMeanLU;
						eMax = eLoopU;
						break;
					}
								
				}
		
				// situation that all three values on branch do not bracket the root
				if( (nodesELoopL == orbNode) && (nodesELoopU == orbNode) 
						&& (matchFuELoopL*matchFuELoopU) > 0 ) {
					
					// compute, which side is closer to the root 
					double distRootL = fabs (0. - matchFuELoopL);
					double distRootU = fabs (0. - matchFuELoopU);

					// set new interval with one boundary from previous interval
					if(distRootL < distRootU) {
						eOnNode = eLoopL;
						eOnDiffNode = eOldMin;
						matchFuEOnNode = matchFuELoopL;
					}
					else{
						eOnNode = eLoopU;
						eOnDiffNode = eOldMax;
						matchFuEOnNode = matchFuELoopU;
					}
 							
					// yet another bisection between two values on different nodes
					do {		
							eLoop = (eOnNode + eOnDiffNode)/2.;
						   matchFuELoop = computeMatchFu (vHartree, vXc, eLoop, z, l);
							
							if(nNodes != orbNode) eOnDiffNode = eLoop;
							else{
								// eLoop still might miss the root
								if( (matchFuELoop*matchFuEOnNode) < 0){
									eLoopL = min (eLoop, eOnNode);
									eLoopU = max (eLoop, eOnNode);
									loopCtrl=false;
								}
								else eOnNode = eLoop;
							}
							
					} while (loopCtrl);
				} 
				
				eMin = eLoopL;
				eMax = eLoopU;

			} while(loopCtrl);
			
			// bisection to narrow the root 
			double matchFuEMax = computeMatchFu (vHartree, vXc, eMax, z, l);

			matchFuELoop = 0.; // we re-use the eLoop variable
 
			while ((eMax - eMin) > 1e-5) {
				eLoop = (eMax + eMin)/2.;
				matchFuELoop = computeMatchFu (vHartree, vXc, eLoop, z, l);

				if (matchFuELoop*matchFuEMax > 0.) eMax = eLoop;
				else eMin = eLoop;

			}

			// use secant method to narrow the root further
			double ePrev, eCur = eMin, eNext = eMax;
			double matchFuDiff, matchFuEPrev = 0., matchFuECur;

			matchFuECur =  computeMatchFu (vHartree, vXc, eCur, z, l);

			while (fabs (matchFuECur - matchFuEPrev) >= 1e-10) {
				
				ePrev = eCur;
				matchFuEPrev = matchFuECur;
				eCur = eNext;
				matchFuECur = computeMatchFu (vHartree, vXc, eCur, z, l);

				matchFuDiff = ((matchFuECur - matchFuEPrev)/(eCur - ePrev));

				eNext = eCur - (matchFuECur/matchFuDiff);
				
				if(fabs (eNext - eCur) < 1e-15) break;

			}

			eigenKS(eVPos) = eNext;

			eUpperVar = eUpperBound;

			phiKSVal(eVPos) = getKSOrbital (vHartree, vXc, eNext, z, l);
         eValCounter++;

         if ( n == 1 ) eLower = eNext;

         if ( n < nMax) n++;
			else {
       
				l++;
            n = nStart;
            nStart++;
            eLowerVar = eLower;
            eUpperVar = eUpper;

			}

         // --- determine position of next eigenvalue
         int nlPosCount = 0;
         for (int i=0; i<=nlPos.getSize () ;i++) {

            eVPos = nlPosCount;

            if ( i == nlPos.getSize ()) {
               
               eVPos = nOrb;
               break;

            }

            if ( (nlPos(i)(0) == n) && (nlPos(i)(1) == l) ) break;
            
            nlPosCount++;

         }

         orbNode = (n - l - 1);

		} // end of loop that one has found one value on actual branch
	
		if((fwdNodes > orbNode) && (eMeanLU > eLowerVar)) {
			eOldUpper = eUpperVar; 
			eUpperVar = eMeanLU;	
		}

		if((fwdNodes < orbNode) && (eMeanLU < eUpperVar)) {
			eOldLower = eLowerVar;
			eLowerVar = eMeanLU;
		}
	}

	return phiKSVal;
}


double SxAtomDFT::getChemPot (double ekT, double chPot, double deltaChPot, 
									   int nElectrons)
{	
	double chPotFixed, fMatchPotFixed, chPotVar = 0.;
	double chPotL = chPot - deltaChPot, chPotU = chPot + deltaChPot;
	double fMatchL = nElectrons 
						  - (getOcc (eigenKS, ekT, chPotL)).sum ();
	double fMatchU = nElectrons 
					     - (getOcc (eigenKS, ekT, chPotU)).sum ();

	// try to find two values bracketing the root
	if (fMatchL*fMatchU > 0) {
			double distRootL = fabs (0. - fMatchL);
			double distRootU = fabs (0. - fMatchU);
			if(distRootL < distRootU) {
				chPotFixed = chPotL;
				fMatchPotFixed = fMatchL;
				deltaChPot *= -1;
			}
			else{
					chPotFixed = chPotU;
					fMatchPotFixed = fMatchU;
			}
 					
			double fMatchChPotVar = 0.;	
			while (fMatchChPotVar*fMatchPotFixed >= 0.) {					
			
				chPotVar = chPotFixed + deltaChPot;
				fMatchChPotVar = nElectrons 
							  		  - (getOcc (eigenKS, ekT, chPotVar)).sum ();
				deltaChPot += deltaChPot;
						
			} 					
			
			chPotL = min (chPotFixed, chPotVar);
			chPotU = max (chPotFixed, chPotVar);
		
		}
		
		// TODO: unify with other root searching problem 
		// bisection to narrow the root 
		fMatchU = nElectrons 
					 - (getOcc (eigenKS, ekT, chPotU)).sum ();
		double fMatchLoop = 0.; 

		while ((chPotU - chPotL) > 0.1) {
			
			double chPotLoop = (chPotL + chPotU)/2.;
			fMatchLoop = nElectrons 
							  - (getOcc (eigenKS, ekT, chPotLoop)).sum ();

			if (fMatchLoop*fMatchU > 0.) chPotU = chPotLoop;
			else chPotL = chPotLoop;

		}

		// use secant method to narrow the root further
		double chPotPrev, chPotCur = chPotL, chPotNext = chPotU;
		double fMatchDiff, fMatchPrev = 0., fMatchCur;

		fMatchCur = nElectrons 
						- (getOcc (eigenKS, ekT, chPotCur)).sum ();

		while (fabs (fMatchCur - fMatchPrev) >= 1e-10) {
			chPotPrev = chPotCur;
			fMatchPrev = fMatchCur;
			chPotCur = chPotNext;
			fMatchCur = nElectrons 
							- (getOcc (eigenKS, ekT, chPotCur)).sum ();

			fMatchDiff = ((fMatchCur - fMatchPrev)/(chPotCur - chPotPrev));

			chPotNext = chPotCur - (fMatchCur/fMatchDiff);
				
			if(fabs (chPotNext - chPotCur) < 1e-15) break;

		}
		
		return chPotNext;
	 
}



int main (int argc, char **argv)
{	

   SxCLI cli (argc, argv); // might be replaced by input-file

	// grid specifications
	double r0 = cli.option ("-r0|--r0", "lower boundary on grid").toDouble(1e-15);
	double rMax = cli.option 
					  ("-rMax|--rMax", "upper boundary on grid").toDouble(5);
	// TODO: Richardson's extrapolation to the limit for nIntervals
	int nIntervals = cli.option 
						  ("-n|--nIntervals", "number of grid intervals").toInt(10000);

	// system specifications	
	int nEl = cli.option ("-l|--nElectrons", "number of electrons").toInt(18);
	int z = cli.option ("-z|--z", "nuclear charge").toInt(18);
	int nOrb = cli.option ("-f|--nOrb", "number of orbitals").toInt(5);
	double ekT = cli.option 
					 ("-t|--ekt", "electronic temperature").toDouble(0.); 
					 //atomic units k*T = k*E [H]/k = E[H]

	// convergence parameter
   double kappa = cli.option 
						("-k|--kappa", "mixing parameter for density").toDouble(0.7);
	double eps = cli.option 
					 ("-eps|--epsilon", "energy convergence value").toDouble(1e-6);
 
   cli.finalize ();

   int nLoops = 0;

   //initialize orbitals
   int n = 1, l  = 0, nMax = 0;
   int nElCount = nEl;
   int it = 0;

   // determine size   
   SxArray<SxDiracVec<Int> > nlPos(nOrb);
	SxDiracVec<Double> fOcc(nOrb), nlTemp(2);
 	

   while ( nElCount > 0 ) {

      if ( l == 0 ) nMax++;

      if ( (nElCount - 2*(2*l + 1)) >= 0 ) {
       
         fOcc(it) = 2*(2*l+1);
      
      }
      else {

         fOcc(it) = nElCount;
      
      }
        
      nlPos(it) = SxList<Int::Type> () << n << l;
      it++;
      nElCount -= 2*(2*l+1);			
      getQNums (&n, &l);

   }

   // case that are more orbital & change input of function approprietly   
   for (int i=it+1; i<=nOrb; i++) {

      if ( l==0 ) nMax++;

      fOcc(i) = 0.;
      nlPos(i) = SxList<Int::Type> () << n << l;
      getQNums (&n, &l);	

   }

   nOrb = int(nlPos.getSize ());
 
	SxAtomDFT atomDFT (r0, rMax, nIntervals, nOrb);
	SxArray<SxDiracVec<Double> > phiKS;
	
	SxDiracVec<Double> vHartree, vXc(nIntervals+1), eXc(nIntervals+1);  
	
	// initial guess for trial density
	// see Acta Phys. Polonica B 38, 3263 (2007)
	SxDiracVec<Double> 
		rho = (pow (nEl, 4.)/(64.*PI))*exp(-0.5*nEl*atomDFT.r),
		rhoIn = rho;
	
	double eTotIn = 1.e50, eTot = 0.;

	// TODO: nan-handling for too large intervals
	double eLower = -200., eUpper = 1; 

	// TODO: better solution for too small interval
	double chPot = -1;
	double deltaChPot = 1.e3;
	

	do {

		rho = (1 - kappa)*rhoIn  + kappa*rho;

		rhoIn = rho;
		
   	vHartree = atomDFT.getHartree (rhoIn, nEl, 0);

		// SxXCFunctional atomXc;
		// for (int i = 0; i < rhoIn.getSize(); i++) {
		//	atomXc.computeLDA (rhoIn(i));
		//	vXc(i) = atomXc.vXc(0);
		//	eXc(i) = atomXc.epsXc;
		// }	
					
		atomDFT.getXc(rhoIn);
		vXc = atomDFT.vX + atomDFT.vC;
		eXc = atomDFT.eX + atomDFT.eC;	

		//check for real values of boundary conditions
		SX_CHECK_VARS (eLower <= (vHartree (0) + vXc(0)), 
	 					   eLower, (vHartree(0), vXc(0)));
		SX_CHECK_VARS (eLower <= (vHartree(nIntervals) + vXc(nIntervals)), 
	 					   eLower, (vHartree(nIntervals), vXc(nIntervals)));
		SX_CHECK_VARS (eUpper <= (vHartree (0) + vXc(0)), 
	 					   eUpper, (vHartree(0), vXc(0)));
		SX_CHECK_VARS (eUpper <= (vHartree(nIntervals) + vXc(nIntervals)), 
	 					   eUpper, (vHartree(nIntervals), vXc(nIntervals)));
		
		phiKS = atomDFT.solveKSeqn (vHartree, vXc, nlPos, eLower, eUpper, z, nOrb, nMax, nEl);

		double chPotTrial = chPot; // use potential from previous iterations
		
		if ( ekT > 1e-5 && (fabs (eTot - eTotIn)) < 1e5*eps ) {
		 			chPot = atomDFT.getChemPot (ekT, chPotTrial, deltaChPot, nEl);
					fOcc = getOcc (atomDFT.eigenKS, ekT, chPot);
		}

		rho=0; // now we are building a new rho
		for (int OrbIt = 0; OrbIt < fOcc.getSize (); OrbIt++) {

			SxDiracVec<Double> rhoSum = fOcc(OrbIt)*phiKS(OrbIt).sqr ();

			rho += rhoSum;
			
		}
    
		rho *= (1/(4.*PI));

		double eigenSum = (fOcc*atomDFT.eigenKS).sum ();
		double potInt = 4.*PI*
							 ((atomDFT.r)*(atomDFT.r)*(atomDFT.r)*rho*
							 (eXc - vXc - (1/2.)*vHartree)).integrate (atomDFT.dx);

		eTotIn = eTot;
		eTot = eigenSum + potInt;

		cout << "eTotTmp: " << eTot << endl;
      nLoops++;
	
		} while((fabs (eTot - eTotIn)) > eps || 
				  (sqrt (((rho - rhoIn).sqr ()).sum ())) > eps);

		cout << "\n\n" << endl;

		cout << setprecision(8) << "eigenValues: " << atomDFT.eigenKS << endl;

		cout << setprecision(8) << "eTot: " << eTot << endl;
		
		cout << setprecision(8) << "eHartree: " << 4.*PI*
							 ((atomDFT.r)*(atomDFT.r)*(atomDFT.r)*rho*
							 ((1/2.)*vHartree)).integrate (atomDFT.dx) << endl;

		cout << setprecision(8) << "eNucleus: " <<  -4.*PI*z*
							 ((atomDFT.r)*(atomDFT.r)*rho).integrate (atomDFT.dx) << endl;

		cout << setprecision(8) << "eXc: " << 4*PI*
								((atomDFT.r)*(atomDFT.r)*(atomDFT.r)
								 *rho*eXc).integrate (atomDFT.dx) << endl;

      cout << "Number of loops: " << nLoops << endl;

      
      return 0;
}
