// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxConfig.h>
#include <SxBinIO.h>
#include <SxParser.h>
#include <SxAtomicStructure.h>
#include <SxNeighbors.h>

int main (int argc, char **argv)
{
   // --- init S/PHI/nX Utilities
   initSPHInXMath ();

   // --- parse command line
   SxCLI cli (argc, argv);
   cli.preUsageMessage =
      "This add-on calculates the average bond directions for every species.\n"
      "Considered is the next nearest neighbor shell.";
      
   cli.authors = "B. Lange";

   bool sxbFile = cli.option ("-b|--sxb", "input file is binary waves file "
                              "rather than S/PHI/nX input file")
                  .toBool ();
   SxString inFile 
      = cli.option ("-i|--input", "input file", 
                    "take original input file")
        .toString (sxbFile ? "waves.sxb" : "input.sx");
   cli.last ().defaultValue = "default: input.sx, or waves.sxb for --sxb flag";

   bool projectXY = cli.option ("--pxy", "flag"
                                "project direction to xy-plane").toBool ();

   
   cli.finalize ();

   // --- read input
   SxAtomicStructure structure;
   SxArray<SxString> chemNames;
   if (sxbFile)  {
      try {
         SxBinIO io (inFile, SxBinIO::BINARY_READ_ONLY);
         structure.read (io);
         SxString chemNameList;
         io.read("chemNames", &chemNameList);
         chemNames = chemNameList.tokenize (',');
         io.close ();
      } catch (SxException e)  {
         e.print ();
         SX_EXIT;
      }
   } else  {
      SxParser parser;
      SxConstPtr<SxSymbolTable> tree;
      tree = parser.read (inFile, "std/structure.std");
      structure = SxAtomicStructure (&*tree);
      chemNames = SxSpeciesData::getElements (&*tree);
   }

   
   cout << "Cell volume: " << structure.cell.volume << " bohr^3" << endl;


   SxVector3<Int> gMesh 
      = SxGrid::suggestMesh (structure.cell, structure.getNAtoms () / 10 + 1);
   SxGrid grid (structure, gMesh);

   SxNeighbors neighbors;

   int nSpecies = structure.getNSpecies ();
   SxVector<Int> nBonds (nSpecies);
   nBonds.set (0);
   SxArray<SxList<int> > bondDirCounter (nSpecies);
   SxArray<SxList<Coord> > bondDir (nSpecies);
   SxArray<SxList<Coord> > bondDirSqr (nSpecies);

   
   for (int iSpecies = 0; iSpecies < nSpecies; iSpecies++)  {
      // Collect Data for all atoms of Species
      for (int iAtom = 0; iAtom < structure.getNAtoms(iSpecies); iAtom++)  {
         // Find neares Neighbor Shell
         double maxDist = 5.0;
         neighbors.compute (grid, structure, structure(iSpecies,iAtom), 
                            maxDist, SxNeighbors::StoreAll);
         int nNeighbors = neighbors.absPositions.getNAtoms ();
         double minDist = sqrt(neighbors.distSqr.minval());
         for (int ja = 0; ja < nNeighbors; ++ja)  {
            Coord tau =  neighbors.absPositions(ja);
            double d = sqrt(neighbors.distSqr(ja));
            Coord dir = (tau - structure(iSpecies,iAtom))/d;
            if (projectXY) {
               dir(2) =  0.0;
               dir /= sqrt(dir.normSqr());
            }
            if ((d-minDist) > 1.0) continue;
            bool newbond = true;
            for (int iBond = 0; iBond< nBonds(iSpecies); iBond++)  {
               double cosPhi = (dir * bondDir(iSpecies)(iBond)).sum();
               //cout << SEPARATOR;
               //dir.print();
               //bondDir(iSpecies)(iBond).print();
               //sxprintf("iSpecies: %i, iAtom: %i, iNeighbor: %i, " 
               //      "dist is %5.3f, Angle is %5.2f\n",
               //      iSpecies, iAtom, ja, d, cosPhi);
               //cout << SEPARATOR;
               if (cosPhi > 0.95) {
                  bondDir(iSpecies)(iBond) =
                     bondDirCounter(iSpecies)(iBond) * bondDir(iSpecies)(iBond)
                     + dir;
                  bondDirSqr(iSpecies)(iBond) =
                     bondDirCounter(iSpecies)(iBond) 
                        * bondDirSqr(iSpecies)(iBond)
                     + dir*dir;
                  bondDirCounter(iSpecies)(iBond)++;
                  bondDir(iSpecies)(iBond) /= bondDirCounter(iSpecies)(iBond);
                  bondDirSqr(iSpecies)(iBond) /=
                     bondDirCounter(iSpecies)(iBond);
                  newbond = false;
                  break;
               }
            }

            if (newbond)  {
               bondDirCounter(iSpecies).append(1);
               bondDir(iSpecies).append(dir);
               bondDirSqr(iSpecies).append(dir*dir);
               nBonds(iSpecies)++;
            }
         }
      }
   }

   //Output
   for (int iSpecies = 0; iSpecies < nSpecies; iSpecies++)  {
      cout << SEPARATOR;
      sxprintf ("For Species %s %i bonds have been found:\n",
            chemNames(iSpecies).ascii(), nBonds(iSpecies));
      for (int iBond = 0; iBond < nBonds(iSpecies); iBond++)   {
         Coord dir = bondDir(iSpecies)(iBond);
         Coord varSqr = (bondDirSqr(iSpecies)(iBond) 
            - dir*dir);
         Coord var; 
         var(0) = varSqr(0) > 0.0 ? sqrt(varSqr(0)) : 0.0;
         var(1) = varSqr(1) > 0.0 ? sqrt(varSqr(1)) : 0.0;
         var(2) = varSqr(2) > 0.0 ? sqrt(varSqr(2)) : 0.0;
         Coord delta = dir+var;
         double cosPhi = (delta*dir).sum ()/sqrt(delta.normSqr()*dir.normSqr());
         if (fabs(cosPhi - 1.0) < 1e-6) cosPhi = 1.0;
         if (fabs(cosPhi + 1.0) < 1e-6) cosPhi = -1.0;
         double deltaDegree = acos(cosPhi) * 180 / PI;
         dir /= sqrt(dir.normSqr());
         sxprintf ("[% 1.3f, % 1.3f, % 1.3f] (+- % 6.2f degree)\n", 
               dir(0), dir(1), dir(2), deltaDegree);
      }
   }

   // Orientation between species
   if (projectXY) {
      cout << endl << endl;
      for (int iSpecies = 0; iSpecies < nSpecies; iSpecies++)  {
         for (int iBond = 0; iBond < nBonds(iSpecies); iBond++)   {
            cout << SEPARATOR;
            Coord dirI = bondDir(iSpecies)(iBond);
            for (int jSpecies = 0; jSpecies < nSpecies; jSpecies++)  {
               for (int jBond = 0; jBond < nBonds(jSpecies); jBond++)   {
                  if ((iSpecies == jSpecies) && (iBond == jBond)) continue;
                  Coord dirJ = bondDir(jSpecies)(jBond);
                  double cosPhi = (dirI*dirJ).sum ()
                     /sqrt(dirI.normSqr()*dirJ.normSqr());
                  if (fabs(cosPhi - 1.0) < 1e-6) cosPhi = 1.0;
                  if (fabs(cosPhi + 1.0) < 1e-6) cosPhi = -1.0;
                  double deltaDegree = acos(cosPhi) * 180 / PI;
                  sxprintf ("%-2s bond %2i rotated against %-2s bond %2i by "
                        "%6.2f degree\n",  chemNames(iSpecies).ascii(),
                         iBond+1, chemNames(jSpecies).ascii(), jBond+1,
                         deltaDegree);
               }
            }
         }
      }
   }
}
