// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <stdlib.h>
#include <iostream>
#include <math.h>
#include <stdio.h>
#include <SxError.h>
#include <SxPrecision.h>
#include <SxConstants.h>
#include <SxString.h>
#include <SxSymbolTable.h>
#include <SxVector.h>
#include <SxBinIO.h>
#include <SxCLI.h>

SxBinIO io;
SxList<SxList<SxVector3<Double> > > tau;
SxList<SxString> speciesNameList;
SxMatrix3<Double> cell;
	
int saveStructureRaw (SxString);
int saveStructurePdb (SxString);
int saveStructureOut (SxString);
int saveStructureSx (SxString);

int loadStructureRaw (SxString);
int loadStructurePdb (SxString); 
int loadStructureOut (SxString); 
int loadStructureSx (SxString); 


void generateCubicLattice (double, SxString); 
void printHelp ();
void printUsage ();


int main (int argc, char **argv)
{
	int i, j;
	SxString          commandLine;
	SxList<SxString>  commandList;
	SxList<SxString>  subCommandList;
	SxString          outputFile, inputFile, inModus, outModus;
	bool hiddenCommand = false;
	double aLat = 0.;

	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			cell (i, j) = 0.;
		}
	}
	
	

	//initSPHInXIO ();
  
	// --- ReadiSx in command-Line 
	for (i = 1; i < argc; i++) {
		 commandLine += SxString(argv[i]);
		 commandLine += SxString(" ");
	 }
 
	commandList = commandLine.tokenize ('-');

	if ((commandList.getSize () != 1) && (commandList.getSize () != 2)){
			printUsage ();
	} else {
	
	// --- ParsiSx command-Line	
	bool correctCommand;	
	subCommandList = commandList(0).tokenize (' ');
	if (commandList.getSize () == 1) {
	  	if (subCommandList(0) == SxString ("h")) {
			printHelp ();
		}
	}

	if (commandList.getSize () == 2) {

		for (i = 0; i < 2; i++) {
			correctCommand = false;
			subCommandList = commandList(i).tokenize (' ');
			
			if (subCommandList(0) == SxString ("c")) {
				correctCommand = true;
				hiddenCommand = true;
				aLat = subCommandList(1).toDouble ();
			}

			
			if (subCommandList(0) == SxString("i")) {
				correctCommand = true;
				inputFile = subCommandList(1);
			}
				
			if (subCommandList(0) == SxString("o")) {
				correctCommand = true;
				outputFile = subCommandList(1);
			}
			
			if (correctCommand == false) {
				cout << "Unknown command: " << subCommandList(0) << endl;
				exit (0);
			}
		}
	}		

	// --- Output of ParsiSx Results
	if (!hiddenCommand) {
	cout << "Input-File";
	cout << " " << inputFile.ascii () ;
	cout << endl;
	}
	
	cout << "Output-File:";
	cout << " " << outputFile.ascii () ;
	cout << endl;

	// --- DeterminiSx Conversion Modus
	SxList<SxString> sides = inputFile.tokenize ('.');
	
if (!hiddenCommand) {
	if (sides.getSize() == 1)  {
		cout << "Please provide a consumerable file endiSx for input file: .raw, .pdb or .out" << endl;
		SX_EXIT;
	}	
	if (sides.getSize() >= 2) {
		SxString end = sides(sides.getSize() - 1);
		if ((end == "raw") || (end == "pdb") || (end == "out") || end == "raw2") {
			inModus = end;
		} else {
			cout << "Format " << end << " not supported !";
			exit (0);
		}
	}
}
	sides = outputFile.tokenize ('.');
	if (sides.getSize() == 1)  {
		cout << "Please provide a consumerable file endiSx for input file: .sx, .raw, .pdb or .out" << endl;
		SX_EXIT;
	}
	if (sides.getSize() >= 2) {
		SxString end = sides(sides.getSize() - 1);
		if ((end == SxString("sx")) || (end == SxString("pdb")) 
				                      || (end == SxString("out")) || ("raw")) {
			outModus = end;
		} else {
			cout << "Format " << end << " not supported !";
			exit (0);
		}
		
	}	
  	int implemented = 0;
	SxString response;


	response = SxString("Converting ") + inModus;
	
	
	if (inModus == SxString("raw2")) {
		implemented = loadStructureRaw(inputFile);
		for (i = 0; i < tau.getSize (); i++) {
			for (j = 0; j < tau(i).getSize (); j++) {
				tau(i)(j) = tau(i)(j)*1.889726;
			}
		}
	}
	if (inModus == SxString("raw")) implemented = loadStructureRaw(inputFile);
	if (inModus == SxString("pdb")) implemented = loadStructurePdb(inputFile);
//	if (inModus == SxString("sx"))  implemented = loadStructureSx(inputFile);
	if (inModus == SxString("out")) implemented = loadStructureOut(inputFile);

		
	response += SxString (" to ");
  	response  += outModus;

	for (i = 0; i < speciesNameList.getSize(); i++) {
		cout << speciesNameList(i);
	}
	
	if (outModus == SxString("raw")) implemented *= saveStructureRaw(outputFile);
	if (outModus == SxString("pdb")) implemented *= saveStructurePdb(outputFile);
	if (outModus == SxString("sx"))  implemented *= saveStructureSx(outputFile);
	if (outModus == SxString("out")) implemented *= saveStructureOut(outputFile);
	
		
	
	if (implemented == 0) response += SxString (" is not implemented yet");
	
	if (hiddenCommand) {
		generateCubicLattice(aLat, outputFile);
	}
	 	
	cout << response.ascii () << endl;
}
}

void generateCubicLattice (double aLat, SxString Filename) {
  	FILE *fpsave;
   int i, j, k;
	double xcoord, ycoord, zcoord;
	SxString head, tail;

 	head = head + SxString("needs species_1;");
	head = head + SxString("\n\n");
	
	tail = tail + SxString("species { \n");
				tail = tail + SxString("   include species_1;\n");

	for (i = 0; i < 6 ; i++) {
		for (j = 0; j < 6 ; j++) {
			for (k = 0; k < 6 ; k++) {
				zcoord = aLat/(double)6/2 + (double)k*aLat/(double)6;
				ycoord = aLat/(double)6/2 + (double)j*aLat/(double)6;
				xcoord = aLat/(double)6/2 + (double)i*aLat/(double)6;

								//tail = tail + speciesNameList(i);
				tail = tail + SxString("   atom { coords = [");
				tail = tail + SxString(xcoord)+SxString(", ");
				tail = tail + SxString(ycoord)+SxString(", ");
				tail = tail + SxString(zcoord)+SxString("]; movable;}\n");
			}
		}
	}
	tail = tail + SxString("}\n\n");
	head = head + tail;
	fpsave = fopen (Filename.ascii (), "w");
	fprintf (fpsave, head.ascii ()); 
	fclose (fpsave);
}

int loadStructureRaw (SxString Filename) 
{
   FILE *fpload;
   const int BUFLEN = 10024;
   char buffer[BUFLEN];
   char *bs;
   int nrSpecies = 0;
	SxVector3<Double> coord;
	SxList<SxVector3<Double> > species;
	SxList<SxString> coordlist;

	if ( !(fpload = fopen (Filename.ascii (), "r")) ) {
      sxprintf ("Can't open file %s",Filename.ascii ()); 
      exit (0);
   } 
	
	SxString line;
	SxList<SxString> tk;
	
	for (bs = fgets (buffer, 10024, fpload); bs != NULL;bs = fgets (buffer, 10024, fpload))
  {
         line = SxString(bs);
			tk = line.tokenize ('/');
			switch (tk.getSize()) { 
				case 1:
					coordlist = tk (0).tokenize (' ');
	 				if (coordlist.getSize () < 3) {
  						cout << "Corrupt File a!"<< endl;
						cout << coordlist.getSize () << endl;
						exit(0);
					} else {
						coord(0) = coordlist (0).toDouble ();
						coord(1) = coordlist (1).toDouble ();
						coord(2) = coordlist (2).toDouble ();
						species.append (coord);
					}
					break;
				case 2:
					if (nrSpecies == 0) (nrSpecies = 1); 
						else {
							tau. append(species);
						}
					speciesNameList.append (tk(1).tokenize('\n')(0));
					species. removeAll();
					coordlist = tk (0).tokenize (' ');
	 				if (coordlist.getSize () < 3) {
  						cout << "Corrupt File b!"<< endl;
						exit(0);
					} else {
						coord(0) = coordlist (0).toDouble ();
						coord(1) = coordlist (1).toDouble ();
						coord(2) = coordlist (2).toDouble ();
						species.append (coord);
					}

					break;
				default:
					cout << "Corrupt File c!"<< endl;
					exit(0);
			}
			
			
  }   
	tau. append(species);


/*	for (i = 0; i < tau.getSize(); i++) {
		cout << speciesNameList(i);
		for (j = 0; j < tau(i).getSize(); j++) {
			cout << tau(i)(j) << endl;
		}
	}
*/
	fclose (fpload);
	return 1;
}


int loadStructureSx (SxString /*Filename*/)
{
	return 0;
}

int loadStructurePdb (SxString Filename)
{
	 FILE *fpload;
   const int BUFLEN = 10024;
   char buffer[BUFLEN];
   char *bs;
	SxVector3<Double> coord;
	SxList<SxVector3<Double> > species;
	SxList<SxString> coordlist;
	bool first = true;
	SxString previousSpecies =  SxString(" ");

	tau.removeAll ();
	speciesNameList.removeAll ();
	
	if ( !(fpload = fopen (Filename.ascii (), "r")) ) {
      sxprintf ("Can't open file %s",Filename.ascii ()); 
      exit (0);
   } 
	
	SxString line;
	SxList<SxString> tk;
	
	for (bs = fgets (buffer, 10024, fpload); bs != NULL;bs = fgets (buffer, 10024, fpload))
  {
         line = SxString(bs);
			coordlist = line.tokenize (' ');
			if (coordlist.getSize () < 6) {
				cout << "Corrupt File a!"<< endl;
				cout << coordlist.getSize () << endl;
				exit(0);
			} else {
				coord(0) = coordlist (3).toDouble () * A2B;
				coord(1) = coordlist (4).toDouble () * A2B;
				coord(2) = coordlist (5).toDouble () * A2B;
						
				if ((coordlist(2) == previousSpecies) || (first)) {
					species.append (coord);
					if (first) {
						speciesNameList.append (coordlist(2)); 
						//speciesNameList.append (coordlist(2) + SxString("\n"));
						previousSpecies = coordlist(2);
						first =false;
					}
				} else {
					tau. append(species);
					speciesNameList.append (coordlist(2) );
					//speciesNameList.append (coordlist(2) + SxString("\n"));
					species. removeAll();
					coord(0) = coordlist (3).toDouble () * A2B;
					coord(1) = coordlist (4).toDouble () * A2B;
					coord(2) = coordlist (5).toDouble () * A2B;
					species.append (coord);
					previousSpecies = coordlist(2);

				}

			}	
  }
	tau.append(species);
				
	fclose (fpload);
	return 1;
}

int loadStructureOut (SxString Filename)
{
   io.open (Filename, SxBinIO::ASCII_READ_ONLY);

   tau = io.loadStructureFHI98 ();
	cell = io.loadCellFHI98 ();
	speciesNameList = io.loadSpeciesFHI98 ();
	return 1;
}



int saveStructureSx (SxString Filename) 
{
   FILE *fpsave;
   int i, j;
	SxString head, tail;

	for (i = 0; i < speciesNameList.getSize (); i++) {
	  	head = head + SxString("needs species_");
		head = head + SxString(i+1);
		head = head + SxString("; //");
		head = head + speciesNameList(i);
		head = head + SxString("\n");
	}
	head = head + SxString("\n\n");
	
	for (i = 0; i < tau.getSize (); i++) {
		tail = tail + SxString("species { \n");
		tail = tail + SxString("   include species_");
		tail = tail + SxString(i + 1);
		tail = tail + SxString(";\n"); 
		//tail = tail + speciesNameList(i);
		for (j = 0; j < tau(i).getSize (); j++) {
				tail = tail + SxString("   atom { coords = [");
				tail = tail + SxString(tau(i)(j)(0))+SxString(", ");
				tail = tail + SxString(tau(i)(j)(1))+SxString(", ");
				tail = tail + SxString(tau(i)(j)(2))+SxString("]; movable;}\n");
		}
		tail = tail + SxString("}\n\n");
	}

	head = head + tail;
	fpsave = fopen (Filename.ascii (), "w");
	fprintf (fpsave, head.ascii ()); 
	fclose (fpsave);
	return 1;
}

int saveStructureRaw (SxString /*Filename*/)
{
	
	// io.saveStructure(Filename, tau, speciesNameList);
	return 0;
}

int saveStructurePdb (SxString /*Filename*/)
{
	sxprintf("%20.10f",  tau(0)(0)(0));

	//io.saveStructurePDB(Filename, tau, speciesNameList);

	return 0;
}

int saveStructureOut (SxString /*Filename*/)
{
	//io.saveStructureAndCell(Filename, tau, speciesNameList, cell);
	return 0;
}



void printHelp() {
	 cout 
					 << "This tools converts structure formats." << endl;
				 cout << "Currently supported are: " 
				 	 << ".sx (sfhiSxx)," 
			 		 << "raw format (just coordinates in lines)\n" 
					 << endl;
				 cout << "Example for convertiSx: \n\n"
					 << "Sxconv.x -i tau -o tau.sx \n"
				    << "... this converts a raw file to the .sx format (the raw-format is idetified by haviSx no endiSx).\n\n"
				    << "In order to allow for different species the raw format must be \nprovided with marks! \n \n \nExample for H2O: \n\n "
					 << "Raw-Format:\n\n  0.  0.  0. //O \n  1.  1.  0. //H \n -1. -1.  0.  \n" <<endl;
					 

				 exit(0);
}

void printUsage () 
{
		cout << "usage: Sxconv.x -i [Input-Structure-Format] -o [Output-Structure-Format] "<< endl;
		cout << "help: Sxconv.x -h" <<
		  endl;	

}
