// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxFermi.h>
#include <SxSpectrum.h>
#include <SxRho.h>
#include <SxMuPW.h>
#include <SxProjector.h>

// --- cutoff-function
inline SxDiracVec<Double> fCut(SxDiracVec<Double> rad, double rcut)
{
      int i = 0, nr = int(rad.getSize ());
      SxDiracVec<Double> fCut(nr);
      fCut.set(0);

      while (rad(i) <= rcut)  {
         fCut(i) = 1.;
         // fCut(i) = sqr((sin(PI*rad(i)/1.25)/(PI*rad(i)/1.25)));
         i++;
      }

      // fCut(i) = (rad(i) < rcut) ? (exp(-rad(i)*rad(i)/0.125)) : 0.;
      // fCut(i) = (rad(i) < rcut) ? 1. : (54.5982*exp(-rad(i)/0.25));

      return fCut;
}

SxArray<SxDiracMat<Complex16> > computePsiG (const SxSymbolTable *table, 
                                             SxPtr<SxGkBasis> GkPtr, 
                                             double rcut, 
                                             int nk, int ia, int is, int l)
{

   SxPseudoPot psPot(&*table);
   SxGkBasis &Gk = *GkPtr;
   SxConstPtr<SxRadBasis> rPtr = 
      SxConstPtr<SxRadBasis>::create(psPot.rad, psPot.logDr);  

   SxArray<SxDiracMat<Complex16> > projMatr (nk);
   SxArray< SxArray<SxDiracVec<TReal8> > > psProj = psPot.pseudoPsi;
  
   for (int iSpecies = 0; iSpecies < psProj.getSize (); iSpecies++) {
      
      for (int j = 0; j < psProj(iSpecies).getSize (); j++) {

         psProj(iSpecies)(j) = 
            fCut (psPot.rad(iSpecies),rcut) * psProj(iSpecies)(j);
         double Norm = 1./ ((psProj(iSpecies)(j).conj () 
                        * psPot.pseudoPsi(iSpecies)(j)
                        * psPot.rad(iSpecies).cub ()
                        ).integrate(psPot.logDr(iSpecies)));
         psProj(iSpecies)(j) *= Norm;

      }

   }

   SxMuPW mu(psProj, rPtr, psPot.pseudoFocc, GkPtr);

   int nOrb = mu.getNStates ();
   SxList<int> muPos;

   cout << "mu orbitals:" << endl;
   for (int i = 0; i < nOrb; i++) {
      if ( (mu.psiOrb(i).iSpecies == is) && (mu.psiOrb(i).iAtom == ia))  {
         if (l == -1 || l == mu.psiOrb(i).l)  {
            muPos << i;
            cout << ' ' << i;
         }
      }
   }
   cout << endl;

   for (int ik = 0; ik < nk; ik++) {

      SxDiracMat<Complex16> mulMatrix(Gk(ik).ng, muPos.getSize ());

      for (int muPosIt = 0; muPosIt < muPos.getSize (); muPosIt++) {

         int muPosTmp = muPos(muPosIt);
         PsiG muI = mu (muPosTmp,0,ik);
         mulMatrix.colRef(muPosIt) <<= muI;

      }  
         projMatr(ik) = mulMatrix;  

   }

   return projMatr;

}

int main (int argc, char **argv)
{

   initSPHInXMath ();
   
   SxCLI cli(argc, argv);

   cli.preUsageMessage 
      = SxString(" Add-on calculates density of states (Total, partial, z-resolved)\n"
                 "  Gaussian broadening applied to each state").wrap ();
   
   int wGroup = cli.newGroup ("sxb");
   SxString wavesFile
            = cli.option ("-w|--waves","file","input waves file")
             .toString   ("waves.sxb");
   int iGroup = cli.newGroup ("sx");
   cli.excludeGroup(wGroup);
   
   SxString inputFile 
      = cli.option("--input","file","read k-vectors from this input file")
        .toString ("input.sx");
   
   cli.setGroup (cli.generalGroup);

   cli.newGroup ("pdos"); 
   bool pdos
      = cli.option ("-p|--pdos", "pdos", 
                    "compute projected density of states ").toBool ();

   // Species of atom
   int is = cli.option ("-is", "iSpecies", "Species of atom")
                 .toInt ();

   // Number of atom
   int ia = cli.option ("-ia", "iAtom", "Number of Atom")
                 .toInt ();

   // l number
   int l = cli.option ("-l", "l-number", "only this l-channel").toInt (-1);
   cli.last ().defaultValue = "default: sum all channels";

   //rCut-radius (rCut of l=0 pseudopotential)
   double rcut = 
      cli.option ("-rcut", "cutoff", "cutoff radius of projector functions")
                 .toDouble ();

   cli.newGroup ("zdos");
   bool zdos
      = cli.option ("-z|--zdos", "zdos", 
               "compute z-resolved density of states ").toBool ();
   
   int dim = cli.option ("--a1|--a2|--a3","which axis to resolve").toChoice ();
   if (dim < 0) dim = 2;
   cli.last ().defaultValue = "default: a3";

   SxString epsFile 
      = cli.option ("-e|--eps", "eps file", 
                    "use energies in <eps file> rather than those in "
                    "waves file").toString ();

   SxString potFile = cli.option("--sigfile","file",
                                 "apply model self energy (z-dependent)")
                      .toString ("");

   SxString outFile
            = cli.option ("-o","file","prefix of the output file;"
                          ".dat is always added to the end of the output file")
             .toString   ("tdos");
 
   SxFFT::plannerCLI (cli);
   double broadening = cli.option ("-b|--broad", "energy [eV]", 
                                   "Gaussian weight factor, eV")
                       .toDouble(0.1) / HA2EV;

   double shift = cli.option ("-s|--shift", "energy [eV]", 
          "shift borders of energy interval from (energyMin, energyMax) "
          "to (energyMin-shift, energyMax+shift), eV")
                  .toDouble(2. * broadening * HA2EV) / HA2EV;
   cli.last ().defaultValue = "default: 2 * broadening";

   SxList<double> range = cli.option("--range","range","Emin:Emax (eV)")
                          .toDoubleList ();
   if (range.getSize () != 2 && range.getSize () != 0 && !cli.error)  {
      cout << "Illegal range." << endl;
      cli.setError ();
   }
   
   bool useOcc = cli.option("--useocc",
                            "weight states with Fermi occupation numbers\n"
                            "otherwise all states have the same weight")
                 .toBool ();
 
   int nPerPeak = cli.option ("--fine", "integer", 
                              "energy resolution parameter, approximately "
                              "number of points used to display 1/2 Gauss peak")
                  .toInt(15);
   
   cli.finalize ();

   // setup for non-default uses
   if (pdos) { 
      is--; ia--;
      outFile = "pdos";
   }
   if (zdos) outFile = "zdos";

   // --- read eps and weights from waves file
   int nk = -1;
   SxVector<Double> kWeights;
   SxFermi fermi;   
   SxParser parser;      
   SxParser::Table table = parser.read(inputFile);
   SxAtomicStructure str(&*table); 

   
   if (cli.groupAvailable(iGroup))  {
      SxKPoints kp(str.cell, &*table);
      nk = kp.getNk ();
      kWeights = kp.weights;
      fermi = SxFermi (1. /* nElec */, 1 /* nStates */,1 /* nSpin */,kp);
   } else {
      try  {
         SxBinIO io (wavesFile, SxBinIO::BINARY_READ_ONLY);
         nk = io.getDimension ("nk");
         kWeights.resize (nk);
         io.read ("kWeights", &kWeights, nk);
         fermi.read (io);
         io.close ();
      } catch (SxException e)  {
         e.print ();
         SX_EXIT;
      }
   }
   
   // --- read waves for partial DoS
   SxPW waves(wavesFile, SxPW::ReadOnDemand);
   SxCell cell;

   if (pdos || zdos)  {
      SxBinIO io (wavesFile, SxBinIO::BINARY_READ_ONLY);
      cell.read (io);
      waves.setGkBasisPtr (SxPtr<SxGkBasis>::create (io));
      waves.getGkBasisPtr ()->changeTau (str);
      io.close ();   
   }

   if (epsFile.getSize () > 0)  {
      int kpInFile, statesInFile;
      SxFermi::peekSpectrumFile (epsFile, &kpInFile, &statesInFile);
      if (kpInFile != kWeights.getSize ())  {
         cout << "'" << epsFile << "' has " << kpInFile;
         cout << " kpoints, but " << nk << " in " << wavesFile << endl;
         SX_QUIT;
      }
      fermi.resizeFermi (statesInFile);
      fermi.readSpectrum (epsFile);
   }

   // --- get minimum and maximum
   int nSpin = fermi.getNSpin ();
   double eMin = fermi.eps(0,0,0),
          eMax = fermi.eps(0,0,0);
   if (range.getSize () == 0)  {
      double eMinK,eMaxK;
      for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {
         for (int ik = 0; ik < nk; ++ik)  {
            eMinK = fermi.eps(iSpin,ik).minval ();
            if (eMinK < eMin) eMin = eMinK;
            eMaxK = fermi.eps(iSpin,ik).maxval ();
            if (eMaxK > eMax) eMax = eMaxK;
         }
      }
   } else {
      eMin = range(0) / HA2EV;
      eMax = range(1) / HA2EV;
      shift = 0.;
   }

   SxMeshR rho;
   SxRBasis R;
   SxVector<Double> sigma;   
   SxVector<Int> zOnly;
   int meshSize = -1, nPts = -1;
   bool selfCorr = false;

   SxVector<Double> zWeight;
   if (zdos) {
      
      if (nSpin == 1) kWeights *= 2.; // to be consistent with TDOS

      SxMesh3D mesh = (*waves.getGkBasisPtr ())(0).fft3d(0).mesh;
      R.set (mesh, cell);

      meshSize = mesh.product ();
      zOnly.resize (meshSize);
      SxVector<Int>::Iterator zOnlyIt = zOnly.begin ();
      for (int iR = 0; iR < meshSize; ++iR)
         *zOnlyIt++ = mesh.getMeshVec(iR, SxMesh3D::Positive)(dim);

      nPts = mesh(dim);

      selfCorr = (potFile.getSize () > 0);
      if (selfCorr)  {
         FILE *fp = fopen(potFile.ascii (), "r");
         if (!fp)  {
            cout << "Invalid sigma file " << potFile << "." << endl;
            SX_QUIT;
         }
         sigma.resize (nPts);
         for (int i = 0; i <  nPts; ++i)  {
            fscanf(fp,"%lf",&sigma(i));
            cout << i << " " << sigma(i) << endl; 
         }
      }
      zWeight.resize (nPts);
   }

   // --- calculate spectrum
   SxSpectrum dos;
   dos.nPerPeak = nPerPeak;
   double weight;
      
   // --- compute projections
   SxArray<SxDiracMat<Complex16> > projMatrix;
   if (pdos) projMatrix = computePsiG (&*table, waves.getGkBasisPtr(), rcut,
                                       nk, ia, is, l);
      
   for (int iSpin = 0; iSpin < nSpin; ++iSpin)  {

      // --- set up new spectrum
      if (zdos) dos.set (eMin - shift, eMax + shift, broadening, nPts);
      else dos.set (eMin - shift, eMax + shift, broadening);
      
      // --- add peaks
      for (int ik = 0; ik < nk; ++ik)  {
         
         for (int i = 0; i < fermi.getNStates(ik); ++i)  {
    
            if (zdos)   {

               if (fermi.eps(i,iSpin,ik) >= eMin && fermi.eps(i,iSpin,ik) <= eMax)
               {
                  double eps = fermi.eps(i,iSpin,ik);
                  rho = (R | waves(i,iSpin,ik)).absSqr ();
                  SxMeshR::Iterator rhoIt = rho.begin ();
                  SxVector<Int>::Iterator zOnlyIt = zOnly.begin ();
                  zWeight.set (0.);
                  for (int iR = 0; iR < meshSize; ++iR)
                     zWeight(*zOnlyIt++) += *rhoIt++;
                  if (selfCorr)  {
                     eps += (zWeight * sigma).sum () * R.dOmega;
                     cout << eps << endl;
                  }
                  zWeight *= kWeights(ik) * R.dOmega;
                  dos.addPeak (eps, zWeight);
               }

            } 
            else   { 

               double sum = 1.;
               if (pdos) {
                  sum = 0.;
                  for (int colIt = 0; colIt < projMatrix(0).nCols (); colIt++)
                     sum += 
                     dot (waves (i,iSpin,ik),projMatrix(ik).colRef(colIt)).absSqr ();
               }

               if (useOcc) weight = fermi.focc(i,iSpin,ik) * kWeights(ik) * sum;
               else weight = 2./double(nSpin) * kWeights(ik) * sum;

               dos.addPeak (fermi.eps(i,iSpin,ik), weight);

            }
         }

      }

      // compute spectrum
      dos.compute ();
      // change from Hartree to eV 
      dos.spectra  /= (iSpin == 0 ? 1. : -1.) * HA2EV;
      dos.energies *= HA2EV;
     
      // output
      try {
         SxString name = outFile;
         if (nSpin == 2)
            name += (SxList<SxString> () << ".up" << ".down")(iSpin);
         name += ".dat"; 
         SxBinIO io(name, SxBinIO::ASCII_WRITE_ONLY);
         dos.fprint (io.fp);
      } catch (SxException e)  {
         e.print ();
         SX_EXIT;
      }
   }

}
