// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxVector.h>
#include <SxMurn.h>
#include <SxMatrix.h>

SxVector<Double> polyFit (const SxVector<Double> &x, const SxVector<Double> &y,
                          double x0, int n, double *errorPtr = NULL)
{
   ssize_t N = x.getSize ();
   SxMatrix<Double> A(n + 1,N);
   SX_LOOP(i)  {
      double xx = x(i) - x0;
      double xn = 1.;
      SX_LOOP(j) { A(j,i) = xn; xn *= xx; }
   }
   SxVector<Double> res = (A ^ A.transpose ()).inverse () ^ (A ^ y);
   if (errorPtr)
      *errorPtr = sqrt( ((res ^ A)- y).normSqr () / double(N));
   return res;
}



void readMurnData (const SxString &inFile, SxVector<Double> *vols, SxVector<Double> *energies)
{
   SX_CHECK (vols);
   SX_CHECK (energies);
   SxStack<double> vList, eList;
   try {
      SxString data = SxString::read (inFile).stripComments();
      SxList<SxString> lines = data.tokenize('\n');
      SxList<SxString>::Iterator lineIt;
      for (lineIt = lines.begin(); lineIt != lines.end(); ++lineIt)  {
         double V, E;
         int nRead = sscanf((*lineIt).ascii (), "%lf%lf", &V, &E);
         if (nRead == 2)  {
            if (V <= 0.)  {
               cout << "Line '" << *lineIt 
                    << "': Volume must be larger than 0!" << endl;
               SX_QUIT;
            }
            vList << V;
            eList << E;
         } else {
            cout << "Failed to parse line '" << (*lineIt) << "'. Skipped.\n";
         }
      }
   } catch (SxException e) {
      e.print ();
      SX_QUIT;
   }
   *vols = SxVector<Double> (vList);
   *energies = SxVector<Double> (eList);
}

int main (int argc, char **argv)
{
   // --- command line parsing
   SxCLI cli(argc, argv);
   cli.preUsageMessage = 
      "This program calculates defect energies in strained material.\n"
      "Units:\n"
      "- volumes in cubic bohr\n"
      "- energies in Hartree\n"
      "- bulk modulus in GPa\n";
       
   SxString bulkFile = cli.option ("-b|--bulk", "file", 
                                   "bulk reference data: <volume> <energy>")
                       .toString ();
   SxString defFile  = cli.option ("-d|--defect", "file", 
                                   "defect supercell data: <volume> <energy>")
                       .toString ();
   int factor = cli.option ("-m|--multiplicity", "number", "supercell multiplicity (with respect to bulk reference)").toInt (1);

   int order = cli.option ("--order", "polynomial order", 
                           "order for polynomial deltaE(V) fit")
               .toInt (2);
   double V0 = cli.option ("-V|--V0", "volume", "bulk equilibrium volume")
               .toDouble ();
   const double au2GPa = 29421.01 /* Hartree/bohr^3 -> GPa*/;
   double B0 = cli.option ("-B0|--modulus", "bulk modulus", "bulk modulus (in GPa)")
               .toDouble () / au2GPa;
   double Bp = cli.option ("-Bp", "bulk modulus", "bulk modulus (in GPa)")
               .toDouble ();
   SxString outFile = cli.option ("-o|--out", "file", 
                                   "output file")
                       .toString ("defectStrain.dat");

   cli.finalize ();
   SxVector<Double> bulkV, bulkE, defV, defE;
   readMurnData (bulkFile, &bulkV, &bulkE);
   readMurnData (defFile, &defV, &defE);
   bulkV *= double(factor);
   bulkE *= double(factor);
   V0    *= double(factor);

   ssize_t nVol = defV.getSize ();

   // mapping
   SxVector<Double> refE(defE.getSize ());
   double deltaVMax = 1e-2 * (defV.maxval ()-defV.minval ())/double(nVol);
   bool ok = true;
   cout << deltaVMax << endl;
   SX_LOOP(iDef)  {
      for (int iBulk = 0; iBulk < bulkV.getSize (); ++iBulk)  {
         double deltaV = defV(iDef) - bulkV(iBulk);
         if (fabs(deltaV) < deltaVMax)  {
            refE(iDef) = bulkE(iBulk);
            if (fabs(deltaV) > 1e-3)  {
               cout << "Warning: mapping defect volume " << defV(iDef)
                    << " to slightly mismatched reference volume " 
                    << bulkV(iBulk) << "(delta = " << deltaV << ")" << endl;
            }
            break;
         } else if (iBulk == bulkV.getSize () - 1)  {
            cout << "Missing bulk reference for supercell V=" << defV(iDef)
                 << endl;
            ok = false;
         }
      }
   }
   if (!ok) { 
      cout << bulkV << endl;
      SX_QUIT; 
   }

   SxVector<Double> deltaE = defE - refE;

   SxVector<Double> fitParam = polyFit (defV, deltaE, V0, order);
   SxMurn bulkMurn;
   bulkMurn.minVolume = V0;
   bulkMurn.bulkModulus = B0;
   bulkMurn.bulkModulusDerivative = Bp;
   bulkMurn.minEnergy = 0.;

   SxVector<Double> pressure(nVol);
   SxVector<Double> bulkVol(nVol);
   SX_LOOP (iVol)  {
      // bulk pressure
      if (fabs(Bp) < 1e-5)  {
         pressure(iVol) = -B0 * log(defV(iVol) / V0);
      } else if (fabs(Bp-1.) < 1e-5)  {
         pressure(iVol) = B0 * (V0/defV(iVol) - 1.);
      } else {
         pressure(iVol) = B0/Bp * (exp(-Bp * log(defV(iVol) / V0)) - 1.);
      }
      // delta E pressure: 1st derivative of fit function
      double dp = 0, dV = defV(iVol) - V0;
      for (int n = order; n > 0; n--)  { // Horner scheme
         dp *= dV;
         dp += fitParam(n) * double(n);
      }
      pressure(iVol) -= dp;
      if (fabs(Bp) < 1e-5)  {
         bulkVol(iVol) = V0 * exp (-pressure(iVol) / B0);
      } else {
         bulkVol(iVol) = V0 * exp(-log(1. + Bp * pressure(iVol) / B0)/Bp);
      }
      cout << defV(iVol) << ' ' << pressure(iVol) << ' ' << bulkVol(iVol) << endl;
   }

   FILE *fp = fopen (outFile.ascii (), "w");
   if (!fp) {
      cout << "Cannot open " << outFile << " for writing " << endl;
      SX_QUIT;
   }
   fprintf (fp, "# <volume> <pressure> <delta E> <delta U> <delta H> <delta V> <bulk strain>\n");
   SX_LOOP(iVol)  {
      double relVol = defV(iVol) - bulkVol(iVol);
      fprintf (fp, "%12.4f\t%10.6f\t%14.8f\t%14.8f\t%14.8f\t%8.4f\t%8.4f\n", 
               defV(iVol), 
               pressure(iVol) * au2GPa, 
               deltaE(iVol), 
               deltaE(iVol) + bulkMurn.getEnergy(defV(iVol)) 
                            - bulkMurn.getEnergy(bulkVol(iVol)),
               deltaE(iVol) + bulkMurn.getEnergy(defV(iVol))
                            - bulkMurn.getEnergy(bulkVol(iVol))
                            + pressure(iVol) * relVol,
               relVol,
               exp(log(bulkVol(iVol)/V0)/3.)-1.);
   }
   fclose (fp);


   



   
}
