// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxFitFunc.h>
#include <SxMatrix3.h>
#include <SxMatrix.h>
#include <SxSimpleParser.h>

const double au2GPa = 29421.01 /* Hartree/bohr^3 -> GPa*/;

/** \brief Polynomial fit to energy-strain curve
  */
class SxStrainFit {
   public:
      /// fit polynomial in scaled, shifted dof space
      SxFitPolynomial P;
      /// offset in dof space
      SxVector<Double> dof0;
      /// scale in dof space
      SxVector<Double> scale;
      /// degree of freedom matrix
      SxVector<Double> dofTo6;
      /// Inverse of reference cell
      SxMatrix3<Double> cell0Inv;

      /// Set the reference cell
      inline void setRefCell (const SxMatrix3<Double> cell0)
      {
         SX_CHECK (fabs(cell0.determinant ()) > 1e-12);
         SX_CHECK (P.paramRef ().getSize () == 0); // must not be initialized
         cell0Inv = cell0.inverse ();
      }

      /** \brief Get stress tensor (as 3x3 matrix)
        @param strain  strain tensor (as 3x3 matrix)
        @param E       optional pointer, if set, energy is saved there
        */
      SxMatrix3<Double> getStress (const SxMatrix3<Double> &strain,
                                   double *E = NULL) const;

      /** \brief Get logarithmic strain tensor
          @param cell Input cell.
          The logarithmic strain tensor is defined as \f$\ln 1+\epsilon\f$.
          It is obtained by diagonalizing the strain matrix, applying
          the "ln 1 + ..." function to the eigenvalues and rotating
          back to the original orientation.
          The strain itself is obtained from the cell \f$A\f$ and the
          reference cell \f$A_0\f$ via
          \f[
          (1 + \epsilon)^2 = (A A_0^{-1})^T (A A_0^-1)
          \f]
          The transpose cancels any rotation between the strained and
          the reference cell.
        */
      SxMatrix3<Double> getLnEps (const SxMatrix3<Double> &cell) const;
      /** \brief Get strain tensor
          @param cell Input cell.
          The strain itself is obtained from the cell \f$A\f$ and the
          reference cell \f$A_0\f$ via
          \f[
          (1 + \epsilon)^2 = (A A_0^{-1})^T (A A_0^-1)
          \f]
          The transpose cancels any rotation between the strained and
          the reference cell.
          The square root is obtained by diagonalizing the (1 + eps)^2 -1 
          matrix, transforming the eigenvalues and rotating back to the
          original orientation.
        */
      SxMatrix3<Double> getStrain (const SxMatrix3<Double> &cell) const;
      /** \brief Remove strain components outside dof space
          @param strain
          This function projects the strain tensor onto the degree-of-freedom
          space for which this fit is defined. Strain components
          outside the degree-of-freedom space do not contribute to
          the energy and therefore could lead to spurious strains
          in optimizations.
        */
      SxMatrix3<Double> cleanStrain (const SxMatrix<Double> &strain) const;

      /** \brief Fit polynomial to set of energy-strain data
          @param lnEps list of logarithmic strain as obtain from getLnEps
          @param energy list of corresponding energies
          @param order polynomial order
        */
      void fit (const SxArray<SxMatrix3<Double> > lnEps,
                const SxArray<double> energy,
                int order);
      /// Constructor
      SxStrainFit ();

      /** \brief Printout in sx format
        */
      void fprint (FILE *fp) const;
      /** \brief Read from symbol table */
      void read (const SxSymbolTable *table);

      /**  \brief Find strain that produces a given stress
           @param stress0 the desired stress
        */
      SxMatrix3<Double> findStrain (const SxMatrix3<Double> &stress0,
                                    const SxMatrix3<Double> &strain0 
                                    = SxMatrix3<Double> (0.,0.,0.,
                                                        0.,0.,0.,
                                                        0.,0.,0.) ) const;
      /** \brief Get dof vector for a given strain */
      SxVector<Double> getDofFromStrain (const SxMatrix3<Double> &strain) const;

      /// Allow to chain contributions to the energy
      SxPtr<SxStrainFit> moreContrib;
};

SxStrainFit::SxStrainFit ()
   : P(0,0)
{
}

SxMatrix3<Double>
SxStrainFit::getStress (const SxMatrix3<Double> &strain, double *E) const
{
   // diagonalize strain matrix
   SxMatrix<Double>::Eigensystem eig
      = SxMatrix<Double>(strain).eigensystem ();
   SX_LOOP(i)  {
      if (fabs(eig.vals(i).re) > 1.) {
         cout << "Strain explosion: strain = " << strain << endl;
         SX_EXIT;
      }
   }
   const SxVector<Double> &U = eig.vecs;

   // --- lnEps = ln (1 + eps): get the 6 independent components
   SxVector<Double> diag = log(1. + eig.vals.real ());
   SxMatrix<Double> xx(3,3);
   xx.set (0.);
   SX_LOOP(i) xx(i,i) = diag(i);
   xx = U ^ xx ^ U.transpose ();
   SxVector<Double> lnEps6(6);
   lnEps6.set (0.);
   for (int id = 0; id < 3; ++id)  {
      lnEps6(0) += U(0, id) * diag(id) * U(0, id);
      lnEps6(1) += U(1, id) * diag(id) * U(1, id);
      lnEps6(2) += U(2, id) * diag(id) * U(2, id);
      lnEps6(3) += U(1, id) * diag(id) * U(2, id);
      lnEps6(4) += U(0, id) * diag(id) * U(2, id);
      lnEps6(5) += U(0, id) * diag(id) * U(1, id);
   }

   // reduce to degree of freedom space
   SxVector<Double> dof = dofTo6.transpose () ^ lnEps6;

   // evaluate
   SxVector<Double> dEdof = P.getDeriv (scale * (dof - dof0), E) * scale;
   // change back to 3x3 in ln (1 + eps)
   SxVector<Double> dElnEps6 = dofTo6 ^ dEdof;
   SxMatrix<Double> dElnEps3x3(3,3);
   dElnEps3x3(0,0) = dElnEps6(0);
   dElnEps3x3(1,1) = dElnEps6(1);
   dElnEps3x3(2,2) = dElnEps6(2);
   dElnEps3x3(2,1) = dElnEps3x3(1,2) = dElnEps6(3);
   dElnEps3x3(2,0) = dElnEps3x3(0,2) = dElnEps6(4);
   dElnEps3x3(1,0) = dElnEps3x3(0,1) = dElnEps6(5);

   /*
   // --- derivative with respect to reference strain
   // go to 3x3 strain eigenspace
   SxMatrix<Double> dElnEps3x3Diag = U.transpose () ^ dElnEps3x3 ^ U;
   // change to dE/d eps(diag)
   for (int i = 0; i < 3; ++i)
      for (int j = 0; j < 3; ++j)
         if (i == j || fabs(eig.vals(i).re - eig.vals(j).re) < 1e-6)
            dElnEps3x3Diag(i,j) /= 1. + eig.vals(i).re;
         else
            dElnEps3x3Diag(i,j) *= (lnEps6(i) - lnEps6(j))
                                 / (eig.vals(i).re - eig.vals(j).re);
   // go back to original coordinate system
   SxMatrix<Double> dEdEps = U ^ dElnEps3x3Diag ^ U.transpose ();
   // inverse volume of reference cell
   double volume0inv = fabs(cell0Inv.determinant ());
   SxMatrix3<Double> res =  dEdEps * volume0inv;
   */

   // --- derivative with infinitesimal strain on top of current strain state
   // inverse volume of reference cell
   double volume0inv = fabs(cell0Inv.determinant ());
   // relative change in inverse volume from hydrostatic strain
   double volEpsInv = exp(-(lnEps6(0) + lnEps6(1) + lnEps6(2)));
   SxMatrix3<Double> res =  dElnEps3x3 * volume0inv * volEpsInv;

   // --- add stress from moreContrib
   if (moreContrib)  {
      SX_CHECK_VAR ((cell0Inv - moreContrib->cell0Inv).absSqr ().sum () < 1e-12,
                    (cell0Inv - moreContrib->cell0Inv).absSqr ().sum ());
      double E2 = 0.;
      res += moreContrib->getStress (strain, E ? (&E2) : NULL);
      if (E) *E += E2;
   }
   return res;
}

SxMatrix3<Double> SxStrainFit::getLnEps (const SxMatrix3<Double> &cell) const
{
   // R ^ (1 + eps) (R is possible rotation);
   SxMatrix3<Double> oneEps = cell ^ cell0Inv;
   // (1+eps)^2
   SxMatrix3<Double> oneEps2 = oneEps.transpose () ^ oneEps;
   // remove 1 from diagonal to improve eigenvector accuracy
   SX_LOOP(i) oneEps2(i,i) -= 1.;
   SxMatrix<Double>::Eigensystem eig
      = SxMatrix<Double> (oneEps2).eigensystem ();
   SxMatrix3<Double> U(eig.vecs);
   SxMatrix3<Double> diag(0.5 * log(1. + eig.vals(0).re), 0, 0,
                          0, 0.5 * log(1. + eig.vals(1).re), 0,
                          0, 0, 0.5 * log(1. + eig.vals(2).re));
   return U ^ diag ^ U.transpose ();
}

SxMatrix3<Double> SxStrainFit::getStrain (const SxMatrix3<Double> &cell) const
{
   // R ^ (1 + eps) (R is possible rotation);
   SxMatrix3<Double> oneEps = cell ^ cell0Inv;
   // (1+eps)^2
   SxMatrix3<Double> oneEps2 = oneEps.transpose () ^ oneEps;
   // remove 1 from diagonal to improve eigenvector accuracy
   SX_LOOP(i) oneEps2(i,i) -= 1.;
   SxMatrix<Double>::Eigensystem eig
      = SxMatrix<Double> (oneEps2).eigensystem ();
   SxMatrix3<Double> U(eig.vecs);
   SxMatrix3<Double> diag (sqrt(eig.vals(0).re + 1.) - 1., 0., 0.,
                           0., sqrt(eig.vals(1).re + 1.) - 1., 0.,
                           0., 0., sqrt(eig.vals(2).re + 1.) - 1.);
   return U ^ diag ^ U.transpose ();
}

SxMatrix3<Double>
SxStrainFit::cleanStrain (const SxMatrix<Double> &cell) const
{
   SxVector<Double> dirty6(6);
   dirty6(0) = cell(0,0);
   dirty6(1) = cell(1,1);
   dirty6(2) = cell(2,2);
   dirty6(3) = 0.5 * (cell(2,1) + cell(1,2));
   dirty6(4) = 0.5 * (cell(2,0) + cell(0,2));
   dirty6(5) = 0.5 * (cell(0,1) + cell(1,0));
   SxVector<Double> clean6 = dofTo6 ^ (dofTo6.transpose () ^ dirty6);
   return SxMatrix3<Double> (clean6(0), clean6(5), clean6(4),
                             clean6(5), clean6(1), clean6(3),
                             clean6(4), clean6(3), clean6(1));
}

SxVector<Double>
SxStrainFit::getDofFromStrain (const SxMatrix3<Double> &strain) const
{
   SxMatrix<Double>::Eigensystem eig
      = SxMatrix<Double> (strain).eigensystem ();
   SxMatrix3<Double> U(eig.vecs);
   SxMatrix3<Double> diag(log(1. + eig.vals(0).re), 0, 0,
                          0, log(1. + eig.vals(1).re), 0,
                          0, 0, log(1. + eig.vals(2).re));
   SxMatrix3<Double> lnEps =  U ^ diag ^ U.transpose ();
   SxVector<Double> dof(6);
   dof(0) = lnEps(0,0);
   dof(1) = lnEps(1,1);
   dof(2) = lnEps(2,2);
   dof(3) = lnEps(1,2);
   dof(4) = lnEps(0,2);
   dof(5) = lnEps(0,1);
   return dofTo6.transpose () ^  dof;
}

void SxStrainFit::fit (const SxArray<SxMatrix3<Double> > lnEps,
                       const SxArray<double> energy,
                       int order)
{
   SX_CHECK_VARS (energy.getSize () == lnEps.getSize (),
                  energy.getSize (), lnEps.getSize ());
   int nConfig = (int)lnEps.getSize ();
   // determine degrees of freedom
   SxMatrix<Double> dofMat(nConfig, 6);
   SX_LOOP(iConf)  {
      dofMat((int)iConf, 0) = lnEps(iConf)(0,0);
      dofMat((int)iConf, 1) = lnEps(iConf)(1,1);
      dofMat((int)iConf, 2) = lnEps(iConf)(2,2);
      dofMat((int)iConf, 3) = lnEps(iConf)(1,2);
      dofMat((int)iConf, 4) = lnEps(iConf)(0,2);
      dofMat((int)iConf, 5) = lnEps(iConf)(0,1);
   }
   //cout << dofMat << endl;
   SxList<SxVector<Double> > dof;
   {
      SxMatrix<Double>::Eigensystem eig 
         = SxMatrix<Double>(dofMat.transpose () ^ dofMat).eigensystem ();
      SX_LOOP(iEig)
         if (eig.vals(iEig).re > 1e-6) dof << eig.vecs.colRef (iEig).getCopy ();
   }
   dofTo6.reformat (6, dof.getSize ());
   SX_LOOP(iEig) dofTo6.colRef (iEig) <<= dof(iEig);

   SxMatrix<Double> confInDof = (dofMat ^ dofTo6).transpose ();

   //cout << dof.getSize () << " degrees of freedom in strain" << endl;
   //cout << confInDof << endl;
   int nDof = (int)dof.getSize ();

   int iMin = 0;
   SX_LOOP(i) if (energy(i) < energy(iMin)) iMin = (int)i;
   
   dof0 = confInDof.colRef (iMin).getCopy ();
   //dof0.set (0.);
   scale.resize (nDof);
   scale.set (0.);
   SX_LOOP(iConf) scale += (confInDof.colRef (iConf)-dof0).normSqr ();
   scale /= double(nConfig);
   scale = 1./sqrt(scale);

   P = SxFitPolynomial (nDof, order);
   int nFitPar = (int)P.getNParam ();
   if (nFitPar > nConfig)  {
      cout << "Not enough data to fit all " << nFitPar 
           << "many free parameters" << endl;
      SX_QUIT;
   }

   SxMatrix<Double> A(nFitPar, nConfig);
   SxVector<Double> b(nFitPar);
   b.set (0.);
   P.paramRef ().set (0.);

   SX_LOOP(iConf)  {
      A.colRef (iConf) <<= P.getParamDeriv (scale * (confInDof.colRef (iConf) - dof0));
      //cout << A.colRef (iConf) << endl;
      b.plus_assign_ax (energy(iConf) - energy(iMin), A.colRef(iConf));
   }

   SxMatrix<Double> X = A ^ A.transpose ();
   cout << "# X_ev=" << X.eigenvalues () << endl;
   P.setParam (X.solve (b));
   P.paramRef()(0) += energy(iMin);
   cout << "#" << P.paramRef () << endl;
}

void SxStrainFit::fprint (FILE *fp) const
{
   SX_CHECK (fp);
   sxfprintf (fp, "strainFit {\n");
   // --- cell
   SxMatrix3<Double> cell0 = cell0Inv.inverse ();
   sxfprintf(fp, "   refCell = [[% 12.8f, % 12.8f, % 12.8f],\n"
                 "              [% 12.8f, % 12.8f, % 12.8f],\n"
                 "              [% 12.8f, % 12.8f, % 12.8f]];\n",
                   cell0(0,0), cell0(1,0), cell0(2,0),
                   cell0(0,1), cell0(1,1), cell0(2,1),
                   cell0(0,2), cell0(1,2), cell0(2,2));
   sxfprintf (fp, "   order = %d;\n", P.getOrder ());
   sxfprintf (fp, "   dof = [");
   for (int j = 0; j < dofTo6.nCols (); ++j)  {
      sxfprintf (fp, "[%+18.16f", dofTo6(0,j));
      for (int i = 1; i < 6; ++i)
         sxfprintf(fp, ",%s%+18.16f", (i % 3) ? "" : "\n           ",
                                      dofTo6(i,j));
      sxfprintf (fp, "]");
      if (j < dofTo6.nCols () - 1) sxfprintf(fp, ",\n /* %d: */ ", j + 2);
   }
   sxfprintf (fp, "];\n");
   int nDof = (int)dof0.getSize ();
   sxfprintf (fp, "   dof0   = [%+12.10f", dof0(0));
   for (int i = 1; i < nDof; ++i) sxfprintf (fp, ",%+12.10f", dof0(i));
   sxfprintf (fp, "];\n");
   sxfprintf (fp, "   scale  = [%+12.10f", scale(0));
   for (int i = 1; i < nDof; ++i) sxfprintf (fp, ",%+14.10f", scale(i));
   sxfprintf (fp, "];\n");

   SxVector<Double> &param = const_cast<SxFitPolynomial&>(P).paramRef ();
   sxfprintf (fp, "   param  = [%+12.10f", param(0));
   for (int i = 1; i < param.getSize (); ++i)
      sxfprintf (fp, ",\n            %+14.10f", param(i));
   sxfprintf (fp, "];\n");
   sxfprintf (fp, "}\n"); 
}

void SxStrainFit::read (const SxSymbolTable *table)
{
   SxMatrix3<Double> cell0;
   SYMBOLPARSE(table)  {
      SYMBOLGROUP("strainFit")  {
         dofTo6 = SYMBOLGET("dof")->toList ();
         scale  = SYMBOLGET("scale")->toList ();
         dof0   = SYMBOLGET("dof0")->toList ();
         cell0 = SxMatrix3<Double> (SYMBOLGET("refCell")->toList ())
                 .transpose ();

         int order = SYMBOLGET("order");
         int nDof = (int)dof0.getSize ();
         dofTo6.reshape (6, nDof);
         P = SxFitPolynomial (nDof, order);
         P.paramRef () = SYMBOLGET("param")->toList ();
      }
   }
   cell0Inv = cell0.inverse ();
}

SxMatrix3<Double> SxStrainFit::findStrain (const SxMatrix3<Double> &stress0,
                                           const SxMatrix3<Double> &strain0) const
{
   SxMatrix<Double> s0 (stress0);
   SxMatrix3<Double> optStrain = strain0;
   SxMatrix<Double> stress, stressTrial, stressOld, search, searchOld;
   double xTrial = 10. , gamma = -1.;
   int nx = 0;
   for (int it = 0; it < 1000; ++it)  {
      stress = getStress (optStrain);
      if ((stress-s0).normSqr () < 1e-18) break;
      if (it == 0)  {
         search = s0-stress;
      } else {
         gamma = (stress-s0).normSqr () / (stressOld-s0).normSqr ();
         if (gamma > 0.9) nx++;
         if (nx > 10) { gamma=0.; nx=0;}
         search = s0-stress + gamma * searchOld;
      }
      SX_LOOP2(i,j) optStrain(i,j) += xTrial * search(i,j);
      stressTrial = getStress (optStrain);
      SxVector<Double> dStress = (stressTrial - stress) / xTrial;
      //cout << "dStress=" << dStress << endl;
      double xOpt = -dot(dStress, stress-s0)/dStress.normSqr ();
      //cout << it << ' ' << xOpt << ' ' << (stress - s0).normSqr ()
      //     << ' ' << gamma << endl;
      //cout << "exp. stress: " << (stress + xOpt * dStress) << endl;
      SX_LOOP2(i,j) optStrain(i,j) += (xOpt - xTrial) * search(i,j);
      searchOld = search;
      stressOld = stress;
   }
   return optStrain;
}

SxMatrix3<Double> getHexCell (double volume, double cOverA)
{
   double a = exp(log(volume / cOverA / sqrt(0.75))/3);
   double c = cOverA * a;
   SxMatrix3<Double> res (      0.5 * a,        0.5 * a, 0,
                        -sqrt(0.75) * a, sqrt(0.75) * a, 0,
                                      0,              0, c);
   SX_CHECK_VARS (fabs(res.determinant () - volume) < 1e-7 * volume,
                  res.determinant (), volume);
   return res;
}

SxMatrix3<Double> getCell (const SxVector<Double> &v, int cellType)
{
   switch (cellType) {
      case 0: // simple cubic
         return SxMatrix3<Double> (v(0), 0., 0.,
                                   0., v(0), 0.,
                                   0, 0., v(0));
      case 1: // fcc
         return 0.5 * v(0)
                * SxMatrix3<Double> (0, 1, 1,
                                     1, 0, 1,
                                     1, 1, 0);
      case 2: // bcc
         return 0.5 * v(0)
                * SxMatrix3<Double> (-1, 1, 1,
                                     1, -1, 1,
                                     1, 1, -1);
      case 3: // hexagonal
         return SxMatrix3<Double> (0.5*v(0), 0.5*v(0), 0.,
                                   -sqrt(0.75) * v(0), sqrt(0.75) * v(0), 0.,
                                   0, 0., v(1));
      case 4: // hexagonal V, c/a
         return getHexCell (v(0), v(1));
      case 5: // tetragonal
         return SxMatrix3<Double> (v(0), 0., 0.,
                                   0., v(0), 0.,
                                   0, 0., v(1));
      case 6: // orthorhombic
         return SxMatrix3<Double> (v(0), 0., 0.,
                                   0., v(1), 0.,
                                   0, 0., v(2));
      default: // generic
         return SxMatrix3<Double> (v(0), v(3), v(6),
                                   v(1), v(4), v(7),
                                   v(2), v(5), v(8));
   }
} 

SxVector<Double> readStrainEnergy(FILE *in, int nr)
{
   SX_CHECK(in);
   SxVector<Double> dataVec;
   fscanf (in, " ");
   int nrNow = 0, line = 0;
   SxList<double> data;
   while (!feof(in)) {
      char nl = 0;
      double d;
      if (fscanf (in, "%lf%*[ \t]", &d) > 0) {
         data << d;
         nrNow++;
      } else {
         break;
      }
      nl = (char)fgetc (in);
      if (nl == '\n')  {
         if (nr == -1) nr = nrNow;
         fscanf(in, "%*[ \t]");
         // --- stop if empty line
//            nl = fgetc (in);
//            if (nl == '\n') break;
//            ungetc (nl, in);
         if (nr != nrNow)  {
            cout << "Wrong column count" << endl;
            cout << "Should be " << nr << endl;
            cout << "Is " << nrNow << endl;
            SX_QUIT;
         }
         nrNow = 0;
         line++;
      } else {
         ungetc (nl, in);
      }
   }
   if (nr == nrNow) line++;
   while (data.getSize () % nr) data.removeLast ();
   dataVec = SxVector<Double> (data);
   dataVec.reshape (nr, data.getSize () / nr);
   return dataVec;
}

int main (int argc, char **argv)
{
   // --- command line parsing
   SxCLI cli(argc, argv);

   int cliRead = cli.newGroup ("Read from sx file");
   SxString inFile = cli.option ("-f|--fitfile", "file", "sx style fit file")
                     .toString ();

   int cliDefect = cli.newGroup ("Defects");
   cli.requireGroup (cliRead);
   SxString bulkFile = cli.option ("-b|--bulk", "file", "bulk reference data")
                       .toString ();
   SxString defFile  = cli.option ("-d|--defect", "file", 
                                   "defect supercell data")
                       .toString ();
   cli.option ("-m|--multiplicity", "supercell multiplicity", "supercell multiplicity (with respect to bulk reference)");
   SxVector3<Int> factor = cli.last ().exists () 
                         ? SxVector3<Int> (cli.last ().toIntList3 ())
                         : SxVector3<Int> (1,1,1);

   SxString sampleFile = cli.option ("-s", "file", "sample defect data for stresses given via bulk cell parameters")
                       .toString ("");

   cli.newGroup ("Fitting");
   cli.excludeGroup (cliRead);
   bool fitting = cli.option ("--fit", "request fitting mode").toBool ();
   SxString input = cli.option ("-i|--input", "file", 
      "input data file. If not given, stdin is used.\n"
      "Each line of the input must give the lattice parameters followed by the "
      "energy. The number of lattice parameters depends on the cell type (see "
      "below). Lattice parameters are given in bohr; energies in Hartree"
          )
                    .toString ("");
   SxString outFile = cli.option ("-o|--output", "file",
                                  "strain fit output file")
                      .toString ("strainFit.sx");

   cli.setGroup (cli.generalGroup);
   int order = cli.option ("-n","int","degree of the fit polynomial")
               .toInt (2,1);
   bool optCell = cli.option ("-O", "optimize cell").toBool ();
   int cellType = cli.option ("--sc|--fcc|--bcc||--hexagonal|--hex-V-c/a|--tetragonal|--orthorhombic",
         "type of cell:\n"
         "sc .......... (a)     simple cubic\n"
         "fcc ......... (a)     face-centered cubic\n"
         "bcc ......... (a)     body-centered cubic\n"
         "hexagonal ... (a,c)   hexagonal\n"
         "hex-V-c/a ... (V,c/a) hexagonal with volume, c/a ratio\n"
         "tetragonal .. (a,c)   tetragonal\n"
         "orthorhombic  (a,b,c) orthorhombic\n"
         "default       (a1x,a1y,a1z,a2x,a2y,a2z,a3x,a3y,a3z)"
         ).toChoice ();
   SxList<int> nCellPar;
   nCellPar << 1 // sc 
            << 1 // fcc 
            << 1 // bcc
            << 2 // hexagonal a, c
            << 2 // hexagonal V, c/a
            << 2 // tetragonal
            << 3 // orthorhombic
            ;

   bool elasticConstants = cli.option ("--elastic", "compute elastic constants").toBool ();
   cli.finalize ();


   // --- 
   int nr = (cellType < nCellPar.getSize () && cellType >=0)
          ? (nCellPar(cellType) + 1) : 10;

   if (!fitting && inFile.getSize () == 0) {
      cout << "Nothing to do." << endl;
      return 0;
   }

   SxStrainFit strainFit;
   SxMatrix3<Double> cell0;
   if (fitting)  {
      FILE *in;
      if (input.getSize () > 0)  {
         in = sxfopen (input.ascii (), "r");
      } else {
         in = stdin;
      }
      SxVector<Double> dataVec = readStrainEnergy(in, nr);
      if (in != stdin) fclose (in);

      int nConfig = (int)dataVec.nCols ();
      SxArray<SxMatrix3<Double> > lnEps(nConfig);
      SxArray<SxMatrix3<Double> > strain(nConfig);
      SxArray<double> energy(nConfig);

      cell0 = getCell (dataVec.colRef (0), cellType);
      strainFit.setRefCell (cell0);
      SX_LOOP(iConf)  {
         SxMatrix3<Double> cell = getCell (dataVec.colRef (iConf), cellType);
         energy(iConf) = dataVec(dataVec.nRows ()-1, (int)iConf);
         lnEps(iConf) = strainFit.getLnEps (cell);
         strain(iConf) = strainFit.getStrain (cell);
      }
      strainFit.fit (lnEps, energy, order);
      if (optCell)  {
         SxMatrix3<Double> optStrain 
            = strainFit.findStrain (SxMatrix3<Double>(0,0,0,0,0,0,0,0,0));
         SxVector<Double> dof = strainFit.getDofFromStrain (optStrain);
         strainFit.P.paramRef () = strainFit.P.evaluate3 (strainFit.scale * (dof - strainFit.dof0), order);
         strainFit.dof0 = dof;
      }
      double err2 = 0.;
      SX_LOOP(iConf)  {
         double E;
         strainFit.getStress (strain(iConf), &E);
         cout << iConf << " fit error=" << (E-energy(iConf)) << endl;
         err2 += sqr(E-energy(iConf));
      }
      cout << "fit rms error = " << sqrt(err2/nConfig) << endl;
      if (nConfig > strainFit.P.paramRef ().getSize ()) {
         int nParam = (int)strainFit.P.paramRef ().getSize ();
         cout << "fit rms conf = " << sqrt(err2/(nConfig - nParam) ) << endl;
      }
   } else if (inFile.getSize () > 0)  {
      SxParser parser;
      SxParser::Table table = parser.read (inFile, "std/strainFit.std");
      strainFit.read (&*table);
      cell0 = strainFit.cell0Inv.inverse ();
   }

   if (optCell)  {
      SxMatrix3<Double> optStrain = strainFit.findStrain (SxMatrix3<Double>(0,0,0,0,0,0,0,0,0));
      double E;
      SxMatrix<Double> stress = strainFit.getStress (optStrain, &E);
      cout << "--- optimization done" << endl;
      cout << "stress=" << stress * au2GPa << endl;
      cout << "strain=" << optStrain << endl;
      sxprintf("energy=%.12f\n", E);

      SxMatrix3<Double> cellOpt = cell0 + (optStrain ^ cell0);
      cout << "cell = " << cellOpt << endl;
      switch (cellType) {
         case 0: // simple cubic
            cout << "a=" << cellOpt(0,0) << endl;
            break;
         case 1: // fcc
         case 2: // bcc
            cout << "a=" << 2.*cellOpt(0,1) << endl;
            break;
         case 3: // hexagonal
         case 4: // hexagonal V, c/a
         case 5: // tetragonal
            cout << "a=" << cellOpt.col(0).norm () << endl;
            cout << "c=" << cellOpt(2,2) << endl;
            break;
         case 6: // orthorhombic
            cout << "a=" << cellOpt(0,0) << endl;
            cout << "b=" << cellOpt(1,1) << endl;
            cout << "c=" << cellOpt(2,2) << endl;
            break;
         default: ;// generic
      }

      if (elasticConstants)  {
         //optStrain.set (0.);
         SxMatrix3<Double> one (1,0,0, 0,1,0, 0,0,1);
         //optStrain = 0.00 * one;
         double dStrain = 1e-6;

         for (int ij = 0; ij < 6; ij++)  {
            SxMatrix3<Double> dEps;
            dEps.set (0.);
            if (ij < 3) dEps(ij,ij) = dStrain;
            else {
               int i = (ij + 1) % 3;
               int j = (ij + 2) % 3;
               dEps(i,j) = dStrain;
            }
            SxMatrix3<Double> dStress = strainFit.getStress (((one + dEps) ^ (one + optStrain)) - one)
                                      - strainFit.getStress (((one - dEps) ^ (one + optStrain)) - one);
            dStress /= 2. * dStrain;
            
            for (int kl = 0; kl < 6; kl++)  {
               int k,l;
               if (kl < 3) k = l = kl;
               else {
                  k = (kl + 1) % 3;
                  l = (kl + 2) % 3;
               }
               cout << "C" << (ij+1) << (kl+1) << "=" << dStress(k,l) * au2GPa << endl;
               //cout << "C" << (ij+1) << (kl+1) << "=" << dStress(l,k) * au2GPa << endl;
            }
         }
      }
   }
   
   if (outFile.getSize () != 0 && fitting)  {
      FILE *fp = sxfopen(outFile.ascii (), "w");
      cout << "Writing fit to '" << outFile << "'..." << endl;
      strainFit.fprint (fp);
      fclose (fp);
   }
   
   if (cli.groupAvailable(cliDefect))  {
      // read bulk data
      FILE *fp = sxfopen (bulkFile, "r");
      SxVector<Double> data = readStrainEnergy (fp, nr);
      fclose (fp);

      // transform into supercells and energy
      SxArray<SxMatrix3<Double> > bulkCells(data.nCols ());
      SxArray<double> bulkEnergy(data.nCols ());
      SX_LOOP(iConf) {
         bulkCells(iConf) = getCell (data.colRef (iConf), cellType);
         bulkEnergy(iConf) = data(nr-1, (int)iConf);
      }

      // --- read defect data
      fp = sxfopen (defFile, "r");
      data = readStrainEnergy (fp, nr);
      fclose (fp);

      // transform into supercells and energy
      SxArray<SxMatrix3<Double> > defCells(data.nCols ());
      SxArray<double> defEnergy(data.nCols ());
      SX_LOOP(iConf) {
         defCells(iConf) = getCell (data.colRef (iConf), cellType);
         defEnergy(iConf) = data(nr-1, (int)iConf);
      }

      // match
      SxArray<double> dE(defCells.getSize ());
      SxStrainFit defFit;
      SX_LOOP2(i,j) defFit.cell0Inv(i,j) 
         = strainFit.cell0Inv(i, j) / double(factor(i));
      SxArray<SxMatrix3<Double> > lnEps(defCells.getSize ());
      SxArray<SxMatrix3<Double> > strain(defCells.getSize ());
      SX_LOOP(iDef) {
         // --- find matching bulk cell
         SxMatrix3<Double> redDefCell = defCells(iDef);
         SX_LOOP2(i,j) redDefCell(i,j) /= factor(j);
         int iBulk;
         for (iBulk = 0; iBulk < bulkCells.getSize (); ++iBulk)
            if ((bulkCells(iBulk) - redDefCell).absSqr ().sum () < 1e-5)
               break;
         if (iBulk == bulkCells.getSize ())  {
            cout << "Missing bulk reference for defect supercell " << (iDef+1)
                 << endl;
            cout << "data = " << data.colRef (iDef) << endl;
            SX_QUIT;
         }
         dE(iDef) = defEnergy(iDef) - bulkEnergy(iBulk) * factor.product ();
         lnEps(iDef) = defFit.getLnEps (defCells(iDef));
         strain(iDef) = defFit.getStrain (defCells(iDef));
      }
      
      // fit dE
      defFit.fit (lnEps, dE, order);

      SxMatrix3<Double> defCell0 = defFit.cell0Inv.inverse ();
      SX_LOOP(iDef) {
         double bulkE, strainedBulkE, fittedDE;
         SxMatrix3<Double> bulkStress, totalStress;
         bulkStress  = strainFit.getStress (strain(iDef), &bulkE);
         //cout << "bulk stress = " << bulkStress << endl;
         totalStress = bulkStress + defFit.getStress (strain(iDef), &fittedDE);
         //cout << "defect stress = " << defFit.getStress (strain(iDef)) << endl;
         SxMatrix3<Double> bulkStrain = strainFit.findStrain (totalStress);
         strainFit.getStress (bulkStrain, &strainedBulkE);
         double dEcorr = dE(iDef) + (bulkE - strainedBulkE)*factor.product ();
         //SxMatrix3<Double> relaxStrain = (strain(iDef) - bulkStrain);
         SxMatrix3<Double> relaxVol = (strain(iDef) - bulkStrain) * defCell0.determinant ();

         SX_LOOP2(i,j) sxfprintf(stdout, "%8.4f\t", strain(iDef)(i,j));
         SX_LOOP2(i,j) sxfprintf(stdout, "%8.4f\t",
                               totalStress(i,j) * au2GPa);
         SX_LOOP2(i,j) sxfprintf(stdout, "%8.4f\t", relaxVol(i,j));

         sxfprintf(stdout, "%14.8f\t%14.8f", dE(iDef), dEcorr);

         sxfprintf(stdout, "\t%14.8f", dEcorr - (relaxVol * totalStress).sum ());
         sxfprintf(stdout, "\t%14.8f", dEcorr - 0.5 * (relaxVol * (totalStress + bulkStress)).sum ());
         sxfprintf(stdout, "\n");
      }

      if (sampleFile.getSize ())  {
         fp = sxfopen (sampleFile, "r");
         data = readStrainEnergy (fp, nr - 1);
         fclose (fp);
         fp = sxfopen (sampleFile + ".out", "w");
         strainFit.P.rescale (strainFit.scale);
         strainFit.scale.set (1.);
         defFit.P.rescale (defFit.scale);
         defFit.scale.set (1.);

         SxStrainFit scFit;
         scFit.cell0Inv = defFit.cell0Inv;
         scFit.dofTo6   = strainFit.dofTo6;
         scFit.dof0     = strainFit.dof0;
         scFit.scale    = strainFit.scale;
         scFit.P = SxFitPolynomial ((int)strainFit.dofTo6.nCols (), order);
         scFit.P.paramRef () = strainFit.P.paramRef () * double(factor.product ());
         scFit.P.paramRef ()(0) = 0.;
         scFit.moreContrib = SxPtr<SxStrainFit>::create (defFit);
         SX_LOOP(iPt)  {
            SxMatrix3<Double> cell = getCell (data.colRef (iPt), cellType);
            SxMatrix3<Double> bulkStrain = strainFit.getStrain (cell);
            double bulkE, deltaE, refE;
            SxMatrix3<Double> stress = strainFit.getStress (bulkStrain, &bulkE);
            SxMatrix3<Double> scStrain = scFit.findStrain (stress);
            
            double E2;
            scFit.getStress (scStrain, &E2);

            defFit.getStress (scStrain, &deltaE);
            strainFit.getStress (scStrain, &refE);
            double deltaU = deltaE + (refE - bulkE)*factor.product ();
            SxMatrix3<Double> relaxVol = (scStrain - bulkStrain) * defCell0.determinant ();
            SxMatrix3<Double> one(1,0,0,0,1,0,0,0,1);
            double relaxVol2 = ((one + scStrain).determinant ()
                              - (one + bulkStrain).determinant ())
                             * defCell0.determinant ();
            double deltaH = deltaU - (relaxVol * stress).sum ();
            SX_LOOP(iPar) fprintf(fp, "%.10g\t", data(iPar, iPt));
            fprintf(fp, "%14.8f\t%14.8f", deltaU, deltaH);
            SX_LOOP2(i,j) fprintf(fp, "\t%8.4f", stress(i,j) * au2GPa);
            SX_LOOP2(i,j) fprintf(fp, "\t%8.4f", relaxVol(i,j));
            fprintf(fp, "\t%14.8f", relaxVol2);
            fprintf(fp, "\n");
         }
         fclose (fp);
      }
   }
}
