// ---------------------------------------------------------------------------
//                                          
//      The ab-initio based multiscale library 
//     
//                  S / P H I / n X
//                  http://www.sphinxlib.de
//     
//      Contact:    Sixten Boeck, boeck@mpie.de
//                  Algorithm Design and Modeling Group
//                  Computational Materials Design
//                  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//     
//      Authors:    see src/AUTHORS
//     
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxAtomicStructure.h>

double angle (SxVector3<Double> v1, SxVector3<Double> v2, SxVector3<Double> v3)
{
   double theta = 0.;
   if (v3.norm () < 1e-5) cout << "|v3| = " << v3.norm () << endl;
   else if (v2.norm () < 1e-5) cout << "|v2| = " << v2.norm () << endl;
   else if (v1.norm () < 1e-5) cout << "|v1| = " << v1.norm () << endl;
   else  {
      double cosTheta = v1 ^ v2 / v1.norm () / v2.norm ();
      if (cosTheta >= 1) theta = 0.;
      else if (cosTheta <= -1) theta = PI;
      else theta = acos(cosTheta);
      if ((theta > 0) && ((cross(v1, v2) ^ v3) < 0)) 
      theta = TWO_PI - theta;
   }
   return (theta);
}

int main (int argc, char **argv)
{
   initSPHInXMath ();
   
   SxCLI cli(argc, argv);

   cli.preUsageMessage 
      = SxString(" Add-on introduces a screw-dislocation dipole to a cell\n"
                 "The input structure should have the third basis vector "
                 "orthogonal to the other two. The dislocation Burger's vector"
                 " is parallel to the third basis; The length is the third "
                 "basis vector times b specified by the option '-b'.").wrap ();
   cli.authors = "Siyuan Zhang";
   
   SxVector3<Double> b(cli.option ("-b", "Burgers vector").toList3 ());

   double nu = cli.option ("--nu", "Poisson's ratio")
              .toDouble (1./3);

   int rep = cli.option ("--rep", "repitition of image dislocations")
            .toInt (10);

   double x0 = cli.option ("--x0", "", "place the first dislocation"
                           " with respect to the structure centre")
              .toDouble (0.);

   double y0 = cli.option ("--y0", "")
              .toDouble (0.);

   double x = cli.option ("-x", "place the second dislocation with respect"
                          " to the first dislocation")
             .toDouble (0.5);

   double y = cli.option ("-y", "")
             .toDouble (0.5);

   bool cut = cli.option ("--cut", "if set value other than 0, cut")
             .toInt (0);

   bool orth = cli.option ("--orthogonal", "orthogonalise the cell")
              .toBool ();

   int disp = cli.option ("--disp", "sampling for displacement field check")
             .toInt (0);

   bool yuja = cli.option ("--yuja", "special testing mode")
              .toBool ();

   bool sxbFile = cli.option ("--sxb", "input file is binary waves file "
                              "rather than S/PHI/nX input file")
                  .toBool ();

   SxString inFile 
      = cli.option ("-i|--input", "input file", 
                    "take original input file")
        .toString (sxbFile ? "waves.sxb" : "input.sx");
   cli.last ().defaultValue = "default: input.sx, or waves.sxb for --sxb flag";
   
   SxString outFile
      = cli.option ("-o","filename", "output file name (screen otherwise)")
        .toString ("dislstructure");

   bool outSxb = cli.option ("--outsxb", "output file is binary "
                              "rather than S/PHI/nX input file").toBool ();
   cli.last ().defaultValue = "default: same format as input file";
   if (sxbFile) outSxb = true;
   if (outSxb && outFile.getSize () == 0)
      outFile = "dislstructure.sxb";
   
   cli.finalize ();

   // --- yuja
   if (yuja) cout << 
   "',':;::,,:..,::'':;:,:::;::,.:;;::::;+:,,.,,,,,,:`..,;,,`.:``,..,..,,,."
   << endl << 
   ":';,;:;;,::,;;:;::;,::;;:,:.:':;,:';;::,,:,.:.,,.:,.;:,,.:,..`,.,...:.."
   << endl << 
   ":;::;':;:,:;;;::;;::;:;::;;;,;:;:::;::;,,:,,;..,,,,;,,..,,..:.:,.`.`:.."
   << endl << 
   "''::;';:',;,:,,,,'::::;;:;;;;;::;,,:;:::,,::,:,,.,,::....,.....,..`.:.."
   << endl << 
   ";',''''#;:.,:;:,:;:'+';;;;':;;,::,::,,;;,,:,,;.,,,::,,.....,`.....,;..."
   << endl << 
   ":;;';;'';,:'':;;;;''';';+;:;;;:,:::'::;;:,:,::,::,:,,,..,,.,,`.....:..."
   << endl << 
   ":;:'+;;:;;''';'::;+;:''+';:;;';::;';:::;:::,::,,,,,:,,,`.,,,.....`.,,;,"
   << endl << 
   "'+++';'';;';;'';':;:;;''''+';;;;;';;:;:;;;,,:,:::;:,:,:,,..,,`.``...;:,"
   << endl << 
   "';+';+++''#'#'+#'+##++++###+++'+''';;;;''';,;;::;';;;;:;,,,:```.....::,"
   << endl << 
   "'''++++++#+##'##+##+#++#####+''+'+';'+++++':;;;''''+++;;';;,.```:,,,,,,"
   << endl << 
   "::;'+'+++#+##+####++++'##++#++''''''''+++'++;+''+++#+#++#';,`,,.:.::..,"
   << endl << 
   ",,:;'''''+#######++'+''+++++'''';';;;''+'++++++++'++#+#+##';;;;:':;;,.,"
   << endl << 
   ".,::'''+'+#######''''++###+'''';;;:;;''''+#######++'+##+###++++'+;:::,,"
   << endl << 
   ".,:;;+++''+#####+;;''##+####'';;;;::;;';'#;+++#+##+''++++########+;;;,,"
   << endl << 
   ".,:,;;+++'+#####+;;'++':###'+;;;;::::;;;##+'#+++'''''''+++########+';;:"
   << endl << 
   "`..:;;'++'+#####':::;;'++++''+;;;::,::;;+''''';;;;;;;;'+''########++':'"
   << endl << 
   ".`.,:;;+'''#####;,.,,::::::::;:::,,,,:::';;;::::::::;:;';;########++';;"
   << endl << 
   "`..::;'''+######:....,,,::::::::,,,,,,:,::::::,,,,,,,:::,;+########++;'"
   << endl << 
   "``.,:;''++#####+,```...,,,,,,,,,,,..,,,,,,,,,,,,,,.,.,.,,:+##########++"
   << endl << 
   "``.,,;''+++####'.````......,.........,.,.................,'##########++"
   << endl << 
   "``..:;;''++##+#'.``````.............................```..,'############"
   << endl << 
   "`..,,:;'#+#####+,``````.`````````..`.......`....```````..,'############"
   << endl << 
   "``...:;+####+###,```````````````````......``````````````.,'############"
   << endl << 
   "``...,:++###+###,````````````````````....```````````````.,'+###########"
   << endl << 
   "`...,::'+###+###.``````````````````..`....``````````````.,;+#####@#####"
   << endl << 
   "`..,,:;+###+++##:` `````````.`````````...,.`````````````.,;############"
   << endl << 
   "``.:,:;++##+#####.`````````..`` ````````....````````````.,;###########+"
   << endl << 
   ".,,,:::'###+#####`````````... ` ``````  `...````````````.,;##########+`"
   << endl << 
   ".,::,::'##+#+####,````````...`````.`.````,...```````````.,'######+###,:"
   << endl << 
   ".,,::,:+####+####+.``````.....,',,,,.'',,....```````````.,'######+##+ '"
   << endl << 
   ",,.,:;'####+######```````...``..,::::,...```````````````.,+#####+#+#' +"
   << endl << 
   ";:,,,'+###+#######.``````..``````.....``````````````````.:######++#+'``"
   << endl << 
   "';:,,:####+#######+`.````````` ````.```````````````````..;######+++':` "
   << endl << 
   "::`.,;############@.`````````` ```..```````````````````.,'######+++'::+"
   << endl << 
   "    ,####+#########;`````````   ```...```````..```````.,:+#####++++': ;"
   << endl << 
   "    '###++##########````````` `.,::::::,.````...``.``.,:'#####+++'+#:,`"
   << endl << 
   "   .+##'++#########@'`````....;''';;;::''';;';,.`.,:::;++#####+++'+#`  "
   << endl << 
   "   ,+#+,;;###+#####@@,```..:+#+#+;..:.,'++'';:,,....,';+#####+++';++: ."
   << endl << 
   "  `+#+,',;############```..,.,,'+;;:':''':;:..,....;,:+#####+#++';+':  "
   << endl << 
   "  '###`.`'##'+########+``.,...`.:;:;:;:,:,.....,;::;'+#####++++''+++`  "
   << endl << 
   " `###;+. ##+;+#########'.......``....,:::,,.`;',;:;'+########++++:++   "
   << endl << 
   " #+.#`` `;+#+#########@#:.,.`..,,,,::;:::,,:.,,,:;++######+##++++##+   "
   << endl << 
   ",++:#; ` '###+#########@#...```.:;';';;;:'.,,::;'+###########+'##;'+   "
   << endl << 
   "'#.#''   '##:'############..````.,:::;;',,,:;;;++########+####+#;;':   "
   << endl << 
   "+# #;'   +##`###########@#+.````....;+,,,',:;''##############+''; ',   "
   << endl << 
   "':`'+   `##''############@#'````...;,.,.:,;;;'+#############+;++. ::   "
   << endl << 
   "+ :'+    ##+###############';.```.;.,,::;:;;+#######+#######++;+  .`   "
   << endl << 
   "# +`     #################+';;.`.;,,,,;;:;:+###############++''+ ,     "
   << endl << 
   "+ +.    `+################+';;;,,,:;;:::''########+########++++' ..    "
   << endl << 
   "; +`    ,#################+'';;;,:;::;;+############+#####++'++',.`    "
   << endl << 
   ".`. `   ##################+';:;;'';''++############+######+++++'''     "
   << endl;

   // --- read input
   SxAtomicStructure structure;
   if (sxbFile)  {
      try {
         SxBinIO io (inFile, SxBinIO::BINARY_READ_ONLY);
         structure.read (io);
         SxString chemNameList;
         io.close ();
      } catch (SxException e)  {
         e.print ();
         SX_EXIT;
      }
   } else  {
      SxParser parser;
      SxConstPtr<SxSymbolTable> tree;
      tree = parser.read (inFile, "std/structure.std");
      structure = SxAtomicStructure (&*tree);
      structure.atomInfo->meta.update (SxAtomicStructure::Elements,
                                       SxSpeciesData::getElements (&*tree));
   }

   structure %= structure.cell;

   SxVector3<Double> e1 = structure.cell (0),
                     e2 = structure.cell (1),
                     e3 = structure.cell (2);
   cout << "Input Cell: " << structure.cell << endl;
   cout << "Inverse Cell: " << structure.cell.inverse () << endl;
   SxVector3<Double> ex = e1 - (e1 ^ e3) * e3 / e3.normSqr (),
                     ey = e2 - (e2 ^ e3) * e3 / e3.normSqr ();
   cout << "ex = " << ex << endl << "ey = " << ey << endl;
   cout << "Input b: " << b << endl;
   b = b(0) * e1 / e1.norm () + b(1) * e2 / e2.norm () + b(2) * e3;
   cout << "b = " << b << endl;
   SxVector3<Double> bs = (b ^ e3) * e3 / e3.normSqr (),
                     be = b - bs, bt;
   if (be.norm () < 1e-5) bt.set (0.);
   else bt = cross (e3, be) / e3.norm ();
   cout << "edge b = " << be 
        << "; screw b = " << bs 
        << "; transverse b = " << bt << endl;
   SxVector3<Double> v = be;
   if (be.norm () < 1e-5) v = x * ex + y * ey;
   cout << "v = " << v << endl;

   if (orth)  {
      structure.cell (0,0) = ex (0);
      structure.cell (1,0) = ex (1);
      structure.cell (2,0) = ex (2);
      structure.cell (0,1) = ey (0);
      structure.cell (1,1) = ey (1);
      structure.cell (2,1) = ey (2);
      cout << "Orthogonised Cell:" << endl << structure.cell << endl;
      structure %= structure.cell;
   }

   // --- find the centres of the structure and dislocations
   SxVector3<Double> sum;
   sum.set (0.);
   int nA = structure.getNAtoms ();
   for (int iA = 0; iA < nA; ++iA) sum += structure.getAtom(iA);
   SxVector3<Double> centre = sum / nA;
   cout << "Structure Centre: " << centre << endl;
   centre += x0 * ex / ex.norm () + y0 * ey / ey.norm ();
   cout << "1. Dislocation Centre: " << centre << endl;
   cout << "2. Dislocation Centre: " << centre + v << endl;

   // --- disp: calculate the displacement field on cell boundaries
   double uex = 0., uey = 0.;
   if (disp)  {
      cout << "Check for boundary condition of displacement field:" << endl;
      SxVector3<Double> pvec;
      for (int pt = 0; pt <= disp; ++pt)  {
         SxVector3<Double> u0;
         u0.set (0.);
         for (int pp = 0; pp <= 3; ++pp)  {
            SxVector3<Double> u;
            u.set (0.);
            SxVector3<Double> exx = ex, eyy = ey;
            /*if (cut)  {
               exx = 0.5 * ex - 0.5 * ey;
               eyy = 0.5 * ex + 0.5 * ey;
            }*/
            if (pp == 0) pvec = double(pt) / disp * eyy;
            else if (pp == 1) pvec = double(pt) / disp * eyy + exx;
            else if (pp == 2) pvec = double(pt) / disp * exx;
            else pvec = double(pt) / disp * exx + eyy;
            for (int i = -rep; i <= rep; ++i)  {
               for (int j = -rep; j <= rep; ++j)  {
                  SxVector3<Double> vij = i * ex + j * ey;
                  int addDisl = 2;
                  //if (cut) addDisl = 4;
                  while (addDisl > 0)  {
                     SxVector3<Double> vector = vij + pvec - centre;
                     if (addDisl % 2 == 1) vector += v;
                     vector -= (vector ^ e3) * e3 / e3.normSqr ();
                     double r = vector.norm (),
                            xx = vector ^ be / be.norm (),
                            yy = vector ^ bt / bt.norm ();
                     SxVector3<Double> ux, uy, uz;
                     ux.set (0.), uy.set (0.), uz.set (0.);
                     if (bs.norm () > 1e-5)
                        uz = bs * angle(v, vector, e3) / TWO_PI;
                     if (be.norm () > 1e-5)  {
                        ux = be/TWO_PI*(atan2(yy, xx) + xx*yy/2/(1-nu)/r/r);
                        uy = -bt/FOUR_PI/(1-nu)*((1-2*nu)*log(r) + xx*xx/r/r);
                     }
                     /*if (addDisl > 2) vector += 0.5 * ex + 0.5 * ey;
                     if (yuja && (i == rep) && (addDisl == 1)) strain = 0.;
                     if (yuja && (j == rep) && (addDisl == 1)) strain = 0.;*/
                     if (addDisl % 2 == 0) u += ux + uy + uz;
                     else u -= ux + uy + uz;
                     addDisl--;
                  }
               }
            }
            //if (abs(u) >= 1) cout << "ATTENTION: u = " << u << endl;
            if (pp % 2 == 0) u0 = u;
            if (pp == 1)  {
               double duex = (u.norm () - u0.norm ()) / e3.norm ();
               cout << pt << ": " << pvec << "; uex = " << duex << endl;
               uex += duex - int(duex);
               if (duex < 0) uex += 1.;
            }
            if (pp == 3)  {
               double duey = (u.norm () - u0.norm ()) / e3.norm ();
               cout << pt << ": " << pvec << "; uey = " << duey << endl;
               uey += duey - int(duey);
               if (duey < 0) uey += 1.;
            }
         }
      }
      uex /= disp + 1.; //calculate average from sum
      uey /= disp + 1.;
      cout << "Suggestions for uex, uey (averaged over above values)" << endl 
           << "uex = " << uex << endl << "uey = " << uey << endl;
   }
   else cout << "ATTENTION: uex (uey) suggestions available "
             << "by rerunning with--disp enabled!" << endl;

   // --- construct cell of the dislocation structure
   cout << "uex = ", cin >> uex;
   cout << "uey = ", cin >> uey;
   SxAtomicStructure dislStruct;
   dislStruct.cell = CellMat (e1 + uex * b * e3,
                              e2 + uey * b * e3,
                              e3).transpose ();
   /*if (cut) dislStruct.cell = CellMat ((e1 - e2) / 2. + uex * b * e3,
                                       (e1 + e2) / 2. + uey * b * e3,
                                       e3).transpose ();*/

   // --- add atoms to the dislocation structure
   for (int is = 0; is < structure.getNSpecies (); ++is)  {
      dislStruct.startCreation ();
      dislStruct.newSpecies ();
      for (int ia = 0; ia < structure.getNAtoms(is); ++ia)  {
         SxVector3<Double> cartesian = structure.getAtom (is, ia);
         SxVector3<Double> 
         relative = cartesian ^ structure.cell.inverse ().transpose ();
         cout << ia << ": " << cartesian << "; " << relative;
         SxVector3<Double> u;
         u.set (0.);
         bool inside = (0 == 0);
         if (cut) inside = (abs (relative(0) + relative(1)) <= 0.5)
                    && (abs (relative(0) - relative(1)) <= 0.5);
         if (inside)  {
            for (int i = -rep; i <= rep; ++i)  {
               for (int j = -rep; j <= rep; ++j)  {
                  SxVector3<Double> vij = i * ex + j * ey;
                  int addDisl = 2;
                  //if (cut) addDisl = 4;
                  while (addDisl > 0)  {
                     SxVector3<Double> vector = vij + cartesian - centre;
                     if (addDisl % 2 == 1) vector += v;
                     vector -= (vector ^ e3) * e3 / e3.normSqr ();
                     //if (addDisl > 2) vector += 0.5 * ex + 0.5 * ey;
                     double r = vector.norm (),
                            xx = vector ^ be / be.norm (),
                            yy = vector ^ bt / bt.norm ();
                     SxVector3<Double> ux, uy, uz;
                     ux.set (0.), uy.set (0.), uz.set (0.);
                     if (bs.norm () > 1e-5)
                        uz = bs * angle(v, vector, e3) / TWO_PI;
                     if (be.norm () > 1e-5)  {
                        ux = be/TWO_PI*(atan2(yy, xx) + xx*yy/2/(1-nu)/r/r);
                        uy = -bt/FOUR_PI/(1-nu)*((1-2*nu)*log(r)+xx*xx/r/r);
                     }
                     /*if (yuja && (i == rep) && (addDisl == 1)) strain = 0.;
                     if (yuja && (j == rep) && (addDisl == 1)) strain = 0.;*/
                     if (addDisl % 2 == 0) u += ux + uy + uz;
                     else u -= ux + uy + uz;
                     addDisl--;
                  }
               }
            }
            cout << "total u = " << u << endl;
            dislStruct.addAtom(is, cartesian + u);
            cout << " -> " << cartesian + u << "; " 
            << ((cartesian + u) ^ structure.cell.inverse().transpose()) 
            << endl;
         }
         else cout << "SB" << endl;
      }
      dislStruct.endCreation ();
      cout << "Number of Atoms of Species" << is << ": " 
           << dislStruct.getNAtoms(is) << endl;
   }
   dislStruct %= dislStruct.cell;

   // --- output
   if (outSxb)  {
      dislStruct.write (outFile);
   } else {
      FILE *file;
      if (outFile.getSize () > 0)  {
         file = fopen(outFile.ascii(),"w");
         if (file == NULL)  {
            cout << "Can't open '" << outFile << "'." << endl;
            SX_EXIT;
         }
      } else {
         file = stdout;
      }

      dislStruct.fprint (file);
      if (file != stdout) fclose (file);

   }
}
