// ---------------------------------------------------------------------------
//                                          
//      The ab-initio based multiscale library 
//     
//                  S / P H I / n X
//                  http://www.sphinxlib.de
//     
//      Contact:    Sixten Boeck, boeck@mpie.de
//                  Algorithm Design and Modeling Group
//                  Computational Materials Design
//                  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//     
//      Authors:    see src/AUTHORS
//     
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxAtomicStructure.h>

double angle (SxVector3<Double> v1, SxVector3<Double> v2, SxVector3<Double> v3)
{
   double theta = 0.;
   if (v3.norm () < 1e-5) cout << "|v3| = " << v3.norm () << endl;
   else if (v2.norm () < 1e-5) cout << "|v2| = " << v2.norm () << endl;
   else if (v1.norm () < 1e-5) cout << "|v1| = " << v1.norm () << endl;
   else  {
      double cosTheta = v1 ^ v2 / v1.norm () / v2.norm ();
      if (cosTheta >= 1) theta = 0.;
      else if (cosTheta <= -1) theta = PI;
      else theta = acos(cosTheta);
      if ((theta > 0) && ((cross(v1, v2) ^ v3) < 0)) 
      theta = TWO_PI - theta;
   }
   return (theta);
}

int main (int argc, char **argv)
{
   initSPHInXMath ();
   
   SxCLI cli(argc, argv);

   cli.preUsageMessage 
      = SxString(" Add-on introduces a screw dislocation to a cell\n"
                 "The dislocation line is parallel to the third basis vector "
                 "of the input structure. ").wrap ();
   cli.authors = "Siyuan Zhang";
   
   SxVector3<Double> b(cli.option ("-b", "Burgers vector").toList3 ());

   double nu = cli.option ("--nu", "Poisson's ratio")
              .toDouble (1./3);

   double x0 = cli.option ("--x0", "", "place the dislocation with respect to "
                           "the structure centre")
              .toDouble (0.);

   double y0 = cli.option ("--y0", "")
              .toDouble (0.);

   double vacuum = cli.option ("--vacuum", "bohr", "thickness of vacuum shell")
             .toDouble (0.);

   bool orth = cli.option ("--orthogonal", "orthogonalise the cell")
              .toBool ();

   bool sxbFile = cli.option ("--sxb", "input file is binary waves file "
                              "rather than S/PHI/nX input file")
                  .toBool ();

   SxString inFile 
      = cli.option ("-i|--input", "input file", 
                    "take original input file")
        .toString (sxbFile ? "waves.sxb" : "input.sx");
   cli.last ().defaultValue = "default: input.sx, or waves.sxb for --sxb flag";
   
   SxString outFile
      = cli.option ("-o","filename", "output file name (screen otherwise)")
        .toString ("dislstructure");

   bool outSxb = cli.option ("--outsxb", "output file is binary "
                              "rather than S/PHI/nX input file").toBool ();
   cli.last ().defaultValue = "default: same format as input file";
   if (sxbFile) outSxb = true;
   if (outSxb && outFile.getSize () == 0)
      outFile = "dislstructure.sxb";
   
   cli.finalize ();

   // --- read input
   SxAtomicStructure structure;
   if (sxbFile)  {
      try {
         SxBinIO io (inFile, SxBinIO::BINARY_READ_ONLY);
         structure.read (io);
         SxString chemNameList;
         io.close ();
      } catch (SxException e)  {
         e.print ();
         SX_EXIT;
      }
   } else  {
      SxParser parser;
      SxConstPtr<SxSymbolTable> tree;
      tree = parser.read (inFile, "std/structure.std");
      structure = SxAtomicStructure (&*tree);
      structure.atomInfo->meta.update (SxAtomicStructure::Elements,
                                       SxSpeciesData::getElements (&*tree));
   }

   structure %= structure.cell;

   SxVector3<Double> e1 = structure.cell (0),
                     e2 = structure.cell (1),
                     e3 = structure.cell (2);
   cout << "Input Cell: " << structure.cell << endl;
   cout << "Inverse Cell: " << structure.cell.inverse () << endl;
   SxVector3<Double> ex = e1 - (e1 ^ e3) * e3 / e3.normSqr (),
                     ey = e2 - (e2 ^ e3) * e3 / e3.normSqr ();
   cout << "ex = " << ex << endl << "ey = " << ey << endl;
   cout << "Input b: " << b << endl;
   b = b(0) * e1 / e1.norm () + b(1) * e2 / e2.norm () + b(2) * e3;
   cout << "b = " << b << endl;
   SxVector3<Double> bs = (b ^ e3) * e3 / e3.normSqr (),
                     be = b - bs, bt;
   if (be.norm () < 1e-5) bt.set (0.);
   else bt = cross (e3, be) / e3.norm ();
   cout << "edge b = " << be 
        << "; screw b = " << bs 
        << "; transverse b = " << bt << endl;
   SxVector3<Double> v = be;
   if (be.norm () < 1e-5) v = ex;
   cout << "v = " << v << endl;

   if (orth)  {
      structure.cell (0,0) = ex (0);
      structure.cell (1,0) = ex (1);
      structure.cell (2,0) = ex (2);
      structure.cell (0,1) = ey (0);
      structure.cell (1,1) = ey (1);
      structure.cell (2,1) = ey (2);
      cout << "Orthogonised Cell:" << endl << structure.cell << endl;
      structure %= structure.cell;
   }

   // --- find the centres of the structure and dislocations
   SxVector3<Double> sum;
   sum.set (0.);
   int nA = structure.getNAtoms ();
   for (int iA = 0; iA < nA; ++iA) sum += structure.getAtom(iA);
   SxVector3<Double> centre = sum / nA;
   cout << "Structure Centre: " << centre << endl;
   centre += x0 * ex / ex.norm () + y0 * ey / ey.norm ();
   cout << "Dislocation Centre: " << centre << endl;

   // --- construct cell of the dislocation structure
   SxAtomicStructure dislStruct;
   dislStruct.cell = CellMat (e1 * (1 + vacuum / e1.norm ()),
                              e2 * (1 + vacuum / e2.norm ()),
                              e3).transpose ();

   // --- add atoms to the dislocation structure
   for (int is = 0; is < structure.getNSpecies (); ++is)  {
      dislStruct.startCreation ();
      dislStruct.newSpecies ();
      for (int ia = 0; ia < structure.getNAtoms(is); ++ia)  {
         SxVector3<Double> cartesian = structure.getAtom (is, ia);
         SxVector3<Double> 
         relative = cartesian ^ structure.cell.inverse ().transpose ();
         cout << ia << ": " << cartesian << "; " << relative;
         SxVector3<Double> u;
         u.set (0.);
         bool inside = (0 == 0);
         if (inside)  {
            SxVector3<Double> vector = cartesian - centre;
            vector -= (vector ^ e3) * e3 / e3.normSqr ();
            double r = vector.norm (),
                   x = vector ^ be / be.norm (),
                   y = vector ^ bt / bt.norm ();
            SxVector3<Double> ux, uy, uz; 
            ux.set (0.), uy.set (0.), uz.set (0.);
            if (bs.norm () > 1e-5)
               uz = bs * angle(v, vector, e3) / TWO_PI;
            if (be.norm () > 1e-5)  {
               ux = be / TWO_PI * (atan2(y, x) + x * y / 2 / (1 - nu) /r/r);
               uy = -bt / FOUR_PI / (1 - nu) * ((1 - 2 * nu)*log(r) + x*x/r/r);
            }
            cout << "ux = " << ux << "; uy = " << uy << "; uz = " << uz << endl;
            u += uz + ux + uy;
            dislStruct.addAtom(is, cartesian + u);
            cout << " -> " << cartesian + u << "; " 
            << ((cartesian + u) ^ structure.cell.inverse().transpose()) 
            << endl;
         }
         else cout << "SB" << endl;
      }
      dislStruct.endCreation ();
      cout << "Number of Atoms of Species" << is << ": " 
           << dislStruct.getNAtoms(is) << endl;
   }
   dislStruct %= dislStruct.cell;

   // --- output
   if (outSxb)  {
      dislStruct.write (outFile);
   } else {
      FILE *file;
      if (outFile.getSize () > 0)  {
         file = fopen(outFile.ascii(),"w");
         if (file == NULL)  {
            cout << "Can't open '" << outFile << "'." << endl;
            SX_EXIT;
         }
      } else {
         file = stdout;
      }

      dislStruct.fprint (file);
      if (file != stdout) fclose (file);

   }
}
