// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxBinIO.h>
#include <SxProjector.h>
#include <SxSymFockGk.h>

namespace Timer
{
   enum FastXTimer { 
      ApplyX,
      ComputeVkq,
      PrepareRot,
      RotateU
   };
}

REGISTER_TIMERS(Timer::FastXTimer)
{
   using namespace Timer;
   regTimer (ApplyX, "X * psi");
   regTimer (ComputeVkq, "Vkq");
   regTimer (PrepareRot, "setup starrot");
   regTimer (RotateU, "rotate u");
}

template<class T>
SxDiracVec<T> getSorted (const SxDiracVec<T> &x, const SxArray<ssize_t> &idx)
{
   SX_CHECK_VARS (x.getSize () == idx.getSize (), x.getSize (), idx.getSize ());
   ssize_t n = x.getSize ();
   SxDiracVec<T> res(n);
   res.setBasis (x.getBasisPtr ());
   for (ssize_t i = 0; i < n; ++i)
      res(i) = x(idx(i));
   return res;
}

PsiR applyX(const SxPW &waves, const SxFermi &fermi,
            const SxStars &stars, const PsiR &psiR,
            double singVal)
{
   CLOCK (Timer::ApplyX);
   const SxGkBasis &gkBasis = waves.getGkBasis ();
   const SxRBasis *rPtr = dynamic_cast<const SxRBasis *>(psiR.getBasisPtr ());
   SX_CHECK (rPtr);
   const SxRBasis &R = *rPtr;
   const SxGBasis &G = R.getGBasis ();
   const SxCell   &cell = R.cell;
   const SxMesh3D mesh = R.fft3d.mesh;

   SxArray<SxDiracVec<Double> > vkq;
   Coord kVec = gkBasis.getK(psiR.handle->auxData.ik);
   SxArray<SxArray<ssize_t> > starRot;
   SxCell meshCell (cell.basis(0) / mesh(0),
                    cell.basis(1) / mesh(1),
                    cell.basis(2) / mesh(2));
   int nSym = cell.symGroupPtr->getNSymmorphic ();

   int nSpin = waves.getNSpin ();
   int nk = waves.getNk ();

   SxDiracVec<TPrecCoeffG> uR, uRs, xPsi(R);
   PsiG wG;
   xPsi.set (0.);

   for (int iq = 0; iq < nk; ++iq)  {
      SxArray<Coord> star = stars.getStar (iq);
      double dq = stars.getWeight(iq);
      vkq.resize (star.getSize ());
      starRot.resize (star.getSize ());
      for (int iStar = 0; iStar < star.getSize (); ++iStar)  {
         int iSym = stars.getRotation (iq, iStar);
         // S rotates ks->(+-)q
         SxMatrix3<Double> S = cell.symGroupPtr->getSymmorphic (iSym % nSym);

         {
            // --- setup v = 4pi/|G+q-k|^2
            CLOCK (Timer::ComputeVkq);
            Coord qVec = S ^ gkBasis.getK(iq);
            if (iSym >= nSym) qVec *= -1.;
            Coord dkq = kVec - qVec;
            SxDiracVec<Double> &vkqStar = vkq(iStar);
            vkqStar.resize (G.ng);
            for (int ig = 0; ig < G.ng; ++ig)  {
               double gkq2 = (G.getG (ig) + dkq).normSqr ();
               vkqStar(ig) = (gkq2 > 1e-10) 
                           ? FOUR_PI / gkq2
                           : singVal / dq;
            }
         }

         //int ng = gkBasis(0).ng;
         //vkqStar(SxIdx(ng, vkqStar.getSize () -1)).set (0.); // Fourier filter

         // --- set up rotation index map
         {
            CLOCK (Timer::PrepareRot);
            // get associated real-space rotation (transpose=inverse)
            SxMatrix3<Int> rot = meshCell.carToRel (S.transpose ());

            SxArray<ssize_t> &idxmap = starRot(iStar);
            idxmap.resize (R.getNElements ());
            SxVector3<Int> rVec;

            for (rVec(0)=0; rVec(0) < mesh(0); rVec(0)++)  {
               for (rVec(1)=0; rVec(1) < mesh(1); rVec(1)++)  {
                  for (rVec(2)=0; rVec(2) < mesh(2); rVec(2)++)  {
                     ssize_t ir = mesh.getMeshIdx(rVec, SxMesh3D::Positive);
                     // compute index after rotation
                     idxmap(ir) = mesh.getMeshIdx(rot^rVec, SxMesh3D::Unknown);
                  }
               }
            }
         }
      }

      // --- compute exchange
      int iSpin = psiR.handle->auxData.iSpin;
      for (int i = 0; i < fermi.getNStates (iq); ++i)  {
         // get occupation number
         double focc = fermi.focc(i, iSpin, iq);
         if (nSpin == 1) focc *= 0.5; // only one spin channel!
         if (focc < 1e-6) continue; // ignore unoccupied states

         // get u in real-space at special q
         uRs = (R | waves(i, iSpin, iq));

         for (int iStar = 0; iStar < star.getSize (); ++iStar)  {
            {
               // --- get u for star member
               CLOCK (Timer::RotateU);
               uR = getSorted (uRs, starRot(iStar));
               if (stars.getRotation (iq, iStar) >= nSym)
                  uR = uR.conj ();
            }
            // w(r') = u*(r') psi(r') in G-space
            wG = (G | (uR.conj () * psiR));
            // u(r) v(r-r')w(r')
            xPsi.plus_assign_ax (-dq * focc, uR * (R | (vkq(iStar) * wG)));
         }
      }
   }
   return xPsi;
}

int main (int argc, char **argv)
{
   // --- init S/PHI/nX Utilities
   initSPHInXMath ();
   // --- init S/PHI/nX timers

   // --- parse command line
   SxCLI cli (argc, argv);
   cli.authors = "C. Freysoldt";
   SxString inFile =
      cli.option ("-i|--input","file","S/PHI/nX input file")
      .toString ("input.sx");

   SxString wavesFile =
      cli.option ("-w|--waves","file","waves file")
      .toString ("waves.sxb");

   bool fock = cli.option ("--fock", "compute fock contr.").toBool ();
   cli.version ("0.1");
   cli.finalize ();

   // --- read input file
   SxParser parser;
   SxParser::Table table = parser.read (inFile);
   
   // --- read waves file
   SxAtomicStructure structure;
   SxMesh3D mesh;
   SxPW waves;
   SxFermi fermi;
   SxBinIO io;
   try  {
      io.open (wavesFile, SxBinIO::BINARY_READ_ONLY);
      waves.read (io);
      fermi.read (io);
      io.read ("meshDim", &mesh);
      structure.read(io);
   } catch (SxException e)  {
      e.print ();
      SX_EXIT;
   }
   
   SxGkBasis &gkBasis = waves.getGkBasis();
   
   // --- setup G and R bases
   SxRBasis R(mesh, structure.cell);
   SxGBasis G(mesh, structure, SxGBasis::getGCut(SxGBasis::getECut(&*table)));
   //G.registerRBasis (R);
   R.registerGBasis (G);

   int ng = gkBasis(0).ng;
   gkBasis.setupN123inv ();
   SxSymFockGk symFock (gkBasis, ng, &*table);


   SxStars stars (structure.cell, gkBasis.kVec, gkBasis.weights, true);
   stars.compute ();

   PsiG psi, fockPsi;
   SxDiracVec<TPrecCoeffG> psiR, xPsi;

   int ik=0;  {
   //for (int ik = 0; ik < waves.getNk (); ++ik)  {
      // switch off singularity treatment
      //symFock.singTable(ik) = 0.;

      if (fock) symFock.compute2 (ik, waves, fermi);
      for (int iSpin = 0; iSpin < waves.getNSpin (); ++iSpin)  {
         for (int i = 0; i < waves.getNStates (); ++i)  {
            psi = waves(i, iSpin, ik);

            if (fock)  {
               fockPsi = symFock * psi;
               cout << "<psi|X-Fock|psi>=" << dot(psi, (symFock * psi)) << endl;
            }

            psiR = (R | psi);
            double singularity = symFock.singTable(ik)
                               * FOUR_PI; 

            xPsi = applyX (waves, fermi, stars, psiR, singularity);
            cout << "<psi|X|psi>=" << dot(psiR, xPsi) * R.dOmega << endl;

            if (fock)  {
               cout << "|(X-Xfock)|psi>| " 
                    << ((gkBasis(ik) | xPsi) - fockPsi).norm () << endl;
            }
         }
      }
   }

   printTiming ();

}

