// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxVector.h>
#include <SxDirac.h>
#include <SxNArray.h>
#include <SxSphereGrid.h>
#include <SxParser.h>
#include <SxYlm.h>
#include <SxUniqueList.h>
#include <SxTimer.h>

enum SxDomTimer { FullDiag, GetOrtho, RaiseDomainLevel, CompleteDiag,
                  CompleteDiagCheck, GetDomainBase, GetDomDiag, GetDomRes};

REGISTER_TIMERS (SxDomTimer)
{
   regTimer (FullDiag, "full diag");
   regTimer (GetOrtho, "get ortho");
   regTimer (RaiseDomainLevel, "raise domain level");
   regTimer (CompleteDiag, "compl. diag");
   regTimer (CompleteDiagCheck, "compl. diag check");
   regTimer (GetDomainBase, "get domain base (DB)");
   regTimer (GetDomDiag, "DB diag");
   regTimer (GetDomRes, "DB residues");
}

class SxDomain;

class SxDomainTreeLeaf 
{
   public:
      /// Hierarchy level
      int level;
      /// Leaf id within level
      int leafId;
      /// Child domains (level + 1)
      SxArray<SxPtr<SxDomainTreeLeaf> > child;
      /// Parent domain (level - 1)
      SxPtr<SxDomainTreeLeaf> parent;
      /// Neighbors
      SxArray<SxPtr<SxDomainTreeLeaf> > neighbor;

   protected:
      /// The actual domains (conjugate id)
      SxArray<SxPtr<SxDomain> > domains;

      /// Level offsets
      SxArray<int> levelOffset;
   public:

      SxDomainTreeLeaf (int levelIn, int idIn)
         : level(levelIn), leafId(idIn)
      {
         // empty
      }
      
      virtual ~SxDomainTreeLeaf () {}

      const SxPtr<SxDomain> & getDomain (int iLevel, int iD) const
      {
         SX_CHECK_VARS (iLevel >=0 && iLevel < levelOffset.getSize (),
                        iLevel, levelOffset.getSize ());
         return domains(levelOffset(iLevel) + iD);
      }

      SxPtr<SxDomain> & getDomain (int iLevel, int iD)
      {
         SX_CHECK_VARS (iLevel >=0 && iLevel < levelOffset.getSize (),
                        iLevel, levelOffset.getSize ());
         return domains(levelOffset(iLevel) + iD);
      }

      /** \brief Resize domain list for a certain level
        \note The domain list of a certain level cannot be changed
              in size later on.
        */
      void resize (int iLevel, int nD)  {
         int nLevel = int(levelOffset.getSize ());
         if (iLevel >= nLevel)  {
            levelOffset.resize (iLevel+1, true);
            for (int i = nLevel; i <= iLevel; ++i)
               levelOffset(i) = 0;
         }
         SX_CHECK (levelOffset(iLevel) == 0);
         if (nD == 0) return;
         int nDomains = int(domains.getSize ());
         domains.resize (nDomains + nD, true);
         levelOffset(iLevel) = nDomains;
      }
};

class SxEnergyDomain : public SxDomainTreeLeaf
{
   public:
      double eMin, eMax;
      SxEnergyDomain (int levelIn, int idIn, double eMinIn, double eMaxIn)
         : SxDomainTreeLeaf (levelIn, idIn), eMin(eMinIn), eMax(eMaxIn)
      {
         SX_CHECK_VARS (eMin < eMax, eMin, eMax);
      }
      void split (int nE);

      SxPtr<SxEnergyDomain> getChild (int iE) const
      {
         return child(iE);
      }
};

void SxEnergyDomain::split (int nE)
{
   SX_CHECK_VAR (child.getSize () == 0, child.getSize ());
   child.resize (nE);
   double dE = (eMax - eMin) / nE;
   for (int iE = 0; iE < nE; ++iE)  {
      child(iE) = SxPtr<SxEnergyDomain>::create (level + 1, iE,
                                                 eMin + iE * dE,
                                                 eMin + (iE + 1) * dE);
   }
}


class SxDomains : public SxArray<SxPtr<SxDomain> > 
{
   public:
      using SxArray<SxPtr<SxDomain> >::operator=;

      void updateNeighbors ();
      void updateNeighborHS ();

      void completeDiag (const SxMatrix<Double> &globalH) const;

      int getTotalSize () const;

      SxPtr<SxEnergyDomain> energyDomain;
};

class SxDomain {
   public:
      /// Neighbor domains
      SxDomains neighbor;

      /// Child domains
      SxDomains child;

      /// Child trafo iChild:(iBaseChild,iBase)
      SxArray<SxMatrix<Complex16> > U;

      /// Hamiltonian
      SxMatrix<Complex16> H;

      /// Hamiltonian with neighbors <i|H|n>
      SxArray<SxMatrix<Complex16> > neighborH,
      /// Overlap with neighbors     <i|S|n>
                                    neighborS;

      /// Constructor
      SxDomain (const SxPtr<SxEnergyDomain> &energy);

      inline int getSize () const {
         if (H.getSize () == 0) return 0;
         SX_CHECK_VARS (H.nRows () == H.nCols (), H.nCols (), H.nRows ());
         return (int)H.nRows ();
      };

      int findNeighbor (const SxDomain &in) const;

      /// Pointer to the spatial domain
      SxPtr<SxDomainTreeLeaf> spatialDomain;

      /// Pointer to the energetic domain
      SxPtr<SxEnergyDomain> energyDomain;

      /// Global id (bottom level only)
      SxArray<int> globalId;

      /// Add this domain's contribution to a global representation
      void addToGlobalVec (SxVector<Complex16> &res, 
                           const SxVector<Complex16> &dWave) const;

      /// Compare Hamiltonian to full representation
      void checkHam (const SxMatrix<Complex16> &H) const;

      /** \brief Get new domain bases for split energy interval
                 from diagonalization in several domains
          @param domains List of connected domains, first entry must
                         be this domain
          @param nE      number of energy intervals
          @param dErel   energy interval overlap
          @result   one basis transformation for each energy interval
                    (nBasisOld x nBasisNew)

          @note The current energy interval is split linearly into
                nE many subintervals

        */
      SxArray<SxMatrix<Complex16> > 
      getDomainBase (const SxArray<SxPtr<SxDomain> > &domains,
                     int nE, double dErel, int pass = 0);

      /** \brief Update Hamiltonian and overlap matrices with neighbors
          @param reciprocal  If true, the matrices on the neighbors are
                             updated, too.
        */
      void updateNeighborHS (bool reciprocal);

      /** \brief Reassemble from children

        This collects all basis functions from the children
        and removes linear dependencies and orthonormalizes.

        The Hamiltonian is recomputed.
        The neighbor matrix elements are NOT recomputed.

        */
      void recompute ();
};

class SxDomainVec {
   public:
      /// Wave function in a certain domain (iD)
      SxArray<SxVector<Complex16> > psiPart;

      /// Domains that support this wave
      SxUniqueList<const SxDomain*> domains;

      SxDomainVec () {}
      SxDomainVec (const SxVector<Complex16> &psi,
                   const SxDomain &D, bool temp = false);
      SxDomainVec (const SxVector<Complex16> &psi,
                   const SxList<const SxDomain*> &D);
      SxDomainVec (const SxVector<Complex16> &psi,
                   const SxArray<SxPtr<SxDomain> > &D);

      SxDomainVec &operator += (const SxDomainVec &in);

      SxDomainVec &plus_assign_ax (const SxComplex16 &a, const SxDomainVec &x);

      SxDomainVec down () const;

      SxDomainVec applyH (bool restricted) const;

      SxDomainVec applyS (bool restricted) const;
};

SxComplex16 dot (const SxDomainVec &a, const SxDomainVec &b)
{
   SxComplex16 res = 0.;
   for (int id = 0; id < a.domains.getSize (); ++id)  {
      const SxDomain &dom1 = *a.domains(id);
      for (int jd = 0; jd < b.domains.getSize (); ++jd)  {
         const SxDomain &dom2 = *b.domains(jd);
         if (&dom1 == &dom2)  {
            res += dot(a.psiPart(id), b.psiPart(jd));
         } else {
            int in = dom1.findNeighbor (dom2);
            if (in >= 0)  {
               res += dot(a.psiPart(id), dom1.neighborS(in) ^ b.psiPart(jd));
            }
         }
      }
   }
   return res;
}

SxDomainVec::SxDomainVec (const SxVector<Complex16> &psi,
                          const SxDomain &D,
                          bool temp)
{
   SX_CHECK (D.getSize () > 0);
   SX_CHECK_VARS (psi.getSize () == D.getSize (), psi.getSize (), D.getSize ());
   psiPart.resize (1);
   if (temp)
      psiPart(0) = psi;
   else
      psiPart(0).copy (psi);
   SX_EXIT;// FIXME
   //domains.append (&D);
}

SxDomainVec::SxDomainVec (const SxVector<Complex16> &psi,
                          const SxList<const SxDomain*> &D)
{
   domains.resize (D.getSize ());
   psiPart.resize (D.getSize ());
   int offset = 0;
   for (int id = 0; id < D.getSize (); ++id)  {
      domains(id) = D(id);
      SX_CHECK (D(id));
      int nBase = D(id)->getSize ();
      psiPart(id).copy (psi(SxIdx(offset, offset + nBase - 1)));
      offset += nBase;
   }
   SX_CHECK_VARS (offset == psi.getSize (), offset, psi.getSize ());
}

SxDomainVec::SxDomainVec (const SxVector<Complex16> &psi,
                          const SxArray<SxPtr<SxDomain> > &D)
{
   domains.resize (D.getSize ());
   psiPart.resize (D.getSize ());
   int offset = 0;
   for (int id = 0; id < D.getSize (); ++id)  {
      domains(id) = D(id).getPtr ();
      int nBase = D(id)->getSize ();
      psiPart(id).copy (psi(SxIdx(offset, offset + nBase - 1)));
      offset += nBase;
   }
   SX_CHECK_VARS (offset == psi.getSize (), offset, psi.getSize ());
}


SxDomainVec & SxDomainVec::operator+= (const SxDomainVec &in)
{
   for (int id = 0; id < in.domains.getSize (); ++id)  {
      SX_EXIT; // FIXME
      //int myId = (domains << in.domains(id));
      //if (myId == psiPart.getSize ())  {
      //   psiPart.resize (myId + 1, true);
      //   psiPart(myId).copy (in.psiPart (id));
      //} else {
      //   psiPart(myId) += in.psiPart (id);
      //}
      //SX_CHECK (psiPart(myId).getSize () > 0);
   }
   return *this;
}

SxDomainVec & SxDomainVec::plus_assign_ax (const SxComplex16 &/*a*/,
                                           const SxDomainVec &x)
{
   for (int id = 0; id < x.domains.getSize (); ++id)  {
      SX_EXIT; // FIXME
      // int myId = (domains << x.domains(id));
      // if (myId == psiPart.getSize ())  {
      //    psiPart.resize (myId + 1, true);
      //    psiPart(myId) = a * x.psiPart (id);
      // } else {
      //    psiPart(myId).plus_assign_ax (a, x.psiPart (id));
      // }
   }
   return *this;
}

SxDomainVec SxDomainVec::down () const
{
   SxDomainVec res;
   for (int id = 0; id < domains.getSize (); ++id)  {
      const SxDomain &dom = *domains(id);
      SX_CHECK (dom.globalId.getSize () == 0);
      SX_CHECK (dom.getSize () > 0);
      for (int ic = 0; ic < dom.child.getSize (); ++ic)  {
         res += SxDomainVec (dom.U(ic) ^ psiPart(id), *dom.child(ic), true);
      }
   }
   return res;
}

SxDomainVec SxDomainVec::applyH (bool restricted) const
{
   SxDomainVec res;
   for (int id = 0; id < domains.getSize (); ++id)  {
      SX_CHECK (domains(id));
      const SxDomain &dom = *domains(id);
      SX_CHECK (psiPart(id).getSize () > 0);
      SX_CHECK_VARS (psiPart(id).getSize () == dom.getSize (),
                     psiPart(id).getSize (), dom.getSize ());

      res += SxDomainVec (dom.H ^ psiPart(id), dom, true);
      for (int in = 0; in < dom.neighbor.getSize (); ++in)  {
         if (!restricted || domains.findPos (dom.neighbor(in).getPtr ()) >= 0)
         {
            res += SxDomainVec (dom.neighborH(in).adjoint () ^ psiPart(id),
                                *dom.neighbor(in), true);
         }
      }
   }
   return res;
}

SxDomainVec SxDomainVec::applyS (bool restricted) const
{
   SxDomainVec res;
   for (int id = 0; id < domains.getSize (); ++id)  {
      SX_CHECK (domains(id));
      const SxDomain &dom = *domains(id);
      SX_CHECK (psiPart(id).getSize () > 0);
      SX_CHECK_VARS (psiPart(id).getSize () == dom.getSize (),
                     psiPart(id).getSize (), dom.getSize ());

      res += SxDomainVec (psiPart(id), dom, true);
      for (int in = 0; in < dom.neighbor.getSize (); ++in)  {
         if (!restricted || domains.findPos (dom.neighbor(in).getPtr ()) >= 0)
         {
            res += SxDomainVec (dom.neighborS(in).adjoint () ^ psiPart(id),
                                *dom.neighbor(in), true);
         }
      }
   }
   return res;
}

int idxN(int i, int N)  {
   int j = i % N;
   if (j < 0) j += N;
   return j;
}

SxMatrix<Complex16> getOrtho (const SxMatrix<Complex16> &S, double eps = 1e-10)
{
   CLOCK (GetOrtho);
   int nBas = int(S.nCols ());
   SX_CHECK_VAR (nBas > 0, nBas);
   SxMatrix<Complex16>::Eigensystem eig;
   //SxMatrix<Complex16> symS = 0.5 * (S.adjoint () + S);
   SxMatrix<Complex16> symS = S;
   eig = symS.eigensystem ();
   SX_CHECK_VAR (eig.vals(nBas - 1).re > 0, eig.vals(nBas-1).re);
   eps *= eig.vals(nBas - 1).re;
   int nFree = 0;
   bool refine = true;
   for (int i = 0; i < nBas; ++i)  {
      if (eig.vals(i).re > eps) {
         nFree++;
         //if (eig.vals(i).re < 1e-3) refine = true;
      }
   }
   SX_CHECK (nFree > 0);
   SxMatrix<Complex16> L(nBas, nFree);
   for (int i = 0, iFree = 0; i < nBas; ++i)  {
      if (eig.vals(i).re > eps)  {
         L.colRef(iFree++) << (1./sqrt(eig.vals(i).re)) * eig.vecs.colRef(i);
      }
   }
   VALIDATE_VECTOR (L);
   if (refine)  {
      // --- reduce the numerical noise by orthonormalizing the
      //     final set a 2nd time
      SxMatrix<Complex16> LSL = L.adjoint () ^ symS ^ L;
      //LSL *= 0.5; LSL += LSL.adjoint ();
      //cout << "LSL:" << (LSL.eigenvalues () - 1.) << endl;

      L = L ^ LSL.choleskyDecomposition ().inverse ().adjoint ();

      /*
      LSL *= -0.5;
      for (int i = 0; i < LSL.nCols (); ++i) {
         LSL(i,i).re += 1.5;
         LSL(i,i).im = 0.;
      }
      L = L ^ LSL;
      */
   }
   /*
   SxMatrix<Complex16> S2 = L.adjoint () ^ symS ^ L;
   for (int i = 0; i < S2.nCols (); ++i) S2(i,i).re -= 1.;
   double s2n = S2.normSqr ();
   cout << "getOrtho check: " << s2n << endl;
   if (s2n > 1e-20)  {
      cout << S2 << endl;
      cout << eig.vals(SxIdx(nBas - nFree, nBas-1)) << endl;
   }
   */
   return L;
}

void SxDomain::addToGlobalVec (SxVector<Complex16> &res, 
                               const SxVector<Complex16> &dWave) const
{
   if (globalId.getSize () > 0)  {
      SX_CHECK_VARS (dWave.getSize () == globalId.getSize (),
                     dWave.getSize (), globalId.getSize ());
      // bottom level: use globalId to map
      for (int ig = 0; ig < globalId.getSize (); ++ig)  {
         res(globalId(ig)) += dWave(ig);
      }
   } else {
      SX_CHECK (child.getSize () >= 0);
      SX_CHECK_VARS (dWave.getSize () == getSize (),
                     dWave.getSize (), getSize ());
      // transform to child basis and continue there
      for (int iChild = 0; iChild < child.getSize (); ++iChild)  {
         if (child(iChild)->getSize () > 0)  {
            child(iChild)->addToGlobalVec (res, U(iChild) ^ dWave);
         }
      }
   }
}

void SxDomain::checkHam (const SxMatrix<Complex16> &globalH) const
{
   ssize_t ng = globalH.nRows (), n = getSize ();
   if (n == 0) return;
   SxMatrix<Complex16> uGlobal (ng, n);
   SxVector<Complex16> uLocal (n);
   uLocal.set (0.);
   for (int i = 0; i < n; ++i)  {
      uLocal(i) = 1.;
      SxVector<Complex16> u = uGlobal.colRef (i);
      u.set (0.);
      addToGlobalVec (u, uLocal);
      uLocal(i) = 0.;
   }
   SxMatrix<Complex16> Sg = uGlobal.adjoint () ^ uGlobal,
                       Hg = uGlobal.adjoint () ^ globalH ^ uGlobal;
   for (int i = 0; i < n; ++i)  {
      Sg(i,i).re -= 1.;
   }
   cout << "H check: " << (Hg - H).normSqr () << endl;
   cout << "S check: " << Sg.normSqr () << endl;
}

void setBlock (SxMatrix<Complex16> *A, int offsetRow, int offsetCol,
               const SxMatrix<Complex16> &B)
{
   SX_CHECK (A);
   SX_CHECK (A->getSize () > 0);
   ssize_t nr = B.nRows (), nc = B.nCols ();
   SX_CHECK_3VARS (A->nCols () >= offsetCol + nc, A->nCols (), offsetCol, nc);
   SX_CHECK_3VARS (A->nRows () >= offsetRow + nr, A->nRows (), offsetRow, nr);
   for (int ic = 0, ico = offsetCol; ic < nc; ++ic, ++ico)
      for (int ir = 0, iro = offsetRow; ir < nr; ++ir, ++iro)
         (*A)(iro, ico) = B(ir, ic);
}

SxMatrix<Complex16> getRowBlock (const SxMatrix<Complex16> &A,
                                 int offset, int nRows)
{
   SxMatrix<Complex16> res (nRows, A.nCols ());
   for (int ic = 0; ic < A.nCols (); ++ic)
      for (int ir = 0, iro = offset; ir < nRows; ++ir, ++iro)
         res(ir, ic) = A(iro, ic);
   VALIDATE_VECTOR (res);
   return res;
}

SxDomain::SxDomain (const SxPtr<SxEnergyDomain> &energy)
   : energyDomain(energy)
{
}

SxDomains raiseDomainLevel (const SxDomains &phi, int nPerD = 2, int ovlp = 1)
{
   CLOCK (RaiseDomainLevel);
   int nPhi = int(phi.getSize ());
   SX_CHECK_VARS (nPhi % nPerD == 0, nPhi, nPerD);
   int nD = nPhi / nPerD;
   SxDomains newDomains;
   newDomains.resize (nD);

   for (int iPhi = 0, iD = 0; iPhi < nPhi; iPhi += nPerD, ++iD)  {

      SxPtr<SxEnergyDomain> energyDomain = phi(iPhi)->energyDomain;

      // --- create new domain
      SxPtr<SxDomain> newDomain = SxPtr<SxDomain>::create (energyDomain);
      newDomains(iD) = newDomain;

      int nChild = nPerD + 2 * ovlp;
      int nBase = 0;
      newDomain->child.resize (nChild);
      int nEmptyChild = 0;
      for (int i = 0, iChild = 0; i < nChild; ++i)  {
         int jD = idxN(iPhi - ovlp + i, nPhi);
         SX_CHECK (phi(jD));
         SX_CHECK (phi(jD)->energyDomain == energyDomain);
         if (phi(jD)->getSize () > 0)  {
            newDomain->child(iChild++) = phi(jD);
            nBase += phi(jD)->getSize ();
         } else {
            nEmptyChild++;
         }
      }
      nChild -= nEmptyChild;
      newDomain->child.resize (nChild, true);
      cout << iPhi << ": ";
      newDomain->recompute ();
   }

   newDomains.updateNeighbors ();

   return newDomains;
   
}

void SxDomain::recompute () 
{
   int nBase = 0;
   ssize_t nChild = child.getSize ();
   for (int iChild = 0; iChild < nChild; ++iChild)  {
      SX_CHECK (child(iChild)->energyDomain == energyDomain);
      nBase += child(iChild)->getSize ();
   }
   //cout << this << " ";
   cout << "nBase = " << nBase;

   if (nBase == 0) { cout << endl; return; }

   // --- setup H and S in new primary basis
   SxMatrix<Complex16> S(nBase,nBase);
   H.reformat (nBase, nBase);
   H.set (0.);
   S.set (0.);
   for (int iBase = 0; iBase < nBase; ++iBase)
      S(iBase, iBase) = 1.;

   for (int iChild = 0, iBase = 0; iChild < nChild; ++iChild)  {
      SxDomain &child1 = *child(iChild);
      int nBase1 = child1.getSize ();
      for (int jChild = 0, jBase = 0; jChild < nChild; ++jChild)  {
         SxDomain &child2 = *child(jChild);
         int nBase2 = child2.getSize ();
         if (iChild == jChild)  {
            setBlock (&H, iBase, jBase, child1.H);
         } else {
            int in = child1.findNeighbor (child2);
            if (in >= 0)  {
               setBlock (&H, iBase, jBase, child1.neighborH (in));
               setBlock (&S, iBase, jBase, child1.neighborS (in));
            }
         }
         jBase += nBase2;
      }
      iBase += nBase1;
   }

   // --- orthonormalize (and reduce)
   SxMatrix<Complex16> L = getOrtho (S);
   cout << " ==> " << L.nCols ();

   // transform H
   H = L.adjoint () ^ H ^ L;

   // S now is unity
   // get transformations
   U.resize (nChild);

   for (int iChild = 0, iBase = 0; iChild < nChild; ++iChild)  {
      int nBaseC = child(iChild)->getSize ();
      U(iChild) = getRowBlock (L, iBase, nBaseC);
      iBase += nBaseC;
      SX_CHECK_VARS (U(iChild).nCols () == getSize (),
                     U(iChild).nCols (), getSize ());
      SX_CHECK_VARS (U(iChild).nRows () == nBaseC,
                     U(iChild).nRows (), nBaseC);
      // POINTER check
      //cout << "recompute " << this << " child "
      //     << child (iChild).getPtr () << ": "
      //     << nBaseC << endl;
   }

   // --- remove far-energy degrees of freedom
   /*
   {
      SxMatrix<Complex16>::Eigensystem eig = H.eigensystem ();
      cout << eig.vals.real () << endl;
      int nH = eig.vals.getSize ();
      double fac = 4.;
      double eMin2 = eMin - fac * (eMax - eMin),
             eMax2 = eMax + fac * (eMax - eMin);
      int nRed = 0;
      for (int i = 0; i < nH; ++i)  {
         if (eig.vals(i).re >= eMin2 && eig.vals(i).re <= eMax2)  {
            nRed++;
         }
      }
      if (nRed < nH)  {
         cout << " ==> " << nRed;
         SxMatrix<Complex16> U2(nH, nRed);
         for (int i = 0, iRed = 0; i < nH; ++i)  {
            if (eig.vals(i).re >= eMin2 && eig.vals(i).re <= eMax2)  {
               U2.colRef(iRed++) << eig.vecs.colRef(i);
            }
         }
         // --- reorthonormalize
         U2 = U2 ^ getOrtho (U2.adjoint () ^ U2);

         // update H
         H = U2.adjoint () ^ H ^ U2;
         // update U
         for (int iChild = 0; iChild < nChild; ++iChild)  {
            U(iChild) = U(iChild) ^ U2;
         }
      }
   }
   */
   cout << endl;
}

void SxDomains::updateNeighbors ()
{
   for (int i = 0; i < getSize (); ++i)  {
      if ((*this)(i)->getSize () == 0) continue;
      // --- find neighbor domains
      ssize_t nChild1 = (*this)(i)->child.getSize ();
      SxStack<SxPtr<SxDomain> > neighbors;
      for (int j = 0; j < getSize (); ++j)  {
         
         if (i == j) continue;

         ssize_t nChild2 = (*this)(j)->child.getSize ();
         bool isNeighbor = false;
         for (int iChild = 0; iChild < nChild1; ++iChild)  {
            SxDomain &child1 = *(*this)(i)->child(iChild);
            for (int jChild = 0; jChild < nChild2; ++jChild)  {
               SxDomain &child2 = *(*this)(j)->child(jChild);
               if (&child1 == &child2 || child1.findNeighbor (child2) >= 0)  {
                  isNeighbor = true;
                  break;
               }
            }
            if (isNeighbor) break;
         }
         if (isNeighbor)
            neighbors << (*this)(j);
      }
      (*this)(i)->neighbor = SxArray<SxPtr<SxDomain> > (neighbors);
   }
   updateNeighborHS ();
}

void SxDomain::updateNeighborHS (bool reciprocal)
{
   ssize_t nChild1 = child.getSize ();
   int nBase1 = getSize ();
   if (nBase1 == 0) return;
   for (int in = 0; in < neighbor.getSize (); ++in)  {
      ssize_t nChild2 = neighbor(in)->child.getSize ();
      int nBase2 = neighbor(in)->getSize ();
      if (nBase2 == 0) return;

      SxMatrix<Complex16> &nH = neighborH(in);
      SxMatrix<Complex16> &nS = neighborS(in);
      nH.reformat (nBase1, nBase2);
      nH.set (0.);
      nS.reformat (nBase1, nBase2);
      nS.set (0.);

      for (int iChild = 0; iChild < nChild1; ++iChild)  {
         SxDomain &child1 = *child(iChild);
         SxMatrix<Complex16> &U1 = U(iChild);

         for (int jChild = 0; jChild < nChild2; ++jChild)  {
            SxDomain &child2 = *neighbor(in)->child(jChild);
            SxMatrix<Complex16> &U2 = neighbor(in)->U(jChild);
            if (&child1 == &child2)  {
               // --- common child
               // set H
               nH += U1.adjoint () ^ child1.H ^ U2;
               // set S (orthonormal)
               nS += U1.adjoint () ^ U2;
            } else {
               // --- child1 and child2 are neighbors?
               int idN = child1.findNeighbor (child2);
               if (idN >= 0)  {
                  SX_CHECK_VARS (U1.nRows () == child1.getSize (),
                                 U1.nRows (), child1.getSize ());
                  SX_CHECK_VARS (U2.nRows () == child2.getSize (),
                                 U2.nRows (), child2.getSize ());
                  SX_CHECK_VARS (U1.nCols () == getSize (),
                                 U1.nCols (), getSize ());
                  SX_CHECK_VARS (U2.nCols () == neighbor(in)->getSize (),
                                 U2.nCols (), neighbor(in)->getSize ());
                  nH += U1.adjoint () ^ child1.neighborH(idN) ^ U2;
                  nS += U1.adjoint () ^ child1.neighborS(idN) ^ U2;
               }
            }
         }
      }
      if (reciprocal)  {
         int jn = neighbor(in)->findNeighbor (*this);
         SX_CHECK (jn >= 0);
         neighbor(in)->neighborH(jn) = nH.adjoint ();
         neighbor(in)->neighborS(jn) = nS.adjoint ();
      }
   }
}

void SxDomains::updateNeighborHS ()
{
   for (int i = 0; i < getSize (); ++i)  {
      int nBase1 = (*this)(i)->getSize ();
      if (nBase1 == 0) continue;
      // --- setup neighborH and neighborsS
      ssize_t nNeighbors = (*this)(i)->neighbor.getSize ();
      (*this)(i)->neighborH.resize (nNeighbors);
      (*this)(i)->neighborS.resize (nNeighbors);
      (*this)(i)->updateNeighborHS (false);
   }
}

void SxDomains::completeDiag (const SxMatrix<Double> &globalH) const
{
   CLOCK (CompleteDiag);
   int nBase = 0;
   for (int iD = 0; iD < getSize (); ++iD)  {
      nBase += (*this)(iD)->getSize ();
   }
   if (nBase == 0) return;
   SxMatrix<Complex16> H(nBase, nBase), S(nBase, nBase);

   H.set (0.);
   S.set (0.);
   for (int iBase = 0; iBase < nBase; ++iBase) S(iBase, iBase) = 1.;

   for (int iD = 0, iBase = 0; iD < getSize (); ++iD)  {
      SxDomain &domain1 = *(*this)(iD);
      int nBase1 = domain1.getSize ();
      for (int jD = 0, jBase = 0; jD < getSize (); ++jD)  {
         SxDomain &domain2 = *(*this)(jD);
         int nBase2 = domain2.getSize ();
         if (iD == jD)  {
            setBlock (&H, iBase, iBase, domain1.H);
         } else {
            int iN = domain1.findNeighbor (domain2);
            if (iN >= 0)  {
               setBlock (&H, iBase, jBase, domain1.neighborH(iN));
               setBlock (&S, iBase, jBase, domain1.neighborS(iN));
            }
         }
         jBase += nBase2;
      }
      iBase += nBase1;
   }
   /*
   {
      int nGlobal = globalH.nRows ();
      SxMatrix<Complex16> fBas(nGlobal, nBase);
      fBas.set (0.);
      for (int iD = 0, iBase = 0; iD < getSize (); ++iD)  {
         SxDomain &domain1 = *(*this)(iD);
         int nBase1 = domain1.getSize ();
         SxVector<Complex16> b(nBase1);
         b.set (0.);
         cout << iD << ": " << iBase;
         for (int j = 0; j < nBase1; ++j, ++iBase)  {
            b(j) = 1.;
            SxVector<Complex16> col = fBas.colRef(iBase);
            (*this)(iD)->addToGlobalVec(col, b);
            b(j) = 0.;
         }
         cout << "-" << (iBase - 1) << endl;
      }
      SxMatrix<Complex16> fS = fBas.adjoint () ^ fBas;
      fS = S - fS;
      for (int i = 0; i < nBase; ++i)  {
         for (int j = 0; j < nBase; ++j)  {
            if (fS(j,i).absSqr () > 1e-20)  {
               sxprintf ("S[%d,%d]: dom=%.2e %.2e <=> full=%.2e %.2e (%.2e %.2e)\n",
                       j,i,S(j,i).re, S(j,i).im,
                       S(j,i).re - fS(j,i).re,
                       S(j,i).im - fS(j,i).im,
                       fS(j,i).re, fS(j,i).im);
            }
         }
      }
   }
   */
   SxMatrix<Complex16> L = getOrtho (S);

   //cout << S << endl;
   H = L.adjoint () ^ H ^ L;
   //cout << (L.adjoint () ^ S ^ L) << endl;

   SxMatrix<Complex16>::Eigensystem eig = H.eigensystem ();

   double eMin = (*this)(0)->energyDomain->eMin,
          eMax = (*this)(0)->energyDomain->eMax;
   cout << "complete diag for " << eMin << "..." << eMax
        << ": " << eig.vals.getSize () << "/" 
        << nBase << " basis in " 
        << getSize () << " domains"
        << endl;

   int nFound = 0;
   for (int i = 0; i < eig.vals.getSize (); ++i)  {
      double eps = eig.vals (i).re;
      if (eps >= eMin && eps <= eMax)  {
         nFound++;
         if (globalH.getSize () == 0)  {
            sxprintf ("=> %.10f\n", eps);
         } else {
            CLOCK (CompleteDiagCheck);
            SxVector<Complex16> vec = L ^ eig.vecs.colRef (i);
            SxVector<Complex16> globalVec(globalH.nRows ());
            globalVec.set (0.);
            for (int iD = 0, iBase = 0; iD < getSize (); ++iD)  {
               SxDomain &domain1 = *(*this)(iD);
               int nBase1 = domain1.getSize ();
               if (nBase1 > 0)  {
                  domain1.addToGlobalVec (globalVec, 
                                           vec(SxIdx(iBase, iBase + nBase1 - 1)));
               }
               iBase += nBase1;
            }
            /*
            cout << "N=" << (globalVec.normSqr ()) << endl;
            cout << globalVec << endl;
            cout << (globalH ^ globalVec) << endl;
            cout << (eps * globalVec) << endl;
            cout << "e=" << dot (globalVec, globalH ^ globalVec) << endl;
            */
            double norm = globalVec.normSqr ();
            if (fabs(1. - norm) > 1e-4)  {
               cout << "N= " << norm << endl;
            }
            double res = ((globalH ^ globalVec) - eps * globalVec).norm ();
            if (res > 1e-4)  {
               sxprintf ("=> %.10f (r=%.2g) !\n", eps, res);
            } else {
               sxprintf ("=> %.10f (r=%.2g)\n", eps, res);
            }
         }
      }
   }
   cout << "completeDiag: " << nFound << "/" << nBase << endl;

}

int SxDomains::getTotalSize () const
{
   int nTot = 0;
   for (int i = 0; i < getSize (); ++i)  {
      nTot += (*this)(i)->getSize ();
   }
   return nTot;
}

int SxDomain::findNeighbor (const SxDomain &in) const
{
   for (int i = 0; i < neighbor.getSize (); ++i)
      if (&in == neighbor(i).getPtr ()) return i;
   return -1;
}

SxArray<SxMatrix<Complex16> > 
SxDomain::getDomainBase (const SxArray<SxPtr<SxDomain> > &domains,
                         int nE, double dERel, int pass)
{
   CLOCK (GetDomainBase);
   cout << "Pass " << pass << endl;
   SX_CHECK (domains(0).getPtr () == this);
   int nBase = 0;
   for (int jD = 0; jD < domains.getSize (); ++jD)  {
      nBase += domains(jD)->getSize ();
   }
   cout << "embedding size = " << nBase << endl;

   // --- setup local range Hloc and Sloc
   SxMatrix<Complex16> Hloc(nBase, nBase), Sloc(nBase, nBase);
   Hloc.set (0.);
   Sloc.set (0.);
   for (int i = 0; i < nBase; ++i) Sloc(i,i) = 1.;
   for (int jD = 0, jBase = 0; jD < domains.getSize (); ++jD)  {
      SxDomain &domain1 = *domains(jD);
      int nBase1 = domain1.getSize ();
      for (int kD = 0, kBase = 0; kD < domains.getSize (); ++kD)  {
         SxDomain &domain2 = *domains(kD);
         int nBase2 = domain2.getSize ();
         if (jD == kD)  {
            setBlock (&Hloc, jBase, jBase, domain1.H);
         } else {
            int iNeighbor = domain1.findNeighbor (domain2);
            if (iNeighbor >= 0)  {
               setBlock (&Hloc, jBase, kBase, domain1.neighborH(iNeighbor));
               setBlock (&Sloc, jBase, kBase, domain1.neighborS(iNeighbor));
            }
         }
         kBase += nBase2;
      }
      jBase += nBase1;
   }
   //cout << "hermite S: " << (Sloc - Sloc.adjoint ()).normSqr () << endl;
   //cout << "hermite H: " << (Hloc - Hloc.adjoint ()).normSqr () << endl;

   // --- solve local range eigenproblem
   START_TIMER (GetDomDiag);
   SxMatrix<Complex16> L = getOrtho (Sloc);
   Hloc = L.adjoint () ^ Hloc ^ L;
   SxMatrix<Complex16>::Eigensystem eig = Hloc.eigensystem ();
   //cout << eig.vals.real () << endl;
   STOP_TIMER (GetDomDiag);

   int nCentral = getSize ();
   SxIdx central (0, nCentral - 1);
   SxMatrix<Complex16> Lcentral = getRowBlock (L, 0, nCentral);

   // residues in child space
   bool baseLevel = child(0)->globalId.getSize () > 0;
   SxArray<SxArray<SxList<SxVector<Complex16> > > > residueC;
   if (!baseLevel) {
      residueC.resize (domains.getSize ());
      for (int jD = 0; jD < domains.getSize (); ++jD)
         residueC(jD).resize (domains(jD)->child.getSize ());
   }

   // --- get bases for the split energy intervals
   double dE = (energyDomain->eMax - energyDomain->eMin) / nE;
   SxArray<SxMatrix<Complex16> > res(nE);
   for (int iE = 0; iE < nE; ++iE)  {
      double eMin1 = energyDomain->getChild(iE)->eMin,
             eMax1 = energyDomain->getChild(iE)->eMax;
      int nNew = 0;
      for (int iEps = 0; iEps < eig.vals.getSize (); ++iEps)  {
         double eps = eig.vals(iEps).re;
         // if (eps >= eMin1 && eps <= eMax1) nNew++;
         double f = 1.;
         if (eps < eMin1)
            f = 1./(1.+exp((eMin1 - eps)/(dE*dERel)));
         else if (eps > eMax1)
            f = 1./(1.+exp((eps - eMax1)/(dE*dERel)));
         if (f > 1e-4)  nNew++;

         // --- evaluate accuracy of lower-level domain base
         //     by computing the lower-level residues
         if (!baseLevel && f > 0.9)  {
            CLOCK (GetDomRes);
            SxDomainVec psi (L ^ eig.vecs.colRef (iEps), domains);
            psi = psi.down ().down ();
            SxDomainVec R = psi.applyH (true);
            R.plus_assign_ax (-eps, psi.applyS (true));
            double normR = sqrt(dot(R, R).re);
            cout << "res(" << eps << ")= " << normR << endl;

            // --- collect child residues
            if (normR > 1e-12)  {
               for (int id = 0; id < R.domains.getSize (); ++id)  {
                  bool found = false;
                  for (int jD = 0; jD < domains.getSize () && !found; ++jD)  {
                     for (int iChild = 0; 
                          iChild < domains(jD)->child.getSize ();
                          ++iChild)  
                     {
                        SX_CHECK_VAR (domains(jD)->child(iChild)->child.getSize () == 1,
                                      domains(jD)->child(iChild)->child.getSize ());
                        SxDomain &cDomain = *domains(jD)->child(iChild)->child(0);
                        if (R.domains(id) == &cDomain)  {
                           residueC(jD)(iChild).append (R.psiPart(id));
                           found = true;
                           break;
                        }
                     }
                  }
               }
            }
         }
      }
      cout << eMin1 << "..." << eMax1 << ": " << nNew; 
      if (nNew == 0) { cout << endl; continue; }

      SxMatrix<Complex16> newBas (nCentral, nNew);
      //newBas.set (0.);
      for (int iEps = 0, iB = 0; iEps < eig.vals.getSize (); ++iEps)  {
         double eps = eig.vals(iEps).re;
         double f = 1.;
         if (eps < eMin1)
            f = 1./(1.+exp((eMin1 - eps)/(dE*dERel)));
         else if (eps > eMax1)
            f = 1./(1.+exp((eps - eMax1)/(dE*dERel)));
         // if (eps >= eMin1 && eps <= eMax1)
         if (f > 1e-4)
            newBas.colRef(iB++) << ((Lcentral ^ eig.vecs.colRef(iEps)) * f);
      }

      // remove singular part (note: central basis is orthonormal)
      if (nNew > 1)
         newBas = newBas ^ getOrtho (newBas.adjoint () ^ newBas);

      cout << " ==>" << newBas.nCols () << endl;
      /*
      if (newBas.nCols () == nCentral)  {
         newBas.set (0.);
         for (int i = 0; i < nCentral; ++i) newBas(i,i).re = 1.;
      }
      */

      res(iE) = newBas;
   }
   bool secondPass = false;
   for (int jD = 0; jD < residueC.getSize (); ++jD)  {
      for (int iChild = 0; iChild < domains(jD)->child.getSize (); ++iChild)  {
         SxDomain &cDomain = *domains(jD)->child(iChild);
         int nRes = (int)residueC(jD)(iChild).getSize ();
         int nBaseC = cDomain.getSize ();
         int nBaseCC = cDomain.child(0)->getSize ();
         if (nRes == 0) continue;
         SxMatrix<Complex16> residues (nBaseCC, nRes);
         for (int ir = 0; ir < nRes; ++ir)
            residues.colRef(ir) << residueC(jD)(iChild)(ir);
         // orthonormalize residues
         SxMatrix<Complex16> R2 = residues.overlap (residues);
         SxMatrix<Complex16>::Eigensystem R2eig = R2.eigensystem (); 
         cout << "domain " << jD 
              << ", child " << iChild << " (" 
              << nBaseC << "/"
              << cDomain.child(0)->getSize () << " bases): "
              << R2eig.vals.real () << endl;

         SxMatrix<Complex16> Sb = cDomain.U(0).overlap (residues);
         cout << "child residue in basis: " << Sb.normSqr () << endl;

         if (pass == 0) continue;
         if (jD != 0) continue;

         int nNew = 0;
         for (int i = nRes - 1; i; --i, nNew++)  {
            if (R2eig.vals(i).re < 1e-20) break;
         }
         if (nNew > 0)  {
            secondPass = true;
            SxMatrix<Complex16> &childU = cDomain.U(0);
            childU.reshape (nBaseCC * nBaseC);
            childU.resize (nBaseCC * (nBaseC + nNew), true);
            childU.reshape (nBaseCC, nBaseC + nNew);
            for (int i = 0; i < nNew; ++i)  {
               int iRes = nRes - nNew + i;
               childU.colRef (nBaseC + i) 
                  << (residues ^ R2eig.vecs.colRef (iRes))
                     / sqrt(R2eig.vals(iRes).re);
            }
            childU = childU ^ getOrtho (childU.overlap (childU));
            cout << "pass " << pass << ": " << nBaseC
                 << "->" << childU.nCols ()
                 << "/" << nBaseCC << endl;
            if (childU.nCols () == nBaseCC)  {
               childU.set (0.);
               for (int i = 0; i < nBaseCC; ++i) childU(i,i).re = 1.;
            }

            SX_CHECK_VARS (childU.nRows () == nBaseCC,
                           childU.nRows (), nBaseCC);
            // POINTER check
            //cout << "getDomainBase residue pass " << &cDomain
            //     << " child " << cDomain.child (0).getPtr () << ": "
            //     << nBaseCC << endl;

            cDomain.H = childU.adjoint () ^ cDomain.child(0)->H ^ childU;
            cDomain.updateNeighborHS (true);
         }
      }
      if (secondPass)  {
         domains(jD)->recompute ();
         domains(jD)->updateNeighborHS (true);
      }
   }
   if (secondPass && pass > 0) 
      return getDomainBase (domains, nE, dERel, pass - 1);
   return res;
}

SxArray<SxDomains> getDomainBase (const SxDomains &in,
                                  double dERel = 0.5, int depth = 1)
{
   int nPass = 1;
   int nE = int(in.energyDomain->child.getSize ());

   SxArray<SxDomains> res(nE);
   int nD = int(in.getSize ());
   for (int iE = 0; iE < nE; ++iE)  {
      res(iE).resize (nD);
   }
   for (int iD = 0; iD < nD; ++iD)  {
      // --- create new domains
      for (int iE = 0; iE < nE; ++iE)  {
         res(iE)(iD) = SxPtr<SxDomain>::create (in.energyDomain->getChild(iE));
      }
      if (in(iD)->getSize () == 0) continue;

      // --- collect local range domains
      SxUniqueList<SxPtr<SxDomain> > domains;
      domains << in(iD);
      SxList<SxPtr<SxDomain> >::Iterator start, end, it;
      start = domains.begin ();
      end = domains.begin ();
      for (int iLevel = 0; iLevel < depth; ++iLevel)  {
         cout << " iLevel = " << iLevel << ": " << domains.getSize () << " ";
         for (it = start; /* done inside */; it++)  {
            cout << &(*it) << ": " << (*it)->neighbor.getSize ()
                 << " neighbors ";
            for (int iNeighbor = 0; iNeighbor < (*it)->neighbor.getSize ();
                  ++iNeighbor)
               domains << (*it)->neighbor(iNeighbor);
            if (it == end) break;
         }
         cout << "=>" << domains.getSize () << endl;
         start = end;
         // update start iterator
         if (start == domains.begin ()) { start = domains.begin (); }
         else                           { start--; start++; } 
         start++;
         if (!start) break;
         end = domains.fromLast ();
      }
      
      SxArray<SxMatrix<Complex16> > newBas
         = in(iD)->getDomainBase (domains, nE, dERel, nPass);
      for (int iE = 0; iE < nE; ++iE)  {
         // --- set on-site stuff on new domain
         SxDomain &newDomain = *res(iE)(iD);
         newDomain.child.resize (1);
         newDomain.child(0) = in(iD);
         if (newBas(iE).getSize () > 0)  {
            newDomain.U.resize (1);
            newDomain.U(0)     = newBas(iE);
            newDomain.H        = newBas(iE).adjoint () ^ in(iD)->H ^ newBas(iE);
            SX_CHECK_VARS (newDomain.U(0).nRows () == newDomain.child(0)->getSize (),
                  newDomain.U(0).nRows (), newDomain.child(0)->getSize ());
            // POINTER check
            //cout << "getDomainBase " << &newDomain << " child "
            //     << newDomain.child (0).getPtr () << ": "
            //     << newDomain.U(0).nRows () << endl;
         }
      }
   }

   for (int iE = 0; iE < nE; ++iE)  {
      res(iE).updateNeighbors ();
   }
   return res;
}


void solveDomain (const SxDomains &in, int nPerD, int ovlp, int level,
                  const SxMatrix<Complex16> &globalH, int depth)
{
   cout << "Level " << level << endl;
   // check Hamiltonian
   //for (int iE = 0; iE < in.getSize (); ++iE) in(iE)->checkHam (globalH);
   SX_CHECK_VAR (level >= 0, level);
   int nTot = in.getTotalSize ();
   cout << "Level " << level
        << " for " << in(0)->energyDomain->eMin << "..." 
        << in(0)->energyDomain->eMax
        << ": " << nTot << " basis functions"
        << endl;
   if (level == 0)  {
      in.completeDiag (globalH);
      return;
   }
   SxDomains domainsPreL2 = raiseDomainLevel (in, nPerD, ovlp);

   int nE = nPerD;
   if (nTot / in.getSize () > 15)  {
      nE *= 2;
   }

   domainsPreL2.energyDomain->split (nE);
   SxArray<SxDomains> domainsL2 
      = getDomainBase (domainsPreL2, 0.1, depth);

   for (int iE = 0; iE < domainsL2.getSize (); ++iE)  {
      solveDomain (domainsL2(iE), nPerD, ovlp, level - 1, globalH, depth);
   }
}

SxPtr<SxDomainTreeLeaf>
setupSpatialDomainTree (SxDomains &domains, int nLevel, int nPerD)
{
   int nD = int(domains.getSize ());
   SxArray<SxPtr<SxDomainTreeLeaf> > leafs(nD);
   
   // --- set up the lowest level
   for (int iD = 0; iD < nD; ++iD)  {
      leafs(iD) = SxPtr<SxDomainTreeLeaf>::create (nLevel - 1, iD);
      leafs(iD)->resize (nLevel - 1, 1);
      leafs(iD)->getDomain (nLevel - 1, 0) = domains(iD);
      domains(iD)->spatialDomain = leafs(iD);
   }
   // --- set up neighbors
   for (int iD = 0; iD < nD; ++iD)  {
      ssize_t nNeighbor = domains(iD)->neighbor.getSize ();
      leafs(iD)->neighbor.resize (nNeighbor);
      for (int in = 0; in < nNeighbor; ++in)  {
         leafs(iD)->neighbor(in) = domains(iD)->neighbor(in)->spatialDomain;
      }
   }
   
   for (int iLevel = nLevel - 2; iLevel; --iLevel)  {
      if (nD % nPerD != 0)  {
         cout << "Cannot generate level " << iLevel
              << "nD=" << nD << " % nPerD=" << nPerD << "="
              << (nD % nPerD) << endl;
         SX_EXIT;
      }
      int nDLevel = nD / nPerD;
      SxArray<SxPtr<SxDomainTreeLeaf> > newLeafs(nDLevel);
      int nChild = nPerD;
      for (int iDOld = 0, iD = 0; iDOld < nD; iDOld += nPerD, ++iD)  {
         newLeafs(iD) = SxPtr<SxDomainTreeLeaf>::create (iLevel, iD);
         newLeafs(iD)->child.resize (nChild);
         for (int iChild = 0; iChild < nChild; ++iChild)  {
            int jD = idxN(iDOld + iChild, nD);
            newLeafs(iD)->child(iChild) = leafs(jD);
            leafs(jD)->parent = newLeafs(iD);
         }
      }
      // --- update neighbors
      for (int iD = 0; iD < nDLevel; ++iD)  {
         SxUniqueList<SxPtr<SxDomainTreeLeaf> > list;
         SxDomainTreeLeaf &leaf = *leafs(iD);
         for (int ic = 0; ic < leaf.child.getSize (); ++ic)  {
            for (int in = 0; in < leaf.child(ic)->neighbor.getSize (); ++in)  {
               SxPtr<SxDomainTreeLeaf> &neighbor 
                  = leaf.child(ic)->neighbor(in)->parent;
               if (neighbor.getPtr () == &leaf) continue;
               list << neighbor;
            }
         }
         leaf.neighbor = list;
      }
      
      // --- proceed to next level
      nD = nDLevel;
      leafs = newLeafs;
   }

   // --- set up top level
   SxPtr<SxDomainTreeLeaf> res 
      = SxPtr<SxDomainTreeLeaf>::create (0, 0);
   res->child = leafs;
   for (int i = 0; i < nD; ++i)  {
      leafs(i)->parent = res;
   }
   return res;
}

int main (int argc, char** argv)
{
   SxCLI cli(argc, argv);
   int N = cli.option ("-N","size", "size").toInt (16,8);
   int nPerD = cli.option ("--nPerD", "size", "domain reduction factor")
               .toInt (2);
   int ovlp = cli.option ("--ovlp", "size", "domain overlap").toInt (0);
   int depth = cli.option ("--depth", "size", "domain embedding").toInt (1);
   int nLevels = cli.option ("-l|--nLevels", "size", "number of levels")
                 .toInt (1);
   int nBasic = cli.option ("--nBasic", "size", "0-level domain size")
                .toInt (4);
   int basicOvlp = cli.option ("--basicOvlp", "size", "0-level domain overlap")
                .toInt (1);
   cli.finalize ();

   SxMatrix<Complex16> H(N,N);
   H.set (0.);
   // kinetic
   for (int i = 0; i < N; ++i)  {
      H(i,i) = 1.;
      H(idxN(i-1,N),i)  = -0.5;
      H(idxN(i+1,N), i) = -0.5;
   }
   // --- potential
   for (int i = 0; i < N / 2; ++i)  {
      H(i,i) += 2. - i / double(N);
   }
   SxMatrix<Complex16>::Eigensystem eig;
   {
      CLOCK (FullDiag);
      eig = H.eigensystem ();
   }
   //cout << eig.vals.real () << endl;
   for (int i = 0; i < N; ++i)  {
      sxprintf ("EXACT: %.10f\n", eig.vals(i).re);
   }
   /*
   for (int i = 0; i < N; ++i)  {
      cout << "# e=" << eig.vals(i).re << endl;
      cout << eig.vecs.colRef (i).absSqr () << endl;
   }
   */

   int nD = N / nBasic;

   SxArray<SxMatrix<Complex16> > phi(nD);
   SxArray<SxVector<Double> > eps(nD);

   // --- set up lowest domain level (SxDomain version)
   SxDomains domains1;
   domains1.resize (nD);
   for (int iD = 0; iD < nD; ++iD)  {
      domains1(iD) = SxPtr<SxDomain>::create (SxPtr<SxEnergyDomain> ());
   }
   int nBasicBase = nBasic + 2 * basicOvlp;
   for (int iD = 0; iD < nD; ++iD)  {
      int off = nBasic * iD;
      domains1(iD)->globalId.resize (nBasicBase);

      SxMatrix<Complex16> Hloc(nBasicBase,nBasicBase);
      for (int i = 0; i < nBasicBase; ++i)  {
         int ii = idxN(off - basicOvlp + i, N);
         domains1(iD)->globalId(i) = ii;
         for (int j = 0; j < nBasicBase; ++j)  {
            int jj = idxN(off - basicOvlp + j, N);
            Hloc(i,j) = H(ii,jj);
         }
      }
      VALIDATE_VECTOR (Hloc);
      domains1(iD)->H = Hloc;
      domains1(iD)->neighbor.resize (2);
      domains1(iD)->neighbor(0) = domains1(idxN(iD - 1, nD));
      domains1(iD)->neighbor(1) = domains1(idxN(iD + 1, nD));

      domains1(iD)->neighborH.resize (2);
      domains1(iD)->neighborS.resize (2);
      for (int iN = 0; iN < 2; ++iN)  {
         domains1(iD)->neighborS (iN).reformat (nBasicBase,nBasicBase);
         domains1(iD)->neighborH (iN).reformat (nBasicBase,nBasicBase);
         domains1(iD)->neighborS (iN).set (0.);
      }
      // --- interdomain overlap
      for (int k = 0; k < 2 * basicOvlp; ++k)  {
         domains1(iD)->neighborS (0)(k,nBasic+k) = 1.;
         domains1(iD)->neighborS (1)(nBasic+k,k) = 1.;
      }
      // --- interdomain Hamiltonian
      for (int i = 0; i < nBasicBase; ++i)  {
         int ii = idxN(off - basicOvlp + i, N);
         for (int j = 0; j < nBasicBase; ++j)  {
            int jj = idxN(off - nBasic - basicOvlp + j, N);
            domains1(iD)->neighborH (0)(i,j) = H(ii, jj);
            jj = idxN(off + nBasic - basicOvlp + j, N);
            domains1(iD)->neighborH (1)(i,j) = H(ii, jj);
         }
      }
      if (nD == 2)  {
         // neighbor 0 and neighbor 1 are the same
         domains1(iD)->neighborS(0) += domains1(iD)->neighborS(1);
         domains1(iD)->neighborS.resize (1, true);
         domains1(iD)->neighborH.resize (1, true);
         domains1(iD)->neighbor.resize (1, true);
      }

      /*
      SxMatrix<Complex16>::Eigensystem eigLoc = Hloc.eigensystem ();
      cout << "Domain " << off << ".." << off + nBasic - 1 << endl;
      cout << eigLoc.vals << endl;
      */
   }
   solveDomain (domains1, nPerD, ovlp, nLevels, H, depth);

   printTiming ();
}

