// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxNeighbors.h>
#include <SxTime.h>

int main (int argc, char **argv)
{
   // --- command line parsing
   SxCLI cli(argc, argv);
   int N = cli.option ("-N", "number", "N for NxN supercell, N must multiple of 3").toInt (12);
   if (N % 3 != 0)  {
      cout << "N must be multiple of 3!" << endl;
      cli.setError ();
   }

   // --- surface
   int surfaceGroup = cli.newGroup ("2D");
   double prefEnergy = cli.option ("-E", "site energy", 
                       "sqrt(3x3) lattice site energy").toDouble (0.);

   // --- bulk
   /*int bulkGroup = */cli.newGroup ("3D");
   cli.excludeGroup (surfaceGroup);
   bool bulk = cli.option ("--bulk", "request bulk model").toBool ();
   int Nz = cli.option ("-M", "number", "M for NxNxM supercell").toInt (6);
   if (!bulk) Nz = 1;

   // --- general 
   int constMu = cli.newGroup ("const. mu");
   double mu = cli.option ("-mu", "mu", "chemical potential").toDouble (); 

   cli.setGroup (SxCLI::generalGroup);
   double xIn = cli.option ("-x", "fraction", "fraction of In").toDouble (1./3., 0., 1.); 
   double kB = 8.6173323e-02;// in meV/K
   double kT = cli.option ("-T", "temperature", "temperature")
              .toDouble (700) * kB; 
   int bSize = cli.option ("-A", "averaging range", 
                           "running averages over N points")
                  .toInt (200);
   ssize_t nIt = cli.option ("--it", "iterations", "number of iterations")
                 .toLong (10000);
   ssize_t eq = cli.option ("--eq", "iterations", 
                            "number of equilibration steps")
                 .toLong (10000);
   ssize_t print = cli.option ("-p", "print steps", 
                               "which steps should be printed").toInt (10);
   int anneal = cli.option ("--anneal", "try simulated annealing")
                 .toInt (0);
   int paramGroup = cli.newGroup ("Set params");
   SxList<double> params = cli.option ("--param", "parameter list",
                                       "parameter list").toDoubleList ();
   cli.finalize ();

   // --- setup a NxN hexagonal supercell
   SxCell prim = SxMatrix3<Double>(1,0,0, -0.5, sqrt(3.)/2.,0, 0, 0, PI).transpose ();
   SxAtomicStructure primStr(SxList<Coord> () << Coord(0,0,0));
   primStr.cell = prim;
   SxPtr<SxAtomInfo> aInfo = aInfo.create (1);
   aInfo->nAtoms(0) = 1;
   aInfo->setupOffset ();
   primStr.replaceInfo (aInfo);
   if (bulk)  {
      primStr.startCreation ();
      primStr.addAtom (prim.relToCar (Coord(1./3., -1./3., 0.5)));
      primStr.endCreation ();
   }

   SxAtomicStructure str = primStr.repeat (N, N, Nz);
   int nAtoms = int(str.getNAtoms ());

   /*
   {
      // --- sort atom by grid cells in order to enhance occ locality
      // experience: not important
      SxGrid grid(str, 100);
      aInfo = SxAtomInfo::derive (str.atomInfo);
      aInfo->parentMap.resize (nAtoms);
      for (int iGrid = 0, ia = 0; iGrid < grid.nAtPerGridCell.getSize (); ++iGrid)  {
         SxVector<Int>::Iterator it = grid.getIdxIt (iGrid),
                                 end = grid.computeItEnd (iGrid);
         for (; it != end; ++it, ++ia)
            aInfo->parentMap(ia) = *it;
      }
      aInfo->nAtoms(0) = nAtoms;
      aInfo->setupOffset ();
      str = SxAtomicStructure (aInfo, str);
   }
   */

   //str.fprint (stdout);

   // --- setup neighbor lists
   SxGrid grid(str, 100);
   SxNeighbors nn;

   int nType = bulk ? 8 : 3;
   SxArray2<int> nList;
   SxVector<Double> dist2(nType), energy(nType);
   SxArray<int> nNeighbor(nType);
   if (bulk)  {
      dist2(0)  = prim.relToCar (Coord(1,0,0)).normSqr (); // 0A
      dist2(1)  = prim.relToCar (Coord(2,1,0)).normSqr (); // 0B
      dist2(2)  = prim.relToCar (Coord(2,0,0)).normSqr (); // 0C
      dist2(3)  = prim.relToCar (Coord(1./3., -1./3., 0.5)).normSqr (); // 1A
      dist2(4)  = prim.relToCar (Coord(4./3., 2./3., 0.5)).normSqr (); // 1B
      dist2(5)  = prim.relToCar (Coord(1./3., 8./3., 0.5)).normSqr (); // 1C
      dist2(6)  = prim.relToCar (Coord(0,0,1)).normSqr (); // 2A
      dist2(7)  = prim.relToCar (Coord(1,0,1)).normSqr (); // 2B
      energy = SxList<double> ()
         <<  45.4  // 0A 
         <<  -17.5  // 0B 
         <<  31.5  // 0C 
         <<  49.1  // 1A 
         <<  -29.2  // 1B 
         <<  -13.0  // 1C 
         <<  -62.2  // 2A 
         <<  15.9; // 2B 
   } else {
      dist2(0)  = prim.basis(0).normSqr (); // 0A
      dist2(1)  = sqr(2) * dist2(0); // 0C
      dist2(2)  = prim.relToCar (Coord(2,1,0)).normSqr (); // 0B
      //energy(0) = 77; // 0A Sangheon
      //energy(1) = 34; // 0C Sangheon
      //energy(0) = 144; // 0A (from Liverios DFT)
      //energy(1) = 22; // 0C (from Liverios DFT)
      //energy(2) = -8; // 0B
      if (cli.groupAvailable(paramGroup))  {
         if (params.getSize () != 3)  {
            cout << "Need exactly 3 params (0A, 0B, 0C)!" << endl;
            SX_QUIT;
         }
         energy(0) = params(0); // 0A
         energy(1) = params(2); // 0C
         energy(2) = params(1); // 0B
      } else {
         energy(0) = 143; // 0A (from Liverios DFT)
         energy(1) = 39; // 0C (from Liverios DFT)
         energy(2) = 9; // 0B
      }
   }
     
   double d2Max = 1.01 * dist2.maxval ();
   SX_LOOP(ia)  {
      nn.compute (grid, str, str.getAtom(ia), sqrt(d2Max), 
                  SxNeighbors::StoreIdx | SxNeighbors::StoreDistSqr);
      if (ia == 0)  {
         int nAll = 0;
         SX_LOOP(iType)  {
            int n = 0;
            SX_LOOP(in)
               if (fabs(nn.distSqr (in) - dist2(iType)) < 1e-10) n++;
            nNeighbor(iType) = n;
            nAll += n;
            cout << "# n(type = " << (iType+1) << ") = " << n 
                 << " dE = " << energy(iType)
                 << " d = " << sqrt(dist2(iType))
                 << endl;
         }
         nList.reformat (nAll, nAtoms);
      }
      int n = 0;
      SX_LOOP(iType)  {
         SX_LOOP(in)  {
            if (fabs(nn.distSqr (in) - dist2(iType)) < 1e-10)  {
               nList(n++,ia) = nn.idx(in);
            }
         }
      }
   }

   // --- surface lattice interaction
   SxCell supl(prim ^ SxMatrix3<Double> ( 2,1,0,
                                         1,2,0,
                                         0,0,1));
   SxArray<bool> prefSite(nAtoms);
   if (!bulk)  {
      SX_LOOP(ia)
         prefSite(ia) = (supl.getMapped (str.ref(ia), SxCell::Origin))
                        .normSqr () < 1e-6;
   }

   // --- prepare simulation
   SxArray<char> occ(nAtoms);
   srand48 ((long)SxTime::getRealTime ());

   occ.set (0);
   ssize_t nIn = lround(xIn * nAtoms);
   // initialize
   for (int iIn = 0; iIn < nIn; ++iIn)  {
      int j;
      do {
         j = int(nAtoms * drand48 ());
      } while (occ(j));
      occ(j) = 1;
   }

   SxArray<long> pairCount(nType);
   pairCount.set (0);

   // --- calculate initial energy and pair count
   double E = 0.;
   SX_LOOP(ia)  {
      if (!occ(ia)) continue;
      int n = nNeighbor(0), iType = 0;
      SX_LOOP(in)  {
         if (!n) n=nNeighbor(++iType);
         if (occ(nList(in,ia))) {
            E += 0.5 * energy(iType);
            pairCount(iType)++;
         }
         n--;
      }
      if (!bulk && prefSite(ia)) E += prefEnergy;
   }
   //cout << "E = " << E/nAtoms << endl;
   //E = 0.;
   //SX_LOOP(iType)
   //   E += 0.5 * pairCount(iType) * energy(iType);
   //cout << "# " << E/nAtoms << endl;

   // --- Monte-Carlo simulations
   SxArray<double> bufE(bSize), bufE2(bSize);
   bufE.set (E);
   bufE2.set (E*E);
   int ib = 0;
   double sumE = E * bSize, sumE2 = E * E * bSize;
   double tsumE = 0., tsumE2 = 0.;
   double tsumx = 0., tsumx2 = 0.;
   SxArray<long> dPairCount(nType), pairCountSum(nType);
   pairCountSum.set (0);
   bool muMode = cli.groupAvailable (constMu);

   cout << "# it 1:E(it) 2:<E>_{window=last " << bSize 
        << " steps} 3:<(E-<E>)^2>_window 4: <E>_{it>0} "
        << " 5:<(E-<E>)^2>_{it>0}"
        << (muMode ? " 6: x(it) 7:<x>_{it>0} 8:sqrt(<(x-<x>)^2>_{it>0})" : "")
        << endl;
   for (ssize_t it = -eq; it < nIt; ++it)  {

      // --- choose an atom pair
      int ia, ja;
      if (muMode)  {
         ia = int( drand48 () * nAtoms); ja = -1;
      } else {
         do {
            ssize_t ij = ssize_t( drand48 () * nAtoms * nAtoms);
            ia = int(ij / nAtoms);
            ja = int(ij - ia * nAtoms);
         } while (occ(ia) == occ(ja));
      }

      // --- get change in energy if ia is Ga and ja is In
      double dE = 0;
      int n = nNeighbor(0), iType = 0, nT = 0;
      int jType = -1;
      if (muMode)  {
         SX_LOOP(in)  {
            if (!n) n=nNeighbor(++iType);
            nT += occ(nList(in,ia));
            if (--n == 0) {
               dE += nT * energy(iType);
               dPairCount(iType) = nT;
               nT = 0;
            }
         }
      } else {
         SX_LOOP(in)  {
            if (!n) n=nNeighbor(++iType);
            nT += occ(nList(in,ia));
            nT -= occ(nList(in,ja));
            if (nList(in,ia) == ja) jType = iType;
            if (--n == 0) {
               dE += nT * energy(iType);
               dPairCount(iType) = nT;
               nT = 0;
            }
         }
      }
      if (!bulk)  {
         if (prefSite(ia)) dE += prefEnergy;
         if (!muMode && prefSite(ja)) dE -= prefEnergy;
      }

      double dMu = muMode ? -mu : 0.;
      if (occ(ia)) {
         dE = -dE;
         dMu = -dMu;
         SX_LOOP(i) dPairCount(i) = -dPairCount(i);
      }

      // if ia/ja are neighbors, correct dE
      if (jType >= 0) {
         dE -= energy(jType);
         dPairCount(jType)--;
      }

      double kT_ = kT;
      if (anneal && it < -eq/2)  {
         // simulated annealing
         double M = anneal * kB / kT;
         kT_ *= 1. -(2. * double(it)/ double(eq) + 1.) * (M-1.);
      }

      // Metropolis Monte-Carlo condition
      if ((dE+dMu) < 0 || exp(-(dE+dMu)/kT_) > drand48 ()) {
         // step is accepted
         E += dE;
         SX_LOOP(i) pairCount(i) += 2 * dPairCount(i);
         if (muMode)
            nIn += (1 - 2 * occ(ia));
         else 
            occ(ja) = char(1 - occ(ja));
         occ(ia) = char(1 - occ(ia));
      }

      // remove old data
      sumE -= bufE(ib);
      sumE2 -= bufE2(ib);
      // add new data
      sumE += (bufE(ib) = E);
      sumE2 += (bufE2(ib) = E * E);

      if (++ib == bSize) ib = 0;

      if (it >= 0)  {
         tsumE += E;
         tsumE2 += E * E;
         SX_LOOP(i) pairCountSum(i) += pairCount(i);
         if (muMode)  {
            double xNow = double(nIn)/double(nAtoms);
            tsumx += xNow;
            tsumx2 += xNow*xNow;
         }
      }
      if (it % print == 0)  {
         double s1 = sumE2 - sumE * sumE / bSize;
         if (s1 < 0.) s1 = 0.;
         double ts = tsumE2 - tsumE * tsumE / double(it + 1);
         if (ts < 0.) ts = 0.;
         // output
         cout << it << ' ' << E/nAtoms << ' ' 
              // average energy
              << (sumE/bSize/nAtoms) << ' '
              // standard deviation 
              << sqrt(s1/(bSize - 1))/nAtoms
              << ' ' << (tsumE/double(it+1))/nAtoms << ' '
              << sqrt(ts/double(it+1))/nAtoms;
         if (muMode)  {
            cout << ' ' << double(nIn)/double(nAtoms) << ' '
                 << (tsumx/double(it+1)) << ' '
                 << sqrt((tsumx2 - tsumx * tsumx/double(it+1))/double(it+1)); 
         }
         cout << endl;
      }
   }

   cout << "# " << E/nAtoms << endl;
   E = 0.;
   SX_LOOP(ia)  {
      if (!occ(ia)) continue;
      int n = nNeighbor(0), iType = 0;
      SX_LOOP(in)  {
         if (!n) n=nNeighbor(++iType);
         if (occ(nList(in,ia))) E += 0.5 * energy(iType);
         n--;
      }
      if (!bulk && prefSite(ia)) E += prefEnergy;
   }
   cout << "# " << E/nAtoms << endl;
   E = 0.;
   SX_LOOP(iType)
      E += 0.5 * double(pairCount(iType)) * energy(iType);
   cout << "# " << E/nAtoms << endl;

   cout << "# T <E/atom> <heat capacity> "
        << (muMode ? "<mu> <x> <(x-<x>)^2> " : "")
        << "<excess occupation>..."
        << endl;
   cout << "# " << kT/kB << ' ' << (tsumE/double(nIt)/nAtoms) 
        << ' ' 
        << (tsumE2 - tsumE * tsumE/double(nIt))/double(nIt)
           / (kT * kT / kB)/nAtoms;
   if (muMode)  {
      cout << ' ' << mu << ' ' << (tsumx/double(nIt)) << ' '
           << sqrt((tsumx2 - tsumx * tsumx/double(nIt))/double(nIt)); 
      xIn = tsumx/double(nIt);
      nIn = ssize_t(nAtoms * xIn);
   }
   // excess occupations
   SX_LOOP(iType)
      cout << ' ' << (double(pairCountSum(iType))
                      / double(nNeighbor(iType) * nIn)
                      / double(nIt)
                      - xIn);
   cout << endl;
}
