// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxConfig.h>
#include <SxBinIO.h>
#include <SxParser.h>
#include <SxAtomicStructure.h>
#include <SxGrid.h>
#include <SxRotation.h>

int main (int argc, char **argv)
{
   // --- init S/PHI/nX Utilities
   initSPHInXMath ();

   // --- parse command line
   SxCLI cli (argc, argv);
   cli.preUsageMessage = "This add-on calculates the irreducable dof "
                                           "of the provided structure";
   cli.authors = "C. Freysoldt, M.A. Uijttewaal";

   SxString inFile = cli.option ("-i|--input", "input file", 
                                 "supercell input file").toString ("input.sx");
   double epsEqual = cli.option ("-e|--epsEqual", "accuracy structure")
                                 .toDouble (2.e-3);
   
   // --- check cartesian directions by default, override with provided dirs
   SxMatrix3<Double> dispDirs (1., 0., 0., 0., 1., 0., 0., 0., 1.);
   if (cli.option ("--dirs","matrix", "9 (comma separated) numbers representing"
                   " 3 displacement directions").exists ())
      dispDirs = SxMatrix3<Double> (cli.last ().toDoubleList ()).transpose ();

   bool debug = cli.option ("--verbose","print some debugging info").toBool ();
   cli.finalize ();

   if (fabs(dispDirs.determinant ()) < 1e-3)  {
      cout << "Displacement directions are linear dependent!" << endl;
      SX_QUIT;
   }
   // --- normalize column vectors
   for (int i = 0; i < 3; ++i)  {
      dispDirs.setCol (i, dispDirs.col (i).normalize ());
      cout << "Displacement " << (i+1) << " along " << dispDirs.col (i) << endl;
   }
   
   // --- read input
   SxParser parser;
   SxConstPtr<SxSymbolTable> tree;
   tree = parser.read (inFile);
   
   // init struc
   SxAtomicStructure struc (&*tree);
   struc.epsEqual = epsEqual;

   SX_EXIT; // the spin stuff does not work
   SxAtomicStructure spins (&*tree, "spin");
   spins.epsEqual = epsEqual;
   spins.replaceInfo (struc.atomInfo);
      
   // get primitive cell with primSyms 
 //cout << struc.getPrimCell (true) << endl;
   SX_EXIT; // the next line does not work
   //SxCell primCell = struc.getSpinSyms (spins, debug);
SxCell primCell; SX_EXIT; // make things compile
   if (debug) cout << struc.cell.volume / primCell.volume 
                   << "#primCell: " << primCell << endl;

   // get inequivalent atoms (in primCell & with supCellComp primSyms)
   SxAtomicStructure strucRed = struc.getInequivalentAtoms (primCell);

   /* --- irredDof are the inequivalent dir's for the inequivalent atoms
     The inequivalent dir's are determined as follows:  
    */
   int nSym = primCell.symGroupPtr->getSize (), 
       nAtom = strucRed.getNAtoms (), found, iBas;
   if (debug) cout << "nAtom" << nAtom << " nSym" << nSym << endl;
   SxCell rot1, rot2; 
   SxVector3<Double> axis1, axis2;
   SxMatrix3<Double> dispM;
   Coord basVec, distVec;
   SxGrid grid (struc, 10);
   SxSymOp symOp;
   cout << "---output IRREDucable Degrees of Freedom----" << endl;
   for (int iAtom = 0; iAtom < nAtom; iAtom++)  {
      /* --use max 2 syms that do not map this atom, are not each inverse or
           the inverse or identity themselves.*/
      cout << "Atom #" << struc.find (strucRed(iAtom), grid) 
           << " coord " << strucRed(iAtom) << endl;
      found = 0;
      for (int iSym = 0; iSym < nSym; iSym++)  {
         symOp = primCell.symGroupPtr->operator() (iSym);
         distVec = strucRed.cell.getMapped(strucRed(iAtom) 
                           - (symOp ^ strucRed(iAtom)), SxCell::Origin);
         if (distVec.normSqr () < 1e-3 && 
           !(  SxRotation (symOp.rot).getType () == SxSymType::None ||
             SxRotation (symOp.rot).getType () == SxSymType::Identity ||
             SxRotation (symOp.rot).getType () == SxSymType::Inversion))
         {
            if (debug) 
               cout << "mapped atom " << (symOp ^ strucRed(iAtom)) << endl;
            if (found == 0)  {//this is first symmetry
               found++;
               rot1 = symOp.rot;
               axis1 = SxRotation (rot1).getType ().opCoord;
               if (debug) 
                  cout << "axis1 " << axis1 << "symm " << symOp << endl;
            }  else if (symOp.rot != -1 * rot1 && symOp.rot != rot1)  
            { //this is 2nd symm. (not the previous or inverse of previous)
               found++;
               rot2 = symOp.rot;
               axis2 = SxRotation (rot2).getType ().opCoord;
               if (debug) 
                  cout << "axis2 " << axis2 << "symm " << symOp << endl;
               break;// 2 is sufficient
            }
         }
      }
      if (found == 2)  {//only one dir-dof      
         // --first try if the provided vectors and their images span space
         for (iBas = 0; iBas < 3; iBas++)  {
            basVec = dispDirs.col (iBas);
            dispM = SxMatrix3<Double> (basVec, rot1 ^ basVec, rot2 ^ basVec);
            cout << "the provided dir " << basVec;
            if (fabs(dispM.determinant ()) > 1e-4) cout << " suffices" << endl;
            else cout << "does not suffice (with the two symmetries found)" << endl;
         }
         // --then try the (primitive!) basis vectors
         for (iBas = 0; iBas < 3; iBas++)  {
            basVec = primCell.basis (iBas).normalize ();//primVec norm'd!
            dispM = SxMatrix3<Double> (basVec, rot1 ^ basVec, rot2 ^ basVec);
            if (fabs(dispM.determinant ()) > 1e-4) break;
         }
         if (debug) cout << "iBas: " << iBas;
         if (iBas == 3)  {//basVec not usefull, try their sum
            basVec = primCell.basis (0) 
                   + primCell.basis (1) + primCell.basis (2); 
            basVec.normalize ();
            dispM = SxMatrix3<Double> (basVec, rot1 ^ basVec, rot2 ^ basVec);
            if (debug) 
               cout << basVec << (rot1 ^ basVec) << (rot2 ^ basVec) << endl;
            if (fabs(dispM.determinant ()) < 1e-4)  {//should hardly occur?
               if (debug) cout << "sum of basVec not useable!" << endl;
               //in that case use a vector with angle ~pi/4 with axes
               if (fabs (axis1 ^ axis2) < .999)  {//extra axis necess.
                  axis2 = axis1.x (Coord ( //cross product
                        SxList<double> () << 1 << sqrt(5.) << PI).normalize ());
               }
               basVec = (axis1 + axis2 + 2 * axis1.x (axis2)).normalize ();
               dispM = SxMatrix3<Double> (basVec, rot1 ^ basVec, rot2 ^ basVec);
               if (fabs(dispM.determinant ()) < 1e-4)  {
                  cout << "Error in SxIrredDof!!! " << endl
                       << "could not find single dir-dof!" << endl;
                  SX_QUIT;
               }
            }
         } 
         cout << "Very usefull dir-dof is " << basVec << endl;
      }  else if (found == 1)  {/*two dir-dof: take 1 that breaks the symmetry 
                                 + 1 orthogonal to that and its image*/
         // --first try if the provided vectors and their images span space
         for (iBas = 0; iBas < 3; iBas++)  {
            basVec = dispDirs.col (iBas);
            if (fabs(basVec ^ (rot1 ^ basVec)) < .999)  {//broken
               axis2 = basVec.x (rot1 ^ basVec); //orth. to that
               cout << "the provided dir " << basVec << " suffices" << endl
                    << "together with the orth. dir " << axis2 << endl;
            } 
         }
         // --then try the (primitive!) basis vectors
         for (iBas = 0; iBas < 3; iBas++)  {
            basVec = primCell.basis (iBas).normalize ();//primVec norm'd!
            if (fabs (basVec ^ (rot1 ^ basVec)) < .999)  { break; }//broken   
         }
         if (debug) cout << "iBas: " << iBas;
         if (iBas == 3)  {//basVec not usefull, try their sum
            basVec = primCell.basis (0) 
                   + primCell.basis (1) + primCell.basis (2);
            basVec.normalize ();
            if (fabs (basVec ^ (rot1 ^ basVec)) >  .999)  {//should hardly occur
               if (debug) cout << "sum of basVec not usable!" 
                               << basVec << (rot1 ^ basVec) << endl;
               //in that case use a vector with angle ~pi/4 with axes
               axis2 = axis1.x (Coord ( //cross product
                       SxList<double> () << 1 << sqrt(5.) << PI).normalize ());
               basVec = (axis1 + axis2).normalize ();
            }
         }
         // 2nd axis is orthogonal to basVec and its image and it is normalized
         axis2 = basVec.x (rot1 ^ basVec); 
         if (axis2.norm () < 1e-4)  {
            cout << "Error in SxIrredDof!!! " << endl
                 << "could not find double dir-dof!" << endl;
            SX_QUIT;
         }
         cout << "Very usefull dir-dof are " << basVec << endl 
              << " and " << axis2 << endl;
         //the third vector is normalized and at least orthogonal to the 2nd one
      }  else  { //found == 0
         cout << "3 (arbitrary) dir-dof; hint take lattice vectors and/or "
              << "cartesian directions" << endl; 
      }//switch found
   }//iAtom
   cout << "NOTE: this is the minimal set of displacements that span the phonon space." << endl
        << "However, for the sake of accuracy (and speed) it is better to take a larger set" << endl
        << "of more symmetric displacements (e.g. cartesian directions) when possible." << endl
        << "The same goes for equivalent atoms in the primitive cell (related e.g. by an inversion):" << endl
        << "for this case two parallel and two anti-parallel displacements are advised." << endl;
   return 0;
}
