// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxRho.h>
#include <SxCLI.h>
#include <SxProjector.h>

int main (int argc, char **argv)  { 
   
   // --- parse command line arguments
   SxCLI cli(argc, argv);
   cli.preUsageMessage 
      = SxString("Create a jellium potential");
   SxString inFile =
      cli.option ("-i|--input","file","S/PHI/nX input file")
      .toString ("input.sx");

   double eCut = cli.option ("--ecut", "number", "cutoff energy (Ry)")
                 .toDouble ();

   SxString outFile = cli.option ("-o|--output","file"," output file name")
                      .toString   ("jellium.sxb");
   
   int dir = cli.option ("--dir","direction",
                         "potential sampling direction (1,2, or 3)")
             .toInt (3,1,3) - 1;
   int densityGroup = cli.newGroup ("specify density");
   double rhoAvg = cli.option ("--rho", "density", "jellium density")
                   .toDouble ();
   double fraction = cli.option ("-x", "fraction", "how much jellium")
                   .toDouble ();
   int chargeGroup = cli.newGroup ("specify charge");
   cli.excludeGroup (densityGroup);
   double from = cli.option ("--from", "position", 
                             "starting position of jellium slab (bohr)")
               .toDouble (0.);
   double to = cli.option ("--to", "position", 
                             "starting position of jellium slab (bohr)")
               .toDouble ();
   double Q = cli.option ("-Q", "number", "number of electrons").toDouble ();
   if (!cli.groupAvailable (densityGroup)&&!cli.groupAvailable (chargeGroup)) {
      if (!cli.error)
         cout << "Either charge or density must be specified" << endl;
      cli.setError ();
   }

   cli.finalize ();

   SxCell cell;
   try {
      cell = SxCell(&*SxParser ().read (inFile));
   } catch (SxException e) {
      e.print ();
      SX_EXIT;
   }

   SxMesh3D dim = SxGBasis::getMeshSize (eCut, cell);

   cout << dim << endl;

   SxVector3<Int> x;

   RhoR pot(1);
   pot(0).resize (dim.product ());

   int iFrom = int(ceil(from/cell.basis(dir).norm () * dim(dir)))%dim(dir);
   int iTo = int(floor(to/cell.basis(dir).norm () * dim(dir))) % dim(dir);
   if (iTo < 0) iTo += dim(dir);
   if (iFrom < 0) iFrom += dim(dir);

   for (x(0) = 0; x(0) < dim(0); ++x(0))  {
      for (x(1) = 0; x(1) < dim(1); ++x(1))  {
         for (x(2) = 0; x(2) < dim(2); ++x(2))  {
            if (cli.groupAvailable (chargeGroup))  {
               double &val = pot(0)(dim.getMeshIdx (x,SxMesh3D::Positive));
               if (iFrom <= iTo)
                  val = (x(dir) >= iFrom && x(dir) <= iTo) ? 1. : 0.;
               else
                  val = (x(dir) <= iTo || x(dir) >= iFrom) ? 1. : 0.;
            } else {
               pot(0)(dim.getMeshIdx (x,SxMesh3D::Positive)) 
                  = (x(dir) <= fraction * dim(dir)) ? -rhoAvg : 0;
            }
         }
      }
   }
   SxFFT::quickFFTPlanner ();
   SxRBasis R(dim, cell);
   SxGBasis G;
   G.set (dim, cell, SxGBasis::getGCut(eCut));
   pot(0).setBasis (&R);
   if (cli.groupAvailable (chargeGroup))  {
      // rescale
      pot(0) *= -Q / pot(0).sum () / R.dOmega;
   }
   cout << "nElectrons = " << -(pot(0).sum () * R.dOmega) << endl;
   SxIdx noZero(1,G.ng);
   SxDiracVec<Complex16> VG = G | pot(0);
   for (ssize_t ig = 1; ig < VG.getSize (); ++ig)
      VG(ig) *= FOUR_PI / G.g2(ig);
   VG(0) = 0.;
   pot(0) = R | VG;

   
   // --- write down charge densities
   try {
      SxBinIO io;
      io.open      (outFile, SxBinIO::BINARY_WRITE_ONLY);
      io.writeMesh (pot, cell, dim);
      io.setMode   (SxBinIO::WRITE_DATA);
      io.writeMesh (pot, cell, dim);
      io.close();
   } catch (SxException e)  {
      e.print ();
      SX_EXIT;
   }

   return 0;
}
