// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxConfig.h>
#include <SxBinIO.h>
#include <SxFourierInterpol.h>

int main (int argc, char **argv)
{
   // --- init S/PHI/nX Utilities
   initSPHInXMath ();

   // --- parse command line
   SxCLI cli (argc, argv);
   cli.preUsageMessage =
      "This interpolates a line through a mesh.";

   // --- line definition
   SxVector3<Double> start (cli.option ("--from","vector", "starting point")
                            .toList3 ());
   cout << "start = " << start << endl;
   
   SxVector3<Double> end (cli.option ("--to","vector","end point").toList3 ());
   cout << "end = " << end << endl;

   int nPoints = cli.option ("-n|--nPoints", "number", 
                             "number of points on line").toInt (100,2);

   bool relative = cli.option ("--relative", "assume relative coordinates")
                   .toBool ();

   SxString outFile
      = cli.option ("-o","filename", "output file name (screen otherwise)")
        .toString ("");
   
   SxFFT::plannerCLI (cli);

   SxString meshFile = 
      cli.option ("-i","mesh file","netcdf mesh file (one mesh only)")
      .toString ();

   cli.finalize ();

   // --- read input

   SxBinIO io;
   try {
      io.open (meshFile, SxBinIO::BINARY_READ_ONLY);
   } catch (SxException e) {
      e.print ();
      SX_EXIT;
   }

   SxFourierInterpol interpol(io);

   interpol.condense ();

   io.close ();

   if (relative)  {
      interpol.cell.changeToCar(&start);
      interpol.cell.changeToCar(&end);
      cout << "start(abs) = " << start << endl;
      cout << "end(abs)   = " << end << endl;
   }

   SxAtomicStructure rVec;
   rVec.resize (nPoints);

   Coord dr = (end - start) / double (nPoints - 1);
   
   for (int i = 0; i < nPoints; i++)
      rVec.setAtom (i, start + i * dr);

   SxVector<Double> result = interpol.interpolate (rVec);
//   result /= interpol.meshInG.getSize ();

   // --- output
   FILE *file;
   if (outFile.getSize () > 0)  {
      file = fopen(outFile.ascii(),"w");
      if (file == NULL)  {
         cout << "Can't open '" << outFile << "'." << endl;
         SX_EXIT;
      }
   } else {
      file = stdout;
   }
      
   double length = sqrt (dr.absSqr ().sum ());
   for (int i = 0; i < nPoints; i++)
      sxfprintf (file, "%.10f %.10f\n", i * length, result(i));

   if (file != stdout) fclose (file);

}

