// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxPotential.h>
#include <SxNeighbors.h>
#include <SxStructOpt.h>


class SxMorse : public SxPotential
{
   public:
      double energy;
      SxSpeciesData speciesData;

      /** \brief Check whether the provided minimization command is known */
      virtual bool isRegistered (const SxSymbolTable *) const { return true; }

      virtual void execute (const SxSymbolTable *cmd, bool calc=true) { SX_EXIT; }

      virtual SxAtomicStructure getForces (const SxAtomicStructure &tau,
                                           const SxSymbolTable *cmd=NULL);
      virtual PrecEnergy getEnergy () const { return energy; }
      virtual SxSpeciesData getSpeciesData () const { return speciesData; }


      double D;
      double re;
      double a;
      double rcut;

      SxMorse (double DIn, double reIn, double aIn);
};

SxMorse::SxMorse (double DIn, double reIn, double aIn)
   : energy (0.),
     D (DIn), re(reIn), a(aIn)
{
   rcut = -log(1e-10) / a;
}

SxAtomicStructure SxMorse::getForces (const SxAtomicStructure &tau,
                                      const SxSymbolTable *)
{
   SxGrid grid(tau, 10);
   SxNeighbors nn;

   SxAtomicStructure forces = tau.getNewStr ();
   forces.set (Coord(0.,0.,0.));

   int mode = SxNeighbors::StoreRel | SxNeighbors::StoreDistSqr;

   energy = 0.;
   SX_LOOP(ia)  {
      nn.compute (grid, tau, tau.constRef (ia), rcut, mode);
      SX_LOOP(in)  {
         double d = sqrt(nn.distSqr (in));
         double dInv = 1./d;
         double expTerm = exp (-a * (d - re));
         double expTerm1 = 1. - expTerm;
         energy += 0.5 * D * (expTerm1 * expTerm1 - 1.);
         double force = D * expTerm1 * expTerm * a;
         Coord f = (force * dInv) * nn.relPositions.constRef(in);
         forces.ref(ia) += f;
         forces.ref(nn.idx(in)) -= f;
      }
   }
   return forces;
}

int main (int argc, char **argv)
{
   // --- init S/PHI/nX Utilities
   initSPHInXMath ();

   // --- parse command line
   SxCLI cli (argc, argv);

   SxString inFile 
      = cli.option ("-i|--input", "input file", 
                    "take original input file")
        .toString ("input.sx");
   //double maxDist = cli.option("-c", "cutoff", "max distance").toDouble (10.);
   
   cli.finalize ();

   // --- read input
   SxParser parser;
   SxConstPtr<SxSymbolTable> tree;
   tree = parser.read (inFile, "std/structure.std");
   SxAtomicStructure structure (&*tree);

   SxMorse morse(0.7102 / HA2EV, 2.8970 * A2B, 1.6047 / A2B);
   morse.speciesData = SxSpeciesData(SxList<SxString> () << "A");
   structure.atomInfo->meta.update (SxAtomicStructure::Elements,
                                    morse.speciesData.chemName);

   cout << morse.D << endl;
   cout << morse.a << endl;
   cout << morse.re << endl;
   cout << morse.rcut << endl;
   cout << morse.getForces (structure)  << endl;

   {
      SxStructOpt structOpt(&morse, structure);
      SxSymbolTable *main = tree->getGroup("main");
      SxSymbolTable *cmd  = NULL;
      for (cmd  = main->begin();
           cmd != NULL;
           cmd  = cmd->nextSibling())
      {
         if (SxStructOpt::isRegistered (cmd))
            structOpt.execute  (cmd);
      }
   }



}

