// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxYlm.h>
#include <SxCLI.h>
#include <SxVector.h>
#include <SxFFT1d.h>
#include <SxNaturalCubicSpline.h>

// TODO: weight-function
double errorNorm (const SxVector<Complex16> &fS,
                  SxFFT1d &fft1d,
                  SxVector<Double> weight,
                  int ns, int gMax,
                  SxVector<Complex16> *gradSPtr
                  )
{
   int nsn = int(fS.getSize ());
   SxVector<Complex16> gradG (nsn), fG (nsn); 
   fft1d.fftForward (nsn, fS.elements, fG.elements);
   // cerr << "norm=" << fG.normSqr () << endl;

   gradG.set (0.);
   double errNorm = 0.;
   for (int ig = 0; ig < gMax; ++ig)  {
      /*
      int x = (ig >= nsn/2) ? (ig - nsn) : ig;
      cout << (x/double(ns)) 
           << '\t' << fG(ig).abs ()
           << '\t' << (fG(ns - ig).absSqr () / fG(ig).absSqr ())
           << endl;
      */
      for (int ii = 1; 2 * ii * ns < nsn; ++ii)  {
         if (ii * ns - ig == nsn) continue;
         errNorm += ((ig > 0) ? 2. : 1. ) 
                  * (fG(ii*ns - ig).absSqr () / fG(ig).absSqr ())
                  * weight(ig);
         gradG(ii*ns - ig) = fG(ii*ns - ig)/fG(ig).absSqr () * weight(ig);
         gradG(ig) += -fG(ii*ns-ig).absSqr () / sqr(fG(ig).absSqr ()) * fG(ig)
                    * weight(ig);
         gradG(nsn -ii*ns + ig) = gradG(ii*ns -ig).conj ();
         if ( ig > 0 && ii * ns < nsn)  {
            errNorm += 2. * (fG(ii*ns + ig).absSqr () / fG(ig).absSqr ())
                      * weight(ig);
            gradG(ii*ns + ig) = fG(ii*ns + ig)/fG(ig).absSqr () * weight(ig);
            gradG(ig) += -fG(ii*ns+ig).absSqr () / sqr(fG(ig).absSqr ()) 
                         * fG(ig) * weight(ig);
            gradG(nsn -ii*ns - ig) = gradG(ii*ns +ig).conj ();
         }
      }
      if (ig > 0)
         gradG(nsn - ig) = gradG(ig).conj ();
   }
   /*
   for (int ig = gMax; ig < nsn/2; ++ig)  {
      int x = (ig >= nsn/2) ? (ig - nsn) : ig;
      cout << (x/double(ns)) 
           << '\t' << fG(ig).abs ()
           << '\t' << 1e-12 << endl;
   }
   cout << "&" << endl;
   */
   gradG.plus_assign_ax ( -dot (fG, gradG), fG);
   if (gradSPtr)  {
      fft1d.fftReverse (nsn, gradG.elements, gradSPtr->elements);
   }
   return errNorm;
}

// TODO: weight function
// TODO: Gaussian smoothening of final result (and use 1/smoothening as weight)
SxVector<Double> 
optimizeShape (int nFFT, int gMax, int nPts,
               SxVector<Complex16> *gFac,
               int N)
{
   SxFFT1d fft1d (SxFFT::Both, nFFT * N);
   fft1d.autoscale=true;
   SxVector<Double> weight(gMax);
   double widthSmooth = nFFT * N / 13.;
   int nEdge = 10;
   for (int ig = 0; ig < gMax; ++ig)  {
      weight(ig) = exp (sqr (ig / widthSmooth));
   }

   // --- initialize with Gaussian
   SxVector<Complex16> fS (nFFT * N);
   fS.set (0.);
   for (int s = nEdge; s < N * nPts-nEdge; ++s)  {
      fS(s) = exp ( -sqr((s-N*nPts*0.5)/(N * nPts)*13.));
   }

   // --- conjugate-gradient optimization
   SxVector<Complex16> gradS(nFFT * N), xS (nFFT * N), xSold;
   xS.set (0.);
   double step = -1.;
   for (int it = 0; it < 100; ++it)  {
      fS /= sqrt (fS.normSqr () * nFFT * N);

      double errNorm = errorNorm (fS, fft1d, weight, nFFT, gMax, &gradS);
      cerr << "errNorm=" << errNorm << endl;
      //if (it > 10 && fabs(errNorm-oldNorm) < 1e-6 * errNorm) break;
      //oldNorm = errNorm;

      /*
      for (int s = 0; s < nFFT * N; ++s)
         cout << (s / double(nPts * N))
              << '\t' << fS(s).re
              << '\t' << gradS(s).re << '\t' << gradS(s).im
              << endl;
      cout << "&" << endl;
      */

      // --- search direction
      for (int s = nEdge; s < N * nPts-nEdge; ++s)  {
         xS(s) = gradS(s);
      }

      if (it > 0)
         xS += (xS.normSqr () / xSold.normSqr ()) * xSold;
      xSold.copy (xS);

      double ddl = 2. * dot(xS, gradS).re * nFFT * N;
      //cerr << "d/dl =" << ddl << endl;

      if (it == 0)  {
         //step = 0.1 * errNorm / ddl;
         step=1e-5;
      }
      SxVector<Complex16> fTrial = fS - step * xS;
      fTrial /= sqrt (fTrial.normSqr () * nFFT * N);

      double eTrial = errorNorm (fTrial, fft1d, weight, nFFT, gMax, NULL);
      //cerr << "eTrial=" << eTrial << endl;
      double c = (eTrial - errNorm + step * ddl)/sqr(step);
      step = ddl/(2. * c);
      //cerr << "c=" << c << "; step = " << step << endl;
      //cerr << "eGuess=" << (errNorm - 0.5 * step * ddl) << endl;

      fS -= step * xS;
   }
   fS /= sqrt (fS.normSqr () * nFFT * N);

   {
      FILE *fp = fopen ("rec-space.dat", "w");
      SxVector<Complex16> fG(N * nFFT);
      fft1d.fftForward (N * nFFT, fS.elements, fG.elements);
      for (int ig = 1; ig <= nFFT * N / 2 + 1; ++ig)  {
         double w = exp (-sqr (ig / widthSmooth));
         fG(ig) *= w;
         fG(N*nFFT - ig) *= w;
         fprintf (fp, "%d\t%.12e\t%.12e\n", ig, fG(ig).abs (), w); 
      }
      fft1d.fftReverse (N * nFFT, fG.elements, fS.elements);
      fclose (fp);
   }

// TODO: Gaussian smoothening of final result (and use 1/smoothening as weight)
   if (gFac)   {
      SxVector<Complex16> fG(N * nFFT);
      fft1d.fftForward (N * nFFT, fS.elements, fG.elements);
      gFac->resize (nFFT);
      gFac->set (0.);
      (*gFac)(0) = 1. / fG(0);
      for (int ig = 1; ig <= gMax; ++ig)  {
         (*gFac)(ig)      = 1. / fG(ig);
         (*gFac)(nFFT-ig) = 1. / fG(nFFT-ig);
      }
      VALIDATE_VECTOR (*gFac);
   }
   SxVector<Double> res(nPts * N + 1);
   res <<= fS(SxIdx(0,nPts * N)).real ();
   return res;

}

int main (int argc, char **argv) 
{
   cout.precision (12);
   SxCLI cli(argc, argv);
   int ns = cli.option ("--ns", "number", "FFT mesh size").toInt (200),
       N = cli.option ("--nFine", "number", "interpolation points").toInt(21),
       nCut = cli.option ("--nCut", "number", "number of s-points").toInt(20),
       ng = cli.option ("--ng", "number", "number of G vectors").toInt (100);

   cli.finalize ();
   if (ng >= ns)  {
      cout << "ng >= ns!" << endl;
      SX_QUIT; 
   }
   SxVector<Complex16> gFac;
   SxVector<Double> fS = optimizeShape (ns, ng/2, nCut, &gFac, N);

   {
      FILE *fp = fopen ("r-space.dat", "w");
      for (int s = 0; s <= N * nCut; ++s)
         fprintf (fp, "%.12e\t%.12e\n", s / double(N), fS(s));
      fclose (fp);
   }

   SxFFT1d fft1d (SxFFT::Both, ns);
   fft1d.autoscale=true;

   /*
   for (int s = 0; s < N * nCut; ++s)  {
      fS(s) /= 1e-6 + exp ( -sqr((s-N*nCut*0.5)/(N * nCut)*13.));
   }
   */
   SxNaturalCubicSpline spline (fS);

   for (double x = 0; x <= 1.; x += 1e-4)
   {
      int s0 = int (ceil(x));
      double t = (x-s0);
      s0 = s0 % ns;
      if (s0 < 0) s0 += ns;

      SxVector<Complex16> peakS(ns), peakG(ns);
      peakS.set (0.);
      for (int s = 0; s < nCut; ++s)
         peakS((s+s0)%ns) = spline.getVal ((s-t) * N) * N
                          //* (1e-6 + exp ( -sqr((s-t-nCut*0.5)/(nCut)*13.)))
                            ;

      //for (int s = 0; s < ns; ++s)  {
      //   cout << s << '\t' << peakS(s) << endl;
      //}
      fft1d.fftForward (ns, peakS.elements, peakG.elements);
      peakG *= gFac;
      double sum = 0.;
      for (int ig = 0; 2 * ig < ng; ++ig)  {
         /*
         cout << ig << '\t' 
              << (peakG(ig) - exp(I * (TWO_PI * (x/ns) * ig) )).abs () << endl; 
      */
         sum += (peakG(ig) - exp(I * (TWO_PI * (x/ns) * ig) )).absSqr ();
      }
      cout << x << '\t' << sqrt(sum/ng) << endl;
   }
}
