// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxYlm.h>
#include <SxPBExHole.h>

SxVector<Double> polyFit (const SxVector<Double> &x, const SxVector<Double> &y,
                          double x0, int n, double *errorPtr = NULL)
{
   int N = x.getSize ();
   SxMatrix<Double> A(n + 1,N);
   SX_LOOP(i)  {
      double xx = x(i) - x0;
      double xn = 1.;
      SX_LOOP(j) { A(j,i) = xn; xn *= xx; }
   }
   SxVector<Double> res = (A ^ A.transpose ()).inverse () ^ (A ^ y);
   if (errorPtr)
      *errorPtr = sqrt( ((res ^ A)- y).normSqr () / double(N));
   return res;
}

SxVector<Double> findMin (const SxVector<Double> &x, const SxVector<Double> &y, int n, double *errorPtr = NULL)
{
   int minIdx;
   y.minval (&minIdx);
   double x0 = x(minIdx);
   for (int i = 0; i < 100; ++i)  {
      SxVector<Double> param = polyFit (x,y,x0,n, errorPtr);
      double dx0 = -0.5 * param(1)/param(2);
      if (fabs(dx0) < 1e-10 || (dx0 < 1e-8 * x0 && i > 90))  {
         param(1) = x0;
         return param;
      }
      x0 += dx0;
      //cout << x0 << ' ' << param << endl;
   }
   SX_QUIT;
}

int main (int argc, char **argv)
{
   // --- command line parsing
   
   SxCLI cli(argc, argv);
   int nPar = cli.option ("-n","int","degree").toInt (4,2);
   bool printError = cli.option ("-e","print error").toBool ();
   cli.finalize ();

   fscanf (stdin, " ");
   while (!feof(stdin))  {
      int nr = -1, nrNow = 0, line = 0;
      SxList<double> data;
      do {
         char nl = 0;
         double d;
         if (fscanf (stdin, "%lf%*[ \t]", &d) > 0) {
            data << d;
            nrNow++;
         } else {
            break;
         }
         nl = fgetc (stdin);
         if (nl == '\n')  {
            if (nr == -1) nr = nrNow;
            fscanf(stdin, "%*[ \t]");
            nl = fgetc (stdin);
            if (nl == '\n') break;
            ungetc (nl, stdin);
            if (nr != nrNow)  {
               cout << "Wrong column count" << endl;
               cout << "Should be " << nr << endl;
               cout << "Is " << nrNow << endl;
               SX_QUIT;
            }
            nrNow = 0;
            line++;
         } else {
            ungetc (nl, stdin);
         }
      } while (!feof(stdin));
      if (nr == nrNow) line++;
      if (line < nPar) continue;
      while (data.getSize () % nr) data.removeLast ();
      SxVector<Double> dataVec(data);
      dataVec.reshape (nr, data.getSize () / nr);
      dataVec = dataVec.transpose ();
      //cout << "# " << nr << " " << line << endl;
      if (dataVec.nCols () < 2) {
         cout << "Found only " << dataVec.nCols () << " columns" << endl;
         SX_QUIT;
      }
      double error;
      SxVector<Double> res = findMin (dataVec.colRef(0), dataVec.colRef (1), nPar, &error);
      sxprintf ("%.8f %.12f", res(1), res(0));
      for (int c = 2; c < dataVec.nCols () ; ++c)  {
         SxVector<Double> fit = polyFit (dataVec.colRef(0), dataVec.colRef (c),
                                         res(1),  nPar);
         sxprintf (" %.12g", fit(0));
      }
      for (int i = 2; i <= nPar; ++i)
         sxprintf (" %.12f", res(i));
      if (printError) sxprintf (" %8.3g", error);
      cout << endl;
   }
      
}
