// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------
#include <SxUtil.h>
#include <SxCLI.h>
#include <SxVector.h>
#include <SxMatrix.h>
#include <SxBinIO.h>
#include <SxMath.h>

int main (int argc, char **argv)
{
   // --- parse command line options
   SxCLI cli (argc, argv);
   
   // --- input and output files
   SxString inFile 
      = cli.option ("-i|--input", "input file", 
                    "take as input file")
        .toString ();
   int order = cli.option ("-n", "Polynomial order", 
                    "Polynomial order of the fit polynomial")
        .toInt ();
   cli.finalize ();

   int dim = order + 1;
   SxMatrix<Double> values;
   try {
      SxBinIO out; 
      out.open(inFile, SxBinIO::ASCII_READ_ONLY);
      values = out.readXYPlot();
      out.close();
   } catch (SxException e) {
      e.print ();
      SX_QUIT;
   }

   cout.precision (16);
   SxVector<Double> x = values.colRef(0);
   SxVector<Double> y = values.colRef(1);
   int nPoints = int(x.getSize());

   // Setup Matrix
   SxMatrix<Double> A(nPoints, dim);
   for(int row = 0; row < nPoints; row++)   {
      for(int col = 0; col < dim; col++)   {
         A(row,col) = 1.0;
         for(int pow = 1; pow <=col; pow++)   {
            A(row,col) *= x(row);
         }
      }
   }
   // Solve set of linear equations
   SxVector<Double> coeff = A.solve(y);

   SxBinIO out;
   out.open("coeff.dat", SxBinIO::ASCII_WRITE_ONLY);
   out.writeXYPlot(coeff);
   out.close();

   if(order == 2 || order == 3)   {
      double xMin, yMin;
      if (order == 2)   {
         xMin = -0.5*coeff(1)/coeff(2);
         yMin = coeff(0) + xMin * (coeff(1) + xMin * coeff(2));
      }
      if (order == 3)   {
         double xMin1 = -(coeff(2) + sqrt(coeff(2) * coeff(2) - 3.*coeff(1)*coeff(3))) / (3.*coeff(3));
         double xMin2 = -(coeff(2) - sqrt(coeff(2) * coeff(2) - 3.*coeff(1)*coeff(3))) / (3.*coeff(3));
         double yMin1 = coeff(0) + xMin1 * (coeff(1) + xMin1 * (coeff(2) + xMin1 * coeff(3)));
         double yMin2 = coeff(0) + xMin2 * (coeff(1) + xMin2 * (coeff(2) + xMin2 * coeff(3)));
         if (yMin1 < yMin2)   {
            xMin = xMin1; 
            yMin = yMin1;
         } else   {   
            xMin = xMin2;
            yMin = yMin2;
         }
      }
      sxprintf("%e %e \n", xMin, yMin);
   } else  {
      cout << "Root finder for higher orders not yet implemented!" << endl;
   }

   return 0;
}
