// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxRho.h>
#include <SxCLI.h>
#include <SxCubicSpline.h>
#include <SxRadRBasis.h>

int main (int argc, char **argv)  { 
   
   // --- parse command line arguments
   SxCLI cli(argc, argv);
   cli.preUsageMessage 
      = SxString("Tabulate mesh as function of radial distance to "
                 "some reference point.").wrap ();
   SxString inFile
      = cli.option ("-i|--input","sxb-file","S/PHI/nX rho.sxb file.")
        .toString ("rho.sxb");
   SxString outFile = cli.option ("-o|--output","file","output file.")
                      .toString ("radial.dat");
   Coord refPoint(cli.option ("-p|--point","coords","reference point")
                  .toList3 ());

   bool useSpline = cli.option ("--spline","returns splined function")
                  .toBool ();

   int radialMesh 
      = cli.option("-n","number", "number of radial points for angular average")
        .toInt (0);
   cli.last ().defaultValue = "default: no average";

   int moment = cli.option("-m","moment", "moment of radial integration")
        .toInt (0);

   double rhoZero =  
      cli.option ("--zeroRho", "double", "zero electronic density")
      .toFloat(0.0,0.0);

   cli.finalize ();

   bool angularAverage = radialMesh > 1;

   initSPHInXMath ();
   
   // --- read in mesh
   SxCell cell;
   SxMesh3D dim;
   RhoR rho;
   try {
      SxBinIO io (inFile, SxBinIO::BINARY_READ_ONLY);
      rho = io.readMesh (&cell, &dim);
      cell.setup ();
   } catch (SxException e)  {
      e.print ();
      SX_EXIT;
   }
   
   int nSpin = int(rho.getSize ()), nR = dim.product ();

   if (angularAverage && nSpin > 1)  {
      cout << "angularAverage for spin-polarized cases not yet implemented.\n";
      SX_EXIT;
   }

   FILE *fp = fopen(outFile.ascii (), "w");
   if (!fp) {
      cout << "Cannot open '" << outFile << "' for writing.\n";
      SX_QUIT;
   } else {
      cout << "Tabulating '" << inFile << "' radially to '" << outFile
           << "'.\n";
   }
   cout.flush ();
   SxVector<Double> dist, radRho;
   if (angularAverage || useSpline) dist.resize (nR);
   if (useSpline) radRho.resize(nR);
   
   // --- wrap mesh into reference point's Wigner-Seitz cell
   Coord d, invDim = 1. / Coord(dim);
   for (int iR = 0; iR < nR; ++iR)  {
      d = cell.relToCar (dim.getMeshVec (iR, SxMesh3D::Positive) * invDim)
        - refPoint;
      cell.map (&d, SxCell::WignerSeitz);
      if (angularAverage)  {
         dist(iR) = d.norm ();
      } else if (useSpline)  {
         dist(iR) = d.norm ();
         radRho(iR) = rho(0)(iR);
      } else {
         fprintf (fp, "%.12f", d.norm ());
         for (int iSpin = 0; iSpin < nSpin; ++iSpin)
            fprintf(fp, "\t%.12f", rho(iSpin)(iR));
         fprintf(fp, "\n");
      }
   }
   
   if (angularAverage)  {
      (cout << "Computing angular average..." << endl).flush ();
      double m = dist.maxval ();
      double step = m / (radialMesh - 1);
      SxVector<Double> avg(radialMesh);
      SxVector<Int> np (radialMesh);
      avg.set (0.);
      np.set (0);
      int idx;
      for (int iR = 0; iR < nR; ++iR)  {
         idx = int(lround(dist(iR) / step));
         avg(idx) += rho(0)(iR);
         np(idx)++;
      }
      for (int i = 0; i < radialMesh; ++i)  {
         if (np(i) > 0) avg(i) /= np(i);
         fprintf (fp, "%.12f\t%.12f\t%d\n", (i+0.5)*step, avg(i), np(i));
      }
   }

   if (useSpline)  {
      double rMin = dist.minval ();
      double rMax = dist.maxval ();
      int nPoints = 200;
      SxRadRBasis radRBasis (rMin, rMax, nPoints);
      SxDiracVec<Double> radR = radRBasis.getRadRFunc ();
      SxCubicSpline<SxDiracVec<Double> > spline (toVector(dist),
            toVector(radRho),
            radR, 
            SxCubicSpline<SxDiracVec<Double> >::Natural,
            SxCubicSpline<SxDiracVec<Double> >::MirrorPlane);
      SxDiracVec<Double> result = spline.getYFit();
      result.setBasis(&radRBasis);
      result.handle->auxData.is = -1;
      result.handle->auxData.ia = -1;
      result.handle->auxData.n  = -1;
      result.handle->auxData.l  = 0;
      result.handle->auxData.m  = NONE_M;
      for (int iR = 0; iR < radR.getSize (); iR++)  {
         if (result(iR) < rhoZero) result(iR) = 0.0;
         sxfprintf (fp, "%.12f\t%.12f\n", radR(iR), result(iR));
      }
      SxDiracVec<Double> radRPowM (radR.getSize());
      radRPowM.set(1.0);
      for(int i = 1; i <=moment; i++) radRPowM *= radR;
      double rhoRM = radRBasis.integrate(result*radRPowM) * FOUR_PI;
      sxprintf("%i.th moment = %.6f\n",moment,rhoRM);
   }

   fclose (fp);
   
   return 0;
}
