// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------
#include <SxCLI.h>
#include <SxConfig.h>
#include <SxBinIO.h>
#include <SxParser.h>
#include <SxAtomicStructure.h>
#include <SxNeighbors.h>
#include <SxStickyFilter.h>
#include <math.h>

SxMatrix3<Double> createRandomRotation ()
{
   //method described in "Fast Random Rotation Matrices - Arvo"
   int dec1 = rand() % 100,
       dec2 = rand() % 100,
       dec3 = rand() % 100;

   double i = 0. + dec1/100.,
          j = 0. + dec2/100.,
          k = 0. + dec3/100.;

   // equation (1)
   SxMatrix3<Double> rMat = 
      SxMatrix3<Double> (cos(2*PI*i),sin(2*PI*i),0, -sin(2*PI*i),cos(2*PI*i),0, 0,0,1);

   SxVector3<Double> vec (cos(2*PI*j)*sqrt(k),sin(2*PI*j)*sqrt(k),sqrt(1-k));

   //equation (2)
   SxMatrix3<Double>
      identity (1.,0.,0., 0.,1.,0., 0.,0.,1.),
      vecOuterProduct ( vec(0)*vec(0), vec(0)*vec(1), vec(0)*vec(2),
                        vec(0)*vec(1), vec(1)*vec(1), vec(1)*vec(2),
                        vec(0)*vec(2), vec(1)*vec(2), vec(2)*vec(2));

   SxMatrix3<Double> hMat = identity - 2 * vecOuterProduct;

   //equation (3)
   SxMatrix3<Double> randomMat = (-1)*rMat ^ hMat;
   
   return randomMat;

}

void decorateStructure(SxAtomicStructure& in, const SxVector3<Int>& gMeshIn, 
                       SxArray<SxString>& chemNamesIn,
                       SxString speciesName, int nAddAtoms, double cutoff)
{
      int nSpecies = in.getNSpecies ();

      int addSpeciesPos = int(chemNamesIn.findPos (speciesName));
 
      if (addSpeciesPos < 0) {
         addSpeciesPos = nSpecies++;
         in.startCreation ();
         in.newSpecies ();
         in.endCreation();
         chemNamesIn.append (speciesName);
      }

      int nAtoms = 0, loopCounter = 0;
      SxNeighbors addAtomsNeighbors;
      while (nAtoms < nAddAtoms) {
          
         int dec1 = rand() % 100, 
             dec2 = rand() % 100, 
             dec3 = rand() % 100;

         double 
             i = (rand() % (int)ceil (in.cell.col(0)(0))) + dec1/100., 
             j = (rand() % (int)ceil (in.cell.col(1)(1))) + dec2/100.,
             k = (rand() % (int)ceil (in.cell.col(2)(2))) + dec3/100.;

         Coord addAtomPos = Coord (SxList<double> () << i << j << k);
                  
         SxGrid grid (in, gMeshIn); // has to be recomputed, when atoms are added        
         addAtomsNeighbors.compute (grid, in, addAtomPos, cutoff, SxNeighbors::StoreIdx);

         if (addAtomsNeighbors.idx.getSize () == 0) {
            in.startCreation();
            in.addAtom(addSpeciesPos, addAtomPos);
            in.endCreation();
            ++nAtoms;  
         }  

         ++loopCounter;
         if (loopCounter > 100000) {
         cout << "ERROR: Cannot distribute atoms after 100000 iterations\n"
              << "Please choose a larger value for cut-off of added atoms"
              << "(or make cell larger)\n" << endl;
         SX_EXIT;
         }  
      }  
}


int main (int argc, char **argv)
{

   SxCLI cli (argc, argv);

   cli.authors = "G. Pfanner";

   cli.preUsageMessage 
               = "This add-on generates random structures (e.g. based on structural filaments).";
   
   int atomOnAxis = cli.option ("--center", "center of rotation axis").toInt (-1,-1);    

   SxString inFile = cli.option ("-i|--input", "input file", 
                    "take original input file").toString ("input.sx");
   SxString outFile = 
               cli.option ("-o","filename", "output file name (screen otherwise)").toString ("");


   bool decorateOnly = 
          cli.option ("-d", "read in atomic structure and only add atoms at random positions").toBool ();

   //repeat structural filament
   int nInitStructRepeat = 
          cli.option ("-n", "number of initial structure to add at random positions").toInt (1);
   double minDist = 
             cli.option ("-c", "cut-off distance between atoms of initial structure").toDouble ();

   //decorate with a species
   int addAtoms = 
          cli.option ("--nadd", "number of additional atoms to insert at random positions").toInt (0,0);
   double minDistaddAtom = 
             cli.option ("--cadd","cut-off distance between additional atoms").toDouble (0.);
   SxString addSpecies = 
               cli.option ("--sadd","Symbol of species to add (from Periodic table)").toString ("");

   //decorate with a second species
   int addMoreAtoms = 
          cli.option ("--naddMore", "number of additional atoms (2nd species) to insert at random positions")
                   .toInt (0,0);
   double minDistaddMoreAtom = 
             cli.option ("--caddMore","cut-off distance between additional atoms (2nd species)").toDouble (0.);
   SxString addMoreSpecies = 
               cli.option ("--saddMore","Symbol of species (2nd) to add (from Periodic table)").toString ("");

   cli.finalize ();

   if (addAtoms > 0 && minDistaddAtom < 0.01) {
      cout << "ERROR: added Atoms are too close to neighors (--cadd is chosen too small)" << endl;  
      SX_EXIT;
   }
   if (addAtoms > 0 && addSpecies.getSize () == 0) {
      cout << "ERROR: species name for Atoms to add is not specified (--sadd \"Element\")" << endl;  
      SX_EXIT;
   }

   if (addMoreAtoms > 0 && minDistaddMoreAtom < 0.01) {
      cout << "ERROR: added Atoms are too close to neighors (--cadd is chosen too small)" << endl;  
      SX_EXIT;
   }
   if (addMoreAtoms > 0 && addMoreSpecies.getSize () == 0) {
      cout << "ERROR: species name for Atoms to add is not specified (--sadd \"Element\")" << endl;  
      SX_EXIT;
   }

   if (decorateOnly && nInitStructRepeat > 1) {
      cout << "ERROR: request for conflicting modes" << endl;  
      SX_EXIT;
   }

   // --- read input (fixed structural element) 
   SxAtomicStructure inStruct, tmpStruct;
   SxArray<SxString> chemNames;
      
   SxParser parser;
   SxConstPtr<SxSymbolTable> tree;
      tree = parser.read (inFile, "std/structure.std");
      inStruct = SxAtomicStructure (&*tree);
      chemNames = SxSpeciesData::getElements (&*tree);
   
   // setup mesh for SxNeighbors
   SxVector3<Int> 
      gMesh = 
      SxGrid::suggestMesh (inStruct.cell, (inStruct.getNAtoms () * nInitStructRepeat + addAtoms) / 10 + 1); 

   // shift structure   
   SxAtomicStructure allStruct;
        
   if (atomOnAxis > -1)  { 
      Coord shift = inStruct (atomOnAxis);
      inStruct -= shift;
      cout << "Rotating structure with respect to atom: " 
           << atomOnAxis << endl;
   }
   else {
      cout << "Rotating structure with respect to [0,0,0]..." << endl;
   }   

   //TODO: SXEXIT if cell is not rectangular
   cout << "Note: Assuming a rectangular cell..." << endl;


   srand (unsigned(time(NULL))); //intialize seed for random numbers
   
   int repIt = 0, loopCounter = 0;
   while (repIt < nInitStructRepeat && !decorateOnly) {

      //Rotate structural element
      SymMat rot = createRandomRotation ();

      if ((rot.transpose () - rot.inverse ()).absSqr ().sum () > 1e-5)  {
         cout << "WARNING: matrix is not unitary." << endl;
      }
      inStruct ^= rot;

      //Translate structural element
      int dec1 = rand() % 100, 
          dec2 = rand() % 100, 
          dec3 = rand() % 100;

      double 
         i = (rand() % (int)ceil (inStruct.cell.col(0)(0))) + dec1/100., 
         j = (rand() % (int)ceil (inStruct.cell.col(1)(1))) + dec2/100.,
         k = (rand() % (int)ceil (inStruct.cell.col(2)(2))) + dec3/100.;

      SxVector3<Double> trans = SxVector3<Double> (i, j, k);

      tmpStruct.copy (inStruct); 
      tmpStruct  += trans;  
      tmpStruct %= tmpStruct.cell; // map atoms back into the original cell

      if (allStruct.getNAtoms() == 0) {
         allStruct.copy (tmpStruct);
         ++repIt;
      }
      else {

         SxNeighbors neighbors;
         bool atomOverlap = false;
         SxGrid grid (allStruct, gMesh);  // has to be recomputed, when atoms are added
         for (int ia = 0; ia < tmpStruct.getNAtoms(); ++ia) {

            neighbors.compute (grid, allStruct, tmpStruct(ia), 
                               minDist, SxNeighbors::StoreAll);

            if (neighbors.idx.getSize() > 0) atomOverlap = true;

         }

         if (!atomOverlap) {

            for (int is = 0; is < tmpStruct.getNSpecies (); ++is)  {
               for (int ia = 0; ia < tmpStruct.getNAtoms(is); ++ia) {

                  allStruct.startCreation ();
                  allStruct.addAtom (is, tmpStruct(is,ia)); 
                  allStruct.endCreation ();

               }
            }

            ++repIt;
         
         }
            
      }
   
      //TODO: better break-condition for unrealistic scenario 
      //      (too many atoms to add in too small cell)
      ++loopCounter;
      if (loopCounter > 100000) {
         cout << "ERROR: Cannot distribute atoms after 100000 iterations\n"
              << "Please choose a larger value for cut-off (or make cell larger)\n"
              << endl;
         SX_EXIT;
      }   
   }

   /*
    TODO: decorate-strategy, i.e. insert fixed number of atoms 
          at random distances for each atom of a given species
   */      

   if (decorateOnly) allStruct.copy (inStruct);

   if (addAtoms > 0) decorateStructure(allStruct, gMesh, chemNames, addSpecies, 
                                       addAtoms, minDistaddAtom);

   if (addMoreAtoms > 0) decorateStructure(allStruct, gMesh, chemNames, addMoreSpecies, 
                                           addMoreAtoms, minDistaddMoreAtom);
            
   //TODO: optimize cell, i.e. shrink it if there is too much space around outermost atoms
   
   FILE *file =  stdout;
   if (outFile.getSize () > 0)  {
      file = fopen(outFile.ascii (),"w");
      if (file == NULL)  {
      cout << "Can't open '" << outFile << "'." << endl;
      SX_EXIT;
      }
   } else {
         file = stdout;
   }

   allStruct.fprint (file);

}   
