// ---------------------------------------------------------------------------
//                                          
//      The ab-initio based multiscale library 
//     
//                  S / P H I / n X
//                  http://www.sphinxlib.de
//     
//      Contact:    Sixten Boeck, boeck@mpie.de
//                  Algorithm Design and Modeling Group
//                  Computational Materials Design
//                  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//     
//      Authors:    see src/AUTHORS
//     
// ---------------------------------------------------------------------------

#include <SxCLI.h>
#include <SxAtomicStructure.h>

double angle (SxVector3<Double> v1, SxVector3<Double> v2, SxVector3<Double> v3)
{
   double theta = 0.;
   if (v3.norm () < 1e-5) cout << "|v3| = " << v3.norm () << endl;
   else if (v2.norm () < 1e-5) cout << "|v2| = " << v2.norm () << endl;
   else if (v1.norm () < 1e-5) cout << "|v1| = " << v1.norm () << endl;
   else  {
      double cosTheta = v1 ^ v2 / v1.norm () / v2.norm ();
      if (cosTheta >= 1) theta = 0.;
      else if (cosTheta <= -1) theta = PI;
      else theta = acos(cosTheta);
      if ((theta > 0) && ((cross(v1, v2) ^ v3) < 0)) 
      theta = TWO_PI - theta;
   }
   return (theta);
}

int main (int argc, char **argv)
{
   initSPHInXMath ();
   
   SxCLI cli(argc, argv);

   cli.preUsageMessage 
      = SxString(" Add-on introduces a screw-dislocation dipole to a cell\n"
                 "The input structure should have the third basis vector "
                 "orthogonal to the other two. The dislocation Burger's vector"
                 " is parallel to the third basis; The length is the third "
                 "basis vector times b specified by the option '-b'.").wrap ();
   cli.authors = "Siyuan Zhang";
   
   double b = cli.option ("-b", "Burger's vector")
             .toDouble (1.);

   int rep = cli.option ("--rep", "repitition of strain fields")
            .toInt (10);

   double x0 = cli.option ("--x0", "", "place the first dislocation"
                           " with respect to the structure centre")
              .toDouble (0.);

   double y0 = cli.option ("--y0", "")
              .toDouble (0.);

   double x = cli.option ("-x", "place the second dislocation with respect"
                          " to the first dislocation")
             .toDouble (0.5);

   double y = cli.option ("-y", "")
             .toDouble (0.5);

   bool cut = cli.option ("--cut", "if set value other than 0, cut")
             .toInt (0);

   bool orth = cli.option ("--orthogonal", "orthogonalise the cell")
              .toBool ();

   int disp = cli.option ("--disp", "sampling for displacement field check")
             .toInt (0);

   bool yuja = cli.option ("--yuja", "special testing mode")
              .toBool ();

   bool sxbFile = cli.option ("--sxb", "input file is binary waves file "
                              "rather than S/PHI/nX input file")
                  .toBool ();

   SxString inFile 
      = cli.option ("-i|--input", "input file", 
                    "take original input file")
        .toString (sxbFile ? "waves.sxb" : "input.sx");
   cli.last ().defaultValue = "default: input.sx, or waves.sxb for --sxb flag";
   
   SxString outFile
      = cli.option ("-o","filename", "output file name (screen otherwise)")
        .toString ("dislstructure");

   bool outSxb = cli.option ("--outsxb", "output file is binary "
                              "rather than S/PHI/nX input file").toBool ();
   cli.last ().defaultValue = "default: same format as input file";
   if (sxbFile) outSxb = true;
   if (outSxb && outFile.getSize () == 0)
      outFile = "dislstructure.sxb";
   
   cli.finalize ();

   // --- yuja
   if (yuja) cout << 
   "',':;::,,:..,::'':;:,:::;::,.:;;::::;+:,,.,,,,,,:`..,;,,`.:``,..,..,,,."
   << endl << 
   ":';,;:;;,::,;;:;::;,::;;:,:.:':;,:';;::,,:,.:.,,.:,.;:,,.:,..`,.,...:.."
   << endl << 
   ":;::;':;:,:;;;::;;::;:;::;;;,;:;:::;::;,,:,,;..,,,,;,,..,,..:.:,.`.`:.."
   << endl << 
   "''::;';:',;,:,,,,'::::;;:;;;;;::;,,:;:::,,::,:,,.,,::....,.....,..`.:.."
   << endl << 
   ";',''''#;:.,:;:,:;:'+';;;;':;;,::,::,,;;,,:,,;.,,,::,,.....,`.....,;..."
   << endl << 
   ":;;';;'';,:'':;;;;''';';+;:;;;:,:::'::;;:,:,::,::,:,,,..,,.,,`.....:..."
   << endl << 
   ":;:'+;;:;;''';'::;+;:''+';:;;';::;';:::;:::,::,,,,,:,,,`.,,,.....`.,,;,"
   << endl << 
   "'+++';'';;';;'';':;:;;''''+';;;;;';;:;:;;;,,:,:::;:,:,:,,..,,`.``...;:,"
   << endl << 
   "';+';+++''#'#'+#'+##++++###+++'+''';;;;''';,;;::;';;;;:;,,,:```.....::,"
   << endl << 
   "'''++++++#+##'##+##+#++#####+''+'+';'+++++':;;;''''+++;;';;,.```:,,,,,,"
   << endl << 
   "::;'+'+++#+##+####++++'##++#++''''''''+++'++;+''+++#+#++#';,`,,.:.::..,"
   << endl << 
   ",,:;'''''+#######++'+''+++++'''';';;;''+'++++++++'++#+#+##';;;;:':;;,.,"
   << endl << 
   ".,::'''+'+#######''''++###+'''';;;:;;''''+#######++'+##+###++++'+;:::,,"
   << endl << 
   ".,:;;+++''+#####+;;''##+####'';;;;::;;';'#;+++#+##+''++++########+;;;,,"
   << endl << 
   ".,:,;;+++'+#####+;;'++':###'+;;;;::::;;;##+'#+++'''''''+++########+';;:"
   << endl << 
   "`..:;;'++'+#####':::;;'++++''+;;;::,::;;+''''';;;;;;;;'+''########++':'"
   << endl << 
   ".`.,:;;+'''#####;,.,,::::::::;:::,,,,:::';;;::::::::;:;';;########++';;"
   << endl << 
   "`..::;'''+######:....,,,::::::::,,,,,,:,::::::,,,,,,,:::,;+########++;'"
   << endl << 
   "``.,:;''++#####+,```...,,,,,,,,,,,..,,,,,,,,,,,,,,.,.,.,,:+##########++"
   << endl << 
   "``.,,;''+++####'.````......,.........,.,.................,'##########++"
   << endl << 
   "``..:;;''++##+#'.``````.............................```..,'############"
   << endl << 
   "`..,,:;'#+#####+,``````.`````````..`.......`....```````..,'############"
   << endl << 
   "``...:;+####+###,```````````````````......``````````````.,'############"
   << endl << 
   "``...,:++###+###,````````````````````....```````````````.,'+###########"
   << endl << 
   "`...,::'+###+###.``````````````````..`....``````````````.,;+#####@#####"
   << endl << 
   "`..,,:;+###+++##:` `````````.`````````...,.`````````````.,;############"
   << endl << 
   "``.:,:;++##+#####.`````````..`` ````````....````````````.,;###########+"
   << endl << 
   ".,,,:::'###+#####`````````... ` ``````  `...````````````.,;##########+`"
   << endl << 
   ".,::,::'##+#+####,````````...`````.`.````,...```````````.,'######+###,:"
   << endl << 
   ".,,::,:+####+####+.``````.....,',,,,.'',,....```````````.,'######+##+ '"
   << endl << 
   ",,.,:;'####+######```````...``..,::::,...```````````````.,+#####+#+#' +"
   << endl << 
   ";:,,,'+###+#######.``````..``````.....``````````````````.:######++#+'``"
   << endl << 
   "';:,,:####+#######+`.````````` ````.```````````````````..;######+++':` "
   << endl << 
   "::`.,;############@.`````````` ```..```````````````````.,'######+++'::+"
   << endl << 
   "    ,####+#########;`````````   ```...```````..```````.,:+#####++++': ;"
   << endl << 
   "    '###++##########````````` `.,::::::,.````...``.``.,:'#####+++'+#:,`"
   << endl << 
   "   .+##'++#########@'`````....;''';;;::''';;';,.`.,:::;++#####+++'+#`  "
   << endl << 
   "   ,+#+,;;###+#####@@,```..:+#+#+;..:.,'++'';:,,....,';+#####+++';++: ."
   << endl << 
   "  `+#+,',;############```..,.,,'+;;:':''':;:..,....;,:+#####+#++';+':  "
   << endl << 
   "  '###`.`'##'+########+``.,...`.:;:;:;:,:,.....,;::;'+#####++++''+++`  "
   << endl << 
   " `###;+. ##+;+#########'.......``....,:::,,.`;',;:;'+########++++:++   "
   << endl << 
   " #+.#`` `;+#+#########@#:.,.`..,,,,::;:::,,:.,,,:;++######+##++++##+   "
   << endl << 
   ",++:#; ` '###+#########@#...```.:;';';;;:'.,,::;'+###########+'##;'+   "
   << endl << 
   "'#.#''   '##:'############..````.,:::;;',,,:;;;++########+####+#;;':   "
   << endl << 
   "+# #;'   +##`###########@#+.````....;+,,,',:;''##############+''; ',   "
   << endl << 
   "':`'+   `##''############@#'````...;,.,.:,;;;'+#############+;++. ::   "
   << endl << 
   "+ :'+    ##+###############';.```.;.,,::;:;;+#######+#######++;+  .`   "
   << endl << 
   "# +`     #################+';;.`.;,,,,;;:;:+###############++''+ ,     "
   << endl << 
   "+ +.    `+################+';;;,,,:;;:::''########+########++++' ..    "
   << endl << 
   "; +`    ,#################+'';;;,:;::;;+############+#####++'++',.`    "
   << endl << 
   ".`. `   ##################+';:;;'';''++############+######+++++'''     "
   << endl;

   // --- read input
   SxAtomicStructure structure;
   SxArray<SxString> chemNames;
   if (sxbFile)  {
      try {
         SxBinIO io (inFile, SxBinIO::BINARY_READ_ONLY);
         structure.read (io);
         SxString chemNameList;
         io.close ();
      } catch (SxException e)  {
         e.print ();
         SX_EXIT;
      }
   } else  {
      SxParser parser;
      SxConstPtr<SxSymbolTable> tree;
      tree = parser.read (inFile, "std/structure.std");
      structure = SxAtomicStructure (&*tree);
      structure.atomInfo->meta.update (SxAtomicStructure::Elements,
                                       SxSpeciesData::getElements (&*tree));
   }

   structure %= structure.cell;

   SxVector3<Double> e1 = structure.cell (0),
                     e2 = structure.cell (1),
                     e3 = structure.cell (2);
   cout << "Input Cell: " << structure.cell << endl;
   cout << "Inverse Cell: " << structure.cell.inverse () << endl;
   SxVector3<Double> ex = e1 - (e1 ^ e3) * e3 / e3.normSqr (),
                     ey = e2 - (e2 ^ e3) * e3 / e3.normSqr ();
   cout << "ex = " << ex << endl << "ey = " << ey << endl;
   if (orth)  {
      structure.cell (0,0) = ex (0);
      structure.cell (1,0) = ex (1);
      structure.cell (2,0) = ex (2);
      structure.cell (0,1) = ey (0);
      structure.cell (1,1) = ey (1);
      structure.cell (2,1) = ey (2);
      cout << "Orthogonised Cell:" << endl << structure.cell << endl;
      structure %= structure.cell;
   }

   // --- find the centres of the structure and dislocations
   SxVector3<Double> sum;
   sum.set (0.);
   int nA = structure.getNAtoms ();
   for (int iA = 0; iA < nA; ++iA) sum += structure.getAtom(iA);
   SxVector3<Double> centre = sum / nA;
   cout << "Structure Centre: " << centre << endl;
   centre += x0 * ex + y0 * ey;
   cout << "1. Dislocation Centre: " << centre << endl;
   SxVector3<Double> v = x * ex + y * ey;
   cout << "2. Dislocation Centre: " << centre + v << endl;

   // --- disp: calculate the displacement field on cell boundaries
   double ux = 0., uy = 0.;
   if (disp)  {
      cout << "Check for boundary condition of displacement field:" << endl;
      SxVector3<Double> pvec;
      for (int pt = 0; pt <= disp; ++pt)  {
         double u0 = 0.;
         for (int pp = 0; pp <= 3; ++pp)  {
            SxVector3<Double> exx = ex, eyy = ey;
            if (cut)  {
               exx = 0.5 * ex - 0.5 * ey;
               eyy = 0.5 * ex + 0.5 * ey;
            }
            double u = 0.;
            if (pp == 0) pvec = double(pt) / disp * eyy;
            else if (pp == 1) pvec = double(pt) / disp * eyy + exx;
            else if (pp == 2) pvec = double(pt) / disp * exx;
            else pvec = double(pt) / disp * exx + eyy;
            for (int i = -rep; i <= rep; ++i)  {
               for (int j = -rep; j <= rep; ++j)  {
                  SxVector3<Double> vij = i * ex + j * ey;
                  int addDisl = 2;
                  if (cut) addDisl = 4;
                  while (addDisl > 0)  {
                     SxVector3<Double> vector = vij + centre - pvec;
                     if (addDisl % 2 == 1) vector += v;
                     if (addDisl > 2) vector += 0.5 * ex + 0.5 * ey;
                     double strain = b * angle(v, vector, e3) / TWO_PI;
                     if (yuja && (i == rep) && (addDisl == 1)) strain = 0.;
                     if (yuja && (j == rep) && (addDisl == 1)) strain = 0.;
                     if (addDisl % 2 == 0) u += strain;
                     else u -= strain;
                     addDisl--;
                  }
               }
            }
            if (abs(u) >= 1) cout << "ATTENTION: u = " << u << endl;
            if (pp % 2 == 0) u0 = u;
            if (pp == 1)  {
               cout << pt << ": " << pvec << "; ux = " << u - u0 << endl;
               ux += u - u0 - int(u - u0);
               if (u < u0) ux += 1.;
            }
            if (pp == 3)  {
               cout << pt << ": " << pvec << "; uy = " << u - u0 << endl;
               uy += u - u0 - int(u - u0);
               if (u < u0) uy += 1.;
            }
         }
      }
      ux /= disp + 1.; //calculate average from sum
      uy /= disp + 1.;
      cout << "Suggestions for u (averaged over above values)" << endl 
           << "ux = " << ux << endl << "uy = " << uy << endl;
   }
   else cout << "ATTENTION: ux (uy) suggestions available "
             << "by rerunning with--disp enabled!" << endl;

   // --- construct cell of the dislocation structure
   cout << "ux = ", cin >> ux;
   cout << "uy = ", cin >> uy;
   SxAtomicStructure dislStruct;
   dislStruct.cell = CellMat (e1 + ux * b * e3,
                              e2 + uy * b * e3,
                              e3).transpose ();
   if (cut) dislStruct.cell = CellMat ((e1 - e2) / 2. + ux * b * e3,
                                       (e1 + e2) / 2. + uy * b * e3,
                                       e3).transpose ();

   // --- add atoms to the dislocation structure
   for (int is = 0; is < structure.getNSpecies (); ++is)  {
      dislStruct.startCreation ();
      dislStruct.newSpecies ();
      for (int ia = 0; ia < structure.getNAtoms(is); ++ia)  {
         SxVector3<Double> cartesian = structure.getAtom (is, ia);
         SxVector3<Double> 
         relative = cartesian ^ structure.cell.inverse ().transpose ();
         cout << ia << ": " << cartesian << "; " << relative;
         double u = 0.;
         bool inside = (0 == 0);
         if (cut) inside = (abs (relative(0) + relative(1)) <= 0.5)
                    && (abs (relative(0) - relative(1)) <= 0.5);
         if (inside)  {
            for (int i = -rep; i <= rep; ++i)  {
               for (int j = -rep; j <= rep; ++j)  {
                  SxVector3<Double> vij = i * ex + j * ey;
                  int addDisl = 2;
                  if (cut) addDisl = 4;
                  while (addDisl > 0)  {
                     SxVector3<Double> vector = vij + centre -cartesian;
                     if (addDisl % 2 == 1) vector += v;
                     if (addDisl > 2) vector += 0.5 * ex + 0.5 * ey;
                     double strain = b * angle(v, vector, e3) / TWO_PI;
                     if (yuja && (i == rep) && (addDisl == 1)) strain = 0.;
                     if (yuja && (j == rep) && (addDisl == 1)) strain = 0.;
                     if (addDisl % 2 == 0) u += strain;
                     else u -= strain;
                     addDisl--;
                  }
               }
            }
            if (abs(u) >= 1) cout << "ATTENTION: u = " << u << endl;
            dislStruct.addAtom(is, cartesian + u * e3);
            cout << " -> " << cartesian + u * e3 << "; " 
            << ((cartesian + u * e3) ^ structure.cell.inverse().transpose()) 
            << endl;
         }
         else cout << "SB" << endl;
      }
      dislStruct.endCreation ();
      cout << "Number of Atoms of Species" << is << ": " 
           << dislStruct.getNAtoms(is) << endl;
   }
   dislStruct %= dislStruct.cell;

   // --- output
   if (outSxb)  {
      dislStruct.write (outFile);
   } else {
      FILE *file;
      if (outFile.getSize () > 0)  {
         file = fopen(outFile.ascii(),"w");
         if (file == NULL)  {
            cout << "Can't open '" << outFile << "'." << endl;
            SX_EXIT;
         }
      } else {
         file = stdout;
      }

      dislStruct.fprint (file);
      if (file != stdout) fclose (file);

   }
}
