// ---------------------------------------------------------------------------
//                                          
//      The ab-initio based multiscale library 
//     
//                  S / P H I / n X
//                  http://www.sphinxlib.de
//     
//      Contact:    Sixten Boeck, boeck@mpie.de
//                  Algorithm Design and Modeling Group
//                  Computational Materials Design
//                  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//     
//      Authors:    see src/AUTHORS
//     
// ---------------------------------------------------------------------------

#ifndef _SX_GRAPH_H_
#define _SX_GRAPH_H_

#include <SxUniqueList.h>
#include <SxArrayList.h>
#include <SxHashedArray.h>
#include <SxPair.h>
#include <SxSet.h>
#include <SxTime.h>
#include <SxPtr.h>

//#include <SxMatrix.h>

/** \brief List Graph

    - The graph uses 4 byte integers instead of 8 byte pointers to link
      everything together. Such structure has lower memory requirements and
      faster lookup. The lookup is faster because less memory is accessed.
    - The edges are stored in SxArrayList to decrease memory taken by each 
      edge and to pack all edges belonging to one node into one chunk of memory.
    - The array with edges can be/is sorted for fast lookup.
    - The nodes are stored in two level array to pack the nodes together and
      to append new nodes faster.
      
    \author Sixten Boeck, boeck@gemmantics.com
    \author Vaclav Bubnik, bubnik@gemmantics.com */
template <class T, class E=SxNull>
class SxGraph : public SxThis<SxGraph<T,E> >
{
   public:
      
      class Node
      {
         public:
            T element;
            SxArrayList<int,int>           in;
            SxArrayList<SxPair<int,E>,int> out;
            int prev;
            int next;
            
            Node () : prev(-1),next(-1) { }
      };

      typedef SxPtr<SxUniqueList<int> > Selection;
      
      class Iterator
      {
         public:
            Iterator ();
            Iterator (SxGraph<T,E>  *graph,
                      int               idx,
                      int               follow,
                      const Selection   &selection);
            
            bool operator!= (const typename SxGraph<T,E>::Iterator &);
            bool valid () const  { return (node != NULL); }
            typename SxGraph<T,E>::Iterator &operator++ ();
            
            // -- element
            T &operator* ();
            T *operator-> ();
            
            // -- edges
            T &in(ssize_t);
            T &out(ssize_t);
            ssize_t getSizeIn () const;
            ssize_t getSizeOut () const;
                        
            inline int getDepth () const { return pathLen; }
            inline Node *getNode () const { return node; }
            inline int getIdx () const { return idx; }
            
            ssize_t findEdgeOut (int idx) const;
            ssize_t findEdgeIn (int idx) const;
            
         protected:
            SxGraph<T,E> *graph;
            Selection selection;
            Node  *node;
            int idx;
            int follow;
            int pathLen;
            typename SxList<int>::ConstIterator selectionIt;
            SxList<SxPair<int,int> > queue;
            SxSet<int> visited;
            
            // --- selected neighbors
            ssize_t selectedSizeIn;
            ssize_t selectedSizeOut;
            SxArrayList<ssize_t> selectedIn;
            SxArrayList<ssize_t> selectedOut;
            
            void next ();
      };

      class EdgeIterator
      {
         public:
            EdgeIterator ();
            EdgeIterator (SxGraph<T,E>  *graph,
                          int               idx,
                          const Selection   &selection);
            
            bool operator!= (const typename SxGraph<T,E>::EdgeIterator &);
            bool valid () const  { return (node != NULL); }
            typename SxGraph<T,E>::EdgeIterator &operator++ ();
            
            E &operator* ();
            E *operator-> ();
            T &from ();
            T &to ();
            
         protected:
            SxGraph<T,E> *graph;
            Selection selection;
            typename SxList<int>::ConstIterator selectionIt;
            Node  *node;
            ssize_t edgeIdx;
            bool singleNode;
            
            void findNextEdge ();
      };
      
      SxGraph ();
      ~SxGraph ();
      
      void setSortEdges (bool); // default false
            
      // -- Insertion
      void append (const T &);
      void appendEdge (const T &, const T &, const E &, bool overwrite=true);
      void prependEdge (const T &, const T &, const E &, bool overwrite=true); // slow
      
      SxGraph<T,E> &append (const SxGraph<T,E> &);
      
      // -- Deletion
      void removeElement (const T &);
      void removeEdge (const T &, const T &);
      void removeAll ();
      
      // -- Search
      bool contains (const T &) const;
      int findPos (const T &) const;
      
      int findPath      (const T         &from,
                         const T         &to,
                         int             maxPathLen=0,
                         const Selection &selection=Selection()) const;
                         
      SxList<T> getPath (const T         &from,
                         const T         &to,
                         int             maxPathLen=0,
                         const Selection &selection=Selection()) const;
      
      SxList<T> topsort (const Selection &selection=Selection()) const;
      
      SxArray<SxArray<T> > getCycles (const Selection &selection=Selection()) const;
      SxList<SxList<int> > getCyclesIdx(const Selection &selection=Selection())const;
      
      // -- Size
      inline ssize_t getSize () const { return nNodes; }
      inline ssize_t getNEdges () const { return nEdges; }
      size_t getNBytes () const;
      
      // Iteratorfunctions
      Iterator begin (const Selection &selection=Selection());
      Iterator begin (const T         &from,
                      int             direction, // 1->,2<-,->3->Both
                      const Selection &selection=Selection());
      Iterator end ();
      EdgeIterator beginEdge (const Selection &selection=Selection());
      EdgeIterator endEdge ();
      
      void print (Iterator, std::ostream &, const SxString &options="") const;
      void printDebug () const;
      
   //protected:
      SxHashedArray<Node> nodes;
      SxArray<int>        hashTable;
      
      int firstElement;
      int lastElement;
      int firstFree;
      bool sortEdges;
      
      ssize_t nNodes;
      ssize_t nEdges;
      
      ssize_t hashSize;
      ssize_t hashUsed;
      ssize_t hashBound;
      
      // --- findPath
      void nextLevel (int                      ptr,
                      int                      pathLen,
                      int                      maxPathLen,
                      const Selection          &selection,
                      SxList<SxPair<int,int> > *queue,
                      SxSet<int>               *visited) const;
                      
      // --- getCycles
      bool circuit (int                        idx,
                    int                        start,
                    SxList<int>                &stack,
                    SxArray<bool>              &visited,
                    SxArray<bool>              &blocked,
                    SxArray<SxArrayList<int> > &B,
                    SxList<SxList<int> >       *circuits) const;
      void unblock (int, SxArray<bool> &, SxArray<SxArrayList<int> >&) const;
      
      bool hasOutEdge (int, int) const;
      ssize_t findEdgeOut (int, int) const;
      ssize_t findEdgeIn (int, int) const;
      void disconnectOut (int, int);
      void disconnectIn (int, int);
      
      // --- hash table
      ssize_t hashFindPos (const T &) const;
      bool hashContains (const T &, ssize_t *);
      void hashResize (ssize_t newSize_);
};

/** \brief Graph Data

    \author Sixten Boeck, boeck@gemmantics.com
    \author Vaclav Bubnik, bubnik@gemmantics.com */
template <class T>
class SxGraph<T,SxNull> : public SxThis<SxGraph<T,SxNull> >
{
   public:
      class Node
      {
         public:
            T element;
            SxArrayList<int,int> in;
            SxArrayList<int,int> out;
            int prev;
            int next;
            
            Node () : prev(-1),next(-1) { }
      };

      typedef SxPtr<SxUniqueList<int> > Selection;
      
      class Iterator
      {
         public:
            Iterator ();
            Iterator (SxGraph<T,SxNull> *graph,
                      int                   idx,
                      int                   follow,
                      const Selection       &selection);
            
            bool operator!= (const typename SxGraph<T,SxNull>::Iterator &);
            bool valid () const  { return (node != NULL); }
            typename SxGraph<T,SxNull>::Iterator &operator++ ();
            
            // -- element
            T &operator* ();
            T *operator-> ();
            
            // -- edges
            T &in(ssize_t);
            T &out(ssize_t);
            ssize_t getSizeIn () const;
            ssize_t getSizeOut () const;
            
            inline int getDepth () const { return pathLen; }
            inline Node *getNode () const { return node; }
            inline int getIdx () const { return idx; }
            
            ssize_t findEdgeOut (int idx) const;
            ssize_t findEdgeIn (int idx) const;
            
            // Returns selection to unvisited nodes
            Selection getUnvisited () const;
            
         protected:
            SxGraph<T,SxNull> *graph;
            Selection selection;
            Node  *node;
            int idx;
            int follow;
            int pathLen;
            typename SxList<int>::ConstIterator selectionIt;
            SxList<SxPair<int,int> > queue;
            SxSet<int> visited;

            // --- selected neighbors
            ssize_t selectedSizeIn;
            ssize_t selectedSizeOut;
            SxArrayList<ssize_t> selectedIn;
            SxArrayList<ssize_t> selectedOut;
                        
            void next ();            
      };

      class ConstIterator
      {
         public:
            ConstIterator ();
            ConstIterator (const SxGraph<T,SxNull> *graph,
                           int                   idx,
                           int                   follow,
                           const Selection       &selection);

            
            bool operator!= (
               const typename SxGraph<T,SxNull>::ConstIterator &) const;
            bool valid () const  { return (node != NULL); }
            typename SxGraph<T,SxNull>::ConstIterator &operator++ ();
            
            // -- element
            const T &operator* () const;
            const T *operator-> () const;

            // -- edges
            const T &in(ssize_t) const;
            const T &out(ssize_t) const;
            ssize_t getSizeIn () const;
            ssize_t getSizeOut () const;
            
            inline int getDepth () const { return pathLen; }
            inline const Node *getNode () const { return node; }
            inline int getIdx () const { return idx; }
            
            // Returns selection to unvisited nodes
            Selection getUnvisited () const;

         protected:
            const SxGraph<T,SxNull> *graph;
            Selection selection;
            const Node  *node;
            int idx;
            int follow;
            int pathLen;
            typename SxList<int>::ConstIterator selectionIt;
            SxList<SxPair<int,int> > queue;
            SxSet<int> visited;

            // --- selected neighbors
            mutable ssize_t selectedSizeIn;
            mutable ssize_t selectedSizeOut;
            mutable SxArrayList<ssize_t> selectedIn;
            mutable SxArrayList<ssize_t> selectedOut;
                        
            void next ();            
      };

      class EdgeIterator
      {
         public:
            EdgeIterator ();
            EdgeIterator (SxGraph<T,SxNull> *graph,
                          int                   idx,
                          const Selection       &selection);
            
           bool operator!=(const typename SxGraph<T,SxNull>::EdgeIterator&);
           bool valid () const  { return (node != NULL); }
            typename SxGraph<T,SxNull>::EdgeIterator &operator++ ();
            
            T &from ();
            T &to ();
            
         protected:
            SxGraph<T,SxNull> *graph;
            Selection selection;
            typename SxList<int>::ConstIterator selectionIt;
            Node  *node;
            ssize_t edgeIdx;
            bool singleNode;
            
            void findNextEdge ();
      };
      
      SxGraph ();
      ~SxGraph ();
      
      void setSortEdges (bool); // default false
      
      // -- Insertion
      void append (const T &);
      void appendEdge (const T &, const T &);
      void prependEdge (const T &, const T &); // slow
      
      SxGraph<T,SxNull> &append (const SxGraph<T,SxNull> &);
      
      // -- Deletion
      void removeElement (const T &);
      void removeEdge (const T &, const T &);
      void removeSelection (const Selection &selection);
      void removeAll ();
      
      // -- Search
      bool contains (const T &) const;
      int findPos (const T &) const;
      
      int findPath      (const T         &from,
                         const T         &to,
                         int             maxPathLen=0,
                         const Selection &selection=Selection()) const;
                         
      SxList<T> getPath (const T         &from,
                         const T         &to,
                         int             maxPathLen=0,
                         const Selection &selection=Selection()) const;

      SxList<T> topsort (const Selection &selection=Selection()) const;
      
      SxArray<SxArray<T> > getCycles (const Selection &selection=Selection()) const;
      SxList<SxList<int> > getCyclesIdx(const Selection &selection=Selection())const;

      /** Returns graph in matrix form.
          graph:      matrix: column M has all neighbor nodes as '1'
             a -> b        a b c
             b -> c      a 0 0 0
                         b 1 0 0  
                         c 0 1 0 */
     // SxMatrix<Int> toMatrix () const;
      
      /** Find all reachable nodes for each node.
          graph:      matrix: column M has all reachable nodes as '1'
             a -> b        a b c
             b -> c      a 0 0 0
                         b 1 0 0  
                         c 1 1 0 */
      //SxMatrix<Int> transitiveClosure () const;
      
      // -- Size
      inline ssize_t getSize () const { return nNodes; }
      inline ssize_t getNEdges () const { return nEdges; }
      size_t getNBytes () const;
      
      // Iteratorfunctions
      Iterator begin (const Selection &selection=Selection());
      Iterator begin (const T         &from,
                      int             direction, // 1->,2<-,->3->Both
                      const Selection &selection=Selection());
      Iterator begin     (const T &from, const Selection &selection=Selection());
      Iterator beginIn   (const T &from, const Selection &selection=Selection());
      Iterator beginBoth (const T &from, const Selection &selection=Selection());
      Iterator end ();
      EdgeIterator beginEdge (const Selection &selection=Selection());
      EdgeIterator endEdge ();

      ConstIterator begin     (const Selection &selection=Selection()) const;
      ConstIterator begin     (const T         &from,
                               const Selection &selection=Selection()) const;
      ConstIterator beginIn   (const T         &from,
                               const Selection &selection=Selection()) const;
      ConstIterator beginBoth (const T         &from,
                               const Selection &selection=Selection()) const;
      ConstIterator end () const;

      void print (Iterator, std::ostream &, const SxString &options="") const;
      void print (ConstIterator, std::ostream &, const SxString &options="") const;
      void printDebug () const;
      
   //protected:
      SxHashedArray<Node> nodes;
      SxArray<int>        hashTable;
      
      int firstElement;
      int lastElement;
      int firstFree;
      bool sortEdges;
      
      ssize_t nNodes;
      ssize_t nEdges;
      
      ssize_t hashSize;
      ssize_t hashUsed;
      ssize_t hashBound;

      void removeNode (int nodeIdx);

      void print (const SxUniqueList<int> &visited,
                  std::ostream            &,
                  const SxString          &options="") const;
      
      // --- findPath
      void nextLevel (int                      ptr,
                      int                      pathLen,
                      int                      maxPathLen,
                      const Selection          &selection,
                      SxList<SxPair<int,int> > *queue,
                      SxSet<int>               *visited) const;
      
      // --- getCycles
      bool circuit (int                        idx,
                    int                        start,
                    SxList<int>                &stack,
                    SxArray<bool>              &visited,
                    SxArray<bool>              &blocked,
                    SxArray<SxArrayList<int> > &B,
                    SxList<SxList<int> >       *circuits) const;
      void unblock (int, SxArray<bool> &, SxArray<SxArrayList<int> >&) const;
      
      bool hasOutEdge (int, int) const;
      ssize_t findEdgeOut (int, int) const;
      ssize_t findEdgeIn (int, int) const;
      void disconnectOut (int, int);
      void disconnectIn (int, int);
      
      // --- hash table
      ssize_t hashFindPos (const T &) const;
      bool hashContains (const T &, ssize_t *);
      void hashResize (ssize_t newSize_);
};

template <class T, class E>
std::ostream& operator<< (std::ostream &, const SxGraph<T,E> &);


#include <SxGraph.hpp>

#endif /* _SX_GRAPH_H_ */
