// ---------------------------------------------------------------------------
//                                          
//      The ab-initio based multiscale library 
//     
//                  S / P H I / n X
//                  http://www.sphinxlib.de
//     
//      Contact:    Sixten Boeck, boeck@mpie.de
//                  Algorithm Design and Modeling Group
//                  Computational Materials Design
//                  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//     
//      Authors:    see src/AUTHORS
//     
// ---------------------------------------------------------------------------

#include <fstream>
#include <SxParserBase.h>
#include <SxFSAction.h>

// --------------------------------------------------------------------------

SxParserBase::SxParserBase ()
   : scannerPtr(NULL),
     inStream(NULL),
     lineOffset(0),
     traces(false),
     curLineIndent(-1),
     prevLineIndent(-1),
     maxIncludeDepth(10)
{
   // empty
}

SxParserBase::~SxParserBase ()
{
   // empty
}

bool SxParserBase::resume () const
{
   return (errors.getSize() < maxErrors);
}

void SxParserBase::errorHandler (const SxString &msg_,
                             ssize_t line0, ssize_t col0,
                             ssize_t line1, ssize_t col1)
{
   SX_TRACE ();
   SxString msg = inFile + ":" + line0 + "." + col0 + "-" + line1 + "." + (col1-1) + ": ";
   msg += msg_.substitute ("TK_", "");
   errors << Error (line0, col0, line1, col1, msg);
}

void SxParserBase::setLineOffset (ssize_t offset)
{
   SX_TRACE ();
   lineOffset = offset;
}

void SxParserBase::setSearchPath (const SxString &path)
{
   SX_TRACE ();
   searchPath << "";  // support absolute filenames
   searchPath << path.tokenize(":");
}

void SxParserBase::setMaxIncludeDepth (ssize_t depth)
{
   SX_TRACE ();
   maxIncludeDepth = depth;
}

void SxParserBase::enableTraces (bool state)
{
   SX_TRACE ();
   traces = state;
}

void SxParserBase::validateIncludes (const SxString &pattern) const
{
   SX_TRACE ();

   if (lexIncludes.getSize() > maxIncludeDepth)  { 
      SX_THROW ("Too many include levels:\n- "
               + SxString::join (lexIncludes, "\n- "));
   }

   SxSortedList<SxString> fqfns = resolveFiles (pattern);
   SxString filename = fqfns.first ();

   // --- is file readable?
   FILE *fp = fopen (filename.ascii(), "r");
   if (!fp)  {
      SX_THROW ("Could not open file " + filename + ".");
   }
   fclose (fp);
}

void SxParserBase::setIncLoc (int line0, int col0, int line1, int col1)
{
   SX_TRACE ();
   lexIncLoc = SxArray<int> (4);
   lexIncLoc(0) = line0; lexIncLoc(1) = col0;
   lexIncLoc(2) = line1; lexIncLoc(3) = col1;
}

bool SxParserBase::processNextInclude ()
{
   SX_TRACE ();
   if (lexIncStack.getSize() == 0)  return false;

   SxString nextFile = lexIncStack.first ();
   lexIncStack.removeFirst ();

   int dummy = 0;
   popInclude   (&dummy, &dummy, &dummy, &dummy);

   int line0 = lexIncLoc(0), col0 = lexIncLoc(1);
   int line1 = lexIncLoc(2), col1 = lexIncLoc(3);

   SX_DBG_MSG ("   about to open " << nextFile);

   pushIncludes (nextFile, line0, col0, line1, col1);
   SX_DBG_MSG ("returning " << (lexIncStack.getSize() > 0));
   return true;
}

bool SxParserBase::pushIncludes (const SxString &pattern,
                             int line0, int col0, int line1, int col1)
{
   SX_TRACE ();
   SX_DBG_MSG ("   opening file " << pattern);

   if (lexIncludes.getSize() > maxIncludeDepth)  { 
      SX_THROW ("Too many include levels.");
   }

   SxSortedList<SxString> fqfns = resolveFiles (pattern);

   SxString filename = fqfns.first ();
   fqfns.removeFirst(); lexIncStack << fqfns;

   // --- is file readable?
   FILE *fp = fopen (filename.ascii(), "r");
   if (!fp)  {
      SX_THROW ("Could not open file " + filename + ".");
   }
   fclose (fp);

   SxArray<int> loc(4);
   loc(0) = line0; loc(1) = col0; loc(2) = line1; loc(3) = col1;

   lexStreams  << inStream;
   lexIncludes << inFile;
   lexLocs     << loc;

   inFile   = filename;
   inStream = new ifstream (filename.ascii(), ios::in|ios::binary);

   return true;
}

bool SxParserBase::popInclude (int *line0, int *col0, int *line1, int *col1)
{
   SX_TRACE ();

   if (lexIncludes.getSize() == 0)  return false;

   delete inStream;
   inStream = lexStreams.last ();
   inFile   = lexIncludes.last ();

   SxArray<int> loc = lexLocs.last();
   *line0 = loc(0); *col0 = loc(1); *line1 = loc(2); *col1 = loc(3);

   lexStreams.removeLast ();
   lexIncludes.removeLast ();
   lexLocs.removeLast ();

   return true;
}

void SxParserBase::pushLex (int col, int line)
{
   SX_TRACE ();
   lexStack.append (SxList<SxString>());
   lexCols << col; lexLines << line;
}

void SxParserBase::popLex (int *colPtr, int *linePtr)
{
   SX_TRACE ();
   lexStack.removeLast ();
   *colPtr  = lexCols.last();  lexCols.removeLast ();
   *linePtr = lexLines.last(); lexLines.removeLast ();
}

void SxParserBase::appendLex (const SxString &in)
{
   SX_TRACE ();
   lexStack.last().append (in);
}

SxList<SxString> SxParserBase::collectLex ()
{
   SX_TRACE ();
   SxList<SxString> res = lexStack.last();
   lexStack.last().removeAll ();
   return res;
}

void SxParserBase::setIndent (int i)
{
   SX_TRACE ();
   curLineIndent = i;
}

void SxParserBase::updateIndent ()
{
   SX_TRACE ();
   prevLineIndent = curLineIndent;
}

int SxParserBase::indentDiff ()
{
   SX_TRACE ();
   return curLineIndent - prevLineIndent;
}

bool SxParserBase::unindent ()
{
   SX_TRACE ();
   return (--prevLineIndent > curLineIndent);
}

SxSortedList<SxString> SxParserBase::resolveFiles (const SxString &pattern) const
{
   SX_TRACE ();
   SxSortedList<SxString> fqfns;
   bool useWildcards = (pattern.contains ("*") || pattern.contains("?"));

   SxString curPath = inFile;
   if (useWildcards && SxFile(curPath).exists())  {  // not STDIN or cin
      curPath = SxFile(curPath).getAbsPath();
   }

   SxList<SxFileInfo> items;
   SxList<SxFileInfo>::ConstIterator infoIt;

   SxList<SxString>::ConstIterator it;
   for (it = searchPath.begin(); it != searchPath.end(); ++it)  {
      SxString fqfn = *it + "/" + pattern;
      try {
         if (useWildcards)  {
            items = SxFSAction::ls (fqfn);
            for (infoIt = items.begin(); infoIt != items.end(); ++infoIt) {
               const SxString path = infoIt->getAbsPath();
               SX_DBG_MSG ("   test -f " << path);
               if (path != curPath)
                  fqfns << path;
            }
            if (fqfns.getSize() > 0)  return fqfns;
         }  else  {
            SX_DBG_MSG ("   test -f " << fqfn);
            if (SxFile(fqfn).exists())  {
               fqfns = SxSortedList<SxString>() << fqfn;
               return fqfns;
            }
         }
      } catch (SxException e)  {
         e.print ();
         SX_EXIT;
      }
   }
   // --- nothing found
   if (!useWildcards)  {
      SX_THROW ("Could not find file " + pattern);
   }
   return SxSortedList<SxString> ();
}


int SxParserBase::readString (const SxString &in, ssize_t maxErrors_)
{
   SX_TRACE ();
   inFile = "(stdin)";
   maxErrors = maxErrors_;
   inStream = new istringstream (in.ascii());
   initScanner (traces);
   int res = parse ();
   destroyScanner ();
   delete inStream;
   handleExceptions ();
   return res;
}

int SxParserBase::readStream (std::istream *in, ssize_t maxErrors_)
{
   SX_TRACE ();
   inFile = "(cin)";
   maxErrors = maxErrors_;
   inStream = in;
   initScanner (traces);
   int res = parse ();
   destroyScanner ();
   inStream = NULL;
   handleExceptions ();
   return res;
}


int SxParserBase::readFile (const SxString &filename, ssize_t maxErrors_)
{
   SX_TRACE ();
   inFile = filename;
   maxErrors = maxErrors_;
   if (!SxFSAction::test_f (filename))  {
      SX_THROW ("Could not find file " + filename + ".");
   }
   // --- is file readable?
   FILE *fp = fopen (filename.ascii(), "r");
   if (!fp)  {
      SX_THROW ("Could not open file " + filename + ".");
   }
   fclose (fp);

   std::ifstream *ifStream = new ifstream (filename.ascii(), ios::in|ios::binary);
   inStream = ifStream;
   initScanner (traces);
   int res = parse ();
   destroyScanner ();
   ifStream->close ();
   delete inStream;
   handleExceptions ();
   return res;
}


int SxParserBase::readFunction (const ReaderFunction &cb, ssize_t maxErrors_)
{
   SX_TRACE ();
   inFile = "(in)";
   maxErrors = maxErrors_;
   inStream = NULL;
   readerCB = cb;
   initScanner (traces);
   int res = parse ();
   destroyScanner ();
   handleExceptions ();
   return res;
}

void SxParserBase::handleExceptions ()
{
   SX_TRACE ();
   if (errors.getSize() > 0)  {
      SxString msg;
      SxList<Error>::ConstIterator it;
      for (it = errors.begin(); it != errors.end(); ++it)  {
         msg += "ERROR: " + it->msg + "\n";
      }
      SX_THROW (msg);
   }
}

