// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_MPI_H_
#define _SX_MPI_H_

#include <SxPtr.h>
#include <SxString.h>
#include <SxIPC.h>
#include <SxProcUnit.h>
#include <SxArray.h>
#include <SxMPIMgr.h>

#include <SxConfig.h>
#ifdef USE_MPI
#  include <mpi.h>
#endif


/** \brief Interface for platform-independent MPI processes

    \b SxMPI = S/PHI/nX Message-Passing Interface Process

    \ingroup group_os
    \author  Sixten Boeck, boeck@mpie.de */
class SX_EXPORT_IPC SxMPI : public SxProcUnit
{
   public:

      enum Destination { Anywhere, GroupMaster, GroupSlave };

      SxMPI (Destination d=Anywhere);
      SxMPI (const SxPtr<SxMPIMgr> &group_, Destination d=Anywhere);

      virtual ~SxMPI ();

      void start ();
      void wait ();

      void barrier ();

      inline int                 broadcast (int);
      inline SxString            broadcast (const SxString &);
      template<class T> inline T broadcast ();

      void terminate ();
      template<class T> void recv (size_t nElem) const;

      virtual void main ();

      bool isMaster () const;
      bool isGroupMaster () const;

   protected:

      friend class SxMPIMgr;

      Destination destination;
      int rank;

      void init ();

      SxPtr<SxMPIMgr> getMPIMgr () const;

};


template<class T>
void SxMPI::recv (size_t nElem) const
{
   SxArray<T> buffer (nElem);
   SX_EXIT;
}

// "sending" broadcast
int SxMPI::broadcast (int data)
{
#  ifdef USE_MPI
      int buffer = data;
      MPI_Bcast (&buffer, 1, MPI_INT, 0, getMPIMgr()->mpiComm); 
      return buffer;
#  else
      return data;
#  endif
}

// "sending" broadcast
SxString SxMPI::broadcast (const SxString &data)
{
#  ifdef USE_MPI
      // communicate size
      ssize_t size = data.getSize ();
      SX_MPI_MASTER_ONLY {
         if (ssize_t(int(size)) != size)  {
            cout << "Cannot broadcast string with " << size
                 << " elements: MPI handles only " << int(size)
                 << endl;
            SX_EXIT;
         }
      }
      MPI_Bcast (&size, (int)sizeof(size), MPI_CHAR, 0, MPI_COMM_WORLD);
      SxString buffer;
      SX_MPI_MASTER_ONLY { buffer = data; }
      SX_MPI_SLAVE_ONLY  { buffer.resize (size); }

//    MPI_Bcast (buffer.elements, data.getSize(), MPI_CHAR, 0, getMPIMgr()->mpiComm); 
      MPI_Bcast (buffer.elements, int(size), MPI_CHAR, 0, MPI_COMM_WORLD);
      return buffer;
#  else
      return data;
#  endif
}

// "receiving"
template<class T>
T SxMPI::broadcast ()
{
#  ifdef USE_MPI
      T buffer;
      return broadcast (buffer);
#  else
      SX_EXIT;
#  endif
}
#endif /* _SX_MPI_H_ */
