// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_THREAD_H_
#define _SX_THREAD_H_

#include <SxConfig.h>
#include <SxPtr.h>
#include <SxIPC.h>
#include <SxProcUnit.h>
#include <SxProcUnitGroup.h>


/** \brief Interface for platform-independent threads

    \b SxThread = SFHIngX Thread Interface

    SxThread is a seperate thread of the program which can run in parallel
    with other threads. Every thread has access to all data of the parent
    process (shared memory). Threadding works only if the underlying operating
    system supports multitasking.
    A new thread object is created by deriving from SxThread and overloading
    of the SxThread::main function. The thread is started using SxThread::start
    and can be joined with other threads with SxThread::wait. 

    \par Example:
\code
   class MyThread : public SxThread
   {
      public:
         int id;

         MyThread (int id_) : SxThread (), id(id_) { }

         virtual void main ()  {
            for (int i=0; i < 10; ++i) {
               printf ("This is thread id %d, i=%d\n", id, i); fflush (stdout);
               SxThread::sleep(1);
            }
         }
   };

   int main ()
   {
      MyThread t1(1), t2(2);
      t1.start ();
      t2.start ();
      t1.wait ();
      t2.wait ();
      printf ("Back in main thread\n");
      return 0;
   }
\endcode
    
    \ingroup group_os
    \author  Sixten Boeck, boeck@sfhingx.de */
class SX_EXPORT_IPC SxThread : public SxProcUnit
{
   public:


      /** \brief Create a new thread

          In order to start the thread the SxThread::start function has to
          be called explicitly. */
      SxThread ();

      /** \brief Destructor.

          The destrcutor does not terminate a thread! Waiting for the 
          end of a thread is done using the barrier function SxThread::wait. */
      virtual ~SxThread ();

      /** \brief Start a new thread

          This function sets up the thread environment and execute the
          SxThread::main function of the derived class. 
          \param priority  The priority used for scheduling the new thread */
      void start ();

      /** \brief Block a thread and wait its termination.

           This function creates a barrier and waits until the thread has 
           (been) terminated. 

           \par Example:
\code
   MyThread t1, t2;

   // --- multi-threadding part starts here
   t1.start (); t2.start ();
   t1.end();    t2.end();
   // --- end of multi-threadding part
\endcode 
       */
      void wait ();


      /** \brief Terminates a thread

          This function terminates the thread by calling the corresponding
          kill function of the underlying thread library. */
      void terminate ();

      /** \brief Slot for computation routine of a thread

          When the thread is executed from SxThread::start this virtual
          function is called. When defining a new thread type by deriving
          from SxThread, overload this function. It should contain the
          computational part which runs threadded. */
      virtual void main ();
      /** send signals before and after calling main()
        */
      virtual void mainWrap ();

      /** \brief Sleep for some secondus

        Calling this function makes the current process/thread sleep until
        a certain time has elapsed. This function can be applied to threads
        or to entire processes.

        \par Example 1
        Making a process sleep for 5 seconds
\code
   int main ()
   {
      for (int i=0; i < 100; ++i)  {
         printf ("i=%d\n", i);
         SxThread::sleep(5);
      }
   }
\endcode
 
       \par Example 2 
       Making a single thread sleep for 1 second for observing changes of 
       a variable (which might be controlled from the slave thread.
\code
   class MyMasterThread : public SxThread
   {
      public:
         MyMasterThread () : SxThread (), f(false)  {  }

         void setFlag (bool f)  {  flag = f; }

         virtual void main ()
         {
             for (;;)  {
                if (flag)  doSomeThing ();
                SxThread::sleep(1);
             }
         }
      protected:
         bool flag;
   };

   class MySlaveThread : public SxThread
   {
      public:
         MySlaveThread (MyMasterThread *m) : SxThread (), master(m);

         virtual void main ()
         {
            // do some heavy calculation and communicate with the master
            for (i=0; i<10000; ++i)  {
               // eat CPU time
               if (...)  master->setFlag (true);  // call master process
            }
         }
         
      protected:
         MyMasterThread *master;
   };

   int main ()
   {
      MyMasterThread tMaster;
      MySlaveThread  tSlave(&tMaster);
      tMaster.start ();
      tSlave.start ();
      tMaster.wait ();
      tSlave.wait ();
      return 0;
   }
\endcode */
      static void sleep (unsigned int seconds);
      static void millisleep (unsigned long mseconds);

      bool isRunning () const;
};

#endif /* _SX_THREAD_H_ */
