// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxThreadPool.h>
#include <SxThread.h>
#include <SxConfig.h>
#include <SxException.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef MACOSX
#  include <sys/sysctl.h>
#endif

/** The thread pool.
    Top level pool depends on hardware!!!
    call initialize() */
static SxThreadPool globalPool;


SxThreadPool::PoolThread::PoolThread (SxThreadPool *pool_, int id_)
   : SxSystemThread (),
     pool(pool_),
     id(id_)
{
   SX_DBG_MSG ("create PoolThread " << id);
}

SxThreadPool::PoolThread::~PoolThread ()
{
   SX_DBG_MSG ("destroy PoolThread " << id);
}

void SxThreadPool::PoolThread::main ()
{
   SX_DBG_MSG ("PoolThread " << id << " main()");
   
   bool workingLoop = true;
   
   pool->mutex.lock ();
   
   while (workingLoop)  {
      while (!pool->terminateCmd && pool->waitingTasks.getSize () < 1)  {
         SX_DBG_MSG ("PoolThread " << id << " waiting for new data");
         pool->condition.wait (&pool->mutex);
      }
      SX_DBG_MSG ("PoolThread " << id << " new data present");
      
      if (pool->terminateCmd)  {
         SX_DBG_MSG ("PoolThread " << id << " terminate command");
         workingLoop = false;
      }  else  {
         SX_DBG_MSG ("PoolThread " << id << " start working");
         // --- move from waiting tasks to running tasks
         SxList<SxTask>::List* task = pool->waitingTasks.first ();
         pool->waitingTasks.removeFirst ();
         pool->runningTasks << task;
         
         // --- run the thread
         SxProcUnit *procUnitPtr = task->element.getProcUnitPtr ();
         SX_CHECK (procUnitPtr);
         SxThread *thread = dynamic_cast<SxThread*>(procUnitPtr);
         SX_CHECK (thread);
         pool->mutex.unlock ();
            thread->mainWrap ();
         pool->mutex.lock ();
         SX_DBG_MSG ("PoolThread " << id << " work finished");
         
         // --- remove the task
         pool->runningTasks.removeElement (task);
         pool->tasks.removeItem (task);
         
         pool->waitCondition.wakeAll ();
      }
   }
   
   SX_DBG_MSG ("PoolThread " << id << " terminated");
   pool->mutex.unlock ();
}

// default for SxThreadMgr was : int offset=0, int multiplier=1
SxThreadPool::SxThreadPool ()
   : terminateCmd (false)
{
   // empty
}

SxThreadPool::~SxThreadPool ()
{
   SX_DBG_MSG ("terminating thread pool");
   
   shutdown ();

   ssize_t i;
   for (i=0; i < pool.getSize(); ++i)  {
      SX_CHECK (pool(i).getPtr ());
      pool(i)->wait ();
   }
   
   SX_DBG_MSG ("thread pool terminated");
}

void SxThreadPool::initialize (int nThreads_)
{
   if (nThreads_ < 1)  {
      nThreads_ = SxThreadPool::getNProcs ();
   }

   SX_DBG_MSG ("creating thread pool nThreads=" << nThreads_);
               
   globalPool.resizePool (nThreads_);
}

SxThreadPool &SxThreadPool::getGlobalObj ()
{
   return globalPool;
}

// --- Expandable thread pools
//bool isExpandable () const { return expandable; }
//void resize (ssize_t nThreads);
//void SxThreadPool::resize (ssize_t nThreads_)
//{
//   SX_CHECK (expandable);
//   resizePool (nThreads_);
//}

ssize_t SxThreadPool::getSize () const
{
   return pool.getSize();
}

void SxThreadPool::submit (const SxTask &task_)
{
   if (getSize () < 1)  {
      SX_THROW ("Can't submit tasks to empty threadpool.");
   }
   if (!task_.getProcUnitPtr ())  {
      SX_THROW ("Can't submit task without process unit.");
   }

   SX_MUTEX (mutex)  {
      tasks << task_;
      waitingTasks << tasks.lastElement;
      condition.wakeOne ();
   }
}

void SxThreadPool::submit (const SxPtr<SxProcUnit> &procUnit_)
{
   if (getSize () < 1)  {
      SX_THROW ("Can't submit tasks to empty threadpool.");
   }
   if (!procUnit_.getPtr ())  {
      SX_THROW ("Can't submit empty process unit.");
   }

   SX_MUTEX (mutex)  {
      tasks << SxTask(procUnit_);
      waitingTasks << tasks.lastElement;
      condition.wakeOne ();
   }
}

void SxThreadPool::operator<< (const SxTask &task)
{
   submit (task);
}

bool SxThreadPool::contains (const SxProcUnit *procUnit_) const
{
   bool result = false;
   
   SX_MUTEX (mutex)  {
      result = containsWaiting (procUnit_) || containsRunning (procUnit_);
   }
   return result;
}

bool SxThreadPool::isRunning (const SxProcUnit *procUnit_) const
{
   bool result = false;

   SX_MUTEX (mutex)  {
      result = containsWaiting (procUnit_) || containsRunning (procUnit_);
   }
   
   return result;
}

void SxThreadPool::wait (const SxProcUnit *procUnit_)
{
   SX_MUTEX (mutex)  {
      while (containsWaiting (procUnit_) || containsRunning (procUnit_))  {
         SX_DBG_MSG ("Task is waiting for result");
         waitCondition.wait (&mutex);
      }
   }
}

ssize_t SxThreadPool::getNTasks () const
{
   ssize_t result = 0;
   
   SX_MUTEX (mutex)  {
      result = waitingTasks.getSize () + runningTasks.getSize ();
   }
   
   return result;
}

int SxThreadPool::getNProcs ()
{
   int nProcs = 0;
   
   // --- try SX_THREADS first
   char *str = getenv ("SX_THREADS");
   if (str) {
      nProcs = atoi (str);
      if (nProcs >= 1)  {
         SX_DBG_MSG ("SX_THREADS=" << nProcs);
         return nProcs;
      }
   }

   // --- use OS calls to find the number of cores
#  ifdef WIN32
      SYSTEM_INFO info;
      GetSystemInfo (&info);
      return info.dwNumberOfProcessors;
#  else
#    ifdef _SC_NPROCESSORS_ONLN
      // or _SC_NPROCESSORS_CONF
      nProcs = static_cast<int>(sysconf (_SC_NPROCESSORS_ONLN));
      SX_DBG_MSG ("_SC_NPROCESSORS_ONLN=" << nProcs);
      return nProcs;
#    elif defined(CTL_HW) && defined(HW_AVAILCPU) && defined(HW_NCPU)
      size_t len = sizeof(nProcs); 
      int name[2];
      name[0] = CTL_HW;
      name[1] = HW_AVAILCPU;
      sysctl (name, 2, &nProcs, &len, NULL, 0);

      if (nProcs < 1)  {
         name[1] = HW_NCPU;
         sysctl (name, 2, &nProcs, &len, NULL, 0);
         
         if (nProcs < 1)  {
            nProcs = 1;
         }
      }
      SX_DBG_MSG ("sysctl(CPU)=" << nProcs);
      return nProcs;
#    else
      return 1;
#    endif
#  endif
}

// lock mutex before calling
bool SxThreadPool::containsWaiting (const SxProcUnit *procUnit_) const
{
   SxList<SxList<SxTask>::List*>::ConstIterator it;
   for (it = waitingTasks.begin(); it != waitingTasks.end(); ++it)  {
      SxProcUnit *procUnitPtr = (*it)->element.getProcUnitPtr ();
      if (procUnitPtr == procUnit_)  {
         return true;
      }
   }
            
   return false;
}

// lock mutex before calling
bool SxThreadPool::containsRunning (const SxProcUnit *procUnit_) const
{
   SxList<SxList<SxTask>::List*>::ConstIterator it;
   for (it = runningTasks.begin(); it != runningTasks.end(); ++it)  {
      SxProcUnit *procUnitPtr = (*it)->element.getProcUnitPtr ();
      if (procUnitPtr == procUnit_)  {
         return true;
      }
   }
            
   return false;
}

void SxThreadPool::resizePool (ssize_t nThreads_)
{
   SX_DBG_MSG ("resizing thread pool from " << pool.getSize ()
               << " to " << nThreads_ << " threads");

   ssize_t nOld = pool.getSize ();
   SX_CHECK_VARS (nThreads_ >= nOld, nThreads_, nOld);

   if (nThreads_ == nOld)  {
      return;
   }

   pool.resize (nThreads_, true);

   for (ssize_t i = nOld; i < nThreads_; ++i)  {
      pool(i) = SxPtr<PoolThread>::create (this, static_cast<int>(i));
      pool(i)->start ();
   }
}

void SxThreadPool::shutdown ()
{
   SX_MUTEX (mutex)  {
      terminateCmd = true;
      condition.wakeAll ();
   }
}


#if 0
// --- handle priority for tasks

SxArray<SxList<SxTask> >  tasks;  // :priority

void SxThreadPool::submit (const SxTask &task)
{
      for (int i=0; i < 100; i++)  {
         SX_CHECK_VAR (tasks(i).getSize() == 0, tasks(i).getSize());
      }
      tasks(task.getPriority()) << task;
      nTasks++;
}

bool SxThreadPool::processNextTask (int poolId, bool useCond)
{
  // SX_MUTEX (qMutex) {
      if (nTasks > 0)  {

         SxList<SxTask>::Iterator it;
         SxArray<int> rank(tasks.getSize());
         for (int p=0; p < SxTask::MaxPriority; ++p)  {
            rank(p) = -1;

            if (tasks(p).getSize() > 0)  
               if (tasks(p).first().spoolingExpired())  {
                  tasks(p).removeFirst ();
                  nTasks--;
               }

            if (tasks(p).getSize() > 0)  
               rank(p) = tasks(p).first().updatePriority();
         }

         // --- determine task with highest rank
         int maxVal    = -1;
         int maxTaskId = -1;
         for (int p=0; p < SxTask::MaxPriority; ++p)  {
            if (rank(p) > maxVal)  {
               maxVal    = rank(p);
               maxTaskId = p;
            }
         }

         // --- spool job
         if (maxTaskId != -1)  {
            cout << "SPOOLING... " << poolId << endl;
            SxPtr<SxThread> unit = tasks(maxTaskId).first().getTask();
            tasks(maxTaskId).removeFirst ();
            nTasks--;

            pool(poolId)->thread = unit;
            if (useCond)
               SX_THREAD_SET_COND(pool(poolId)->runCond);
            cout << "SPOOLING... " << poolId << " DONE\n";
            return true;
         } else {
            cout << "NOT SPOOLING...\n";
         }
         cout.flush ();
      }
   //}
      return false;
}

#endif /* if 0 */
