// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_ALGORITHM_H_
#define _SX_ALGORITHM_H_

namespace SxAlgorithm  {
   /// Things that can go wrong in numerical algorithms
   enum Exception {
      /// Tried to find a root in an interval, but there is none
      NoZeroInInterval,
      /// Some numerical limit is reached
      NumericalLimit
   };
}

/// Numerical limits for algorithms
class SxLimits {
   public:
      /// Numerical limit
      double epsX, epsY;
      /// Constructor
      SxLimits (double epsXIn = 1e-10, double epsYIn = 1e-10)
         : epsX(epsXIn), epsY (epsYIn)
      {
         SX_CHECK_VAR (epsXIn >= 0., epsX);
         SX_CHECK_VAR (epsYIn >= 0., epsY);
      }
};

/// Specify an interval
class SxRange {
   public:
      /// Lower bound
      double from;
      /// Upper bound
      double to;
      SxRange (double fromIn, double toIn)
         : from(fromIn), to (toIn)
      {
         SX_CHECK_VARS (from <= to, from, to);
      }
};

namespace SxAlgorithm
{
   /// The base class for the Function classes
   class FunctionBase {
      protected:
         FunctionBase () {}
   };

   /// Wrapper class for double func(double)
   class DoubleFunction : public FunctionBase
   {
      protected:
         double (*funcPtr)(double);
      public:
         inline double operator () (double x) { return (*funcPtr)(x); }
         DoubleFunction (double (*fPtr)(double)) : funcPtr (fPtr) {}
   };

   /// Wrapper class for double T::func(double)
   template <class T>
   class MemberFunction : public FunctionBase
   {
      protected:
         const T& obj;
         double (T::*funcPtr)(double) const;
      public:
         inline double operator() (double x) const
         {
            return (obj.*funcPtr) (x);
         }
         MemberFunction (const T &obj_, double (T::*fPtr)(double) const)
            : obj(obj_), funcPtr(fPtr) {}
   };

   /// Wrapper class for double T::func(double, A)
   template <class T, class A>
   class MemberFunction1Arg : public FunctionBase
   {
      protected:
         T& obj;
         double (T::*funcPtr)(double, A) const;
         const A &arg;
      public:
         inline double operator() (double x) const
         {
            return (obj.*funcPtr) (x, arg);
         }
         inline MemberFunction1Arg (T& obj_,
                                    double (T::*fPtr)(double, A) const,
                                    const A& arg_)
            : obj(obj_), funcPtr(fPtr), arg(arg_)
         { /* empty */}
   };

   /// Wrapper class for double T::func(double, A, B)
   template <class T, class A, class B>
   class MemberFunction2Arg : public FunctionBase
   {
      protected:
         T& obj;
         double (T::*funcPtr)(double, A, B);
         const A &arg1;
         const B &arg2;
      public:
         inline double operator() (double x) const
         {
            return (obj.*funcPtr) (x, arg1, arg2);
         }

         inline MemberFunction2Arg (T& obj_,
                                    double (T::*fPtr)(double, A, B),
                                    const A& arg1_, const B& arg2_)
            : obj(obj_), funcPtr(fPtr), arg1(arg1_), arg2(arg2_)
         {
            // empty
         }
   };

   /// Wrapper class for double T::func(double, A, B, C)
   template <class T, class A, class B, class C>
   class MemberFunction3Arg : public FunctionBase
   {
      protected:
         T& obj;
         double (T::*funcPtr)(double, A, B, C);
         const A &arg1;
         const B &arg2;
         const C &arg3;
      public:
         inline double operator() (double x) const
         {
            return (obj.*funcPtr) (x, arg1, arg2, arg3);
         }
         inline MemberFunction3Arg (T& obj_,
                                    double (T::*fPtr)(double, A, B, C),
                                    const A& arg1_, const B& arg2_,
                                    const C& arg3_)
            : obj(obj_), funcPtr(fPtr), arg1(arg1_), arg2(arg2_), arg3(arg3_)
         {
            // empty
         }
   };
}

/// Generate a SxAlgorithm::DoubleFunction object
inline SxAlgorithm::DoubleFunction
SxFunction (double (*fPtr)(double))
{
   return SxAlgorithm::DoubleFunction (fPtr);
}

/// Generate a SxAlgorithm::MemberFunction object
template<class T>
inline SxAlgorithm::MemberFunction<T>
SxFunction (T& obj, double (T::*funcPtr)(double))
{
   return SxAlgorithm::MemberFunction<T> (obj, funcPtr);
}

/// Generate a SxAlgorithm::MemberFunction1Arg object
template<class T, class A>
inline SxAlgorithm::MemberFunction1Arg<T, A>
SxFunction (T& obj, double (T::*funcPtr)(double, A), const A& arg)
{
   return SxAlgorithm::MemberFunction1Arg<T, A> (obj, funcPtr, arg);
}

/// Generate a SxAlgorithm::MemberFunction2Arg object
template<class T, class A, class B>
inline SxAlgorithm::MemberFunction2Arg<T, A, B>
SxFunction (T& obj, double (T::*funcPtr)(double, A, B),
            const A& arg1, const B& arg2)
{
   return SxAlgorithm::MemberFunction2Arg<T, A, B> (obj, funcPtr, arg1, arg2);
}

/// Generate a SxAlgorithm::MemberFunction3Arg object
template<class T, class A, class B, class C>
inline SxAlgorithm::MemberFunction3Arg<T, A, B, C>
SxFunction (T& obj, double (T::*funcPtr)(double, A, B, C),
            const A& arg1, const B& arg2, const C& arg3)
{
   return SxAlgorithm::MemberFunction3Arg<T, A, B, C> (obj, funcPtr,
                                                       arg1, arg2, arg3);
}

/// Generate a SxAlgorithm::MemberFunction object (const)
template<class T>
inline SxAlgorithm::MemberFunction<const T>
SxFunction (const T& obj, double (T::*funcPtr)(double) const)
{
   return SxAlgorithm::MemberFunction<const T> (obj, funcPtr);
}

/// Generate a SxAlgorithm::MemberFunction1Arg object (const)
template<class T, class A>
inline SxAlgorithm::MemberFunction1Arg<const T, A>
SxFunction (const T& obj, double (T::*funcPtr)(double, A) const, const A& arg)
{
   return SxAlgorithm::MemberFunction1Arg<const T, A> (obj, funcPtr, arg);
}

/// Generate a SxAlgorithm::MemberFunction2Arg object (const)
template<class T, class A, class B>
inline SxAlgorithm::MemberFunction2Arg<const T, A, B>
SxFunction (const T& obj, double (T::*funcPtr)(double, A, B) const,
            const A& arg1, const B& arg2)
{
   return SxAlgorithm::MemberFunction2Arg<const T, A, B> (obj, funcPtr,
                                                          arg1, arg2);
}

/// Generate a SxAlgorithm::MemberFunction3Arg object (const)
template<class T, class A, class B, class C>
inline SxAlgorithm::MemberFunction3Arg<const T, A, B, C>
SxFunction (const T& obj, double (T::*funcPtr)(double, A, B, C) const,
            const A& arg1, const B& arg2, const C& arg3)
{
   return SxAlgorithm::MemberFunction3Arg<const T, A, B, C> (obj, funcPtr,
                                                             arg1, arg2, arg3);
}

// ----------------------------------------------------------------------

/** \brief Find a zero of the function in the provided interval

  @param func     a SxFunction wrapper for the function, or a functor
                  derived from SxAlgorithm::FunctionBase and having
                  the double operator() (double) overloaded.
  @param interval interval to search (see note)
  @param limit    epsY: max. function value
  @return a root of the equation |func(x)| < epsY

  @note
  The sign of the function values at the two initial interval boundaries
  must differ.

  \par Exceptions
  - NoZeroInInterval: the function does not change sign at the
                      interval boundaries
  - NumericalLimit:   the algorithm did not find the root in the given
                      numerical limits

  \example
\code
double foo(double x) { return (x*x + x - 1.); }

...
try {
   goldenRatio = findZeroInInterval (SxFunction (foo), SxRange(0., 1.));
} catch (SxAlgorithm::Exception e)  {
   // exception handling
}
\endcode
*/

template<class T>
double findZeroInInterval (const T& func, const SxRange &interval,
                           const SxLimits &limit = SxLimits(1e-10, 1e-10))
throw (enum SxAlgorithm::Exception)
{
   // check that T is derived from FunctionBase
   SX_CHECK (static_cast<const SxAlgorithm::FunctionBase *>(&func));
   SX_CHECK_VAR (limit.epsX > 0., limit.epsX);
   SX_CHECK_VAR (limit.epsY > 0., limit.epsY);

   double xMin = interval.from, xMax = interval.to;
   double yMin = func (xMin);
   double yMax = func (xMax);
   if (   (fabs(yMin) > limit.epsY)
       && (fabs(yMax) > limit.epsY)
       && (yMin * yMax > 0.))       throw SxAlgorithm::NoZeroInInterval;

   double x0, y0;
   for (int it = 200; it; --it)  {
      // this reduces the (xMin,xMax) interval by up to 60 orders of magnitude
      // if this is not sufficient, provide a better starting interval
      x0 = (xMax + xMin) / 2.;
      y0 = func (x0);

      if (fabs (y0) < limit.epsY)  return x0;

      if ( yMin * y0 > 0. )  {
         xMin = x0;
         yMin = y0;
      }  else  {
         xMax = x0;
         yMax = y0;
      }
      if (   fabs(xMin - xMax) < limit.epsX
          && fabs(yMin - yMax) < limit.epsY ) return x0;
   }
   throw SxAlgorithm::NumericalLimit;
}

/** \brief Find a zero of the function

  \note This finds some root func(x) = 0, so  you should be
  sure that there is only a single one. This is true for monotonic functions.

  The algorithm starts from an (arbitrarily chosen) interval -100 to +100,
  which is increased until the function values' signs at the boundaries differ.
  Then, findZeroInInterval is used to locate the root.

  @param func     a SxFunction wrapper for the function, or a functor
                  derived from SxAlgorithm::FunctionBase and having
                  the double operator() (double) overloaded.
  @param limit    epsY: max. function value
  @return a root of the equation func(x) = 0

  \par Exceptions
  - NoZeroInInterval: failed to find an appropriate interval
  - NumericalLimit:   the algorithm did not find the root in the given
                      numerical limits

\endcode
*/
template<class T>
double findZero (const T& func, const SxLimits &limit = SxLimits(1e-10, 1e-10))
throw (enum SxAlgorithm::Exception)
{
   // check that T is derived from FunctionBase
   SX_CHECK (static_cast<const SxAlgorithm::FunctionBase *>(&func));
   SX_CHECK_VAR (limit.epsX > 0., limit.epsX);
   SX_CHECK_VAR (limit.epsY > 0., limit.epsY);

   double xMin =-100., xMax = 100.;
   double yMin = func (xMin);
   double yMax = func (xMax);
   if ((fabs(yMin) > limit.epsY) && (fabs(yMax) > limit.epsY))  {
      int nSteps = 200; // 60 orders of magnitude
      while ((fabs(yMin) > limit.epsY) && (fabs(yMax) > limit.epsY)
             && yMin * yMax > 0. && nSteps--)  {
         // yMin and yMax on same side of y=0 axis.
         // Double the interval, keep the center...
         double x0 = 0.5 * (xMin + xMax);
         xMin = 2. * xMin - x0;
         xMax = 2. * xMax - x0;
         // ... hope that this contains the zero now
         yMin = func (xMin);
         yMax = func (xMax);
      }
      if (nSteps <= 0) throw SxAlgorithm::NoZeroInInterval;
   }
   return findZeroInInterval (func, SxRange(xMin, xMax), limit);
}


#endif /* _SX_ALGORITHM_H_ */
