// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_BOUND_PTR_H_
#define _SX_BOUND_PTR_H_

#include <SxPtr.h>
#include <SxCBoundPtr.h>
#include <SxSigSlots.h>


/** \brief Safebound pointer class

    \b SxBoundPtr = S/PHI/nX Bound Pointer

    Bound pointers is a type- and thread-safe way of handling function
    pointers in S/PHI/nX. In contrast to C function pointers this class
    is also able to point to static and even non-static member functions
    of classes.

    This class is a safe wrapper for the (dangerous) low-level class
    SxCBoundPtr.  In contrast to SxCBoundPtr this class can keep track of 
    the life time of the objects in case of pointer-to-members kind of 
    bound pointers.

\code
#include <SxBoundPtr.h>

// --- function pointer to a global function
// --- function pointer to a static member function
// --- function pointer to a non-static member function
\endcode


    \author Sixten Boeck, boeck@mpie.de */
template<class RES, 
         class P1 = SxNull, class P2 = SxNull, class P3 = SxNull,
         class P4 = SxNull, class P5 = SxNull, class P6 = SxNull
        >
class SxBoundPtr : public SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>
{
   public:

      typedef SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>  TBoundPtr;
      typedef SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6> TCBoundPtr;

   signals:

      SxSignal<void *> SX_SIGNAL (sigDeleting);

   public:

      class Receiver
      {
         public:
            Receiver (bool valid_=false) : valid (valid_) { }
            bool isValid () const { return valid; }

         public slots:
            void slotDeleteObj (void *) { printf ("DELETING\n"); valid = false; }
            
         protected:
            bool valid;
      };
      SxPtr<Receiver> receiver;

      SxBoundPtr () 
         : SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6> (),
           receiver (SxPtr<Receiver>::create (false))
      { }

     ~SxBoundPtr () { sigDeleting.send (receiver.getPtr()); }

      SxBoundPtr (const SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6> &in)
         : SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6> (in), 
           receiver (SxPtr<Receiver>::create (true))
      {  }

      // --------------------------------------------------------------------

      template<class TRES,class TP1>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (TRES (*f)(TP1))
      {
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (f)
         );
         return bPtr;
      }

      template<class TRES,class TP1,class TP2>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (TRES (*f)(TP1,TP2))
      {
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (f)
         );
         return bPtr;
      }

      template<class TRES,class TP1,class TP2,class TP3>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (TRES (*f)(TP1,TP2,TP3))
      {
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (f)
         );
         return bPtr;
      }

      template<class TRES,class TP1,class TP2,class TP3,class TP4>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (TRES (*f)(TP1,TP2,TP3,TP4))
      {
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (f)
         );
         return bPtr;
      }

      template<class TRES,class TP1,class TP2,class TP3,class TP4,class TP5>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (TRES (*f)(TP1,TP2,TP3,TP4,TP5))
      {
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (f)
         );
         return bPtr;
      }

      template<class TRES,
               class TP1,class TP2,class TP3,
               class TP4,class TP5,class TP6>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (TRES (*f)(TP1,TP2,TP3,TP4,TP5,TP6))
      {
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (f)
         );
         return bPtr;
      }

      // --------------------------------------------------------------------

      template <class Callee,class TRES,class Class,class TP1>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,TRES (Class::* const &f)(TP1))
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,class TP1,class TP2>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,TRES (Class::* const &f)(TP1,TP2))
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,
                class TP1,class TP2,class TP3>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,TRES (Class::* const &f)(TP1,TP2,TP3))
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,
                class TP1,class TP2,class TP3,class TP4>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,TRES (Class::* const &f)(TP1,TP2,TP3,TP4))
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,
                class TP1,class TP2,class TP3,class TP4,class TP5>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,
              TRES (Class::* const &f)(TP1,TP2,TP3,TP4,TP5))
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,
                class TP1,class TP2,class TP3,class TP4,class TP5,class TP6>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,
              TRES (Class::* const &f)(TP1,TP2,TP3,TP4,TP5,TP6))
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      // --------------------------------------------------------------------

      template <class Callee,class TRES,class Class,class TP1>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,TRES (Class::* const &f)(TP1)const)
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,class TP1,class TP2>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,TRES (Class::* const &f)(TP1,TP2)const)
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,
                class TP1,class TP2,class TP3>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,
              TRES (Class::* const &f)(TP1,TP2,TP3)const)
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,
                class TP1,class TP2,class TP3,class TP4>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,
              TRES (Class::* const &f)(TP1,TP2,TP3,TP4)const)
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,
                class TP1,class TP2,class TP3,class TP4,class TP5>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,
              TRES (Class::* const &f)(TP1,TP2,TP3,TP4,TP5)const)
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      template <class Callee,class TRES,class Class,
                class TP1,class TP2,class TP3,class TP4,class TP5,class TP6>
      static SxBoundPtr<RES,P1,P2,P3,P4,P5,P6>
      create (const SxPtr<Callee> &c,
              TRES (Class::* const &f)(TP1,TP2,TP3,TP4,TP5,TP6)const)
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> bPtr;
         bPtr = SxBoundPtr<RES,P1,P2,P3,P4,P5,P6> (
            SxCBoundPtr<RES,P1,P2,P3,P4,P5,P6>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }


      // --------------------------------------------------------------------

      operator int () const
      {
         return receiver->isValid() && TCBoundPtr::operator int();
      }

      // --------------------------------------------------------------------


      RES operator() () const 
      {
         SX_CHECK (receiver->isValid()); // object has been deleted meanwhile!
         return TCBoundPtr::operator() ();
      }

      RES operator() (P1 p1) const 
      {
         SX_CHECK (receiver->isValid()); // object has been deleted meanwhile!
         return TCBoundPtr::operator() (p1);
      }


      RES operator() (P1 p1, P2 p2) const 
      {
         SX_CHECK (receiver->isValid()); // object has been deleted meanwhile!
         return TCBoundPtr::operator() (p1, p2);
      }

      RES operator() (P1 p1, P2 p2, P3 p3) const 
      {
         SX_CHECK (receiver->isValid()); // object has been deleted meanwhile!
         return TCBoundPtr::operator() (p1, p2, p3);
      }

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4) const 
      {
         SX_CHECK (receiver->isValid()); // object has been deleted meanwhile!
         return TCBoundPtr::operator() (p1, p2, p3, p4);
      }

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4, P5 p5) const 
      {
         SX_CHECK (receiver->isValid()); // object has been deleted meanwhile!
         return TCBoundPtr::operator() (p1, p2, p3, p4, p5);
      }

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4, P5 p5, P6 p6) const 
      {
         SX_CHECK (receiver->isValid()); // object has been deleted meanwhile!
         return TCBoundPtr::operator() (p1, p2, p3, p4, p5, p6);
      }
};



// SPECIAL CASE: void

template<class RES>
class SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull> 
   : public SxCBoundPtr<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
{
   public:

      typedef SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull>  TBoundPtr;
      typedef SxCBoundPtr<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull> TCBoundPtr;

   signals:

      SxSignal<void *> SX_SIGNAL (sigDeleting);

   public:

      class Receiver
      {
         public:
            Receiver (bool valid_=false) : valid (valid_) { }
            bool isValid () const { return valid; }

         public slots:
            void slotDeleteObj (void *) { printf ("DELETING\n"); valid = false; }
            
         protected:
            bool valid;
      };
      SxPtr<Receiver> receiver;

      SxBoundPtr () 
         : SxCBoundPtr<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull> (),
           receiver (SxPtr<Receiver>::create (false))
      { }

     ~SxBoundPtr () { sigDeleting.send (receiver.getPtr()); }

      SxBoundPtr (const SxCBoundPtr<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull> &in)
         : SxCBoundPtr<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull> (in), 
           receiver (SxPtr<Receiver>::create (true))
      {  }

      // --------------------------------------------------------------------

      template<class TRES>
      static SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull>
      create (TRES (*f)(void))
      {
         SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull> bPtr;
         bPtr = SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull> (
            SxCBoundPtr<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>::create (f)
         );
         return bPtr;
      }

      // --------------------------------------------------------------------

      template <class Callee,class TRES,class Class>
      static SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull>
      create (const SxPtr<Callee> &c,TRES (Class::* const &f)(void))
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull> bPtr;
         bPtr = SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull> (
            SxCBoundPtr<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }

      // --------------------------------------------------------------------

      template <class Callee,class TRES,class Class>
      static SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull>
      create (const SxPtr<Callee> &c,TRES (Class::* const &f)(void)const)
      {
         SX_CHECK (c.getPtr());
         SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull> bPtr;
         bPtr = SxBoundPtr<RES,void,SxNull,SxNull,SxNull,SxNull,SxNull> (
            SxCBoundPtr<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>::create (*c.getPtr(), f)
         );
         c.receiver->sigDeleting.connect (bPtr.receiver.getPtr(), 
                                 &TBoundPtr::Receiver::slotDeleteObj);
         bPtr.sigDeleting.connect (c.receiver, 
                                   &SxPtr<Callee>::Receiver::slotDeleting);
         return bPtr;
      }


      // --------------------------------------------------------------------

      operator int () const
      {
         return receiver->isValid() && TCBoundPtr::operator int();
      }

      // --------------------------------------------------------------------


      RES operator() () const 
      {
         SX_CHECK (receiver->isValid()); // object has been deleted meanwhile!
         return TCBoundPtr::operator() ();
      }

};



#endif /* _SX_BOUND_PTR_H_ */
