// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_CALC_H_
#define _SX_CALC_H_

#include <SxString.h>
#include <SxArray.h>
#include <SxUtil.h>

#ifdef WIN32
#  include <sxfileno.h>
#endif

/** \brief Simple pocket calculator

    \b SxCalc = S/PHI/nX Pocket Calculator

    SxCalc provides a simple "pocket" calculator. It is able to convert an
    expression given as SxString to a double. Furthermore, variables (such as
    pi) can be used. In order to speed up the parsing/evaluation performance
    instead of strings variable names are given by numbers, such as 1st,
    2nd, 3rd variable. 

    \par Examples:
\code
   // --- setup a view variables
   SxList<double> varList; varList << 3.14 << 6.28 << -50.0;
   SxCalc calc (varList);

   // --- compile expression: using $0 := pi, $2 := -50
   SxString expr = "(10e5 + $2) * sin(0.5 * $0) / sqrt(27.)";

   cout << "result = " << calc.eval (expr + ";") << endl;
\endcode
   \author Sixten Boeck, boeck@mpie.de
   \ingroup group_utils */
class SX_EXPORT_UTIL SxCalc
{
   public:

      /** \brief Initialize a pocket calcuator without built-in variables 
        */
      SxCalc ();

      /** \brief Initialize a pocket calcuator with built-in variables 
        */
      SxCalc (const SxArray<double> &);
     ~SxCalc ();

      /** \brief Evaluate an expression

          Call this function to evaluate the expression. Make sure that 
          the expression is terminated with a ';'. */
      double eval (const SxString &);

      /** \brief Returns the currently evaluated expression.
        */
      SxString getExpr () const;

      /** \brief Returns the value of the i-th built-in variable */
      double getVar (int i) const;


#     ifdef WIN32
         static double cbrt (double);
#     endif

   protected:

      SxString buffer;
      SxArray<double> vars;
};

extern void   SxCalc_init ();
extern void   SxCalc_error (const SxString &);
extern void   SxCalc_parseExpr (const SxString &);
extern int    SxCalc_lex ();
extern int    SxCalc_pos;
extern double SxCalc_res;
extern SxCalc *sxCalc;

#endif /* _SX_CALC_H_ */
