// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxError.h>
#include <SxUtil.h>
#include <SxTime.h>
#include <SxRedirect.h>
#include <stdio.h>
#include <iostream>
#include <errno.h>

#include <stdarg.h>

#ifdef SX_ANDROID
#  include <android/log.h>
#endif /* SX_ANDROID */

typedef void (*SxExitCB)(void) ATTR_NO_RETURN;
static SxExitCB sxExitCB = NULL;

void ATTR_NO_RETURN sxExitExceptionCB (void) throw (SxExit)
{
   throw SxExit(); 
}

void sxExitException ()
{
   sxExitCB = sxExitExceptionCB;
}

void sxExitAbort ()
{
   sxExitCB = NULL;
}

void sxExit (int) throw (SxExit)
{
   if (sxExitCB)  (*sxExitCB)(); 
   else  {
      fflush (stdout);
      fflush (stderr);
      std::cout.flush ();
      std::cerr.flush ();
#ifdef NDEBUG
      exit (1);
#else
      abort ();
#endif
   }
   
#ifdef SX_IOS
   __builtin_unreachable();
#endif /* SX_IOS */
}

void sxQuit () throw (SxExit)
{
   if (sxExitCB)  (*sxExitCB)();
   else {
      exit (1);
   }
   
#ifdef SX_IOS
   __builtin_unreachable();
#endif /* SX_IOS */
}

void sxOutOfMemoryHandler ()
{
   sxprintf ("\nSPHInX: memory allocation failed.\n");
   sxExit (2);
}

void sxUncaughtExceptionHandler ()
{
   sxprintf ("\nSPHInX: unexpected exception\n");
   sxExit (3);
}


SxString sxstrerror ()
{
#  ifdef WIN32
      return sxstrerror (GetLastError ());
#  else
      return sxstrerror (errno);
#  endif /* WIN32 */
}

SxString sxstrerror (int errnum)
{
   SxString msg;
#  ifdef WIN32
      char buf[1024];
      if (::strerror_s (buf, sizeof (buf), errnum) != 0)  msg = "Unknown error";
      else  msg = SxString(buf);
#  else
#     ifdef HAVE_STRERROR_R
         char buf[1024]; // message is truncated if buf is too small
#        if defined(STRERROR_R_CHAR_P) || defined(__USE_GNU)
            char const *str = strerror_r (errnum, buf, sizeof (buf));
            if (str)  msg = str;
            else      msg = "Unknown error";
#        else
            strerror_r (errnum, buf, sizeof (buf));
            msg = buf;
#        endif /* STRERROR_R_CHAR_P */
#     else
         msg = strerror (errnum);
#     endif /* HAS_STRERROR_R */
#  endif /* WIN32 */

   return msg;
}

#ifdef WIN32
SxString sxstrerror (DWORD sysErrCode)
{
   LPVOID lpMsgBuf;
   DWORD ret = FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER |
                              FORMAT_MESSAGE_FROM_SYSTEM |
                              FORMAT_MESSAGE_IGNORE_INSERTS,
                              NULL,
                              sysErrCode,
                              MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                              (LPTSTR) &lpMsgBuf,
                              0, NULL);
   if (ret)  {
      SxString msg = SxString((LPCSTR)lpMsgBuf, static_cast<int>(ret));
      LocalFree (lpMsgBuf);
      return msg;
   }  else  {
      return "FormatMessage () error";
   }
}
#endif /* WIN32 */

#ifdef SX_IOS
SxString sxstrerror (CFErrorRef error)
{
   SxString result;
   if (error != NULL) {
      CFIndex errorCode = CFErrorGetCode(error);
      if (errorCode != 0) {
         result = "error " + SxString(errorCode);
         CFStringRef desc = CFErrorCopyDescription(error);
         if (desc) {
            result += ": " + SxString(desc);
            CFRelease(desc);
         }
      }
   }
   return result;
}
#endif /* SX_IOS */

SxTraceLog::SxTraceLog (int        level_,
                        const char *tag_,
                        const char *function_)
: level(level_)
{
   strncpy (tag, tag_, 79);
   tag[79] = '\0';

   strncpy (functionName, function_, 79);
   functionName[79] = '\0';

   sxlog (level, tag, functionName, "BEGIN");

   t1 = SxTime::getRealTime ();
}
   
SxTraceLog::~SxTraceLog ()
{
   t2 = SxTime::getRealTime ();
   float dt = static_cast<float>((t2 - t1) * 1.0e3);
   sxlog (level, tag, functionName, "END %.3f", dt);
}

void sxlog (int        level_,
            const char *tag_,
            const char *function_,
            const char *fmt_, ...)
{
   if (fmt_)  {
#     ifdef SX_ANDROID
         int level = 0;
         switch (level_) {
            case 2:
               level = ANDROID_LOG_VERBOSE;
               break;
            case 3:
               level = ANDROID_LOG_DEBUG;
               break;
            case 4:
               level = ANDROID_LOG_INFO;
               break;
            case 5:
               level = ANDROID_LOG_WARN;
               break;
            case 6:
               level = ANDROID_LOG_ERROR;
               break;
         }
#     else
         SX_UNUSED(level_);
#     endif /* SX_ANDROID */

      va_list arg;
      va_start (arg, fmt_);
#     ifdef SX_ANDROID
         __android_log_vprint (level, tag_, fmt_, arg);
#     else
         sxprintf ("___%s: %s: ", tag_, function_);
         vfprintf (stdout, fmt_, arg);
         sxprintf ("\n");
#     endif /* SX_ANDROID */
      va_end (arg);
   }
}

// This construct installs the above handlers for
// - new_handler: new failed to allocate memory
// - unexpected : an exception was not caught
// before the main program starts
static class SxInstallHandlers {
   public:
      SxInstallHandlers ()  {
         std::set_new_handler (sxOutOfMemoryHandler);
         std::set_unexpected (sxUncaughtExceptionHandler);
      }
} theHandlerInstaller;

#ifndef NDEBUG
SxTrace::SxTrace (int        enabled_,
                  const char *filename_,
                  const char *function_,
                  const char *functionShort_)
: enabled (enabled_)
{
   if (enabled > 1)  {
      strncpy (filename, filename_, 79);
      filename[79] = '\0';
      strncpy (functionName, function_, 79);
      functionName[79] = '\0';
      strncpy (functionNameShort, functionShort_, 79);
      functionNameShort[79] = '\0';

      std::cout << "___" << filename <<" "<< functionNameShort <<": BEGIN"
                << std::endl;
      std::cout.flush ();

      // --- wall clock time 1
      t1 = SxTime::getRealTime ();
   }
}
   
   
SxTrace::~SxTrace ()
{
   if (enabled > 1)  {
      // --- wall clock time 2
      t2 = SxTime::getRealTime ();

      // --- Print time spend in procedure in milliseconds.
      //     general performance is in digit place ([xy]ms vs [z]ms)
      std::cout << "___"
                << filename << " " << functionNameShort << ": END " 
                << std::fixed << std::setprecision(3) <<
                   ((t2 - t1) * 1.0e3) << " ms" << std::endl;

      std::cout.flush ();
   }
}

#endif
