// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------
#ifndef _SX_ERROR_H_
#define _SX_ERROR_H_

#include <SxConfig.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <iostream>
#include <iomanip>
#include <signal.h>
#include <SxException.h>
#include <SxUtil.h>

#ifdef WIN32
#  define _WINSOCKAPI_ // prevent winsock.h to be included from windows.h
                       // otherwise later include of winsock2.h would show
                       // redefinition errors
#  include <windows.h> // DWORD, GetLastError()
#endif /* WIN32 */

#ifdef SX_IOS
// CFErrorRef, CFStringRef, CFStringGetLength(), CFErrorGetCode(), ...
#  include <CoreFoundation/CoreFoundation.h>
#endif /* SX_IOS */

// ATTR_NO_RETURN indicates that the function never returns
// this removes many false alarms about possibly uninitialized vars
#ifdef ATTR_NO_RETURN
#  error ATTR_NO_RETURN macro name collision - find a better name!
#endif
#ifdef WIN32
#  define ATTR_NO_RETURN
#  define DECL_NO_RETURN __declspec(noreturn)
#else
#  define DECL_NO_RETURN
#  ifdef __GNUC__
#    define ATTR_NO_RETURN __attribute__ ((noreturn))
#  else
#    define ATTR_NO_RETURN
#  endif
#endif


class SX_EXPORT_UTIL SxExit : public SxException { };
extern void SX_EXPORT_UTIL DECL_NO_RETURN sxExit (int) throw (SxExit) ATTR_NO_RETURN;
extern void SX_EXPORT_UTIL DECL_NO_RETURN sxQuit () throw (SxExit) ATTR_NO_RETURN;

void sxExitException ();
void sxExitAbort ();

class SxCatchExit 
{
   public:
      inline SxCatchExit ()  { sxExitException (); }
      inline ~SxCatchExit () { sxExitAbort (); }

};

#define CATCH_SX_EXIT SxCatchExit sxCatchExitl

// --- GCC >= 2.0 provides these two magic variables which hold the name
//     of the current function, as a string.
#define SX_FUNCTION __FUNCTION__
#ifdef WIN32
#  define SX_PRETTY_FUNCTION __FUNCTION__  // __FUNCSIG__
#else
#  define SX_PRETTY_FUNCTION __PRETTY_FUNCTION__
#endif

#ifdef NDEBUG
#   define SX_CHECK(expr)                      ((void)0)
#   define SX_CHECK_VAR(expr,var)              ((void)0)
#   define SX_CHECK_VARS(expr,var1,var2)       ((void)0)
#   define SX_CHECK_3VARS(expr,var1,var2,var3) ((void)0)
#   define SX_CHECK_RANGE(idx,size)            ((void)0)
#   define SX_BREAK                            ((void)0)
#   define SX_EXIT                                                           \
         { std::cout << "\n+---------------------------------------------"   \
                     << "--------------------------------\n";                \
           std::cout << "| Abnormal program stop in file " << __FILE__       \
                     << " at line " << __LINE__ << ".\n";                    \
           std::cout <<"+---------------------------------------------"      \
                     << "--------------------------------\n";                \
           sxExit(1); }                                                       
#   define SX_EXIT_INFO  SX_EXIT
#   define SX_QUIT                                                           \
         { std::cout << "\n+---------------------------------------------"   \
                     << "--------------------------------\n";                \
           std::cout << "| Program stopped in file " << __FILE__             \
                     << " at line " << __LINE__ << "\n"                      \
                     << "| Invalid or inconsistent input as described "      \
                     << "above.\n";                                          \
           std::cout << "+---------------------------------------------"     \
                     << "--------------------------------\n";                \
           sxQuit (); }
#   define SX_DEPRECATE(msg)                  ((void)0)
#   define SX_TRACE()                         ((void)0)
#   define SX_DBG_MSG(msg)                    ((void)0)

#else  /* NDEBUG */                                                         
#   define PROMPT_ERR0(expr)                                                 \
           std::cout << std::endl << "ASSERTATION FAILED in "                \
                     << __FILE__ << ", line " << __LINE__ << "!\n";          \
           std::cout << #expr "\n\n";
#   define PROMPT_ERR1(expr,var)                                             \
           std::cout << std::endl << "ASSERTATION FAILED in "                \
                     << __FILE__ << ", line " << __LINE__ << "!\n";          \
           std::cout << #expr << ", " << #var << "=" << (var) << "\n\n";
#   define PROMPT_ERR2(expr,var1,var2)                                       \
           std::cout << std::endl << "ASSERTATION FAILED in "                \
                     << __FILE__ << ", line " << __LINE__ << "!\n";          \
           std::cout << #expr << ", " << #var1 << "=" << (var1)              \
                              << ", " << #var2 << "=" << (var2) << "\n\n";
#   define PROMPT_ERR3(expr,var1,var2,var3)                                  \
           std::cout << std::endl << "ASSERTATION FAILED in "                \
                     << __FILE__ << ", line " << __LINE__ << "!\n";          \
           std::cout << #expr << ", " << #var1 << "=" << (var1)              \
                              << ", " << #var2 << "=" << (var2)              \
                              << ", " << #var3 << "=" << (var3) << "\n\n";
#   define SX_EXIT_INFO(line)                                                \
         { std::cout << "\n+---------------------------------------------"   \
                     << "--------------------------------\n";                \
           std::cout << "| Fatal program stop in file " << __FILE__          \
                     << " at line "<<line<<".\n" \
                     << "| Function: " << SX_PRETTY_FUNCTION << "\n";        \
           std::cout <<"+---------------------------------------------"      \
                     << "--------------------------------\n";                \
           sxExit(1);          /* dump core file */                          \
        }    
#   define SX_EXIT       SX_EXIT_INFO(__LINE__)                       
#   define SX_CHECK(expr)                                                    \
          if ( !(expr) ) { PROMPT_ERR0(expr); SX_EXIT; }                       
#   define SX_CHECK_VAR(expr,var)                                            \
          if ( !(expr) ) { PROMPT_ERR1(expr,var); SX_EXIT; }                   
#   define SX_CHECK_VARS(expr,var1,var2)                                     \
          if ( !(expr) ) { PROMPT_ERR2(expr,var1,var2); SX_EXIT; }             
#   define SX_CHECK_3VARS(expr,var1,var2,var3)                               \
          if ( !(expr) ) { PROMPT_ERR3(expr,var1,var2,var3); SX_EXIT; }
#   define SX_CHECK_RANGE(idx,size)                                          \
          SX_CHECK_VARS((idx) >= 0 && (idx) < (size), idx, size)
#   define SX_BREAK                                                          \
        { std::cout << "Hit <Enter> to continue";                            \
           std::cout.flush ();                                               \
          if (getchar() == 'c')  SX_EXIT;  }
#   define SX_QUIT                                                           \
         { std::cout << "\n+---------------------------------------------"   \
                     << "--------------------------------\n";                \
           std::cout << "| Program stopped in file " << __FILE__             \
                     << " at line " << __LINE__ << ".\n";                    \
           std::cout <<"+---------------------------------------------"      \
                     << "--------------------------------\n";                \
           sxQuit (); }
#   define SX_DEPRECATE(msg)                                                 \
         { std::cout << "\n+---------------------------------------------"   \
                     << "--------------------------------\n";                \
           std::cout << "| Deprecate function call found in file "<<__FILE__ \
                     << " at line " << __LINE__ << ".\n"                     \
                     << "| " << msg << endl;                                 \
           std::cout <<"+---------------------------------------------"      \
                     << "--------------------------------\n";                \
           sxQuit (); }

class SX_EXPORT_UTIL SxTrace
{
   public:
      SxTrace (int        enabled_,
               const char *filename_,
               const char *function_,
               const char *functionShort_);
      ~SxTrace ();
      
   protected:
      int enabled;
      char filename[80];
      char functionName[80];
      char functionNameShort[80];
      double t1;
      double t2;
};

// Function-like Macro SX_TRACE ();
// http://gcc.gnu.org/onlinedocs/cpp/Function_002dlike-Macros.html
#   define SX_TRACE()                                                         \
    static int sxTraceEnabled = sxIsDebugEnabled (__SX_FILE__,SX_FUNCTION,NULL); \
    SxTrace sxTrace (sxTraceEnabled,__SX_FILE__,SX_PRETTY_FUNCTION,SX_FUNCTION);
    
#   define SX_DBG_MSG(msg)                                                    \
    {                                                                         \
       static int debugEnabled = sxIsDebugEnabled(__SX_FILE__,SX_FUNCTION,NULL); \
       if (debugEnabled == 1)  {                                              \
          std::cout << "___" << msg << std::endl;                             \
          std::cout.flush ();                                                 \
       }  else if (debugEnabled == 2)  {                                      \
          std::cout << "___"                                                  \
                    << __SX_FILE__<<" "<<SX_FUNCTION<< ": " << msg << std::endl; \
          std::cout.flush ();                                                 \
       }                                                                      \
    }

#endif /* NDEBUG */

class SX_EXPORT_UTIL SxTraceLog
{
   public:
      SxTraceLog (int        level,
                  const char *tag,
                  const char *function);
      ~SxTraceLog ();
      
   protected:
      int level;
      char tag[80];
      char functionName[80];
      double t1;
      double t2;
};

#  define __SX_FILE__ (strrchr(__FILE__, '/') ? strrchr(__FILE__, '/')+1 : __FILE__)

#  define SX_TRACED(LOG_TAG) SxTraceLog sxTrace(3,LOG_TAG,SX_PRETTY_FUNCTION)

// Variadic macro
//    C99 and C++11 standard
//    gcc, Visual Studio support
//    sxcompflags.m4 -Wno-variadic-macros gcc option to hide warnings
//    http://en.wikipedia.org/wiki/Variadic_macro
#  define SX_LOGV(LOG_TAG,...) sxlog(2, LOG_TAG, SX_FUNCTION, __VA_ARGS__)
#  define SX_LOGD(LOG_TAG,...) sxlog(3, LOG_TAG, SX_FUNCTION, __VA_ARGS__)
#  define SX_LOGI(LOG_TAG,...) sxlog(4, LOG_TAG, SX_FUNCTION, __VA_ARGS__)
#  define SX_LOGW(LOG_TAG,...) sxlog(5, LOG_TAG, SX_FUNCTION, __VA_ARGS__)
#  define SX_LOGE(LOG_TAG,...) sxlog(6, LOG_TAG, SX_FUNCTION, __VA_ARGS__)

SX_EXPORT_UTIL void sxlog (int        level,
                           const char *tag,
                           const char *function,
                           const char *fmt, ...);


#define SVN_HEAD                                                              \
     if (BRANCH == SVN_STABLE_ID)  {                                          \
        std::cout << "Feature has not been approved for the STABLE branch.\n";\
        SX_EXIT; }                                                            \
     if (BRANCH == SVN_RELEASE_ID)  {                                         \
        std::cout << "Feature not yet available.\n";                          \
        SX_EXIT; }

#define SVN_STABLE                                                            \
     if (BRANCH == SVN_RELEASE_ID)  {                                         \
        std::cout << "This feature is not yet available.\n";                  \
        SX_EXIT; }

/// Check error code
#define SX_CHECK_ERR(expr)                                                    \
    {                                                                         \
       int errcode = expr;                                                    \
       if (errcode < 0)  {                                                    \
          int err = errno;                                                    \
          std::cout << "ERROR CODE " << errcode << " in "                     \
                    << __FILE__ << ", line " << __LINE__ << "!\n";            \
          if (errcode == -1)  errcode = err;                                  \
          std::cout << strerror (errcode) << std::endl;                       \
                                                                              \
          SX_EXIT;                                                            \
       }                                                                      \
    }

class SxString;
/** Get an error message from a Windows-specific system error code resp. a POSIX
    errno value. Beware that intermediate library calls, syscalls etc. may
    overwrite the values. It is better to store the value immediately after a
    failure in a local variable 'err' and call sxstrerror (err). */
extern SxString SX_EXPORT_UTIL sxstrerror ();
/** Get an error message from an errno error code (formally errno_t) */
extern SxString SX_EXPORT_UTIL sxstrerror (int errnum);

#ifdef WIN32
/** Get an error message for a Windows-specific system error code (last-error
    code;
    <http://msdn.microsoft.com/en-us/library/ms680347(v=vs.85).aspx>,
    <http://msdn.microsoft.com/en-us/library/ms681381(v=vs.85).aspx>)
\code
   BOOL status = WriteFile (...);
   if (!status)  {
      DWORD err = ::GetLastError ();
      cout << "ERROR: " << sxstrerror (err) << endl;
   }
\endcode
*/
extern SxString SX_EXPORT_UTIL sxstrerror (DWORD sysErrCode);
#endif /* WIN32 */

#ifdef SX_IOS
extern SxString SX_EXPORT_UTIL sxstrerror (CFErrorRef error);
#endif /* SX_IOS */

/// Remove "unused parameter" warnings
#define SX_UNUSED(var) (void)var;

/// This auxiliary function states that we ran out of memory and sxExits
extern void SX_EXPORT_UTIL DECL_NO_RETURN sxOutOfMemoryHandler () ATTR_NO_RETURN;

#endif /* _SX_ERROR_H_ */

/**
   \def SX_CHECK_RANGE
   SX_CHECK_RANGE has no effect in release mode.
   \par Example:
\code
void set (int idx, char x)
{
   const int BUFFER_LEN = 10;
   static char buffer[BUFFER_LEN];
   
   SX_CHECK_RANGE (idx, BUFFER_LEN);
   buffer[idx] = x;
}

int main (int, char **)
{
   set (0, 'a');
   set (11, 'x');
   
   return 0;
}
\endcode

   \par Example output:
\verbatim
ASSERTATION FAILED in main.cpp, line 6!
idx>=0 && idx<BUFFER_LEN, idx=11, BUFFER_LEN=10
\endverbatim */

/**
   \def SX_TRACE
   SX_TRACE has no effect in release mode.
   \par Example file main.cpp:
\code
int main (int, char **)
{
   SX_TRACE ();
   
   sxprintf ("abc\n");
   
   return 0;
}
\endcode

   \par Example output:
\verbatim
main.cpp main: BEGIN
abc
main.cpp main: END 1.359 ms
\endverbatim

   \par Application:
\code
void foo (void)
{
   SX_TRACE ();
   // ...
}

// instead of:
void foo (void)
{
   sxprintf ("in foo ()\n");
   
   // ...
   
   sxprintf ("foo done\n");
}
\endcode */

/**
   \def SX_DBG_MSG
   SX_DBG_MSG has no effect in release mode.
   \par Example file main.cpp:
\code
int main (int argc, char **)
{
   SX_DBG_MSG ("argc=" << argc);
   return 0;
}
\endcode

   \par Example output:
\verbatim
main.cpp main: argc=1
\endverbatim */
