// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

// note1:
//    The SxException class is used by almost all other service classes.
//    Hence, here we cannot use those service classes (like SxString, SxArray).
//    That's why here conventional C memory management has to be applied.


#include <SxException.h>
#include <SxString.h>
#include <SxList.h>

bool SxException::dumpCoreFile = false;


SxException::SxException ()
   : message(NULL),
     type (SxException::Unknown),
     senderFile (NULL),
     senderLine (-1),
     next(NULL)
{
   message = new char [1];  // see note1
   message[0] = '\0';
}

SxException::SxException (const SxException &in)
   : message(NULL),
     type (in.type),
     senderFile (NULL),
     senderLine (in.senderLine),
     next(NULL)
{
   copy (in);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}


SxException::SxException (const char *message_,
                          const char senderFile_[], int senderLine_)
   : message(NULL),
     type (SxException::Unknown),
     senderFile (NULL),
     senderLine (senderLine_),
     next(NULL)
{
   init (SxException::Unknown, message_, senderFile_, senderLine_);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}


SxException::SxException (int type_, const char *message_,
                          const char senderFile_[], int senderLine_)
   : message(NULL),
     type (type_),
     senderFile (NULL),
     senderLine (senderLine_),
     next(NULL)
{
   init (type_, message_, senderFile_, senderLine_);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}

SxException::SxException (const SxException &chain_,
                          int type_, const char *message_,
                          const char senderFile_[], int senderLine_)
   : message(NULL),
     type (type_),
     senderFile (NULL),
     senderLine (senderLine_),
     next(NULL)
{
   init (type_, message_, senderFile_, senderLine_);

   next = new SxException();
   next->copy (chain_);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}

SxException::~SxException ()
{
   destroy ();
}

SxException &SxException::operator= (const SxException &in)
{
   if (&in != this)  {
      copy (in);
   }
   return *this;
} 

void SxException::init (const SxException &in)
{
   init (in.type, in.message, in.senderFile, in.senderLine);
}

void SxException::init (int        type_,
                        const char *message_, 
                        const char *senderFile_,
                        int        senderLine_)
{
   if (message_)  {
      message = new char [strlen (message_)+1];  // see note1
      strcpy (message, message_);
   }
   if (senderFile_)  {
      senderFile = new char [strlen (senderFile_)+1];  // see note1
      strcpy (senderFile, senderFile_);
   }
   type       = type_;
   senderLine = senderLine_;
}

void SxException::destroy ()
{
   if (message)  {
      delete [] message;
      message = NULL;
   }
   if (senderFile)  {
      delete [] senderFile;
      senderFile = NULL;
   }
   if (next)  {
      delete next;
      next = NULL;
   }
}

void SxException::copy (const SxException &in)
{
   destroy ();
   init (in);

   SxException *dst = this;
   SxException *src = in.next;
   while (src)  {
      dst->next = new SxException();
      dst->next->init (*src);
      dst = dst->next;
      src = src->next;
   }
}

void SxException::causeSegFault ()
{
//#  ifdef NDEBUG
//      std::cout ("Exceptions can be traced only in the SFHIngX DEBUG mode.\n");
//      dumpCoreFile = false;
//#  else
      std::cout <<  "Exceptions will cause segmentation faults and "
              "dump a core file.\n";
      dumpCoreFile = true;
//#  endif /* NDEBUG */
}


void SxException::throwExceptions ()
{
   dumpCoreFile = false;
}

void SxException::print (bool printSender) const
{
   const SxException *ptr = this;
   while (ptr)  {
      std::cout << std::endl << ptr->message << std::endl;
      if (printSender)
         std::cout << "This Exception was emitted from "
                   << ptr->senderFile
                   << " (line " << ptr->senderLine << ")\n";

      ptr = ptr->next;
   }
}

SxString SxException::toString (bool printSender_,
                                const char delimiter_[]) const
{
   SxList<SxString> list;

   const SxException *ptr = this;
   while (ptr)  {
      SxString str;
      if (printSender_)  {
         str = ptr->senderFile;
         ssize_t i = str.getSize () - 1;
         while (i >= 0 && str(i) != '\\' && str(i) != '/') i--;
         if (i >= 0) str = str.subString (i + 1);
         str = " (" + str + ":" + SxString(ptr->senderLine) + ")";
      }
      str = ptr->message + str;
      list << str;
      ptr = ptr->next;
   }

   return SxString::join (list, delimiter_);
}

