// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_EXCEPTION_BASE_H_
#define _SX_EXCEPTION_BASE_H_

#include <SxUtil.h>

extern bool dumpCoreFile;

class SxString;

/** \brief SFHIngX Exception class

    \b SxException = SFHIngX Exception Handling Class

    This is a standard exception class for all SFHIngX exceptions. Exception
    handling in C++ can cause spurious bus errors when they are not caught
    properly in a try-catch block. If SFHIngX aborts with a bus error which
    could be related to such a missing try-catch block this class allows
    to handle the exception with an SFHIngX SX_EXIT statement instead. In the
    DEBUG mode the resulting core file can be analyzed (\ref dbg_analysis).

    \author Sixten Boeck
  */
class SX_EXPORT_UTIL SxException
{
   public:

      /// \brief Exception types
      enum Type  {
         /// unknown exception
         Unknown = 0,
         /** Interruption due to timeout signal. */
         TimeOut,
         // --- SxSymbolTable
         /** The specified group could not be found in the input file
             thrown by SxSymbol or SxSymbolTable */
         MissingGroup,   
         /** The specified symbol could not be found in the input file
             thrown by SxSymbol or SxSymbolTable */
         MissingSymbol,
         /** netCDF I/O error */
         IoError,
         // --- SxFile
         NoPermission,
         NoSuchFile,
         FileExists,
         // --- SxString
         ConversionFailed,
         RegExpFailed,
         // --- SxProcess
         /** Process could not be forked. This usually happens when memory
             is not sufficient to copy process. */
         ForkFailed,
         /** Process could not be started. */
         ExecFailed,
         /** Could not create pipe */
         PipeFailed
      };
      

      /// \brief Standard constructor
      SxException ();
      /// \brief Copy constructor
      SxException (const SxException &);
      SxException (const char *message, 
                   const char senderFile[]="unknown", int senderLine=0);
      /** \brief Constructor.
 
          This is the usual way to generate an exception.
          \code
             if (...) 
                throw SxException (SxException::IO_ERROR,
                                   "File not found", __FILE__, __LINE__);
          \endcode
        
          \param type        the enumeration type of the exception
          \param message     the complete message string
          \param senderFile  Ususally the cpp string constant __FILE__
          \param senderLine  Usually the cpp line number __LINE__ */
      SxException (int type_, const char *message, 
                   const char senderFile[]="unknown", int senderLine=0);

      SxException (const SxException &chain, int type_, const char *message,
                   const char senderFile[]="unknown", int senderLine=0);

      /// \brief Destructor
      ~SxException ();

      /// \brief assignment operator
      SxException &operator= (const SxException &);

      /** \brief Print the message body

          if \em printSender is true additionally the file and the line
          from where this exception was trown will be printed. */
      void print (bool printSender=false) const;

      SxString toString (bool printSender=false,
                         const char delimiter[]=": ") const;


      /// \brief returns the message body.
      const char *getMessage () const  { return message; }

      /** \brief Return type of the exception */
      int getType () const { return type; }

      /// \brief returns the sender file.
      const char *getFileName () const  { return senderFile; }

      /// \brief returns the sender line.
      int getLineNumber () const { return senderLine; }


      /// \brief returns next exception in the chain.
      const SxException *getNext () const { return next; }


      /** \brief Treat all exception as SX_EXIT statements.

          Usually an uncaught exception causes an "Abort" signal. So no
          core file will be prompted and backtracing in the debug mode 
          becomes impossible.
          By calling this static function every exception violates the
          memory explicily in order to dump a qualified core file.
          For debugging only
          \sa SxException::throwExceptions
          \sa SxException::dumpCoreFile */ 
      static void causeSegFault   ();
      /** \brief Treat all exceptions normally (no SX_EXIT emulation)
        
          For debugging only. Switch off dumping of core files rather
          than thwing exceptions.
          \sa SxException::causeSegFault
          \sa SxException::dumpCodeFile */
      static void throwExceptions ();

      /** Program global variable.

          If SxException::dumpCoreFile is set to true all exception are
          treated with a SFHIngX SX_EXIT statement instead. This allows the
          developer to trace missing try-catch blocks.

          This variable shouldn't be set directly. Use better
          SxException::causeSegFaults or SxException::throwExceptions.

          \code
             SxException::dumpSegFaults();
          \endcode

          \sa \ref page_debug  */
      static bool dumpCoreFile;

   protected:

      /// \brief The exception message body
      char *message;

      /// \brief Contains the exception type
      int  type;

      /// \brief Contains the file name from where the exception was trown.
      char *senderFile;

      /// \brief Contains the line from where the exception was trown.
      int senderLine;

      /// \brief Next exception in the chain
      SxException *next;

      void init (const SxException &);
      void init (int        type_,
                 const char *message, 
                 const char *senderFile,
                 int        senderLine);
      void destroy ();
      void copy (const SxException &);
};

#endif /* _SX_EXCEPTION_BASE_H_ */
