// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_FUNCTOR_H_
#define _SX_FUNCTOR_H_

#include <string.h>
#include <SxError.h>

/** \brief Empty data type to handle variadic template classes

    Use this class to define variatic template classes. For example
\code
template<class A, class B=SxNull,class C=SxNull>
class MyClass { };

template<class A>
class MyClass<A,SxNull,SxNull>
{
   public:
      someCode
};

template<class A, class B>
class MyClass<A,B,SxNull>
{
   public:
      someCode
};
\endcode
\author Sixten Boeck, boeck@mpie.de */
class SxNull { };

// --------------------------------------------------------------------------


/** \brief Basis class for all functors

    \b SxFunctorBase = S/PHI/nX Functor Base class

    This functor class is a container for bound function pointers. It stores
    the object (callee), the member function or the static global function.
    The information is anonymized by casting all types to void.

    \author Sixten Boeck, boeck@mpie.de */
class SxFunctorBase
{
   public:
      /** \brief Type of a pointer to member functions
        */
      typedef void (SxFunctorBase::*MemFunc)();

      /** \brief Type of a pointer to static global functions
        */
      typedef void (*Func)();

      SxFunctorBase () : callee(0), func(0) { /* empty */ }

      /** \param func_    Pointer to global function
          */
      SxFunctorBase (const Func *func_)
         : callee(0), 
           func(func_)
      {
         // empty
      }

     /** \param callee_  Pointer to object or NULL for global functions
          \param func_    Pointer to member function
          \param size     size of function pointer */
      SxFunctorBase (const void *callee_, const MemFunc *func_, size_t size)
         : callee(const_cast<void *>(callee_)), 
           func(0)
      {
         SX_CHECK (callee_);
         if (callee_)  { // member function pointer
#           ifdef MSVC
               memFunc = *func_;
#           else
               memcpy (memFunc, func_, size);
#           endif
         }
      }


      /** \brief support 'if (callbackPtr)'
        */
      operator int() const
      {
         return func || callee;
      }

      bool operator== (const SxFunctorBase &in)
      {
         return callee == in.callee && func == in.func;
      }

      bool operator!= (const SxFunctorBase &in)
      {
         return !operator== (in);
      }

      /** \brief Pointer to object or NULL for global functions
        */
      void *callee;

      /** \brief function or member function data
        */
      union {
         const void *func;
#        ifdef MSVC
            MemFunc memFunc;
#        else
            char memFunc[sizeof(MemFunc)];
#        endif
      };
};


// --------------------------------------------------------------------------

template<class RES, 
         class P1 = SxNull, class P2 = SxNull, class P3 = SxNull, 
         class P4 = SxNull, class P5 = SxNull, class P6 = SxNull,
         class Terminator = SxNull
        >
class SxFunctor : public SxFunctorBase
{
   // empty
};

// --- SxFunctor<RES>
template<class RES>
class SxFunctor<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() () const {
         SX_CHECK (func || callee);  // not initialized! use SxFuncBinder<T>::create!!!
         return metaFunc (*this);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_)
         : SxFunctorBase (f_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, SxFunctorBase::MemFunc *f_, size_t s_)
         : SxFunctorBase (c_, f_, s_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1>
template<class RES, class P1>
class SxFunctor<RES,P1,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1) const {
         SX_CHECK (func || callee);  // not initialized! use SxFuncBinder<T>::create!!!
         return metaFunc (*this, p1);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES (*MetaFunc)(const SxFunctorBase &, P1);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_)
         : SxFunctorBase (f_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, SxFunctorBase::MemFunc *f_, size_t s_)
         : SxFunctorBase (c_, f_, s_), metaFunc(mf_) { /* empty */ }


   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2>
template<class RES, class P1, class P2>
class SxFunctor<RES,P1,P2,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2) const {
         SX_CHECK (func || callee);  // not initialized! use SxFuncBinder<T>::create!!!
         return metaFunc (*this, p1, p2);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES (*MetaFunc)(const SxFunctorBase &, P1, P2);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_)
         : SxFunctorBase (f_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, SxFunctorBase::MemFunc *f_, size_t s_)
         : SxFunctorBase (c_, f_, s_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2,P3>
template<class RES, class P1, class P2, class P3>
class SxFunctor<RES,P1,P2,P3,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2, P3 p3) const {
         SX_CHECK (func || callee);  // not initialized! use SxFuncBinder<T>::create!!!
         return metaFunc (*this, p1, p2, p3);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &, P1, P2, P3);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_)
         : SxFunctorBase (f_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, SxFunctorBase::MemFunc *f_, size_t s_)
         : SxFunctorBase (c_, f_, s_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2,P3,P4>
template<class RES, class P1, class P2, class P3, class P4>
class SxFunctor<RES,P1,P2,P3,P4,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4) const {
         SX_CHECK (func || callee);  // not initialized! use SxFuncBinder<T>::create!!!
         return metaFunc (*this, p1, p2, p3, p4);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &, P1, P2, P3 , P4);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_)
         : SxFunctorBase (f_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, SxFunctorBase::MemFunc *f_, size_t s_)
         : SxFunctorBase (c_, f_, s_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2,P3,P4,P5>
template<class RES, class P1, class P2, class P3, class P4, class P5>
class SxFunctor<RES,P1,P2,P3,P4,P5,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4, P5 p5) const {
         SX_CHECK (func || callee);  // not initialized! use SxFuncBinder<T>::create!!!
         return metaFunc (*this, p1, p2, p3, p4, p5);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &, P1, P2, P3, P4, P5);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_)
         : SxFunctorBase (f_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, SxFunctorBase::MemFunc *f_, size_t s_)
         : SxFunctorBase (c_, f_, s_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2,P3,P4,P5,P6>
template<class RES, class P1, class P2, class P3, class P4, class P5, class P6>
class SxFunctor<RES,P1,P2,P3,P4,P5,P6,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4, P5 p5, P6 p6) const {
         SX_CHECK (func || callee);  // not initialized! use SxFuncBinder<T>::create!!!
         return metaFunc (*this, p1, p2, p3, p4, p5, p6);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &, P1, P2, P3, P4, P5, P6);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_)
         : SxFunctorBase (f_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, SxFunctorBase::MemFunc *f_, size_t s_)
         : SxFunctorBase (c_, f_, s_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};




// translate function pointers to functors

// SxFuncTranslator<Func, RES,...>
template<class Func, class RES, 
         class P1 = SxNull, class P2 = SxNull, class P3 = SxNull, 
         class P4 = SxNull, class P5 = SxNull, class P6 = SxNull,
         class Terminator = SxNull
        >
class SxFuncTranslator : public SxFunctor<RES,P1,P2,P3,P4,P5,P6,Terminator>
{ /* empty */ };

// SxFuncTranslator<Func, RES>
template<class Func, class RES>
class SxFuncTranslator<Func,RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES>
{
   public:
      SxFuncTranslator (Func func_) 
         : SxFunctor<RES> (metaFunc, 0, (const void *)func_, 0) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_)
      {
         SX_CHECK (functor_.func);
         return (Func (functor_.func))();
      }
};

// SxFuncTranslator<Func,RES,P1>
template<class Func, class RES, class P1>
class SxFuncTranslator<Func,RES,P1,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1>
{
   public:

      SxFuncTranslator (RES (*func_)(P1)) 
         : SxFunctor<RES,P1> (metaFunc, (SxFunctorBase::Func *)(const void *)func_) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1)
      {
         SX_CHECK (functor_.func);
#        ifdef MSVC
            return (reinterpret_cast<RES (*)(P1)>(functor_.func))(p1);
#        else
            return (Func(functor_.func))(p1);
#        endif
      }
};

// SxFuncTranslator<Func,RES,P1,P2>
template<class Func, class RES, class P1, class P2>
class SxFuncTranslator<Func,RES,P1,P2,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor< RES,P1,P2>
{
   public:
      SxFuncTranslator (Func func_) 
         : SxFunctor<RES,P1,P2> (metaFunc, 0, (const void *)func_, 0) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1, P2 p2)
      {
         SX_CHECK (functor_.func);
         return (Func (functor_.func))(p1, p2);
      }
};

// SxFuncTranslator<Func,RES,P1,P2,P3>
template<class Func, class RES, class P1, class P2, class P3>
class SxFuncTranslator<Func,RES,P1,P2,P3,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3>
{
   public:
      SxFuncTranslator (Func func_) 
         : SxFunctor<RES,P1,P2,P3> (metaFunc, 0, (const void *)func_, 0) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1, P2 p2, P3 p3)
      {
         SX_CHECK (functor_.func);
         return (Func (functor_.func))(p1, p2, p3);
      }
};

// SxFuncTranslator<Func,RES,P1,P2,P3,P4>
template<class Func, class RES, class P1, class P2, class P3, class P4>
class SxFuncTranslator<Func,RES,P1,P2,P3,P4,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4>
{
   public:
      SxFuncTranslator (Func func_) 
         : SxFunctor<RES,P1,P2,P3,P4> (metaFunc, 0, (const void *)func_, 0) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4)
      {
         SX_CHECK (functor_.func);
         return (Func (functor_.func))(p1, p2, p3, p4);
      }
};

// SxFuncTranslator<Func,RES,P1,P2,P3,P4,P5>
template<class Func, class RES, class P1, class P2, class P3, class P4, class P5>
class SxFuncTranslator<Func,RES,P1,P2,P3,P4,P5,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5>
{
   public:
      SxFuncTranslator (Func func_) 
         : SxFunctor<RES,P1,P2,P3,P4,P5> (metaFunc, 0, (const void *)func_, 0) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5)
      {
         SX_CHECK (functor_.func);
         return (Func (functor_.func))(p1, p2, p3, p4, p5);
      }
};

// SxFuncTranslator<Func,RES,P1,P2,P3,P4,P5,P6>
template<class Func, class RES, class P1, class P2, class P3, class P4, class P5, class P6>
class SxFuncTranslator<Func,RES,P1,P2,P3,P4,P5,P6,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5,P6>
{
   public:
      SxFuncTranslator (Func func_) 
         : SxFunctor<RES,P1,P2,P3,P4,P5,P6> (metaFunc, 0, (const void *)func_, 0) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5, P6 p6)
      {
         SX_CHECK (functor_.func);
         return (Func (functor_.func))(p1, p2, p3, p4, p5, p6);
      }
};



// ---------------------------------------------------------------------------

// translate member function pointers to functors

// SxMemTranslator<Callee, MemFunc, RES, ...>
template<class Callee, class MemFunc, class RES, 
         class P1 = SxNull, class P2 = SxNull, class P3 = SxNull,
         class P4 = SxNull, class P5 = SxNull, class P6 = SxNull,
         class Terminator = SxNull
        >
class SxMemTranslator : public SxFunctor<RES,P1,P2,P3,P4,P5,P6,Terminator>
{ /* empty */ };


// SxMemTranslator<Callee, MemFunc, RES>
template<class Callee, class MemFunc, class RES>
class SxMemTranslator<Callee,MemFunc,RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES>
{
   public:
      typedef RES (Callee::* MF)();
      SxMemTranslator (Callee &callee_, const MF &memFunc_)
         : SxFunctor<RES> (metaFunc, &callee_, 
                           (SxFunctorBase::MemFunc *)(&memFunc_), 
                           sizeof(MemFunc))
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_)
      {
         Callee *callee = (Callee *)functor_.callee;
#        ifdef MSVC
            MF memFunc = reinterpret_cast<RES (Callee::*)()>(functor_.memFunc);
#        else
            MemFunc &memFunc (*( MemFunc *)(RES *)(functor_.memFunc));
#        endif
         return (callee->*memFunc)();
      }
};

// SxMemTranslator<Callee, MemFunc, RES, P1>
template<class Callee, class MemFunc, class RES, 
         class P1>
class SxMemTranslator<Callee,MemFunc,RES,P1,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1>
{
   public:
      typedef RES (Callee::* MF)(P1);
      SxMemTranslator (Callee &callee_, const MF &memFunc_)
         : SxFunctor<RES,P1> (metaFunc, &callee_, 
                              (SxFunctorBase::MemFunc *)(&memFunc_), 
                              sizeof(MemFunc))
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1)
      {
         Callee *callee = (Callee *)functor_.callee;
#        ifdef MSVC
            MF memFunc = reinterpret_cast<RES (Callee::*)(P1)>(functor_.memFunc);
#        else
            MemFunc &memFunc (*( MemFunc *)(RES *)(functor_.memFunc));
#        endif
         return (callee->*memFunc)(p1);
      }
};

// SxMemTranslator<Callee, MemFunc, RES, P1, P2>
template<class Callee, class MemFunc, class RES, 
         class P1, class P2>
class SxMemTranslator<Callee,MemFunc,RES,P1,P2,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2>
{
   public:
      typedef RES (Callee::* MF)(P1,P2);
      SxMemTranslator (Callee &callee_, const MF &memFunc_)
         : SxFunctor<RES,P1,P2> (metaFunc, &callee_, 
                                 (SxFunctorBase::MemFunc *)(&memFunc_), 
                                 sizeof(MemFunc))
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1, P2 p2)
      {
         Callee *callee = (Callee *)functor_.callee;
#        ifdef MSVC
            MF memFunc = reinterpret_cast<RES (Callee::*)(P1,P2)>(functor_.memFunc);
#        else
            MemFunc &memFunc (*( MemFunc *)(RES *)(functor_.memFunc));
#        endif
         return (callee->*memFunc)(p1, p2);
      }
};

// SxMemTranslator<Callee, MemFunc, RES, P1, P2, P3>
template<class Callee, class MemFunc, class RES, 
         class P1, class P2, class P3>
class SxMemTranslator<Callee,MemFunc,RES,P1,P2,P3,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3>
{
   public:
      typedef RES (Callee::* MF)(P1,P2,P3);
      SxMemTranslator (Callee &callee_, const MF &memFunc_)
         : SxFunctor<RES,P1,P2,P3> (metaFunc, &callee_, 
                                    (SxFunctorBase::MemFunc *)(&memFunc_),
                                    sizeof(MemFunc))
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3)
      {
         Callee *callee = (Callee *)functor_.callee;
#        ifdef MSVC
            MF memFunc = reinterpret_cast<RES (Callee::*)(P1,P2,P3)>(functor_.memFunc);
#        else
            MemFunc &memFunc (*( MemFunc *)(RES *)(functor_.memFunc));
#        endif
         return (callee->*memFunc)(p1, p2, p3);
      }
};

// SxMemTranslator<Callee, MemFunc, RES, P1, P2, P3, P4>
template<class Callee, class MemFunc, class RES, 
         class P1, class P2, class P3, class P4>
class SxMemTranslator<Callee,MemFunc,RES,P1,P2,P3,P4,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4>
{
   public:
      typedef RES (Callee::* MF)(P1,P2,P3,P4);
      SxMemTranslator (Callee &callee_, const MF &memFunc_)
         : SxFunctor<RES,P1,P2,P3,P4> (metaFunc, &callee_, 
                                       (SxFunctorBase::MemFunc *)(&memFunc_),
                                       sizeof(MemFunc))
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4)
      {
         Callee *callee = (Callee *)functor_.callee;
#        ifdef MSVC
            MF memFunc = reinterpret_cast<RES (Callee::*)(P1,P2,P3,P4)>(functor_.memFunc);
#        else
            MemFunc &memFunc (*( MemFunc *)(RES *)(functor_.memFunc));
#        endif
         return (callee->*memFunc)(p1, p2, p3, p4);
      }
};

// SxMemTranslator<Callee, MemFunc, RES, P1, P2, P3, P4, P5>
template<class Callee, class MemFunc, class RES, 
         class P1, class P2, class P3, class P4, class P5>
class SxMemTranslator<Callee,MemFunc,RES,P1,P2,P3,P4,P5,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5>
{
   public:
      typedef RES (Callee::* MF)(P1,P2,P3,P4,P5);
      SxMemTranslator (Callee &callee_, const MF &memFunc_)
         : SxFunctor<RES,P1,P2,P3,P4,P5> (
               metaFunc, &callee_, 
               (SxFunctorBase::MemFunc *)(&memFunc_), 
               sizeof(MemFunc)
           )
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5)
      {
         Callee *callee = (Callee *)functor_.callee;
#        ifdef MSVC
            MF memFunc = reinterpret_cast<RES (Callee::*)(P1,P2,P3,P4,P5)>(functor_.memFunc);
#        else
            MemFunc &memFunc (*( MemFunc *)(RES *)(functor_.memFunc));
#        endif
         return (callee->*memFunc)(p1, p2, p3, p4, p5);
      }
};

// SxMemTranslator<Callee, MemFunc, RES, P1, P2, P3, P4, P5, P6>
template<class Callee, class MemFunc, class RES, 
         class P1, class P2, class P3, class P4, class P5, class P6>
class SxMemTranslator<Callee,MemFunc,RES,P1,P2,P3,P4,P5,P6,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5,P6>
{
   public:
      typedef RES (Callee::* MF)(P1,P2,P3,P4,P5,P6);
      SxMemTranslator (Callee &callee_, const MF &memFunc_)
         : SxFunctor<RES,P1,P2,P3,P4,P5,P6> (
               metaFunc, &callee_, 
               (SxFunctorBase::MemFunc *)(&memFunc_), 
               sizeof(MemFunc)
           )
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5, P6 p6)
      {
         Callee *callee = (Callee *)functor_.callee;
#        ifdef MSVC
            MF memFunc = reinterpret_cast<RES (Callee::*)(P1,P2,P3,P4,P5,P6)>(functor_.memFunc);
#        else
            MemFunc &memFunc (*( MemFunc *)(RES *)(functor_.memFunc));
#        endif
         return (callee->*memFunc)(p1, p2, p3, p4, p5, p6);
      }
};



#endif /* _SX_FUNCTOR_H_ */
