// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------



#ifndef _SX_LIST_H_
#define _SX_LIST_H_

#include <SxError.h>
#include <SxMemConsumer.h>
#include <SxUtil.h>
#include <stdio.h>
#include <iostream>

// SxList-class
/** This class has been implemented in case the standard template library
    (STL) is not available on some platforms.
    It just provides a doubly linked list.
    @ingroup Tools
    @author Sixten Boeck
    */
template<class T>
class SxList : public SxMemConsumer
{
   public:
      class List {
         public:
            List () { prev = next = NULL; }
            T element;
            List *prev;
            List *next;
      };
      List *firstElement;
      List *lastElement;
      ssize_t size;

      // in case class T has no operator== 
//      bool isEqual (const T &, const T &) const;

   public:
      // SxList::Iterator-class
      class Iterator
      {
         public:
            // Constructor
            Iterator (typename SxList<T>::List *p=NULL)  {
               ptr = p;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
            }
            // Operations
            // -- Dereferencation   
            T &operator* () const  {
               SX_CHECK (ptr);
               return ptr->element; 
            }
            
            T *operator-> () const  {
               SX_CHECK (ptr);
               return &ptr->element; 
            }

            // -- Equality
            bool operator== (const typename SxList<T>::Iterator &it) const  {
               return ( it.ptr == ptr );
            }
               
            bool operator!= (const typename SxList<T>::Iterator &it) const  {
               return ( it.ptr != ptr );
            }

            bool valid () const  { return (ptr != NULL); }
            
            bool operator! () const {
               return !valid ();
            }

            // -- Assignment
            typename SxList<T>::Iterator &operator= (const typename SxList<T>::Iterator &rhs){
               if ( this == &rhs)  return *this;
               ptr = rhs.ptr;
               ptrNext = rhs.ptrNext;
               ptrPrev = rhs.ptrPrev;
               return *this;
            }
            
            // -- Incrementation
            const typename SxList<T>::Iterator operator++ (int)  { // postfix
               SX_CHECK (ptr);
               typename SxList<T>::Iterator copy (ptr);
               ptr = ptrNext;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
               return copy;
            }

            typename SxList<T>::Iterator      &operator++ ()  {  // prefix
               SX_CHECK (ptr);
               ptr = ptrNext;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
               return *this;
            }

            void                               operator+= (ssize_t id)  {
               for (ssize_t i=0; i<id; i++)  {
                  SX_CHECK (ptr);
                  ptr = ptrNext;
                  if (ptr)  {
                     ptrNext = ptr->next;
                     ptrPrev = ptr->prev;
                  }  else  {
                     ptrNext = ptrPrev = NULL;
                  }
               }
            }

            // -- Decrementation
            const typename SxList<T>::Iterator operator-- (int)  { // postfix
               SX_CHECK (ptr);
               typename SxList<T>::Iterator copy;
               copy = typename SxList<T>::Iterator (ptr);
               ptr = ptrPrev;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
               return copy;
            }
               
            typename SxList<T>::Iterator      &operator-- ()  {  // prefix
               SX_CHECK (ptr);
               ptr = ptrPrev;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
               return *this;
            }


         //protected:
            // Members
            typename SxList<T>::List *ptr;
            typename SxList<T>::List *ptrNext;
            typename SxList<T>::List *ptrPrev;
      };
      
      // SxList::ConstIterator-class
      class ConstIterator
      {
         public:
            // Constructor
            ConstIterator (const typename SxList<T>::List *p=NULL)  {
               ptr = p;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
            }
            // Typecast
            ConstIterator (const Iterator &it)  {
               ptr = it.ptr;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
            }

            // Operations
            // -- Dereferencation   
            const T &operator* () const  {
               SX_CHECK (ptr);
               return ptr->element; 
            }

            const T *operator-> () const  {
               SX_CHECK (ptr);
               return &ptr->element; 
            }

            
            // -- Equality
            bool operator== (const typename SxList<T>::ConstIterator &cIter
                  ) const 
            {
               return ( cIter.ptr == ptr );
            }   

            bool operator!= (const typename SxList<T>::ConstIterator &cIter
                  ) const
            {
               return ( cIter.ptr != ptr );
            }

            bool valid () const  { return (ptr != NULL); }

            bool operator! () const {
               return !valid ();
            }

            // -- Assignment
            typename SxList<T>::ConstIterator &operator= (const typename SxList<T>::ConstIterator &rhs){
               if ( this == &rhs)  return *this;
               ptr = rhs.ptr;
               ptrNext = rhs.ptrNext;
               ptrPrev = rhs.ptrPrev;
               return *this;
            }

            // -- Incrementation
            const typename SxList<T>::ConstIterator operator++ (int)  { // postfix
               SX_CHECK (ptr);
               typename SxList<T>::ConstIterator copy (ptr);
               ptr = ptrNext;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
               return copy;
            }

            typename SxList<T>::ConstIterator      &operator++ ()  {  // prefix
               SX_CHECK (ptr);
               ptr = ptrNext;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
               return *this;
            }

            void                               operator+= (ssize_t id)  {
               for (ssize_t i=0; i<id; i++)  {
                  SX_CHECK (ptr);
                  ptr = ptrNext;
                  if (ptr)  {
                     ptrNext = ptr->next;
                     ptrPrev = ptr->prev;
                  }  else  {
                     ptrNext = ptrPrev = NULL;
                  }
               }
            }

            // -- Decrementation
            const typename SxList<T>::ConstIterator operator-- (int)  { // postfix
               SX_CHECK (ptr);
               typename SxList<T>::ConstIterator copy;
               copy = typename SxList<T>::ConstIterator (ptr);
               ptr = ptrPrev;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
               return copy;
            }
               
            typename SxList<T>::ConstIterator      &operator-- ()  {  // prefix
               SX_CHECK (ptr);
               ptr = ptrPrev;
               if (ptr)  {
                  ptrNext = ptr->next;
                  ptrPrev = ptr->prev;
               }  else  {
                  ptrNext = ptrPrev = NULL;
               }
               return *this;
            }

            friend class SxList<T>;


         //protected:
            // Members
            const typename SxList<T>::List *ptr;
            const typename SxList<T>::List *ptrNext;
            const typename SxList<T>::List *ptrPrev;
      };
      
      // Constructors
      // -- Constructor
      SxList ();
      // -- Copyconstructor
      SxList (const SxList<T> &);
      // -- Destructor
      ~SxList ();

      // Operations
      // -- Randomaccess
      inline T &operator() (ssize_t);
      inline const T &operator() (ssize_t) const;

      // -- Assignment
      SxList<T> &operator=  (const SxList<T> &);

      // -- Equality
      bool       operator== (const SxList<T> &) const;
      
      bool       operator!= (const SxList<T> &) const;

      // -- Insertion
      inline SxList<T> &operator<< (const T &);
      inline SxList<T> &operator<< (const SxList<T> &);


      // Methods
      // -- Insertion
      SxList<T> &append (const T &);
      SxList<T> &append (const SxList<T> &);
      SxList<T> &append (const Iterator &it, const T &);
      SxList<T> &insert (ssize_t newPos, const T &);
      SxList<T> &prepend (const T &);
      SxList<T> &prepend (const Iterator &it, const T &);

      // -- Search
      ssize_t findPos (const T &) const;
      bool contains (const T &) const;

      // --Size 
      inline ssize_t getSize () const { return size; }
      void resize (ssize_t);

      // -- Deletion
      inline void remove (ssize_t);
      inline void removeItem (List *ptr);
      inline void removeItem (Iterator *it);
      inline void removeElement (const T &);
      inline void removeFirst ();
      inline void removeLast ();
      inline void removeAll ();

      // Accessfunctions
      inline T &first ();
      inline const T &first () const;
      inline T &last ();
      inline const T &last () const;

      // Iteratorfunctions
      inline typename SxList<T>::Iterator begin ();
      inline typename SxList<T>::ConstIterator begin () const;
      inline typename SxList<T>::Iterator end ();   
      inline typename SxList<T>::ConstIterator end () const;
      inline typename SxList<T>::Iterator fromLast ();
      inline typename SxList<T>::ConstIterator fromLast () const;
      inline typename SxList<T>::Iterator toFirst ();
      inline typename SxList<T>::ConstIterator toFirst () const;
      inline typename SxList<T>::Iterator find (const T &) const;

};

//------------------------------------------------------------------------------
// SxList-class
//------------------------------------------------------------------------------
// Constructors
// -- Constructor
template<class T>
SxList<T>::SxList () : SxMemConsumer ()
{
   size        = 0;
   lastElement = firstElement = NULL;
}

// -- Copyconstructor
template<class T>
SxList<T>::SxList (const SxList<T> &in) : SxMemConsumer (in)
{
   SX_CHECK_VAR (in.size >= 0, in.size);

   firstElement = lastElement = NULL;
   size = in.size;
   if (in.size > 0)  {
      List *srcPtr   = in.firstElement;
      List *prevDest = NULL;
      firstElement   = new List;
      List *destPtr  = firstElement;
      while ( srcPtr )  {
         destPtr->element = srcPtr->element;
         destPtr->prev    = prevDest;
         destPtr->next    = NULL;
         prevDest = destPtr;

         srcPtr = srcPtr->next;
         if (srcPtr)  destPtr = destPtr->next = new List;
         lastElement = destPtr;
      }
   }
   TRACK_MALLOC (*this, 1);
}

// -- Destructor
template<class T>
SxList<T>::~SxList ()
{
   removeAll ();
}

// Operations
// -- Assignment
template<class T>
SxList<T> &SxList<T>::operator= (const SxList<T> &in)
{
   SX_CHECK_VAR (in.size >= 0, in.size);
   if ( this == &in ) return *this;

   removeAll ();

   firstElement = lastElement = NULL;
   size = in.size;
   if (in.size > 0)  {
      List *srcPtr   = in.firstElement;
      List *prevDest = NULL;
      firstElement   = new List;
      lastElement = firstElement;
      List *destPtr  = firstElement;
      while ( srcPtr )  {
         destPtr->element = srcPtr->element;
         destPtr->prev    = prevDest;
         destPtr->next    = NULL;
         prevDest = destPtr;

         srcPtr = srcPtr->next;
         if (srcPtr)  destPtr = destPtr->next = new List;
         lastElement = destPtr;
      }
   }

   TRACK_MALLOC (*this, 1);
   return *this;
}

// -- Equality
template<class T>
bool SxList<T>::operator== (const SxList<T> &rhs) const
{
   if (getSize () != rhs.getSize ())  {
      return false;
   } else  {
      ConstIterator it;
      ConstIterator itRhs;
      for (itRhs = rhs.begin (), it = begin (); it != end (); ++itRhs, ++it)  {
         if ( (*it) != (*itRhs))  return false;
      }
      return true;
   }
   
}

template<class T>
bool SxList<T>::operator!= (const SxList<T> &rhs) const
{
   return !operator== (rhs);
}

// -- Randomaccess
template<class T>
T &SxList<T>::operator() (ssize_t i)
{
   SX_CHECK_VARS (i >= 0 && i < size, i, size);
   List *ptr;
   ssize_t j = i;
// if ( i+1 <= size /2 )  {   
   if ( i+1 <= size>>1 )  {   // forward 
      ptr = firstElement;
      for (j=0; j<i; j++)  ptr = ptr->next;
   }  else  {             // reverse
      ptr = lastElement;
      for (j=size; j>i+1; j--)  ptr = ptr->prev;
   }
   return ptr->element;
}

template<class T>
const T &SxList<T>::operator() (ssize_t i) const
{
   SX_CHECK_VARS (i >= 0 && i < size, i, size);
   List *ptr;
   ssize_t j = i;
// if ( i+1 <= size /2 )  {   
   if ( i+1 <= size>>1 )  {   // forward 
      ptr = firstElement;
      for (j=0; j<i; j++)  ptr = ptr->next;
   }  else  {             // reverse
      ptr = lastElement;
      for (j=size; j>i+1; j--)  ptr = ptr->prev;
   }
   return ptr->element;
}

// -- Insertion
template<class T>
SxList<T> &SxList<T>::operator<< (const T &t)
{
   return append (t);
}

template<class T>
SxList<T> &SxList<T>::operator<< (const SxList<T> &l)
{
   return append (l);
}

// Methods
// -- Insertion
template<class T>
SxList<T> &SxList<T>::append (const T &e)
{
   List *entry    = new List;
   entry->element = e;

   entry->prev    = lastElement;
   entry->next    = NULL;
   
   if ( lastElement )  lastElement->next = entry;
   lastElement = entry;
   
   if ( !firstElement ) firstElement = entry;
   size++;
   TRACK_MALLOC (*this, 1);
   return *this;
}

template<class T>
SxList<T> &SxList<T>::append (const SxList<T> &in)
{
   typename SxList<T>::ConstIterator it;
   for (it = in.begin(); it != in.end(); it++)   append (*it);
   return *this;
}

template<class T>
SxList<T> &SxList<T>::append (const Iterator &it_, const T &in)
{
   if (getSize() == 0)  return append (in);
   if (*it_ == first()) return insert (1, in);
   if (*it_ == last())  return append (in);

   Iterator it = it_;
   it++;

   List *ptr = it.ptr;

   List *entry    = new List;
   entry->element = in;
   entry->prev    = ptr->prev;
   entry->next    = ptr;
   ptr->prev->next = entry;
   ptr->prev       = entry;
   size++;
   return *this;
}

template<class T>
SxList<T> &SxList<T>::insert (ssize_t newPos, const T &in)
{
   SX_CHECK_VARS (newPos >= 0 && newPos <= getSize(), newPos, getSize());
   if (newPos == 0)  {
      return prepend (in);
   }
   if (newPos == getSize())  {
      return append (in);
   }

   List *ptr = firstElement;
   for (ssize_t i=0; i < newPos; i++)  {
      ptr = ptr->next;
   }

   List *entry    = new List;
   entry->element = in;
   entry->prev    = ptr->prev;
   entry->next    = ptr;
   ptr->prev->next = entry;
   ptr->prev       = entry;
   size++;
   return *this;
}

template<class T>
SxList<T> &SxList<T>::prepend (const Iterator &it, const T &in)
{
   if (getSize() == 0)  return append (in);

   if (*it == first())  return prepend (in);

   List *ptr = it.ptr;

   List *entry    = new List;
   entry->element = in;
   entry->prev    = ptr->prev;
   entry->next    = ptr;
   ptr->prev->next = entry;
   ptr->prev       = entry;
   size++;
   return *this;
}

template<class T>
SxList<T> &SxList<T>::prepend (const T &e)
{
   List *entry    = new List;
   entry->element = e;
   entry->prev    = NULL;
   entry->next    = firstElement;

   if ( firstElement )  firstElement->prev = entry;
   firstElement = entry;

   if ( !lastElement ) lastElement = entry;

   size++;
   TRACK_MALLOC (*this, 1);
   return *this;
}

// -- Search
template<class T>
ssize_t SxList<T>::findPos (const T &e) const
{
   ssize_t i=0;
   List *ptr = firstElement;
   while ( ptr )  {
      //if ( ptr->element ==  e )  return i;
//      if ( isEqual (ptr->element, e) )  return i;
      if (ptr->element == e)  return i;
      ptr = ptr->next;
      i++;
   }
   
   return -1;
}

template<class T>
bool SxList<T>::contains (const T &in) const
{
   return findPos (in) == -1 ? false : true;
}

//template<class T>
//bool SxList<T>::isEqual (const T &a, const T &b) const
//{
//   ssize_t nBytes = sizeof (a);
//   char *aPtr = (char *)&a;
//   char *bPtr = (char *)&b;
//   for (ssize_t i=0; i<nBytes; i++)  {
//      if ( *aPtr != *bPtr ) return false;
//      aPtr++;
//      bPtr++;
//   }
//
//   return true;
//}

// -- Size
template<class T>
void SxList<T>::resize (ssize_t newSize)
{
   SX_CHECK_VAR ( newSize >= 0, newSize );

   List *entry = firstElement;

   if (size == newSize)  return;

   if        ( newSize == 0)     {  // clear entire list

      removeAll ();

   } else if ( newSize < size )  {  // truncate end of list

      for (ssize_t i=0; i < newSize-1; i++)  {
         entry = entry->next;
      }
      List *ptr = entry->next, *tmp = NULL;
      while (ptr)  {
         tmp = ptr;
         ptr = ptr->next;
         delete tmp;
      }
      lastElement = entry;
      lastElement->next = NULL;

   } else {                         // add elements to the end

      ssize_t diffSize = newSize - size;
      for (ssize_t i=0; i < diffSize; i++)  {
         entry       = new List;
         entry->prev = lastElement;
         entry->next = NULL;
         
         if ( lastElement )  lastElement->next = entry;
         lastElement = entry;
         
         if ( !firstElement ) firstElement = entry;
      }

   }
   size = newSize;
   TRACK_MALLOC (*this, 1);
}

// -- Deletion
template<class T>
void SxList<T>::removeItem (List *ptr)
{
   if (ptr)  {
      List *prevPtr = ptr->prev;
      List *nextPtr = ptr->next;

      if (ptr == firstElement) firstElement = nextPtr;
      if (ptr == lastElement)  lastElement = prevPtr;
      if (prevPtr) prevPtr->next = nextPtr;
      if (nextPtr) nextPtr->prev = prevPtr;
                  
      delete ptr;
      size--;
   }
}

template<class T>
void SxList<T>::removeItem (Iterator *it)
{
   SX_CHECK (it);
   removeItem (it->ptr);
   *it = Iterator();
}

template<class T>
void SxList<T>::remove  (ssize_t idx)
{
   SX_CHECK_VARS (idx >= 0 && idx < size, idx, size);

   List *ptr = firstElement;
   for (ssize_t i=0; i < idx; i++)  {
      ptr = ptr->next;
   }
   removeItem (ptr);
}

template<class T>
void SxList<T>::removeElement  (const T &e)
{
   List *ptr = firstElement;
   while ( ptr )  {
      if (ptr->element == e)  {
         removeItem (ptr);
         return;
      }
      ptr = ptr->next;
   }
}

template<class T>
void SxList<T>::removeFirst ()
{
   removeItem (firstElement);
}

template<class T>
void SxList<T>::removeLast ()
{
   removeItem (lastElement);
}

template<class T>
void SxList<T>::removeAll ()
{
   if (size > 0)  {
      List *ptr = lastElement;
      List *tmp;
      // --- delete entire list
//      while ( ptr )  {
//         tmp = ptr;
//         ptr = ptr->prev;
//         delete tmp;
//      }
      while (size > 1)
      {
         tmp = ptr;
         ptr = ptr->prev;
         delete tmp;
         size--;
      }
      delete ptr;
   }
   size = 0;
   firstElement = lastElement = NULL;
}

// Accessfunctions
template<class T>
T &SxList<T>::first ()
{
   SX_CHECK_VAR (size > 0, size);
   SX_CHECK     (firstElement);
   List *ptr = firstElement;
   return ptr->element;
}

template<class T>
const T &SxList<T>::first () const
{
   SX_CHECK_VAR (size > 0, size);
   SX_CHECK     (firstElement);
   List *ptr = firstElement;
   return ptr->element;
}


template<class T>
T &SxList<T>::last ()
{
   SX_CHECK_VAR (size > 0, size);
   SX_CHECK     (lastElement);
   List *ptr = lastElement;
   return ptr->element;
}

template<class T>
const T &SxList<T>::last () const
{
   SX_CHECK_VAR (size > 0, size);
   SX_CHECK     (lastElement);
   List *ptr = lastElement;
   return ptr->element;
}

// Iterator-functions
template<class T>
typename SxList<T>::Iterator SxList<T>::begin ()
{
   return typename SxList<T>::Iterator (firstElement);
}

template<class T>
typename SxList<T>::ConstIterator SxList<T>::begin () const
{
   return typename SxList<T>::ConstIterator (firstElement);
}

template<class T>
typename SxList<T>::Iterator SxList<T>::end ()
{
   return typename SxList<T>::Iterator (NULL);
}

template<class T>
typename SxList<T>::ConstIterator SxList<T>::end () const
{
   return typename SxList<T>::ConstIterator (NULL);
}

template<class T>
typename SxList<T>::Iterator SxList<T>::fromLast () 
{
   if ( lastElement )  return typename SxList<T>::Iterator (lastElement);
   else                return typename SxList<T>::Iterator (NULL);
}

template<class T>
typename SxList<T>::ConstIterator SxList<T>::fromLast () const
{
   if ( lastElement )  return typename SxList<T>::ConstIterator (lastElement);
   else                return typename SxList<T>::ConstIterator (NULL);
}

template<class T>
typename SxList<T>::Iterator SxList<T>::toFirst ()
{
   return typename SxList<T>::Iterator (NULL);
}


template<class T>
typename SxList<T>::ConstIterator SxList<T>::toFirst () const
{
   return typename SxList<T>::ConstIterator (NULL);
}
template<class T>
typename SxList<T>::Iterator SxList<T>::find (const T &e) const
{
   List *ptr = firstElement;
   while ( ptr )  {
      if ( ptr->element == e )
         return typename SxList<T>::Iterator (ptr);
      ptr = ptr->next;
   }
   
   return typename SxList<T>::Iterator (NULL);
}

//------------------------------------------------------------------------------
// global functions
//------------------------------------------------------------------------------
template<class T>
std::ostream &operator<< (std::ostream &s, const SxList<T> &in)
{
   ssize_t i=0;
   typename SxList<T>::ConstIterator it;
   for (it=in.begin(); it!=in.end(); it++, i++)
      s << i << ": " << *it << std::endl;
   return s;
}

template<class T>
size_t getNBytes (const SxList<T> &in)
{
   typename SxList<T>::ConstIterator it;
   size_t nBytes = 0;
   for (it = in.begin(); it != in.end(); ++it)  {
      nBytes += getNBytes (*it)
              + sizeof (typename SxList<T>::List *)   // SxList<T>::List::prev
              + sizeof (typename SxList<T>::List *);  // SxList<T>::List::next
   }
   nBytes += sizeof (in);  // = sizeof (SxList<T>::firstElement)
                           // + sizeof (SxList<T>::lastElement)
                           // + sizeof (SxList<T>::size)
                           // + sizeof (SxMemConsumer::className)
                           // + sizeof (SxMemConsumer::parent)
                           // + sizeof (SxMemConsumer::memCounterId)

   return nBytes;
}

#endif /* _SX_LIST_H_ */
