// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_MAP_H_
#define _SX_MAP_H_

#include <SxException.h>
#include <SxHashTable.h>

/** \brief Maps keys to values

    \b SxMap = SPHInX Key-Value Map

    This template class creates key-value pairs and manages them.
    The data types of both key and value are template arguments. It can
    be instantiated for all data types which define
    - a default constructor
    - a copy constructor
    - an operator==
    
    \par Example:

\code
   SxMap<SxString,double> hash;
   hash("PI") = 3.14;
   hash("2PI") = 6.28;
   SxMap<SxString,double>::Iterator it;
   for (it = hash.begin(); it != hash.end(); ++it)  {
      cout << it->getKey() << " = " << it->getValue() << endl;
   }
   cout << "2pi = " << hash.getValue("2pi") << endl;
\endcode

    \author Sixten Boeck, boeck@gemmantics.com
    \author Vaclav Bubnik, bubnik@gemmantics.com */
template<class K, class V, class H=SxHashFunction>
class SxMap
{
   public:
   
      typedef K  Key;
      typedef V  Value;

      class Iterator;
      
      /** \brief Iterator for SxMap objecs

          Using this class it is possible to iterate over all key-value-pairs 
          of an SxMap object.

          \par Example
\code
   SxMap<SxString,double> hash;
   hash("PI") = 3.14;
   hash("2PI") = 6.28;
   SxMap<SxString,double>::Iterator it;
   for (it = hash.begin(); it != hash.end(); ++it)  {
      cout << it.getKey << "=" << it.getValue();
   }
\endcode
           If all keys/values have to be accessed the iteration is significantly
           faster then a direct access.  */
      class ConstIterator
      {
         public:
            /** \brief Default constructor */
            ConstIterator ();
            /** \brief Typecast constructor */
            ConstIterator (const Iterator &it);
            /** \brief Creates a new iterator */
            ConstIterator (const typename SxList<K>::ConstIterator &,
                           const typename SxList<V>::ConstIterator &);
            /** \brief Destroy iterator */
           ~ConstIterator ();

            /** \brief Check if two iterators are equal */
            bool operator== (const typename SxMap<K,V,H>::ConstIterator &)const;
            /** \brief Check if two iterators are not equal */
            bool operator!= (const typename SxMap<K,V,H>::ConstIterator &)const;
            /** \brief Postfix incremental operator of an iterator */
            const typename SxMap<K,V,H>::ConstIterator  operator++ (int);
            /** \brief Prefix incremental operator of an iterator */
                  typename SxMap<K,V,H>::ConstIterator &operator++ ();
            /** \brief Postfix decremental operator of an iterator */
            const typename SxMap<K,V,H>::ConstIterator  operator-- (int);
            /** \brief Prefix decremental operator of an iterator */
                  typename SxMap<K,V,H>::ConstIterator &operator-- ();
                  
            /** \brief Extract the key from the key-value pair */
            const K &getKey () const;
            /** \brief Extract the value from the key-value pair */
            const V &getValue () const;
            /** \brief Dereferencing operator */
            const V &operator->() const;

         protected:
            typename SxList<K>::ConstIterator itK;
            typename SxList<V>::ConstIterator itV;
      };

      class Iterator
      {
         public:
            /** \brief Default constructor */
            Iterator ();
            /** \brief Creates a new iterator */
            Iterator (const typename SxList<K>::Iterator &,
                      const typename SxList<V>::Iterator &);
            /** \brief Destroy iterator */
           ~Iterator ();

            /** \brief Check if two iterators are equal */
            bool operator== (const typename SxMap<K,V,H>::Iterator &);
            /** \brief Check if two iterators are not equal */
            bool operator!= (const typename SxMap<K,V,H>::Iterator &);
            /** \brief Postfix incremental operator of an iterator */
            const typename SxMap<K,V,H>::Iterator  operator++ (int);
            /** \brief Prefix incremental operator of an iterator */
                  typename SxMap<K,V,H>::Iterator &operator++ ();
            /** \brief Postfix decremental operator of an iterator */
            const typename SxMap<K,V,H>::Iterator  operator-- (int);
            /** \brief Prefix decremental operator of an iterator */
                  typename SxMap<K,V,H>::Iterator &operator-- ();
                  
            /** \brief Extract the key from the key-value pair */
            const K &getKey ();
            /** \brief Extract the value from the key-value pair */
            V &getValue ();
            /** \brief Dereferencing operator */
            V &operator->();

            friend class ConstIterator;

         protected:
            typename SxList<K>::Iterator itK;
            typename SxList<V>::Iterator itV;
      };

      /** \brief Default constructor 
       */
      SxMap ();
      SxMap (const SxMap<K,V,H> &);

      /** \brief Destructor 
        */
      ~SxMap ();

      /** \brief Assign a key.

          This is the default way of assigning a key.

          \par Example: 
\code          
   SxMap<SxString,double> hash;
   hash("PI") = 3.14;
\endcode */
      inline V &operator() (const K &key);
      inline const V &operator() (const K &key) const;
      
      // -- Assignment
      SxMap<K,V,H> &operator= (const SxMap<K,V,H> &);
      
      /** \brief Find hash table index of a key.

          \par Example: 
\code
   ssize_t idx = table.findKey ("PI");
   if (idx >= 0)  {
      return table.getValueIdx (idx);
   }
\endcode

          Which is faster than:
\code
   if (table.containsKey ("PI"))  {
      return table("PI");
   }
\endcode */
      ssize_t findKey (const K &) const;
            V &getValueIdx (ssize_t idx);
      const V &getValueIdx (ssize_t idx) const;
      
      /** \return number of key-value pairs */
      ssize_t getSize () const;
      
      /** \brief Checks whether a certain key is present in the hash */
      bool containsKey (const K &) const;

      /** \brief Checks whether a certain value is present in the hash */
      bool containsValue (const V &) const;

      /** \brief Removes a key-value pair 
        */
      bool removeKey (const K &);

      /** \brief Removes a key-value pair by value
        */
      bool removeValue (const V &);
            
      /** \brief Removes all key-value pairs
        */
      void removeAll ();
      
      /** \brief Returns the list of all defined keys in the map.

          \par Example:
\code          
   SxMap<SxString,SxString> env;
   env("SX_INCLUDE_PATH") = "/home/johndoe";
   env("PATH") = "/bin:/usr/bin";
   cout << env.getKeys() << endl;
\endcode
         yields the output "SX_INCLUDE_PATH" and "PATH".  
         \sa getValues */
      const SxList<K> &getKeys () const;

      /** \brief Returns the list of all defined values in the map.

          \par Example:
\code          
   SxMap<SxString,SxString> env;
   env("SX_INCLUDE_PATH") = "/home/johndoe";
   env("PATH") = "/bin:/usr/bin";
   cout << env.getValues() << endl;
\endcode
         yields the output "/home/johndoe" and "/bin:/usr/bin".  
         \sa getKeys
       */
      const SxList<V> &getValues () const;

      /** \brief Create iterator pointing to the begin of the map

          Use this function find the first key-value pair.
          \par Example:
\code
   SxMap<SxString,SxString> hash = ...;
   SxMap<SxString,SxString>::Iterator it;
   for (it = hash.begin(); it != hash.end(); ++it)  {
      cout << (*it).getKey() << " = " << (*it).getValue () << endl;
   }
\endcode */
      typename SxMap<K,V,H>::ConstIterator begin () const;
      typename SxMap<K,V,H>::Iterator begin ();


      /** \brief Create iterator pointing to the end of the map

          Use this function find the last key-value pair.
          \par Example:
\code
   SxMap<SxString,SxString> hash = ...;
   SxMap<SxString,SxString>::Iterator it;
   for (it = hash.begin(); it != hash.end(); ++it)  {
      cout << (*it).getKey() << " = " << (*it).getValue () << endl;
   }
\endcode */
      typename SxMap<K,V,H>::ConstIterator end () const;
      typename SxMap<K,V,H>::Iterator end ();

      /** \brief append map to existing map

          This function extends an existing map. It appends key/value pairs of
          the existing map. If a key exists already in the current map it can 
          either be overwritten or ignored. */
      SxMap<K,V,H> &append (const SxMap<K,V,H> &map, bool overwrite=true);

      /** \brief print the content of a map 

          This function can be used to print all key-value pairs to
          stdout. Useful for debugging purposes. */
      void print () const;

   protected:
      SxHashTable<K,V,H> table;
      SxList<K> keys;
      SxList<V> values;
};

#include <SxMap.hpp>

#endif /* _SX_MAP_H_ */
