// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

template<class K, class V, class H>
SxMap<K,V,H>::SxMap ()
{
   // empty
}

template<class K, class V, class H>
SxMap<K,V,H>::SxMap (const SxMap<K,V,H> &map_)
{
   append (map_, true);
}

template<class K, class V, class H>
SxMap<K,V,H> &SxMap<K,V,H>::operator= (const SxMap<K,V,H> &map_)
{
   removeAll ();
   return append (map_, true);
}

template<class K, class V, class H>
SxMap<K,V,H>::~SxMap ()
{
   removeAll ();
}

template<class K, class V, class H>
inline V &SxMap<K,V,H>::operator() (const K &key_)
{
   ssize_t idx = -1;
   if (!table.contains (key_, &idx))  {
      keys.append (key_);
      table.elements[idx] = keys.lastElement;
      values.append (V ());
      table.value(idx) = values.lastElement;
      return values.lastElement->element;
   }  else  {
      return table.value(idx)->element;
   }
}

template<class K, class V, class H>
inline const V &SxMap<K,V,H>::operator() (const K &key_) const
{
   ssize_t idx = table.findPos (key_);
   if (idx >= 0)  {
      return table.value(idx)->element;
   }
   throw SxException("The map has no element for this key.",__FILE__,__LINE__);
}

template<class K, class V, class H>
ssize_t SxMap<K,V,H>::findKey (const K &key_) const
{
   return table.findPos (key_);
}

template<class K, class V, class H>
V &SxMap<K,V,H>::getValueIdx (ssize_t idx_)
{
   return table.value(idx_)->element;
}

template<class K, class V, class H>
const V &SxMap<K,V,H>::getValueIdx (ssize_t idx_) const
{
   return table.value(idx_)->element;
}

template<class K, class V, class H>
ssize_t SxMap<K,V,H>::getSize () const
{
   return keys.getSize ();
}

template<class K, class V, class H>
inline bool SxMap<K,V,H>::containsKey (const K &key_) const
{
   return table.findPos (key_) >= 0;
}

template<class K, class V, class H>
bool SxMap<K,V,H>::containsValue (const V &value_) const
{
   typename SxList<V>::ConstIterator it = values.begin ();
   typename SxList<V>::ConstIterator itEnd = values.end ();
      
   for (; it != itEnd; ++it)  {
      if (*it == value_)  {
         return true;
      }
   }
   return false;
}

template<class K, class V, class H>
bool SxMap<K,V,H>::removeKey (const K &key_)
{
   ssize_t idx = table.findPos (key_);
   if (idx >= 0)  {
      values.removeItem (table.value(idx));
      keys.removeItem (table.elements[idx]);
      table.remove (idx);
      return true;
   }
   return false;
}

template<class K, class V, class H>
bool SxMap<K,V,H>::removeValue (const V &value_)
{
   typename SxList<V>::ConstIterator it = values.begin ();
   typename SxList<K>::ConstIterator itK = keys.begin ();
   typename SxList<V>::ConstIterator itEnd = values.end ();
      
   for (; it != itEnd; ++it, ++itK)  {
      if (*it == value_)  {
         return removeKey (*itK);
      }
   }
   return false;
}

template<class K, class V, class H>
void SxMap<K,V,H>::removeAll ()
{
   table.removeAll ();
   keys.removeAll ();
   values.removeAll ();
}

template<class K, class V, class H>
const SxList<K> &SxMap<K,V,H>::getKeys () const
{
   return keys;
}

template<class K, class V, class H>
const SxList<V> &SxMap<K,V,H>::getValues () const
{
   return values;
}

template<class K, class V, class H>
typename SxMap<K,V,H>::ConstIterator SxMap<K,V,H>::begin () const
{
   return ConstIterator (keys.begin (), values.begin ());
}

template<class K, class V, class H>
typename SxMap<K,V,H>::Iterator SxMap<K,V,H>::begin ()
{
   return Iterator (keys.begin (), values.begin ());
}

template<class K, class V, class H>
typename SxMap<K,V,H>::ConstIterator SxMap<K,V,H>::end () const
{
   return ConstIterator (keys.end (), values.end ());
}

template<class K, class V, class H>
typename SxMap<K,V,H>::Iterator SxMap<K,V,H>::end ()
{
   return Iterator (keys.end (), values.end ());
}

template<class K, class V, class H>
SxMap<K,V,H> &SxMap<K,V,H>::append (const SxMap<K,V,H> &map, bool overwrite)
{
   ConstIterator it;
   for (it = map.begin(); it != map.end(); ++it)  {
      if (overwrite || !containsKey (it.getKey ()))  {
         operator()(it.getKey ()) = it.getValue ();
      }
   }
   return *this;
}

template<class K, class V, class H>
void SxMap<K,V,H>::print () const
{
   ConstIterator it;
   for (it = begin(); it != end(); ++it)  {
      std::cout << it.getKey () << " = " << it.getValue () << std::endl;
   }
}

// --------------------------------------------------------------------------
template<class K, class V, class H>
SxMap<K,V,H>::ConstIterator::ConstIterator ()
{
   // empty
}

template<class K, class V, class H>
SxMap<K,V,H>::ConstIterator::ConstIterator (const Iterator &it_)
   : itK(it_.itK),
     itV(it_.itV)
{
   // empty
}

template<class K, class V, class H>
SxMap<K,V,H>::ConstIterator::ConstIterator (
   const typename SxList<K>::ConstIterator &itK_,
   const typename SxList<V>::ConstIterator &itV_)
   : itK(itK_),
     itV(itV_)
{
   // empty
}

template<class K, class V, class H>
SxMap<K,V,H>::ConstIterator::~ConstIterator ()
{
   // empty
}

template<class K, class V, class H>
bool 
SxMap<K,V,H>::ConstIterator::operator== (
      const typename SxMap<K,V,H>::ConstIterator &in) const
{
   return itK == in.itK;
}

template<class K, class V, class H>
bool SxMap<K,V,H>::ConstIterator::operator!= (
      const typename SxMap<K,V,H>::ConstIterator &in) const
{
   return !operator== (in);
}

template<class K, class V, class H>
const typename SxMap<K,V,H>::ConstIterator
SxMap<K,V,H>::ConstIterator::operator++ (int)
{
   typename SxMap<K,V,H>::ConstIterator copy (itK, itV);
   ++itK;
   ++itV;
   return copy;
}

template<class K, class V, class H>
typename SxMap<K,V,H>::ConstIterator &SxMap<K,V,H>::ConstIterator::operator++ ()
{
   ++itK;
   ++itV;
   return *this;
}

template<class K, class V, class H>
const typename SxMap<K,V,H>::ConstIterator
SxMap<K,V,H>::ConstIterator::operator-- (int)
{
   typename SxMap<K,V,H>::ConstIterator copy (itK, itV);
   --itK;
   --itV;
   return copy;
}

template<class K, class V, class H>
typename SxMap<K,V,H>::ConstIterator &SxMap<K,V,H>::ConstIterator::operator-- ()
{
   --itK;
   --itV;
   return *this;
}

template<class K, class V, class H>
const K &SxMap<K,V,H>::ConstIterator::getKey () const
{
   return *itK;
}

template<class K, class V, class H>
const V &SxMap<K,V,H>::ConstIterator::getValue () const
{
   return *itV;
}

template<class K, class V, class H>
const V &SxMap<K,V,H>::ConstIterator::operator-> () const
{
   return *itV;
}

// --------------------------------------------------------------------------

template<class K, class V, class H>
SxMap<K,V,H>::Iterator::Iterator ()
{
   // empty
}

template<class K, class V, class H>
SxMap<K,V,H>::Iterator::Iterator (const typename SxList<K>::Iterator &itK_,
                                  const typename SxList<V>::Iterator &itV_)
   : itK(itK_),
     itV(itV_)
{
   // empty
}

template<class K, class V, class H>
SxMap<K,V,H>::Iterator::~Iterator ()
{
   // empty
}

template<class K, class V, class H>
bool SxMap<K,V,H>::Iterator::operator== (
   const typename SxMap<K,V,H>::Iterator &in)
{
   return itK == in.itK;
}

template<class K, class V, class H>
bool SxMap<K,V,H>::Iterator::operator!= (
   const typename SxMap<K,V,H>::Iterator &in)
{
   return !operator== (in);
}

template<class K, class V, class H>
const typename SxMap<K,V,H>::Iterator
SxMap<K,V,H>::Iterator::operator++ (int)
{
   typename SxMap<K,V,H>::Iterator copy (itK, itV);
   ++itK;
   ++itV;
   return copy;
}

template<class K, class V, class H>
typename SxMap<K,V,H>::Iterator &SxMap<K,V,H>::Iterator::operator++ ()
{
   ++itK;
   ++itV;
   return *this;
}

template<class K, class V, class H>
const typename SxMap<K,V,H>::Iterator
SxMap<K,V,H>::Iterator::operator-- (int)
{
   typename SxMap<K,V,H>::Iterator copy (itK, itV);
   --itK;
   --itV;
   return copy;
}

template<class K, class V, class H>
typename SxMap<K,V,H>::Iterator &SxMap<K,V,H>::Iterator::operator-- ()
{
   --itK;
   --itV;
   return *this;
}

template<class K, class V, class H>
const K &SxMap<K,V,H>::Iterator::getKey ()
{
   return *itK;
}

template<class K, class V, class H>
V &SxMap<K,V,H>::Iterator::getValue ()
{
   return *itV;
}

template<class K, class V, class H>
V &SxMap<K,V,H>::Iterator::operator-> ()
{
   return *itV;
}

// --------------------------------------------------------------------------
/** \brief Maps keys to values

    \b SxMap = SPHInX Key-Value Map

    \author Sixten Boeck, boeck@gemmantics.com
    \author Vaclav Bubnik, bubnik@gemmantics.com */
template<class K, class V>
class SxMap<K,V,SxNull>
{
   public:
      typedef K  Key;
      typedef V  Value;

      class Iterator;
      
      class ConstIterator
      {
         public:
            ConstIterator ();
            ConstIterator (const Iterator &it);
            ConstIterator (const typename SxList<K>::ConstIterator &,
                           const typename SxList<V>::ConstIterator &);
           ~ConstIterator ();

            bool operator== (const typename SxMap<K,V,SxNull>::ConstIterator &)
                    const;
            bool operator!= (const typename SxMap<K,V,SxNull>::ConstIterator &)
                    const;
            const typename SxMap<K,V,SxNull>::ConstIterator  operator++ (int);
                  typename SxMap<K,V,SxNull>::ConstIterator &operator++ ();
            const typename SxMap<K,V,SxNull>::ConstIterator  operator-- (int);
                  typename SxMap<K,V,SxNull>::ConstIterator &operator-- ();
                  
            const K &getKey () const;
            const V &getValue () const;
            const V &operator->() const;

         protected:
            typename SxList<K>::ConstIterator itK;
            typename SxList<V>::ConstIterator itV;
      };
      
      class Iterator
      {
         public:
            Iterator ();
            Iterator (const typename SxList<K>::Iterator &,
                      const typename SxList<V>::Iterator &);
           ~Iterator ();

            bool operator== (const typename SxMap<K,V,SxNull>::Iterator &);
            bool operator!= (const typename SxMap<K,V,SxNull>::Iterator &);
            const typename SxMap<K,V,SxNull>::Iterator  operator++ (int);
                  typename SxMap<K,V,SxNull>::Iterator &operator++ ();
            const typename SxMap<K,V,SxNull>::Iterator  operator-- (int);
                  typename SxMap<K,V,SxNull>::Iterator &operator-- ();
                  
            const K &getKey ();
            V &getValue ();
            V &operator->();

            friend class ConstIterator;

         protected:
            typename SxList<K>::Iterator itK;
            typename SxList<V>::Iterator itV;
      };
      
      SxMap ();
      SxMap (const SxMap<K,V,SxNull> &);
      ~SxMap ();
      
      inline V &operator() (const K &key);
      inline const V &operator() (const K &key) const;
      
      SxMap<K,V,SxNull> &operator= (const SxMap<K,V,SxNull> &);

      ssize_t findKey (const K &) const;
            V &getValueIdx (ssize_t idx);
      const V &getValueIdx (ssize_t idx) const;

      bool containsKey (const K &) const;
      bool containsValue (const V &) const;
      
      ssize_t getSize () const { return keys.getSize (); }
      
      bool removeKey (const K &);
      bool removeValue (const V &);
      void removeAll ();
      
      const SxList<K> &getKeys () const { return keys; }
      const SxList<V> &getValues () const { return values; }
      
      typename SxMap<K,V,SxNull>::ConstIterator begin () const;
      typename SxMap<K,V,SxNull>::Iterator begin ();
      typename SxMap<K,V,SxNull>::ConstIterator end () const;
      typename SxMap<K,V,SxNull>::Iterator end ();
      
      SxMap<K,V,SxNull> &append (const SxMap<K,V,SxNull> &map,
                                 bool                    overwrite=true);
      
      void print () const;

   protected:
      SxList<K> keys;
      SxList<V> values;
};

template<class K, class V>
SxMap<K,V,SxNull>::SxMap ()
{
   // empty
}

template<class K, class V>
SxMap<K,V,SxNull>::SxMap (const SxMap<K,V,SxNull> &map_)
{
   append (map_, true);
}

template<class K, class V>
SxMap<K,V,SxNull> &SxMap<K,V,SxNull>::operator= (const SxMap<K,V,SxNull> &map_)
{
   removeAll ();
   return append (map_, true);
}

template<class K, class V>
SxMap<K,V,SxNull>::~SxMap ()
{
   removeAll ();
}

template<class K, class V>
inline V &SxMap<K,V,SxNull>::operator() (const K &key_)
{
   typename SxList<K>::Iterator itKey = keys.begin ();
   typename SxList<K>::Iterator itEnd = keys.end ();
   typename SxList<V>::Iterator itValue = values.begin ();
      
   for (; itKey != itEnd; ++itKey, ++itValue)  {
      if (*itKey == key_)  {
         return *itValue;
      }
   }
   
   keys << key_;
   values << V();
   
   return values.lastElement->element;
}

template<class K, class V>
inline const V &SxMap<K,V,SxNull>::operator() (const K &key_) const
{
   typename SxList<K>::ConstIterator itKey = keys.begin ();
   typename SxList<K>::ConstIterator itEnd = keys.end ();
   typename SxList<V>::ConstIterator itValue = values.begin ();
      
   for (; itKey != itEnd; ++itKey, ++itValue)  {
      if (*itKey == key_)  {
         return *itValue;
      }
   }

   throw SxException("The map has no element for this key.",__FILE__,__LINE__);
}

template<class K, class V>
ssize_t SxMap<K,V,SxNull>::findKey (const K &key_) const
{
   return keys.findPos (key_);
}

template<class K, class V>
V &SxMap<K,V,SxNull>::getValueIdx (ssize_t idx_)
{
   return values(idx_);
}

template<class K, class V>
const V &SxMap<K,V,SxNull>::getValueIdx (ssize_t idx_) const
{
   return values(idx_);
}

template<class K, class V>
inline bool SxMap<K,V,SxNull>::containsKey (const K &key_) const
{
   return keys.contains (key_);
}

template<class K, class V>
bool SxMap<K,V,SxNull>::containsValue (const V &value_) const
{
   return values.contains (value_);
}

template<class K, class V>
bool SxMap<K,V,SxNull>::removeKey (const K &key_)
{
   ssize_t idx = keys.findPos (key_);
   if (idx >= 0)  {
      keys.remove (idx);
      values.remove (idx);
      return true;
   }
   return false;
}

template<class K, class V>
bool SxMap<K,V,SxNull>::removeValue (const V &value_)
{
   ssize_t idx = values.findPos (value_);
   if (idx >= 0)  {
      keys.remove (idx);
      values.remove (idx);
      return true;
   }
   return false;
}

template<class K, class V>
void SxMap<K,V,SxNull>::removeAll ()
{
   keys.removeAll ();
   values.removeAll ();
}

template<class K, class V>
typename SxMap<K,V,SxNull>::ConstIterator SxMap<K,V,SxNull>::begin () const
{
   return ConstIterator (keys.begin (), values.begin ());
}

template<class K, class V>
typename SxMap<K,V,SxNull>::Iterator SxMap<K,V,SxNull>::begin ()
{
   return Iterator (keys.begin (), values.begin ());
}

template<class K, class V>
typename SxMap<K,V,SxNull>::ConstIterator SxMap<K,V,SxNull>::end () const
{
   return ConstIterator (keys.end (), values.end ());
}

template<class K, class V>
typename SxMap<K,V,SxNull>::Iterator SxMap<K,V,SxNull>::end ()
{
   return Iterator (keys.end (), values.end ());
}

template<class K, class V>
SxMap<K,V,SxNull> &SxMap<K,V,SxNull>::append (const SxMap<K,V,SxNull> &map_,
                                              bool                   overwrite_)
{
   ConstIterator it = map_.begin ();
   ConstIterator itEnd = map_.end ();
   for (; it != itEnd; ++it)  {
      if (overwrite_ || !containsKey (it.getKey ()))  {
         operator()(it.getKey ()) = it.getValue ();
      }
   }
   return *this;
}

template<class K, class V>
void SxMap<K,V,SxNull>::print () const
{
   ConstIterator it;
   for (it = begin(); it != end(); ++it)  {
      std::cout << it.getKey () << " = " << it.getValue () << std::endl;
   }
}

// --------------------------------------------------------------------------

template<class K, class V>
SxMap<K,V,SxNull>::ConstIterator::ConstIterator ()
{
   // empty
}

template<class K, class V>
SxMap<K,V,SxNull>::ConstIterator::ConstIterator (const Iterator &it_)
   : itK(it_.itK),
     itV(it_.itV)
{
   // empty
}

template<class K, class V>
SxMap<K,V,SxNull>::ConstIterator::ConstIterator (
   const typename SxList<K>::ConstIterator &itK_,
   const typename SxList<V>::ConstIterator &itV_)
   : itK(itK_),
     itV(itV_)
{
   // empty
}

template<class K, class V>
SxMap<K,V,SxNull>::ConstIterator::~ConstIterator ()
{
   // empty
}

template<class K, class V>
bool 
SxMap<K,V,SxNull>::ConstIterator::operator== (
      const typename SxMap<K,V,SxNull>::ConstIterator &in) const
{
   return itK == in.itK;
}

template<class K, class V>
bool SxMap<K,V,SxNull>::ConstIterator::operator!= (
      const typename SxMap<K,V,SxNull>::ConstIterator &in) const
{
   return !operator== (in);
}

template<class K, class V>
const typename SxMap<K,V,SxNull>::ConstIterator
SxMap<K,V,SxNull>::ConstIterator::operator++ (int)
{
   typename SxMap<K,V,SxNull>::ConstIterator copy (itK, itV);
   ++itK;
   ++itV;
   return copy;
}

template<class K, class V>
typename SxMap<K,V,SxNull>::ConstIterator
&SxMap<K,V,SxNull>::ConstIterator::operator++ ()
{
   ++itK;
   ++itV;
   return *this;
}

template<class K, class V>
const typename SxMap<K,V,SxNull>::ConstIterator
SxMap<K,V,SxNull>::ConstIterator::operator-- (int)
{
   typename SxMap<K,V,SxNull>::ConstIterator copy (itK, itV);
   --itK;
   --itV;
   return copy;
}

template<class K, class V>
typename SxMap<K,V,SxNull>::ConstIterator
&SxMap<K,V,SxNull>::ConstIterator::operator-- ()
{
   --itK;
   --itV;
   return *this;
}

template<class K, class V>
const K &SxMap<K,V,SxNull>::ConstIterator::getKey () const
{
   return *itK;
}

template<class K, class V>
const V &SxMap<K,V,SxNull>::ConstIterator::getValue () const
{
   return *itV;
}

template<class K, class V>
const V &SxMap<K,V,SxNull>::ConstIterator::operator-> () const
{
   return *itV;
}

// --------------------------------------------------------------------------

template<class K, class V>
SxMap<K,V,SxNull>::Iterator::Iterator ()
{
   // empty
}

template<class K, class V>
SxMap<K,V,SxNull>::Iterator::Iterator (const typename SxList<K>::Iterator &itK_,
                                       const typename SxList<V>::Iterator &itV_)
   : itK(itK_),
     itV(itV_)
{
   // empty
}

template<class K, class V>
SxMap<K,V,SxNull>::Iterator::~Iterator ()
{
   // empty
}

template<class K, class V>
bool SxMap<K,V,SxNull>::Iterator::operator== (
   const typename SxMap<K,V,SxNull>::Iterator &in)
{
   return itK == in.itK;
}

template<class K, class V>
bool SxMap<K,V,SxNull>::Iterator::operator!= (
   const typename SxMap<K,V,SxNull>::Iterator &in)
{
   return !operator== (in);
}

template<class K, class V>
const typename SxMap<K,V,SxNull>::Iterator
SxMap<K,V,SxNull>::Iterator::operator++ (int)
{
   typename SxMap<K,V,SxNull>::Iterator copy (itK, itV);
   ++itK;
   ++itV;
   return copy;
}

template<class K, class V>
typename SxMap<K,V,SxNull>::Iterator &SxMap<K,V,SxNull>::Iterator::operator++ ()
{
   ++itK;
   ++itV;
   return *this;
}

template<class K, class V>
const typename SxMap<K,V,SxNull>::Iterator
SxMap<K,V,SxNull>::Iterator::operator-- (int)
{
   typename SxMap<K,V,SxNull>::Iterator copy (itK, itV);
   --itK;
   --itV;
   return copy;
}

template<class K, class V>
typename SxMap<K,V,SxNull>::Iterator &SxMap<K,V,SxNull>::Iterator::operator-- ()
{
   --itK;
   --itV;
   return *this;
}

template<class K, class V>
const K &SxMap<K,V,SxNull>::Iterator::getKey ()
{
   return *itK;
}

template<class K, class V>
V &SxMap<K,V,SxNull>::Iterator::getValue ()
{
   return *itV;
}

template<class K, class V>
V &SxMap<K,V,SxNull>::Iterator::operator-> ()
{
   return *itV;
}
