// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_MARKER_H_
#define _SX_MARKER_H_

#include <SxUtil.h>
#include <SxString.h>
#include <SxMap.h>

/** \brief substitute markers

    \b SxMarker = S/PHI/nX Text Marker Substitution

    The S/PHI/nX Text markers have been inspired by typoscript, typo3.org.
    Search for the string "***marker***" and replace it with the given
    value:
\code
       ***marker***
\endcode
    Search for the string "***marker***". If the corresponding value is
    not provided replace it with the default marker value:
\code        
       ***marker:default***
\endcode
    Search for the string "***marker***" and replace according to an 
    expression. The value of the marker can be accessed with the
    [#] marker.
\code
       The value is ***marker<0:less:greater than zero (value=[#])***
\endcode
   Here a source code snippet:
\code
SxString marker, value;
SxString cmd = "qsub ***email:johndoe@mail*** -l ***mem_free=***mem:4***gb
   cout << ">" << cmd << "<\n";
   SxMarker markers
   markers("mem") = "10";
   markers("np") = "abc";
   markers("name") = "myEmail@abc.de";
   cout << ">" << markers.apply(cmd) << "<\n";

\endcode
    \author Sixten Boeck */
class SX_EXPORT_UTIL SxMarker
{
   public:
      typedef SxMap<SxString,SxString>::Iterator      Iterator;
      typedef SxMap<SxString,SxString>::ConstIterator ConstIterator;

      SxMarker (const SxString &border_="*");
      SxMarker (const SxMap<SxString,SxString> &, const SxString &border_="*");
      virtual ~SxMarker ();


      virtual SxString function (const SxString &func,
                                 const SxString &arg) const;

      SxString &operator() (const SxString &);        // set new marker
      SxString  operator() (const SxString &) const;  // read marker
      // apply marker
      SxString apply (const SxString &, const SxString &border_="") const; 

      Iterator      begin ();
      ConstIterator begin () const;
      Iterator      end ();
      ConstIterator end () const;

      SxList<SxString> getKeys () const;
      SxList<SxString> getValues () const;

      void append (const SxMarker &);

      bool containsKey (const SxString &) const;
      static SxList<SxString> getUnresolved (const SxString &tmpl, 
                                             const SxString &border_="*",
                                             ssize_t maxSize=20);

      void print () const;

   //protected:

      SxString borderDef;
      SxMap<SxString,SxString> markers;

};

#endif /* _SX_MARKER_H_ */
