// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_MEMORY_H_
#define _SX_MEMORY_H_

#include <SxError.h>
#include <SxUtil.h>
#include <SxString.h>
#include <iostream>
#include <stdlib.h>


extern SX_EXPORT_UTIL size_t maxCachedSize[2];



/** \brief Memory handler for the SPHInX Algebra Library.

    The vector class allocates many temporary objects of the same size
    during the evaluations of algebraic equations. These memory blocks
    allocated in this class. When delete/free is invoked from the vector
    class the existing memory block is actually cached rather than removed.
    The next allocation call which requests the same number of bytes returns
    the cached memory block. No further OS function call is performed.
    On some platforms this procedure seems to be significantly faster then
    the corresponding malloc/free calls from the OS. 

    \author Sixten Boeck, boeck@mpie.de */
template<class T>
class SxMemory
{
   public:

      SxMemory () { }
      ~SxMemory () { }

      static SxMemory<T> &getObj () {
         static SxMemory<T> obj;
         return obj;
      }

      inline T *malloc (size_t nElem);
      inline void *malloc (size_t nElem, size_t bytesPerElem);
      inline void free (T *ptr, size_t nElem);
      inline void free (void *ptr, size_t nElem, size_t bytesPerElem);

      void init ();
      void clean (unsigned short id);
      void destroy ();
      long getAccesses (unsigned short id);
      size_t getMaxMemory (unsigned short id);
      
   protected:
      void  *buffer[2];
      size_t nBytes[2];
      long   accessCounter[2];

      /// Default alignment (16)
      static inline size_t getAlignment ()
      {
         return 16;
      }
};



template<> 
inline size_t SxMemory<int>::getAlignment ()
{ 
   return sizeof(void*); 
}

template<> 
inline size_t SxMemory<long>::getAlignment ()
{ 
   return sizeof(void*); 
}

template<class T>
T *SxMemory<T>::malloc (size_t nElem)
{
   const short id = 0;   //  slot for SxVector<?>::elements
   T *resPtr;
   size_t arraySize = nElem * sizeof(T);
   if (buffer[id])  {
      if (nBytes[id] == arraySize)  {
         resPtr = (T *)buffer[id];
         buffer[id] = NULL;
         accessCounter[id]++;
         return resPtr;
      }  else  {
         // --- operations with a new vector size started: flush buffer
         clean (id);
      }
   }	
#ifdef HAVE_POSIX_MEMALIGN
   void *ptr;
   if ( posix_memalign(&ptr, getAlignment (), arraySize) == 0)
      return (T *)ptr;
#else
   // warning "unaligned malloc may break high-performance code"
   resPtr = (T *)::malloc (arraySize);
   if (resPtr)  return resPtr;
#endif

   cout << SEPARATOR;
   cout << "|Insufficient memory.\n";
   cout << "| Can't allocate further " << nElem << " elements. "
      << "(" << arraySize << " bytes).";
   SX_EXIT;
   return NULL;
}


template<class T>
void *SxMemory<T>::malloc (size_t nElem, size_t bytesPerElem)
{
   const short id = 1;  // slot for SxVector<?>::handle
   void *resPtr;
   size_t arraySize = nElem * bytesPerElem;
   if (buffer[id])  {
      if (nBytes[id] == arraySize)  {
         resPtr = (void *)buffer[id];
         buffer[id] = NULL;
         accessCounter[id]++;
         return resPtr;
      }  else  {
         // --- operations with a new vector size started: flush buffer
         clean (id);
      }
   }	
   resPtr = ::malloc (arraySize);
   if (resPtr)  return resPtr;

   cout << SEPARATOR;
   cout << "|Insufficient memory.\n";
   cout << "| Can't allocate further " << nElem << " elements. "
      << "(" << arraySize << " bytes).";
   SX_EXIT;
   return NULL;
}



template<class T>
void SxMemory<T>::free (T *ptr, size_t nElem) 
{
   const short id = 0;  // SxVector<?>::elements slot
   if (!buffer[id])  {
      size_t arraySize = nElem * sizeof(T);
      buffer[id]  = (void *)ptr;
      nBytes[id]  = arraySize;
      if (maxCachedSize[id] < arraySize)  maxCachedSize[id] = arraySize;
   }  else  {
      ::free (ptr);
   }
}


template<class T>
void SxMemory<T>::free (void *ptr, size_t nElem, size_t bytesPerElem)
{
   SX_CHECK (ptr);
   const short id = 1;  // SxVector<?>::handle slot
   if (!buffer[id])  {
      size_t arraySize = nElem * bytesPerElem;
      buffer[id]  = (void *)ptr;
      nBytes[id]  = arraySize;
      if (maxCachedSize[id] < arraySize)  maxCachedSize[id] = arraySize;
   }  else  {
    ::free (ptr);
   }
}


template<class T>
void SxMemory<T>::init ()
{
   buffer[0] = buffer[1] = NULL;
   nBytes[0] = nBytes[1] = (size_t)0;
   maxCachedSize[0] = maxCachedSize[1] = (size_t)0;
   accessCounter[0] = accessCounter[1] = 0;
}


template<class T>
void SxMemory<T>::clean (unsigned short id)
{
   SX_CHECK_VAR (id < 2, id);
   if (buffer[id])  ::free (buffer[id]);
   buffer[id] = NULL;
   nBytes[id] = (size_t)0;
}


template<class T>
void SxMemory<T>::destroy ()
{
   clean (1);
   clean (0);
}


template<class T>
long SxMemory<T>::getAccesses (unsigned short id)
{
   SX_CHECK_VAR (id < 2, id);
   return accessCounter[id];
}


template<class T>
size_t SxMemory<T>::getMaxMemory (unsigned short id)
{
   SX_CHECK_VAR (id < 2, id);
   return maxCachedSize[id];
}




#endif /* _SX_MEMORY_H_ */
