// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_SIGSLOTS_H_
#define _SX_SIGSLOTS_H_

#include <SxSignals.h>
#include <SxPtr.h>

/** \brief Type safe signal slot mechanism

    connect signals and slots:

\code
#include <SxSigSlots.h>

   class MySlider
{
   public:
      MySlider () { }

   public slots:
 
      void slotABC (double val)
      {
         printf ("MySlider::slotABC received: %g\n", val);
      }

   signals:

      SxSignal<void,double> SX_SIGNAL(valueChanged);
};




class MyIsosurface
{
   public:
      MyIsosurface () { }
 
      void generateIso (double threshold)
      {
         printf ("MyIsosurface::slotABC received: %g\n", threshold);
      }

};


int main ()
{
   MySlider slider;
   SxPtr<MyIsosurface> isoPtr;

   {
      SxPtr<MyIsosurface> iso = SxPtr<MyIsosurface>::create ();
      sxconnect (&slider, valueChanged, iso, MyIsosurface::generateIso);

      isoPtr = iso;
      slider.valueChanged.send (5.0);
   }
\endcode
    \author Sixten Boeck, boeck@mpie.de */
#define sxconnect(SRC,SIGNAL,DST,SLOT)                                      \
   SxSigConnector::registerPtr(DST,SxSigConnector::getPtr(SRC)->SIGNAL);    \
   SxSigConnector::getPtr(SRC)->SIGNAL.connect (SxSigConnector::getPtr(DST),\
                                                &SLOT);

#define sxdisconnect(SRC,SIGNAL,DST,SLOT) \
   SxSigConnector::deregisterPtr(DST,SxSigConnector::getPtr(SRC)->SIGNAL);     \
   SxSigConnector::getPtr(SRC)->SIGNAL.disconnect (SxSigConnector::getPtr(DST),\
                                                   &SLOT);

/* \brief Signal connector

   \b SxSigConnector = S/PHI/nX Signal Slot Connector

   This wrapper class is used by the sxconnect macro.

   \author Sixten Boeck, boeck@mpie.de */
struct SxSigConnector  
{

   template<class T>
   static void *getPtr(int)
   {
      return NULL;
   }

   template<class T>
   static T *getPtr(T *p)
   { 
      SX_CHECK (p);
      return p; 
   }

   template<class T>
   static T *getPtr(const SxPtr<T> &p)
   {
      SX_CHECK (p.getPtr());
      return p.getPtr(); 
   }

   template<class T>
   static void registerPtr (int) { /* empty */ }

   template<class T>
   static void registerPtr (T *) { /* empty */ }

   template<class SIGNAL,class T>
   static void registerPtr (const SxPtr<T> &obj, SIGNAL &signal)
   {
      SX_CHECK (obj.getPtr());
      obj.receiver->sigDeleting.connect (&signal, &SIGNAL::slotDeregister);
   }

   template<class T>
   static void deregisterPtr (int) { /* empty */ }

   template<class T>
   static void deregisterPtr (T *) { /* empty */ }

   template<class SIGNAL,class T>
   static void deregisterPtr (const SxPtr<T> &obj, SIGNAL &signal)
   {
      SX_CHECK (obj.getPtr());
      obj.receiver->sigDeleting.disconnect (&signal, &SIGNAL::slotDeregister);
   }
};



#endif /* _SX_SIGSLOTS_H_ */
