
// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_SIGNALS_H_
#define _SX_SIGNALS_H_

#include <SxCBoundPtr.h>
#include <SxList.h>

// --- support QT-like doxygen documentation 
#ifdef BUILD_DOXYGEN
#   define SX_SIGNAL(name)  name()
#else
#   define slots    /* empty */
#   undef signals
#   define signals  public /* empty */
#   define SX_SIGNAL(name)  name
#endif

/** \brief Signal/Slot handling

    \b SxSignals = S/PHI/nX Signals

    This class provides the base for a Signal/Slot mechanism in S/PHI/nX. 
\code
#include <SxSigSlots.h>
class MySlider
{
   public:
      MySlider ()
      {
         sxconnect (this, valueChanged, this, MySlider::slotABC);
      }

   public slots:
 
      void slotABC (double val)
      {
         printf ("AAA::slotABC received: %g\n", val);
      }

   public signals:
      SxSignal<void,double> SIGNAL(valueChanged);
};

class MyIsosurface
{
   public:
      MyIsosurface () { }
 
      void generateIso (double threshold)
      {
         printf ("BBB::slotABC received: %g\n", threshold);
      }

};


int main ()
{
   MySlider slider;
   MyIsosurface iso;

   sxconnect (&slider, valueChanged, &iso, MyIsosurface::generateIso);

// slider.valueChanged.connect (&fooIso);


   sxconnect (&slider, valueChanged, NULL, fooIso);
   slider.valueChanged.send (0.5);
}
\endcode
    \author Sixten Boeck, boeck@mpie.de */
// ---------------------------------------------------------------------------
template<class P1=SxNull, class P2=SxNull, class P3=SxNull, class P4=SxNull, class P5=SxNull, class P6=SxNull>
class SxSignal
{
   public:
      typedef SxCBoundPtr<void,P1,P2,P3,P4,P5,P6>  TBoundPtr;

      // --- global function slots
      template<class TP1>
      void connect (void *dst, void (*slot)(TP1))
      {
         SX_CHECK (!dst);
         receivers << TBoundPtr::create (slot);
      }
      template<class TP1,class TP2>
      void connect (void *dst, void (*slot)(TP1,TP2))
      {
         SX_CHECK (!dst);
         receivers << TBoundPtr::create (slot);
      }
      template<class TP1,class TP2,class TP3>
      void connect (void *dst, void (*slot)(TP1,TP2,TP3))
      {
         SX_CHECK (!dst);
         receivers << TBoundPtr::create (slot);
      }
      template<class TP1,class TP2,class TP3,class TP4>
      void connect (void *dst, void (*slot)(TP1,TP2,TP3,TP4))
      {
         SX_CHECK (!dst);
         receivers << TBoundPtr::create (slot);
      }
      template<class TP1,class TP2,class TP3,class TP4,class TP5>
      void connect (void *dst, void (*slot)(TP1,TP2,TP3,TP4,TP5))
      {
         SX_CHECK (!dst);
         receivers << TBoundPtr::create (slot);
      }
      template<class TP1,class TP2,class TP3,class TP4,class TP5,class TP6>
      void connect (void *dst, void (*slot)(TP1,TP2,TP3,TP4,TP5,TP6))
      {
         SX_CHECK (!dst);
         receivers << TBoundPtr::create (slot);
      }


      // --- member function slots
      template<class DST,class Class,class TP1>
      void connect (DST *dst, void (Class::* const &slot)(TP1))
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2))
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2,class TP3>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3))
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4))
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4,class TP5>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5))
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4,class TP5,class TP6>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5,TP6))
      {
         receivers << TBoundPtr::create (*dst, slot);
      }

      // --- const member function slots
      template<class DST,class Class,class TP1>
      void connect (DST *dst, void (Class::* const &slot)(TP1)const)
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2)const)
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2,class TP3>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3)const)
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4)const)
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4,class TP5>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5)const)
      {
         receivers << TBoundPtr::create (*dst, slot);
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4,class TP5,class TP6>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5,TP6)const)
      {
         receivers << TBoundPtr::create (*dst, slot);
      }


      void slotDeregister (void *p)
      {
         for (typename SxList<TBoundPtr>::Iterator it = receivers.begin();
              it != receivers.end (); 
              ++it)
         {
            if (it->callee == p) receivers.removeElement (*it);
         }
      }

      void disconnect (const TBoundPtr &cb)  
      {
         SX_CHECK (receivers.contains (cb));
         receivers.removeElement (cb);
      }

      template<class DST,class Class,class TP1>
      void disconnect (DST *dst, void (Class::* const &slot)(TP1))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,class TP1,class TP2>
      void disconnect (DST *dst, void (Class::* const &slot)(TP1,TP2))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,class TP1,class TP2,class TP3>
      void disconnect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4>
      void disconnect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,
               class TP1,class TP2,class TP3,
               class TP4,class TP5>
      void disconnect (DST *dst, 
                       void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,
               class TP1,class TP2,class TP3,
               class TP4,class TP5,class TP6>
      void disconnect (DST *dst, 
                       void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5,TP6))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class TP1>
      void send (const TP1 &p1)
      {
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1);
      }

      template<class TP1,class TP2>
      void send (const TP1 &p1, const TP2 &p2)
      {
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1,p2);
      }

      template<class TP1,class TP2,class TP3>
      void send (const TP1 &p1, const TP2 &p2, const TP3 &p3)
      {
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, p3);
      }

      template<class TP1,class TP2,class TP3, class TP4>
      void send (const TP1 &p1, const TP2 &p2, const TP3 &p3, const TP4 &p4)
      {
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, p3, p4);
      }

      template<class TP1,class TP2,class TP3, class TP4, class TP5>
      void send (const TP1 &p1, const TP2 &p2, const TP3 &p3, 
                 const TP4 &p4, const TP5 &p5)
      {
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, p3, p4, p5);
      }

      template<class TP1,class TP2,class TP3, class TP4, class TP5, class TP6>
      void send (const TP1 &p1, const TP2 &p2, const TP3 &p3, 
                 const TP4 &p4, const TP5 &p5, const TP6 &p6)
      {
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, p3, p4, p5, p6);
      }




   protected:
      SxList<TBoundPtr> receivers;
};


#endif /* _SX_SIGNALS_H_ */
