// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------
#ifndef _SX_UTIL_H_
#define _SX_UTIL_H_

#include <math.h>
#include <stdlib.h>
#include <SxConfig.h>

#ifdef WIN32
#  if defined(_EXPORT_sxutil)
#     define SX_EXPORT_UTIL __declspec(dllexport)
#  else
#     define SX_EXPORT_UTIL __declspec(dllimport)
#  endif
#else
#  define SX_EXPORT_UTIL
#endif

// --- define unique identifiers
//     Example:  int SX_UNIQUE_ID(abc)  => int abc##__LINE##
#define SX_UNIQUE_ID_NAME(var,x)  var ## x
#define SX_UNIQUE_ID_WRAP(var,x)  SX_UNIQUE_ID_NAME(var,x)
#define SX_UNIQUE_ID(var)         SX_UNIQUE_ID_WRAP(var,__LINE__)

// --- define unit blocks
//     Example: SERIALIZE {
//                 ...
//              }
//              define SX_SERIALIZE  SX_BLOCK(SxMutex)
#define SX_BLOCK(type) \
   bool SX_UNIQUE_ID(cond)=true; for (type SX_UNIQUE_ID(var); SX_UNIQUE_ID(cond); SX_UNIQUE_ID(cond) = false)

#define SX_BLOCK_ARG(type,arg) \
   bool SX_UNIQUE_ID(cond)=true; for (type SX_UNIQUE_ID(var)(arg); SX_UNIQUE_ID(cond); SX_UNIQUE_ID(cond) = false)



class SX_EXPORT_UTIL SxUtil
{
   public:
      SxUtil ();
     ~SxUtil ();

#ifndef SX_IOS
     static SxUtil &getGlobalInstance ();
#endif /* SX_IOS */

     int nProcs;
};

inline int SX_EXPORT_UTIL toInt (float arg)  { return (int)(floor(arg+.5)); }
inline int SX_EXPORT_UTIL toInt (double arg) { return (int)(floor(arg+.5)); }



extern char __flags_;
extern void SX_EXPORT_UTIL setNProcs (int n);
extern int  SX_EXPORT_UTIL getNProcs ();
extern long SX_EXPORT_UTIL sxChunkSize;

extern void SX_EXPORT_UTIL sxGetExecPath (char *res, size_t len); // #exception

extern void SX_EXPORT_UTIL sxchmod (const char *path, int mode); // #exception

/** \brief Test if debugging is enabled.
    \return 1 if debugging is enabled, 0 otherwise.
    
    \par Environment variables:
    \li SX_DEBUG_FILE specifies a comma separated list of filenames
    \li SX_DEBUG_FUNCTION specifies a comma separated list of function names
    
    \par Disable all debug messages:
\verbatim
> export SX_DEBUG_FILE=none
\endverbatim
    
    \par Example file main.cpp:
\code
#include <SxError.h>

class SxMyDate
{
   public:
      int year;
      SxMyDate ();
      void setYear (int year);
      int getYear () const;
};

SxMyDate::SxMyDate ()
   : year (1970)
{
   SX_DBG_MSG ("default year="<<year);
}

void SxMyDate::setYear (int year_)
{
   year = year_;
   SX_DBG_MSG ("year="<<year);
}

int SxMyDate::getYear () const
{
   SX_DBG_MSG ("year="<<year);
   return year;
}

int main (int, char **)
{
   SX_TRACE ();
   
   SxMyDate date;
   date.setYear (2010);
   date.getYear ();
   
   return 0;
}
\endcode

   \par Output:
   \li with > export SX_DEBUG_FILE=
   \li or with > export SX_DEBUG_FILE=main.cpp
   \li or with > export SX_DEBUG_FILE=main.cpp,SxError.h
\verbatim
main.cpp main: BEGIN
main.cpp SxMyDate: default year=1970
main.cpp setYear: year=2010
main.cpp getYear: year=2010
main.cpp main: END 1.359 ms
\endverbatim

   \li with > export SX_DEBUG_FUNCTION=setYear
\verbatim
main.cpp setYear: year=2010
\endverbatim

   \li with > export SX_DEBUG_FUNCTION=setYear:getYear
\verbatim
main.cpp setYear: year=2010
main.cpp getYear: year=2010
\endverbatim */
SX_EXPORT_UTIL extern int sxIsDebugEnabled (const char *filename,
                                            const char *function,
                                            const char *feature);

#endif /* _SX_UTIL_H_ */
